/* LaHaShem HaAretz U'Mloah */

#ifndef _CONTROL_H
#define _CONTROL_H

#include "data.h"
#include "hmm.h"

typedef struct {
  Hmm *hmm;			/* Hidden Markov Model */
  Ffm **ffm;			/* Data models for HMM states */
  int uu;			/* number of states in HMM */
} Model;

/* -- function definition -- */

/* -- utility functions -- */

extern Model *allocModel(int ii, int uu);
extern void defineContFeat(Model *m, int i, Real priorVar);
extern void defineRadialFeat(Model *m, int i);
extern void displayModel(FILE *f, Model *m);
extern void freeModel(Model *m);
void print_model(Model *m);
extern void randomiseModel(Model *m, int ut, int rab);

/* -- processing functions -- */

extern Real logLike(Model *m, RMat data);
extern int *best_state_sequence(Model *m, RMat data);
extern Real update(Model *m, RMat *data, int ll);

/* -- accessor functions -- */

extern int model_ii(Model *m);
extern int model_feature_type(Model *m, int i);
extern Real model_parameter(Model *m, int u, int i, int n);
extern int model_uu(Model *m);
extern Real model_a(Model *m, int u, int v);
extern Real model_b(Model *m, int u);
extern void set_model_parameter(Model *m, int u, int i, int n, Real x);
extern void set_model_a(Model *m, int u, int v, Real x);
extern void set_model_b(Model *m, int u, Real x);

#endif

/* Tam V'Nishlam Shevah L'El Borei Olam */
