/* LaHaShem HaAretz U'Mloah */

#ifndef _DATA_H
#define _DATA_H

#include <stdio.h>
#include "hlib-c.h"

struct ContFI {Real initialSigma;};

typedef Real (*LogLikeF)(Real, void *);
typedef void (*MaxF)(void *, Real **, RVec *, int);
typedef int FeatType;

typedef struct {Real mu; Real sigma;} Param;
typedef struct {Real mean; Real kappa;} VMParam;
typedef struct {
  void **p;			/* model parameters */
  LogLikeF *lf;			/* likelihood functions */
  MaxF *mf;			/* maximisation functions */
  FeatType *ft;			/* feature types */
  void **fi;			/* additional feature info */
  int ii;			/* number of features */
} Ffm;

enum {FT_CONTINUOUS, FT_RADIAL, FT_OTHER}; /* Feature types */

/* utility functions */

extern Ffm *allocFFM(int ii);
extern void displayFFM(FILE *f, Ffm *m);
extern void freeFFM(Ffm *m);

/* standard evaluation and maximisation functions */

extern Real logGauss(Real x, void *param);
extern Real logVonMises(Real x, void *param);
extern void maxGauss(void *param, Real **x, RVec *w, int ll);
extern void maxVonMises(void *param, Real **x, RVec *w, int ll);

/* accessor functions */

extern FeatType getFeatType(Ffm *m, int i);
extern void *getFeatInfo(Ffm *m, int i);
extern void *getParam(Ffm *m, int i);
extern int getFeatNum(Ffm *m);
extern Real getInitialSigma(void);
extern void setInitialSigma(Real x);

/* setup functions */

extern void randomiseParams(Ffm *m);
extern void *setFeatType(Ffm *m, int i, FeatType ft, void *fi);
extern void setMaxFunc(Ffm *m, int i, MaxF f);
extern void setLogLikeFunc(Ffm *m, int i, LogLikeF f);
extern void *setParam(Ffm *m, int i, void *p);

/* model evaluation and maximisation functions */

extern void FFM_logL(Ffm *m, RVec logL, RMat data);
extern void FFM_maximise(Ffm *m, RMat *data, RVec *r, int ll);

#endif
/* Tam V'Nishlam Shevah L'El Borei Olam */
