#ifndef _HLIBC_SAFE_H
#define _HLIBC_SAFE_H

#include "rand.h"

#ifndef BOOLEAN
#define BOOLEAN int
#endif
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define DEG_TO_RAD(x) (PI/180.0*(x))
#define RAD_TO_DEG(x) (180.0/PI*(x))
#define RGB_TO_GREY(r, g, b) (0.299*(r)+0.587*(g)+0.114*(b))
#define MAX_HUE_VALUE 359
#define MAX_FRAMES 5400
#define MAX_OBJECTS 20
#define MAX_VERTICES 300
#define RED_BLOCK_RED_LOW 88.0	     /* needs work: make parameter */
#define RED_BLOCK_RED_HIGH 255.0     /* needs work: make parameter */
#define RED_BLOCK_GREEN_LOW 24.0     /* needs work: make parameter */
#define RED_BLOCK_GREEN_HIGH 56.0    /* needs work: make parameter */
#define RED_BLOCK_BLUE_LOW 24.0      /* needs work: make parameter */
#define RED_BLOCK_BLUE_HIGH 56.0     /* needs work: make parameter */
#define RED_BLOCK_HUE_LOW 8.0	     /* needs work: make parameter */
#define RED_BLOCK_HUE_HIGH 338.0     /* needs work: make parameter */
#define RED_BLOCK_SATURATION_LOW 0.4 /* needs work: make parameter */
#define PERSISTENCE 2		     /* needs work: make parameter */
#define RING_BUF_SIZE 5

/* macros which are constants */
#define CLIP_MIN 20.0         /* min eigenvalue in ellipse covariance matrix */
#define INFINITY (1.0/0.0)
#define LOG_MATH_PRECISION 35.0
#define MACH_EPS 1e-7
#define MINIMUM_SIGMA 0.01                        /* minimum std dev allowed */
#define NEGATIVE_INFINITY (-1.0/0.0)
#ifndef PI
#define PI 3.14159265358979323846
#endif
#define TRACE_MAX_INDEX 25

#define safe_malloc(x) _safe_malloc((x), __LINE__, __FILE__)

/* useful macros */
#ifdef _______FLOAT_REAL
typedef float Real;
#else
typedef double Real;
#endif

#ifndef SQR
static Real ___sqrarg, ___sqrarg2;
#define SQR(a) (___sqrarg=(a),___sqrarg*___sqrarg)
#define SQR2(a) (___sqrarg2=(a),___sqrarg2*___sqrarg2)
#endif

/* macros which take arguments */
#ifndef MIN
#define MIN(a,b) ((a)>(b)?(b):(a))
#endif
#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif
#define IS_ZERO(x) (-MACH_EPS<(x)&&(x)<MACH_EPS)

#define RANDOM (rand_uniform())
#define SIGMOID(x) (1.0/(1.0+exp(-(x))))

/* data accessor macros */
#define VLENGTH(v) ((v)->x)
#define ROWS(m) ((m)->y)
#define COLUMNS(m) ((m)->x)
#define VECTOR(x) ((x)->v)
#define MATRIX(x) ((x)->m)

/* datatype definitions */

struct RVecStruct {size_t x; Real *v;};
struct RMatStruct {size_t x; size_t y; Real **m;};
struct RMat3dStruct {size_t x; size_t y; size_t z; Real ***m;};
typedef struct RVecStruct *RVec;
typedef struct RMatStruct *RMat;
typedef struct RMat3dStruct *RMat3d;
struct line {Real x1, y1, x2, y2;};

enum HLibSafeDebugFlags {WARN, MALLOC, MATHERR, VERSION_PRINT,
			 TOP_HLIB_DEBUG_FLAG};
#endif
