(MODULE HLIB-CHANGE-MOVIE)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-change-movie.sch")
(include "c-externals.sc")

;;; Constants

(define jframe1 '#((#(100 120) #(130 140) #(150 170))(#(180 190)
						      #(160 150))))

(define jframe2 '#((#(200 120) #(230 140) #(250 170))(#(280 190)
						      #(260 150))))

(define polygon1 '(#(1 2) #(2 3) #(3 4)))

(define polygon2 '(#(5 6) #(3 4) #(1 2)))

(define jframe `#(,polygon1 ,polygon2))

(define jeff `#(,jframe1 ,jframe2 ,jframe2))

;;; Structures

(define-structure movie frames)

(define-structure frame number objects contacts models)

(define-structure object name flag points m i r v w v. w.)

;;; Procedures

(define (debugging-hero p q r)
 (let* ((a (distance p q))
	(b (distance q r))
	(c (distance r p))
	(s (/ (+ a b c) 2)))
  (sqrt (max 0.0 (* s (- s a) (- s b) (- s c))))))

(define (debugging-area-of-convex-polygon points)
 (if (or (null? points) (null? (rest points)) (null? (rest (rest points))))
     0
     (let ((root (first points)))
      (let loop ((points (rest points)))
       (+ (debugging-hero root (first points) (second points))
	  (if (null? (rest (rest points))) 0 (loop (rest points))))))))

(define (jmovie->rmovie jmovie)
 (make-movie
  (map-n-vector (lambda (frame-number)
		 (jframe->rframe frame-number
				 (vector-ref jmovie frame-number)
				 (if (zero? frame-number)
				     #f
				     (vector-ref jmovie (- frame-number 1)))
				 (if (>= frame-number
					 (- (vector-length jmovie) 1))
				     #f
				     (vector-ref jmovie (+ frame-number 1)))))
		(vector-length jmovie))))

(define (jframe->rframe frame-number jframe previous-jframe next-jframe)
 (make-frame frame-number
	     (map-n (lambda (i)
		     (polygon->object (vector-ref jframe i)
				      (if (eq? previous-jframe #f)
					  '(#(0.0 0.0))
					  (vector-ref previous-jframe i))
				      (if (eq? next-jframe #f)
					  '(#(0.0 0.0))
					  (vector-ref next-jframe i))))
		    (vector-length jframe))
	     #f				;unfilled
	     #f))			;unfilled

(define (polygon->object polygon previous-polygon next-polygon)
 (make-object
  #f					;unfilled
  #t
  (map vertex-2d->vertex-3d polygon)
  (debugging-area-of-convex-polygon (map vertex-2d->vertex-3d polygon))
  '#(#(1.0 0.0 0.0) #(0.0 1.0 0.0) #(0.0 0.0 1.0))
  (mean (list->vector (map vertex-2d->vertex-3d polygon)))
  (let ((velocity-2d
	 (calculate-velocity
	  (list->vector previous-polygon) (list->vector next-polygon))))
   (vector (x velocity-2d) (y velocity-2d) 0.0))
  (vector
   0.0
   0.0
   (calculate-angular-velocity
    (list->vector previous-polygon) (list->vector next-polygon)))
  (let ((acceleration-2d
	 (calculate-acceleration
	  (list->vector previous-polygon) (list->vector polygon)
	  (list->vector next-polygon))))
   (vector (x acceleration-2d) (y acceleration-2d) 0.0))
  (vector
   0.0
   0.0
   (calculate-angular-acceleration
    (list->vector previous-polygon) (list->vector polygon)
    (list->vector next-polygon)))))

(define (vertex-2d->vertex-3d vertex-2d)
 (vector (exact->inexact (x vertex-2d)) (exact->inexact (y vertex-2d)) 0.0))

(define (calculate-motion polygon) (eigenvector-angle1 (covariance polygon)))

(define (calculate-velocity previous-polygon next-polygon)
 (k*v 0.5 (v- (centroid next-polygon) (centroid previous-polygon))))

(define (calculate-angular-velocity previous-polygon next-polygon)
 (- (eigenvector-angle1 (covariance next-polygon))
    (eigenvector-angle1 (covariance previous-polygon))))

(define (calculate-acceleration previous-polygon polygon next-polygon)
 (v- (v- (centroid next-polygon) (centroid polygon))
     (v- (centroid polygon) (centroid previous-polygon))))

(define (calculate-angular-acceleration previous-polygon polygon
					next-polygon)
 (- (- (eigenvector-angle1 (covariance next-polygon))
       (eigenvector-angle1 (covariance polygon)))
    (- (eigenvector-angle1 (covariance polygon))
       (eigenvector-angle1 (covariance previous-polygon)))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
