(MODULE HLIB-EM)
;;; LaHaShem HaAretz U'Mloah

;;; Fit a set of K ellipses to a set of lines using the EM algorithm.
;;; needs work: EM:EM and EM:GROUP prefixed to avoid name conflicts

(include "QobiScheme.sch")
(include "hlib-em.sch")
(include "c-externals.sc")

;;; Structures

(define-structure group ellipse lines)

;;; Variables

(define *n* #f)
(define *lines* #f)
(define *s* #f)
(define *z* #f)
(define *log-pi* #f)
(define *mu-x* #f)
(define *mu-y* #f)
(define *sigma-xx* #f)
(define *sigma-xy* #f)
(define *sigma-yy* #f)
(define *again?* #f)
(define *p?* #f)
(define *best-log-likelihood* #f)

;;; Procedures

(define (log-f i j)
 (let* ((l (vector-ref *lines* i))
	(x1 (x (p l)))
	(y1 (y (p l)))
	(x2 (x (q l)))
	(y2 (y (q l)))
	(mu-x (vector-ref *mu-x* j))
	(mu-y (vector-ref *mu-y* j))
	(sigma-xx (vector-ref *sigma-xx* j))
	(sigma-xy (vector-ref *sigma-xy* j))
	(sigma-yy (vector-ref *sigma-yy* j))
	(delta (- (* sigma-xx sigma-yy) (* sigma-xy sigma-xy))))
  (* (vector-ref *s* i)
     (/ (+ (*  3.0        mu-y sigma-xx     mu-y)
	   (* -3.0          y1 sigma-xx     mu-y)
	   (*               y1 sigma-xx       y1)
	   (* -3.0          y2 sigma-xx     mu-y)
	   (*               y2 sigma-xx       y1)
	   (*               y2 sigma-xx       y2)
	   (* -6.0        mu-x sigma-xy     mu-y)
	   (*  3.0        mu-x sigma-xy       y1)
	   (*  3.0        mu-x sigma-xy       y2)
	   (*  3.0          x1 sigma-xy     mu-y)
	   (* -2.0          x1 sigma-xy       y1)
	   (* -1.0          x1 sigma-xy       y2)
	   (*  3.0          x2 sigma-xy     mu-y)
	   (* -1.0          x2 sigma-xy       y1)
	   (* -2.0          x2 sigma-xy       y2)
	   (*  3.0        mu-x sigma-yy     mu-x)
	   (* -3.0          x1 sigma-yy     mu-x)
	   (*               x1 sigma-yy       x1)
	   (* -3.0          x2 sigma-yy     mu-x)
	   (*               x2 sigma-yy       x1)
	   (*               x2 sigma-yy       x2)
	   (*  3.0 (log delta) sigma-xx sigma-yy)
	   (*  6.0    (log pi) sigma-xx sigma-yy)
	   (*  6.0     (log 2) sigma-xx sigma-yy)
	   (* -3.0 (log delta) sigma-xy sigma-xy)
	   (* -6.0    (log pi) sigma-xy sigma-xy)
	   (* -6.0     (log 2) sigma-xy sigma-xy))
	(* -6.0 delta)))))

(define (initial-s!)
 (let ((n (reduce
	   min
	   (map-n (lambda (i) (line-segment-length (vector-ref *lines* i)))
		  *n*)
	   #f)))
  (for-each-n
   (lambda (i)
    (vector-set! *s* i (/ (line-segment-length (vector-ref *lines* i)) n)))
   *n*)))

(define (initial-z!)
 (for-each-n
  (lambda (i) (for-each-n (lambda (j) (matrix-set! *z* i j 0.0)) *m*)) *n*)
 ;; This is really a kludge to handle the case where there are fewer line
 ;; segments then requested mixture components.
 (let ((p (random-partition-of-size *m* (enumerate (max *m* *n*)))))
  (for-each-n (lambda (j)
	       (for-each (lambda (i) (matrix-set! *z* (remainder i *n*) j 1.0))
			 (list-ref p j)))
	      *m*))
 (when *p?* (write 'z) (newline) (write *z*) (newline)))

(define (initial-pi!)
 (for-each-n (lambda (j) (vector-set! *log-pi* j (log (/ *m*)))) *m*)
 (when *p?* (write 'pi) (newline) (write (map-vector exp *log-pi*)) (newline)))

(define (lem-log-likelihood)
 (sum (lambda (i)
       (log (sum (lambda (j) (exp (+ (vector-ref *log-pi* j) (log-f i j))))
		 *m*)))
      *n*))

(define (compute-z!)
 (when *p?*
  (write 'log-pi/log-fij/log-pi+log-fij/exp-log-pi+log-fij)
  (newline)
  (for-each-n
   (lambda (i)
    (write (map-n
	    (lambda (j)
	     (list (vector-ref *log-pi* j)
		   (log-f i j)
		   (+ (vector-ref *log-pi* j) (log-f i j))
		   (exp (+ (vector-ref *log-pi* j) (log-f i j)))))
	    *m*))
    (newline))
   *n*))
 ;; two calls to (LOG-F I J)
 (for-each-n
  (lambda (i)
   (let ((n (sum (lambda (j) (exp (+ (vector-ref *log-pi* j) (log-f i j))))
		 *m*)))
    (if (zero? n)
	(let ((j1 (argmin-n
		   (lambda (j)
		    (sqrt
		     (+ (sqr (- (vector-ref *mu-x* j)
				(* 0.5
				   (+ (x (p
					  (vector-ref *lines* i)))
				      (x (q
					  (vector-ref *lines* i)))))))
			(sqr (- (vector-ref *mu-y* j)
				(* 0.5
				   (+ (y (p
					  (vector-ref *lines* i)))
				      (y (q
					  (vector-ref *lines* i))))))))))
		   *m*)))
	 (for-each-n
	  (lambda (j)
	   (let ((new-z (if (= j j1) 1.0 0.0)))
	    (when (> (abs (- new-z (matrix-ref *z* i j))) *epsilon*)
	     (set! *again?* #t))
	    (matrix-set! *z* i j new-z)))
	  *m*))
	(let ((n (log n)))
	 (for-each-n
	  (lambda (j)
	   (let ((new-z (exp (- (+ (vector-ref *log-pi* j) (log-f i j)) n))))
	    (when (> (abs (- new-z (matrix-ref *z* i j))) *epsilon*)
	     (set! *again?* #t))
	    (matrix-set! *z* i j new-z)))
	  *m*)))))
  *n*)
 (when *p?* (write 'z) (newline) (write *z*) (newline)))

(define (compute-pi!)
 (for-each-n
  (lambda (j)
   (vector-set!
    *log-pi* j (log (/ (sum (lambda (i) (matrix-ref *z* i j)) *n*) *n*))))
  *m*)
 (when *p?* (write 'pi) (newline) (write (map-vector exp *log-pi*)) (newline)))

(define (compute-mu!)
 (for-each-n
  (lambda (j)
   (let ((n (sum (lambda (i) (* (matrix-ref *z* i j) (vector-ref *s* i)))
		 *n*)))
    (vector-set! *mu-x* j
		 (/ (sum (lambda (i)
			  (* 0.5
			     (matrix-ref *z* i j)
			     ;; from here on is not a function of J and can be
			     ;; factored out
			     (vector-ref *s* i)
			     (+ (x (p (vector-ref *lines* i)))
				(x (q (vector-ref *lines* i))))))
			 *n*)
		    n))
    (vector-set! *mu-y* j
		 (/ (sum (lambda (i)
			  (* 0.5
			     (matrix-ref *z* i j)
			     ;; from here on is not a function of J and can be
			     ;; factored out
			     (vector-ref *s* i)
			     (+ (y (p (vector-ref *lines* i)))
				(y (q (vector-ref *lines* i))))))
			 *n*)
		    n))))
  *m*)
 (when *p?*
  (write 'mu)
  (newline)
  (write *mu-x*)
  (newline)
  (write *mu-y*)
  (newline)))

(define (compute-sigma!)
 (for-each-n
  (lambda (j)
   (let* ((n (sum (lambda (i) (* (matrix-ref *z* i j) (vector-ref *s* i)))
		  *n*))
	  (xm (vector-ref *mu-x* j))
	  (ym (vector-ref *mu-y* j))
	  (xx (/ (sum (lambda (i)
		       (let ((x1 (x (p (vector-ref *lines* i))))
			     (x2 (x (q (vector-ref *lines* i)))))
			(/ (* (matrix-ref *z* i j)
			      (vector-ref *s* i)
			      (+ (*  3.0 xm xm)
				 (*      x1 x1)
				 (*      x1 x2)
				 (*      x2 x2)
				 (* -3.0 x1 xm)
				 (* -3.0 x2 xm)))
			   3.0)))
		      *n*)
		 n))
	  (xy (/ (sum (lambda (i)
		       (let ((x1 (x (p (vector-ref *lines* i))))
			     (y1 (y (p (vector-ref *lines* i))))
			     (x2 (x (q (vector-ref *lines* i))))
			     (y2 (y (q (vector-ref *lines* i)))))
			(/ (* (matrix-ref *z* i j)
			      (vector-ref *s* i)
			      (+ (*  6.0 xm ym)
				 (*  2.0 x1 y1)
				 (*      x1 y2)
				 (*      x2 y1)
				 (*  2.0 x2 y2)
				 (* -3.0 x1 ym)
				 (* -3.0 y1 xm)
				 (* -3.0 x2 ym)
				 (* -3.0 y2 xm)))
			   6.0)))
		      *n*)
		 n))
	  (yy (/ (sum (lambda (i)
		       (let ((y1 (y (p (vector-ref *lines* i))))
			     (y2 (y (q (vector-ref *lines* i)))))
			(/ (* (matrix-ref *z* i j)
			      (vector-ref *s* i)
			      (+ (*  3.0 ym ym)
				 (*      y1 y1)
				 (*      y1 y2)
				 (*      y2 y2)
				 (* -3.0 y1 ym)
				 (* -3.0 y2 ym)))
			   3.0)))
		      *n*)
		 n))
	  (sigma (clip-eigenvalues (vector (vector xx xy) (vector xy yy))
				   '#(20.0 20.0))))
    (vector-set! *sigma-xx* j (matrix-ref sigma 0 0))
    (vector-set! *sigma-xy* j (matrix-ref sigma 0 1))
    (vector-set! *sigma-yy* j (matrix-ref sigma 1 1))))
  *m*)
 (when *p?*
  (write 'sigma) (newline)
  (write *sigma-xx*) (newline)
  (write *sigma-xy*) (newline)
  (write *sigma-yy*) (newline)))

(define (em:em m lines)
 (set! *n* (vector-length lines))
 (set! *lines* lines)
 (set! *s* (make-vector *n*))
 (initial-s!)
 (when m (set! *m* m))
 (set! *z* (make-matrix *n* *m*))
 (initial-z!)
 (when m
  (set! *log-pi* (make-vector *m*))
  (set! *mu-x* (make-vector *m*))
  (set! *mu-y* (make-vector *m*))
  (set! *sigma-xx* (make-vector *m*))
  (set! *sigma-xy* (make-vector *m*))
  (set! *sigma-yy* (make-vector *m*))
  (initial-pi!)
  (compute-mu!)
  (compute-sigma!))
 (let ((l (lem-log-likelihood)))
  (when *p?* (format #t "Initial log likelihood: ~s~%" l))
  (let loop ()
   (set! *again?* #f)
   (compute-z!)
   (compute-pi!)
   (compute-mu!)
   (compute-sigma!)
   (when *p?*
    (let ((new-l (lem-log-likelihood)))
     (format #t "New log likelihood: ~s" new-l)
     (when (< new-l l) (format #t " decreasing"))
     (newline)
     (set! l new-l)))
   (when *again?* (loop)))))

(define (em:group m lines l)
 (let ((best-result (em:em m lines)))
  (set! *best-log-likelihood* (lem-log-likelihood))
  (for-each-n
   (lambda (i)
    (let* ((this-result (em:em m lines))
	   (this-log-likelihood (lem-log-likelihood)))
     (when (> this-log-likelihood *best-log-likelihood*)
      (set! best-result this-result)
      (set! *best-log-likelihood* this-log-likelihood))))
   (- l 1))
  (list->vector
   (map-n (lambda (j)
	   (make-group (create-ellipse (vector-ref *mu-x* j)
				       (vector-ref *mu-y* j)
				       (vector-ref *sigma-xx* j)
				       (vector-ref *sigma-xy* j)
				       (vector-ref *sigma-yy* j))
		       (removeq #f
				(map-n (lambda (i)
					(if (>= (matrix-ref *z* i j) 0.5)
					    (vector-ref lines i)
					    #f))
				       *n*))))
	  *m*))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
