(MODULE HLIB-FIND-LINES)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-find-lines.sch")
(include "c-externals.sc")

;;; TTMTTD
;;;  1. Look at Kanade's work
;;;  2. Should add generic functions to define-structure and use them
;;;     throughout.
;;;  3. Should have way of displaying chains, line segments, and image points.
;;;  4. Read Basri's work from Winston.
;;;  5. Get UMD TR.
;;;  6. Fix line-segments.
;;;  7. Have filter-chain remove chains which have only one element.

;;; Target verbs:
;;;  1. pick up
;;;  2. put down
;;;  3. push
;;;  4. pull
;;;  5. slide
;;;  6. drop
;;;  7. fall
;;;  8. bounce
;;;  9. collide
;;; 10. collapse
;;; 11. knock over

;;; Variables

(define *line-threshold* 9)
(define *neighbour-threshold* 4)
(define *length-threshold* 144)
(define *point-proximity-threshold* 4)
(define *flag* #f)

;;; Procedures

(define (set-line-threshold! x)
 (when (negative? x) (panic "Line threshold must not be less than zero"))
 (set! *line-threshold* (* x x)))

(define (set-neighbour-threshold! x)
 (when (negative? x) (panic "Neighbour threshold must not be less than zero"))
 (set! *neighbour-threshold* (* x x)))

(define (set-length-threshold! x)
 (when (negative? x) (panic "Length threshold must not be less than zero"))
 (set! *length-threshold* (* x x)))

(define (set-point-proximity-threshold! x)
 (when (negative? x)
  (panic "Point proximity threshold must not be less than zero"))
 (set! *point-proximity-threshold* (* x x)))

(define (pgm-chains pgm . parity)
 (let* ((m (pnm-height pgm))
	(n (pnm-width pgm))
	(grey (pgm-grey pgm))
	(a (make-matrix n m #f))
	(chains '()))
  (define (gather-chains x-direction y-direction first-check)
   (let* ((right-border (- n 1))
	  (bottom-border (- m 1))
	  (x-constraint
	   (if (eq? x-direction -)
	       (lambda (x y) (and (positive? x)
				  (matrix-ref a (x-direction x 1) y)))
	       (lambda (x y) (and (< x right-border)
				  (matrix-ref a (x-direction x 1) y)))))
	  (y-constraint
	   (if (eq? y-direction -)
	       (lambda (x y) (and (positive? y)
				  (matrix-ref a x (y-direction y 1))))
	       (lambda (x y) (and (< y bottom-border)
				  (matrix-ref a x (y-direction y 1))))))
	  (x-looper (if (eq? x-direction -)
			for-each-n
			for-each-n-decreasing))
	  (y-looper (if (eq? y-direction -)
			for-each-n
			for-each-n-decreasing)))
    (x-looper
     (lambda (x)
      (y-looper
       (lambda (y)
	(unless (zero? (matrix-ref grey y x))
	 (if (eq? first-check 'x)
	     (cond ((and (x-constraint x y) (matrix-ref a (x-direction x 1) y))
		    (matrix-set! a x y (cons (vector x y)
					     (matrix-ref a
							 (x-direction x 1)
							 y)))
		    (matrix-set! a (x-direction x 1) y #f))
		   ((and (y-constraint x y) (matrix-ref a x (y-direction y 1)))
		    (matrix-set! a x y (cons (vector x y)
					     (matrix-ref a
							 x
							 (y-direction y 1))))
		    (matrix-set! a x (y-direction y 1) #f))
		   (else (matrix-set! a x y (list (vector x y)))))
	     (cond ((and (y-constraint x y) (matrix-ref a x (y-direction y 1)))
		    (matrix-set! a x y (cons (vector x y)
					     (matrix-ref a
							 x
							 (y-direction y 1))))
		    (matrix-set! a x (y-direction y 1) #f))
		   ((and (x-constraint x y) (matrix-ref a (x-direction x 1) y))
		    (matrix-set! a x y (cons (vector x y)
					     (matrix-ref a
							 (x-direction x 1)
							 y)))
		    (matrix-set! a (x-direction x 1) y #f))
		   (else (matrix-set! a x y (list (vector x y))))))))
       m))
     n)))
  (if (or (null? parity) (not (first parity)))
      (gather-chains + + 'x)
      (case (first parity)
       ((down) (gather-chains - - 'x))
       ((up) (gather-chains - + 'x))
       ((top) (gather-chains - - 'y))
       ((bottom) (gather-chains - + 'y))
       ((charm) (panic "This program is not for you"))))
  (for-each-n (lambda (x)
	       (for-each-n (lambda (y)
			    (when (matrix-ref a x y)
			     (set! chains (cons (matrix-ref a x y) chains))))
			   m))
	      n)
  chains))

(define (square-of-distance-from-2d-point-to-line r p q)
 ;; worst case:
 ;; 4 +
 ;; 1 unary -
 ;; 5 binary -
 ;; 9 *
 (let* ((rx (x r))
	(ry (y r))
	(px (x p))
	(py (y p))
	(qx (x q))
	(qy (y q))
	(c1 (- qx px))
	(c2 (- qy py))
	(c3 (* c1 c1))
	(c4 (* c2 c2))
	(c5 (- rx px)))
  (cond ((zero? c1) (* c5 c5))
	((zero? c2) (* (- ry py) (- ry py)))
	(else
	 (let* ((dy (- ry (/ (+ (* py c3) (* c5 c1 c2) (* ry c4)) (+ c3 c4))))
		(dx (- (/ (* dy c2) c1))))
	  (+ (* dx dx) (* dy dy)))))))

;;; determines whether two points are near enough neighbours in the image that
;;; the chains they make up should be connected
(define (neighbours? p q) (<= (distance-squared p q) *neighbour-threshold*))

;;; determines if two chains are connected in the image, returns the new
;;; connected chain if there are, fails otherwise
(define (connected x y)
 (when (or (null? x) (null? y))
  (panic "Null chain detected in function CONNECTED"))
 (let ((xp (first x))
       (yp (first y))
       (xq (last x))
       (yq (last y)))
  (cond ((neighbours? xp yq) (append y x))
	((neighbours? xq yp) (append x y))
	((neighbours? xp yp) (append (reverse x) y))
	((neighbours? xq yq) (append x (reverse y)))
	(else #f))))

;;; Examines a list of chains and returns a list with adjacent chains
;;; connected, to the extent to which this is possible
(define (connected-chains chains)
 (define (loop chains results)
  (cond ((null? chains) results)
	((null? (cdr chains)) (append chains results))
	(else
	 (let ((chain-pair (find-first-connection chains)))
	  ;; if a chain doesn't connect with any chains, it will not ever
	  ;; connect with conglomerations of those chains
	  (if chain-pair
	      (loop (cons (first chain-pair)
			  (remove (second chain-pair) (rest chains)))
		    results)
	      (loop (rest chains) (cons (first chains) results)))))))
 (define (find-first-connection chains)
  (define (loop x l)
   (cond ((null? l) #f)
	 ((connected x (first l)) => (lambda (z) (list z (first l))))
	 (else (loop x (rest l)))))
  (loop (first chains) (rest chains)))
 (loop chains '()))

(define (over-threshold-length? l)
 (>= (distance-squared (p l) (q l)) *length-threshold*))

(define (filter-chain chain)
 (let ((last-point (- (length chain) 1))
       (points (list->vector chain)))
  (define (loop start end)
   (cond
    ((= start end) '())
    ((= 1 (- start end)) (list (vector-ref points start)))
    (else
     (let* ((p (vector-ref points start))
	    (q (vector-ref points end))
	    (s (+ start 1))
	    (maximum-distance (square-of-distance-from-2d-point-to-line
			       (vector-ref points s) p q))
	    (r (+ s 1)))
      (for-each-n
       ;; needs work: potentially buggy if (- end r) isn't evaluated
       ;;             immediately
       (lambda (i)
	(let* ((d (square-of-distance-from-2d-point-to-line
		   (vector-ref points r) p q)))
	 (when (> d maximum-distance) (set! s r) (set! maximum-distance d)))
	(set! r (+ r 1)))
       (- end r))
      (if (> maximum-distance *line-threshold*)
	  ;; needs work: reduce excessive append calls
	  (append (loop start s) (loop s end))
	  (list p))))))
  (append (loop 0 last-point) (list (vector-ref points last-point)))))

(define (efficient-connected-chains chains)
 (connected-chains (map filter-chain (sort chains > length))))

(define (chain->line-segments chain)
 (if (null? (rest chain))
     '()
     (cons (make-line-segment (first chain) (second chain))
	   (chain->line-segments (rest chain)))))

(define (adjacent? p q)
 (and (not (equal? p q))
      (<= (distance-squared p q) *point-proximity-threshold*)))

(define (adjacent-line x y)
 (let ((xp (p x))
       (yp (p y))
       (xq (q x))
       (yq (q y)))
  (cond ((adjacent? xp yq) (list (make-line-segment yp xp) y))
	((adjacent? xq yp) (list (make-line-segment xq yq) y))
	((adjacent? xp yp) (list (make-line-segment xp yq) y))
	((adjacent? xq yq) (list (make-line-segment yp xq) y))
	(else #f))))

(define (find-adjacent-line x l)
 (cond ((null? l) #f)
       ((adjacent-line x (first l)) => (lambda (z) z))
       (else (find-adjacent-line x (rest l)))))

(define (merge-adjacent-endpoints lines)
 (define (loop l c)
  (cond ((null? l) c)
	((null? (cdr l)) (append l c))
	(else (let* ((x (first l))
		     (y-pair (find-adjacent-line x (rest l))))
	       (if y-pair
		   (loop (cons (first y-pair) (remove (second y-pair) l)) c)
		   (loop (rest l) (cons x c)))))))
 (loop lines '()))

(define (line-segments pgm)
 (merge-adjacent-endpoints
  (remove-if-not
   over-threshold-length?
   (reduce append
	   (map (lambda (x) (chain->line-segments (filter-chain x)))
		(efficient-connected-chains (pgm-chains pgm)))
	   '()))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
