(MODULE HLIB-HULL)
;;; LaHaShem HaAretz U'Mloah

;;; Things to try:
;;;  1. move-three
;;;  2. best-of-n(best-of-n(hull);EM)
;;;  3. correspondence
;;;  4. message to give total cost

;;; needs work: HULL:GROUP prefixed to avoid name conflicts

(include "QobiScheme.sch")
(include "hlib-hull.sch")
(include "c-externals.sc")

;;; Variables

(define *hull-level* 0)

(define *hull-metric* 'area)

;;; Procedures

(define (polygon-cost points)
 ((case *hull-metric*
   ((area) area-of-polygon)
   ((perimeter) perimeter-of-polygon)
   (else (fuck-up)))
  points))

(define (cost-of-partition groups)
 (reduce
  +
  (map (lambda (group)
	(polygon-cost (append (map p group) (map q group))))
       groups)
  0))

(define (convex-hull-of-line-segments line-segments)
 (convex-hull (append (map p line-segments) (map q line-segments))))

(define (points->line-segments points)
 (if (or (null? points) (null? (rest points)))
     '()
     (map make-line-segment
	  points
	  (cons (last points) (reverse (rest (reverse points)))))))

(define (hull:group k line-segments)
 (let* ((groups (random-partition-of-size k line-segments))
	(hulls (map convex-hull-of-line-segments groups))
	(areas (map polygon-cost hulls)))
  (let loop ()
   (let ((again? #f))
    (call-with-current-continuation
     (lambda (return)
      (for-each
       (lambda (group1 hull1 area1)
	(for-each
	 (lambda (line-segment)
	  (when (some (lambda (point)
		       (or (eq? point (p line-segment))
			   (eq? point (q line-segment))))
		      hull1)
	   (let* ((group1-prime (removeq line-segment group1))
		  (hull1-prime (convex-hull-of-line-segments group1-prime))
		  (area1-prime (polygon-cost hull1-prime)))
	    (for-each
	     (lambda (group2 area2)
	      (when (not (eq? group1 group2))
	       (let* ((group2-prime (cons line-segment group2))
		      (hull2-prime
		       (convex-hull-of-line-segments group2-prime))
		      (area2-prime (polygon-cost hull2-prime)))
		(when (> (- area1 area1-prime) (- area2-prime area2))
		 (let loop ((groups groups)
			    (hulls hulls)
			    (areas areas))
		  (unless (null? groups)
		   (cond ((eq? (first groups) group1)
			  (set-car! groups group1-prime)
			  (set-car! hulls hull1-prime)
			  (set-car! areas area1-prime))
			 ((eq? (first groups) group2)
			  (set-car! groups group2-prime)
			  (set-car! hulls hull2-prime)
			  (set-car! areas area2-prime)))
		   (loop (rest groups) (rest hulls) (rest areas))))
		 (set! again? #t)
		 (return #f)))))
	     groups
	     areas))))
	 group1))
       groups
       hulls
       areas)
      (when (> *hull-level* 0)
       (for-each
	(lambda (group1 hull1 area1)
	 (for-each
	  (lambda (line-segment1)
	   (when (some (lambda (point)
			(or (eq? point (p line-segment1))
			    (eq? point (q line-segment1))))
		       hull1)
	    (for-each
	     (lambda (line-segment2)
	      (when (and
		     (not (eq? line-segment1 line-segment2))
		     (some (lambda (point)
			    (or (eq? point (p line-segment2))
				(eq? point (q line-segment2))))
			   hull1))
	       (let* ((group1-prime
		       (removeq line-segment2 (removeq line-segment1 group1)))
		      (hull1-prime (convex-hull-of-line-segments group1-prime))
		      (area1-prime (polygon-cost hull1-prime)))
		(for-each
		 (lambda (group2 area2)
		  (when (not (eq? group1 group2))
		   (let* ((group2-prime
			   (cons line-segment2 (cons line-segment1 group2)))
			  (hull2-prime
			   (convex-hull-of-line-segments group2-prime))
			  (area2-prime (polygon-cost hull2-prime)))
		    (when (> (- area1 area1-prime) (- area2-prime area2))
		     (let loop ((groups groups)
				(hulls hulls)
				(areas areas))
		      (unless (null? groups)
		       (cond ((eq? (first groups) group1)
			      (set-car! groups group1-prime)
			      (set-car! hulls hull1-prime)
			      (set-car! areas area1-prime))
			     ((eq? (first groups) group2)
			      (set-car! groups group2-prime)
			      (set-car! hulls hull2-prime)
			      (set-car! areas area2-prime)))
		       (loop (rest groups) (rest hulls) (rest areas))))
		     (set! again? #t)
		     (return #f)))))
		 groups
		 areas)
		(when (> *hull-level* 1)
		 (for-each
		  (lambda (group2 area2)
		   (when (not (eq? group1 group2))
		    (let* ((group2-prime (cons line-segment1 group2))
			   (hull2-prime
			    (convex-hull-of-line-segments group2-prime))
			   (area2-prime (polygon-cost hull2-prime)))
		     (for-each
		      (lambda (group3 area3)
		       (when (and (not (eq? group1 group3))
				  (not (eq? group2 group3)))
			(let* ((group3-prime (cons line-segment2 group3))
			       (hull3-prime
				(convex-hull-of-line-segments group3-prime))
			       (area3-prime (polygon-cost hull3-prime)))
			 (when (> (- area1 area1-prime)
				  (+ (- area2-prime area2)
				     (- area3-prime area3)))
			  (let loop ((groups groups)
				     (hulls hulls)
				     (areas areas))
			   (unless (null? groups)
			    (cond ((eq? (first groups) group1)
				   (set-car! groups group1-prime)
				   (set-car! hulls hull1-prime)
				   (set-car! areas area1-prime))
				  ((eq? (first groups) group2)
				   (set-car! groups group2-prime)
				   (set-car! hulls hull2-prime)
				   (set-car! areas area2-prime))
				  ((eq? (first groups) group3)
				   (set-car! groups group3-prime)
				   (set-car! hulls hull3-prime)
				   (set-car! areas area3-prime)))
			    (loop (rest groups) (rest hulls) (rest areas))))
			  (set! again? #t)
			  (return #f)))))
		      groups
		      areas))))
		  groups
		  areas)))))
	     group1)))
	  group1))
	groups
	hulls
	areas))))
    (when again? (loop))))
  groups))

;;; Tam V'Nishlam Shevah L'El Borei Olam
