(MODULE HLIB-MOST)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-most.sch")
(include "c-externals.sc")

;;; Structures

(define-structure pixel x y hue saturation value red green blue)

;;; Variables

(define *width* 640)			;needs work: should set from video file

(define *height* 480)			;needs work: should set from video file

(define *max-rgb* (vector *max-red* *max-blue* *max-green*))

(define *m* #f)

(define *motion-threshold* 1.0)

(define *words* '("clipped" "corresponded"))

(define *really-words* '("clipped"
			 "corresponded"
			 "left"
			 "right"
			 "qobi"
			 "mann"
			 "piroska"
			 "michael"
			 "fellbaum"
			 "sebastien"
			 "tjan"
			 "dwj"
			 "mary"
			 "waltz"
			 "allan"))

;;; Procedures

(define (randomize) (for-each-n (lambda (i) (random-boolean)) (getpid)))

(define (strip pathname words trailing-digits?)
 (cond
  ((has-extension? pathname)
   (strip (strip-extension pathname) words trailing-digits?))
  ((has-directory? pathname)
   (strip (strip-directory pathname) words trailing-digits?))
  ((and trailing-digits?
	(char-numeric? (string-ref pathname (- (string-length pathname) 1))))
   (strip (list->string (but-last (string->list pathname)))
	  words
	  trailing-digits?))
  (else (let ((word
	       (find-if
		(lambda (word)
		 (let ((n (string-length word)))
		  (and (>= (string-length pathname) (+ n 1))
		       (char=? (string-ref
				pathname (- (string-length pathname) (+ n 1)))
			       #\-)
		       (string=? (substring pathname
					    (- (string-length pathname) n)
					    (string-length pathname))
				 word))))
		words)))
	 (if word
	     (strip (substring
		     pathname
		     0
		     (- (string-length pathname) (+ (string-length word) 1)))
		    words
		    trailing-digits?)
	     pathname)))))

(define (mean x)
 (when (null? x) (fuck-up))
 (k*v (/ (vector-length x)) (reduce-vector v+ x #f)))

(define (covariance x)
 (when (null? x) (fuck-up))
 (let ((mu (mean x)))
  (k*m (/ (vector-length x))
       (self-outer-product
	dot (transpose (map-vector (lambda (x) (v- x mu)) x))))))

(define (min-n f n)
 (let loop ((n (- n 2)) (best-f (f (- n 1))))
  (if (negative? n)
      best-f
      (let ((f-i (f n)))
       (if (< f-i best-f) (loop (- n 1) f-i) (loop (- n 1) best-f))))))

(define (max-n f n)
 (let loop ((n (- n 2)) (best-f (f (- n 1))))
  (if (negative? n)
      best-f
      (let ((f-i (f n)))
       (if (> f-i best-f) (loop (- n 1) f-i) (loop (- n 1) best-f))))))

(define (argmin-n f n)
 (let loop ((n (- n 2)) (best-f (f (- n 1))) (best-i (- n 1)))
  (if (negative? n)
      best-i
      (let ((f-i (f n)))
       (if (< f-i best-f)
	   (loop (- n 1) f-i n)
	   (loop (- n 1) best-f best-i))))))

(define (argmax-n f n)
 (let loop ((n (- n 2)) (best-f (f (- n 1))) (best-i (- n 1)))
  (if (negative? n)
      best-i
      (let ((f-i (f n)))
       (if (> f-i best-f)
	   (loop (- n 1) f-i n)
	   (loop (- n 1) best-f best-i))))))

(define (argmin f l)
 (let loop ((l (rest l)) (best-cost (f (first l))) (best (first l)))
  (if (null? l)
      best
      (let ((this-cost (f (first l))))
       (if (< this-cost best-cost)
	   (loop (rest l) this-cost (first l))
	   (loop (rest l) best-cost best))))))

(define (argmax f l)
 (let loop ((l (rest l)) (best-cost (f (first l))) (best (first l)))
  (if (null? l)
      best
      (let ((this-cost (f (first l))))
       (if (> this-cost best-cost)
	   (loop (rest l) this-cost (first l))
	   (loop (rest l) best-cost best))))))

(define (mode l)
 (let* ((cs (remove-duplicates l))
	(ds (map (lambda (c) (count c l)) cs)))
  (list-ref cs (position (reduce max ds 0) ds))))

(define (boolean-vector->c-int-vector v)
 (let* ((length (vector-length v))
	(c (make-c-int-vector length)))
  (do ((i 0 (+ i 1))) ((>= i length))
   (c-int-vector-set! c i (if (vector-ref v i) 1 0)))
  c))

(define (c-int-vector->boolean-vector c length)
 (let* ((v (make-vector length)))
  (do ((i 0 (+ i 1))) ((>= i length))
   (vector-set! v i (if (zero? (c-int-vector-ref c i)) #f #t)))
  v))

(define (boolean-matrix->c-int-matrix m)
 (let* ((height (matrix-rows m))
	(width (matrix-columns m))
	(c (make-c-int-matrix height width)))
  (do ((i 0 (+ i 1))) ((>= i height))
   (do ((j 0 (+ j 1))) ((>= j width))
    (c-int-matrix-set! c i j (if (matrix-ref m i j) 1 0))))
  c))

(define (c-int-matrix->boolean-matrix c height width)
 (let ((m (make-matrix height width)))
  (do ((i 0 (+ i 1))) ((>= i height))
   (do ((j 0 (+ j 1))) ((>= j width))
    (matrix-set! m i j (if (zero? (c-int-matrix-ref c i j)) #f #t))))
  m))

(define (vector->c-int-vector v)
 (let* ((length (vector-length v))
	(c (make-c-int-vector length)))
  (do ((i 0 (+ i 1))) ((>= i length))
   (c-int-vector-set! c i (vector-ref v i)))
  c))

(define (c-int-vector->vector c length)
 (let* ((v (make-vector length)))
  (do ((i 0 (+ i 1))) ((>= i length))
   (vector-set! v i (c-int-vector-ref c i)))
  v))

(define (matrix->c-int-matrix m)
 (let* ((height (matrix-rows m))
	(width (matrix-columns m))
	(c (make-c-int-matrix height width)))
  (do ((i 0 (+ i 1))) ((>= i height))
   (do ((j 0 (+ j 1))) ((>= j width))
    (c-int-matrix-set! c i j (matrix-ref m i j))))
  c))

(define (c-int-matrix->matrix c height width)
 (let ((m (make-matrix height width)))
  (do ((i 0 (+ i 1))) ((>= i height))
   (do ((j 0 (+ j 1))) ((>= j width))
    (matrix-set! m i j (c-int-matrix-ref c i j))))
  m))

(define (vector->c-float-vector v)
 (let* ((length (vector-length v))
	(c (make-c-float-vector length)))
  (do ((i 0 (+ i 1))) ((>= i length))
   (c-float-vector-set! c i (vector-ref v i)))
  c))

(define (c-float-vector->vector c length)
 (let* ((v (make-vector length)))
  (do ((i 0 (+ i 1))) ((>= i length))
   (vector-set! v i (c-float-vector-ref c i)))
  v))

(define (matrix->c-float-matrix m)
 (let* ((height (matrix-rows m))
	(width (matrix-columns m))
	(c (make-c-float-matrix height width)))
  (do ((i 0 (+ i 1))) ((>= i height))
   (do ((j 0 (+ j 1))) ((>= j width))
    (c-float-matrix-set! c i j (matrix-ref m i j))))
  c))

(define (c-float-matrix->matrix c height width)
 (let ((m (make-matrix height width)))
  (do ((i 0 (+ i 1))) ((>= i height))
   (do ((j 0 (+ j 1))) ((>= j width))
    (matrix-set! m i j (c-float-matrix-ref c i j))))
  m))

(define (vector->c-real-vector v)
 (let* ((length (vector-length v))
	(c (make-c-real-vector length)))
  (do ((i 0 (+ i 1))) ((>= i length))
   (c-real-vector-set! c i (vector-ref v i)))
  c))

(define (c-real-vector->vector c length)
 (let* ((v (make-vector length)))
  (do ((i 0 (+ i 1))) ((>= i length))
   (vector-set! v i (c-real-vector-ref c i)))
  v))

(define (matrix->c-real-matrix m)
 (let* ((height (matrix-rows m))
	(width (matrix-columns m))
	(c (make-c-real-matrix height width)))
  (do ((i 0 (+ i 1))) ((>= i height))
   (do ((j 0 (+ j 1))) ((>= j width))
    (c-real-matrix-set! c i j (matrix-ref m i j))))
  c))

(define (c-real-matrix->matrix c height width)
 (let ((m (make-matrix height width)))
  (do ((i 0 (+ i 1))) ((>= i height))
   (do ((j 0 (+ j 1))) ((>= j width))
    (matrix-set! m i j (c-real-matrix-ref c i j))))
  m))

(define (a-subset-of-size k set)
 (when (or (< k 0) (< (length set) k)) (fail))
 (cond ((= (length set) k) set)
       ((null? set) (fail))
       (else (either (cons (first set) (a-subset-of-size (- k 1) (rest set)))
		     (a-subset-of-size k (rest set))))))

(define (create-ellipse mu-x mu-y sigma-xx sigma-xy sigma-yy)
 (let* ((sigma (vector (vector sigma-xx sigma-xy)
		       (vector sigma-xy sigma-yy)))
	(e (eigenvalues sigma)))
  (make-ellipse mu-x
		mu-y
		(eigenvector-angle1 sigma)
		(sqrt (* 2.0 (vector-ref e 0)))
		(sqrt (* 2.0 (vector-ref e 1))))))

(define (quantize x quanta) (inexact->exact (* quanta (floor (/ x quanta)))))

(define (classify x mus distance)
 (let ((distances (map (lambda (mu) (distance mu x)) mus)))
  (positionv (reduce min distances 0) distances)))

(define (similar-xs y ys xs distance-threshold f distance)
 (let ((index (position y ys)))
  (remove-if-not (lambda (x)
		  (and (<= (distance (f x) y) distance-threshold)
		       (= index (classify (f x) ys distance))))
		 xs)))

(define (histogram-clusterer
	 xs count-threshold distance-threshold mu quantize f distance)
 (let ((ys (map (lambda (class)
		 (mu (reduce append (map second class) '())))
		(equivalence-classesp
		 (lambda (x y)
		  (<= (distance (first x) (first y)) distance-threshold))
		 (map (lambda (quantized-x)
		       (let ((xs (remove-if-not
				  (lambda (x)
				   (equal? (quantize x) quantized-x)) xs)))
			(list (mu xs) xs)))
		      (let ((quantized-xs (map quantize xs)))
		       (map first
			    (remove-if
			     (lambda (quantized-x-count)
			      (< (second quantized-x-count) count-threshold))
			     (map (lambda (quantized-x)
				   (list quantized-x
					 (count quantized-x quantized-xs)))
				  (remove-duplicates quantized-xs))))))))))
  (remove-if (lambda (y)
	      (< (length (similar-xs y ys xs distance-threshold f distance))
		 count-threshold))
	     ys)))

(define (hsv-hue hsv) (vector-ref hsv 0))
(define (hsv-saturation hsv) (vector-ref hsv 1))
(define (hsv-value hsv) (vector-ref hsv 2))

(define (rgb-red rgb) (vector-ref rgb 0))
(define (rgb-green rgb) (vector-ref rgb 1))
(define (rgb-blue rgb) (vector-ref rgb 2))

(define (ppm-rgb-ref ppm y x)
 (vector (matrix-ref (ppm-red ppm) y x)
	 (matrix-ref (ppm-green ppm) y x)
	 (matrix-ref (ppm-blue ppm) y x)))

(define (ppm-rgb-set! ppm y x rgb)
 (matrix-set! (ppm-red ppm) y x (rgb-red rgb))
 (matrix-set! (ppm-green ppm) y x (rgb-green rgb))
 (matrix-set! (ppm-blue ppm) y x (rgb-blue rgb)))

(define (pnm->ppm pnm)
 (cond ((ppm? pnm) pnm)
       ((pgm? pnm) (pgm->ppm pnm))
       ((pbm? pnm) (pbm->ppm pnm))
       (else (fuck-up))))

(define (pnm->pgm pnm)
 (cond ((ppm? pnm) (ppm->pgm pnm))
       ((pgm? pnm) pnm)
       ((pbm? pnm) (pbm->pgm pnm))
       (else (fuck-up))))

(define (pnm->pgm-edges pnm sigma a1 a2 t1 t2 l w)
 (let* ((pgm (if (zero? sigma)
		 (pnm->pgm pnm)
		 (pgm-gaussian-smooth (pnm->pgm pnm) sigma)))
	(viff (create-image (pnm-height pgm) (pnm-width pgm))))
  (do ((y 0 (+ y 1))) ((>= y (pnm-height pgm)))
   (do ((x 0 (+ x 1))) ((>= x (pnm-width pgm)))
    (viff-set-pixel! viff y x (matrix-ref (pgm-grey pgm) y x))))
  (detect-edges! a1 a2 t1 t2 l w viff)
  (let ((grey (make-matrix (pnm-height pgm) (pnm-width pgm))))
   (do ((y 0 (+ y 1))) ((>= y (pnm-height pgm)))
    (do ((x 0 (+ x 1))) ((>= x (pnm-width pgm)))
     (matrix-set! grey y x (edge-pixel viff y x))))
   (free-xvimage viff)
   (make-pgm (pgm-raw? pgm) (pgm-maxval pgm) grey))))

(define (pnm->pgm-line-segments pnm sigma a1 a2 t1 t2 l w)
 (let* ((pgm (if (zero? sigma)
		 (pnm->pgm pnm)
		 (pgm-gaussian-smooth (pnm->pgm pnm) sigma)))
	(viff (create-image (pnm-height pgm) (pnm-width pgm))))
  (do ((y 0 (+ y 1))) ((>= y (pnm-height pgm)))
   (do ((x 0 (+ x 1))) ((>= x (pnm-width pgm)))
    (viff-set-pixel! viff y x (matrix-ref (pgm-grey pgm) y x))))
  (detect-edges! a1 a2 t1 t2 l w viff)
  (let ((grey (make-matrix (pnm-height pgm) (pnm-width pgm))))
   (do ((y 0 (+ y 1))) ((>= y (pnm-height pgm)))
    (do ((x 0 (+ x 1))) ((>= x (pnm-width pgm)))
     (matrix-set! grey y x (edge-pixel viff y x))))
   (free-xvimage viff)
   (make-pgm (pgm-raw? pgm) (pgm-maxval pgm) grey))))

(define (pnm->line-segments pnm sigma a1 a2 t1 t2 l w)
 (let* ((pgm (if (zero? sigma)
		 (pnm->pgm pnm)
		 (pgm-gaussian-smooth (pnm->pgm pnm) sigma)))
	(viff (create-image (pnm-height pgm) (pnm-width pgm))))
  (do ((y 0 (+ y 1))) ((>= y (pnm-height pgm)))
   (do ((x 0 (+ x 1))) ((>= x (pnm-width pgm)))
    (viff-set-pixel! viff y x (matrix-ref (pgm-grey pgm) y x))))
  ;; needs work: seuil, coef, and longmin are hardwired
  (detect-line-segments! a1 a2 t1 t2 l w viff)
  (let ((line-segments
	 (map-n (lambda (i)
		 (make-line-segment
		  (vector (line-segment-x1 i) (line-segment-y1 i))
		  (vector (line-segment-x2 i) (line-segment-y2 i))))
		(number-of-line-segments))))
   (free-xvimage viff)
   line-segments)))

(define (video-file-frame->pgm pathname video-type frame)
 (pnm->pgm (video-file-frame->ppm pathname video-type frame)))

(define (video-file-frame->pgm-edges
	 pathname video-type frame sigma a1 a2 t1 t2 l w)
 (pnm->pgm-edges
  (video-file-frame->ppm pathname video-type frame) sigma a1 a2 t1 t2 l w))

(define (video-file-frame->pgm-line-segments
	 pathname video-type frame sigma a1 a2 t1 t2 l w)
 (pnm->pgm-line-segments
  (video-file-frame->ppm pathname video-type frame) sigma a1 a2 t1 t2 l w))

(define (video-file-frame->line-segments
	 pathname video-type frame sigma a1 a2 t1 t2 l w)
 (pnm->line-segments
  (video-file-frame->ppm pathname video-type frame) sigma a1 a2 t1 t2 l w))

(define (video-file->pgm-movie pathname video-type)
 (call-with-video-input-file
  pathname
  video-type
  (lambda (video-input-port)
   (let loop ((pgm-movie '()))
    (let ((ppm (read-ppm-from-video-input-port video-input-port)))
     (if ppm
	 (loop (cons (pnm->pgm ppm) pgm-movie))
	 (list->vector (reverse pgm-movie))))))))

(define (video-file->pgm-edges-movie pathname video-type sigma a1 a2 t1 t2 l w)
 (call-with-video-input-file
  pathname
  video-type
  (lambda (video-input-port)
   (let loop ((pgm-edges-movie '()))
    (let ((ppm (read-ppm-from-video-input-port video-input-port)))
     (if ppm
	 (loop (cons (pnm->pgm-edges ppm sigma a1 a2 t1 t2 l w)
		     pgm-edges-movie))
	 (list->vector (reverse pgm-edges-movie))))))))

(define (video-file->pgm-line-segments-movie
	 pathname video-type sigma a1 a2 t1 t2 l w)
 (call-with-video-input-file
  pathname
  video-type
  (lambda (video-input-port)
   (let loop ((pgm-line-segments-movie '()))
    (let ((ppm (read-ppm-from-video-input-port video-input-port)))
     (if ppm
	 (loop (cons (pnm->pgm-line-segments ppm sigma a1 a2 t1 t2 l w)
		     pgm-line-segments-movie))
	 (list->vector (reverse pgm-line-segments-movie))))))))

(define (video-file->line-segments-movie
	 pathname video-type sigma a1 a2 t1 t2 l w)
 (call-with-video-input-file
  pathname
  video-type
  (lambda (video-input-port)
   (let loop ((line-segments-movie '()))
    (let ((ppm (read-ppm-from-video-input-port video-input-port)))
     (if ppm
	 (loop (cons (pnm->line-segments ppm sigma a1 a2 t1 t2 l w)
		     line-segments-movie))
	 (list->vector (reverse line-segments-movie))))))))

(define (pbm-pixels pbm ppm)
 ;; Selects the pixels from PPM that correspond to white (#T) pixels from PBM.
 (unless (pbm? pbm) (panic "Argument to PBM-PIXELS is not a PBM"))
 (let ((width (pnm-width pbm))
       (height (pnm-height pbm))
       (bitmap (pbm-bitmap pbm))
       (red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm))
       (pixels '()))
  (for-each-n
   (lambda (x)
    (for-each-n
     (lambda (y)
      (when (matrix-ref bitmap y x)
       (let* ((red (matrix-ref red y x))
	      (green (matrix-ref green y x))
	      (blue (matrix-ref blue y x))
	      (hsv (rgb->hsv (vector red green blue))))
	(set! pixels
	      (cons (make-pixel
		     x y
		     (hsv-hue hsv) (hsv-saturation hsv) (hsv-value hsv)
		     red green blue)
		    pixels)))))
     height))
   width)
  pixels))

(define (subsample-ppm ppm subsample left)
 (let ((pixels '())
       (width (pnm-width ppm))
       (height (pnm-height ppm))
       (red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (do ((x 0 (+ x subsample))) ((>= x width))
   (do ((y 0 (+ y subsample))) ((>= y height))
    (when (>= x left)
     (set! pixels (cons (vector (matrix-ref red y x)
				(matrix-ref green y x)
				(matrix-ref blue y x))
			pixels)))))
  pixels))

(define (subsample-ppm-that p ppm subsample left)
 (let ((pixels '())
       (width (pnm-width ppm))
       (height (pnm-height ppm))
       (red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (do ((x 0 (+ x subsample))) ((>= x width))
   (do ((y 0 (+ y subsample))) ((>= y height))
    (let ((pixel (vector (matrix-ref red y x)
			 (matrix-ref green y x)
			 (matrix-ref blue y x))))
     (when (and (>= x left) (p pixel)) (set! pixels (cons pixel pixels))))))
  pixels))

(define (inject-pnm! pnm ppm)
 (let* ((width (pnm-width pnm))
	(height (pnm-height pnm))
	(red2 (ppm-red ppm))
	(green2 (ppm-green ppm))
	(blue2 (ppm-blue ppm)))
  (cond
   ((ppm? pnm)
    (let ((red (ppm-red pnm))
	  (green (ppm-green pnm))
	  (blue (ppm-blue pnm)))
     (for-each-n (lambda (y)
		  (for-each-n (lambda (x)
			       (matrix-set! red2 y x (matrix-ref red y x))
			       (matrix-set! green2 y x (matrix-ref green y x))
			       (matrix-set! blue2 y x (matrix-ref blue y x)))
			      width))
		 height)))
   ((pgm? pnm)
    (let ((grey (pgm-grey pnm)))
     (for-each-n (lambda (y)
		  (for-each-n (lambda (x)
			       (let ((pixel (matrix-ref grey y x)))
				(matrix-set! red2 y x pixel)
				(matrix-set! green2 y x pixel)
				(matrix-set! blue2 y x pixel)))
			      width))
		 height)))
   ((pbm? pnm)
    (let ((bitmap (pbm-bitmap pnm)))
     (for-each-n (lambda (y)
		  (for-each-n (lambda (x)
			       (cond
				((matrix-ref bitmap y x)
				 (matrix-set! red2 y x *max-red*)
				 (matrix-set! green2 y x *max-green*)
				 (matrix-set! blue2 y x *max-blue*))
				(else
				 (matrix-set! red2 y x 0)
				 (matrix-set! green2 y x 0)
				 (matrix-set! blue2 y x 0))))
			      width))
		 height)))
   (else (panic "Argument to INJECT-PNM! is not a PNM")))))

(define (overwrite-pbm! pbm ppm rgb)
 (unless (pbm? pbm) (panic "Argument to OVERWRITE-PBM! is not a PBM"))
 (let ((width (pnm-width pbm))
       (height (pnm-height pbm))
       (red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm))
       (bitmap (pbm-bitmap pbm)))
  (for-each-n (lambda (y)
	       (for-each-n (lambda (x)
			    (when (matrix-ref bitmap y x)
			     (matrix-set! red y x (rgb-red rgb))
			     (matrix-set! green y x (rgb-green rgb))
			     (matrix-set! blue y x (rgb-blue rgb))))
			   width))
	      height)))

(define (overwrite-pixels! pixels ppm)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (for-each
   (lambda (pixel)
    (matrix-set! red (pixel-y pixel) (pixel-x pixel) (pixel-red pixel))
    (matrix-set! green (pixel-y pixel) (pixel-x pixel) (pixel-green pixel))
    (matrix-set! blue (pixel-y pixel) (pixel-x pixel) (pixel-blue pixel)))
   pixels)))

(define (overwrite-pixels-as! pixels ppm rgb)
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (for-each
   (lambda (pixel)
    (matrix-set! red (pixel-y pixel) (pixel-x pixel) (rgb-red rgb))
    (matrix-set! green (pixel-y pixel) (pixel-x pixel) (rgb-green rgb))
    (matrix-set! blue (pixel-y pixel) (pixel-x pixel) (rgb-blue rgb)))
   pixels)))

(define (ppm->coloured-pbm ppm saturation-threshold value-threshold)
 (let* ((width (pnm-width ppm))
	(height (pnm-height ppm))
	(bitmap (make-matrix height width)))
  (do ((y 0 (+ y 1))) ((>= y height))
   (do ((x 0 (+ x 1))) ((>= x width))
    (let ((hsv (rgb->hsv (ppm-rgb-ref ppm y x))))
     (matrix-set! bitmap y x
		  (and (> (hsv-saturation hsv) saturation-threshold)
		       (> (hsv-value hsv) value-threshold))))))
  (make-pbm (ppm-raw? ppm) bitmap)))

(define (c-threshold-normal-flow-magnitude
	 pgm1 pgm2 epsilon sigma threshold)
 ;; The moving pixels become white (#T) in the result.
 (unless (and (pgm? pgm1)
	      (pgm? pgm2)
	      (eq? (pgm-raw? pgm1) (pgm-raw? pgm2))
	      (= (pnm-width pgm1) (pnm-width pgm2))
	      (= (pnm-height pgm1) (pnm-height pgm2)))
  (panic
   "Arguments to C-THRESHOLD-NORMAL-FLOW-MAGNITUDE are not matching PGMs"))
 (let* ((width (pnm-width pgm1))
	(height (pnm-height pgm1))
	(e1 (pgm-grey pgm1))
	(e2 (pgm-grey pgm2))
	(ce1 (matrix->c-real-matrix e1))
	(ce2 (matrix->c-real-matrix e2))
	(ct (make-c-int-matrix height width))
	(t (make-matrix height width #f)))
  (c-threshold-normal-flow-magnitude-internal
   width height ce1 ce2 ct epsilon sigma threshold)
  (let ((pbm (make-pbm (pgm-raw? pgm1)
		       (c-int-matrix->boolean-matrix ct height width))))
   (free-c-real-matrix! ce1 height)
   (free-c-real-matrix! ce2 height)
   (free-c-int-matrix! ct height)
   pbm)))

(define (c-pbm-proximity-clusterer pbm proximity-threshold)
 ;; Clusters white (#T) pixels.
 (unless (pbm? pbm)
  (panic "Argument to C-PBM-PROXIMITY-CLUSTERER is not a PBM"))
 (let* ((width (pnm-width pbm))
	(height (pnm-height pbm))
	(bitmap (pbm-bitmap pbm))
	(m (make-c-int-matrix height width)))
  (do ((y 0 (+ y 1))) ((>= y height))
   (do ((x 0 (+ x 1))) ((>= x width))
    (c-int-matrix-set! m y x (if (matrix-ref bitmap y x) 0 1))))
  (let* ((i (c-pbm-proximity-clusterer-internal
	     m height width proximity-threshold 0 -1 -1))
	 (clusters (make-vector (- (- i) 1) '())))
   (do ((y 0 (+ y 1))) ((>= y height))
    (do ((x 0 (+ x 1))) ((>= x width))
     (let ((i (c-int-matrix-ref m y x)))
      (when (negative? i)
       (vector-set! clusters
		    (- (- i) 1)
		    (cons (vector x y) (vector-ref clusters (- (- i) 1))))))))
   (free-c-int-matrix! m height)
   (vector->list clusters))))

(define (convert-video-file
	 input-pathname
	 input-video-type
	 output-pathname
	 output-video-type
	 from
	 to)
 (call-with-video-input-file
  input-pathname
  input-video-type
  (lambda (video-input-port)
   (do ((i 0 (+ i 1))) ((>= i from))
    (read-ppm-from-video-input-port video-input-port))
   (call-with-video-output-file
    output-pathname
    output-video-type
    5000
    (lambda (video-output-port)
     (do ((i 0 (+ i 1))) ((>= i (+ (- to from) 1)))
      (write-pnm-to-video-output-port
       (read-ppm-from-video-input-port video-input-port)
       video-output-port)))))))

;;; Experimental stuff.

(define (processed-video-file->pixmaps pathname video-type processor)
 (call-with-video-input-file
  pathname
  video-type
  (lambda (video-input-port)
   (let loop ((last-ppm #f) (pixmaps '()))
    (let ((ppm (read-ppm-from-video-input-port video-input-port)))
     (if ppm
	 (if last-ppm
	     (loop ppm (cons (pnm->pixmap (processor last-ppm ppm)) pixmaps))
	     (loop ppm pixmaps))
	 (list->vector (reverse pixmaps))))))))

(define (pnm->pbm-edges pnm sigma a1 a2 t1 t2 l w)
 ;; Edge pixels are white (#T) in the result.
 (pgm->pbm (pnm->pgm-edges pnm sigma a1 a2 t1 t2 l w) 128))

(define (ppm-thresholded-absolute-difference ppm1 ppm2 threshold)
 ;; Pixels above the distance threshold are white (#T) in the result.
 (unless (and (ppm? ppm1)
	      (ppm? ppm2)
	      (= (ppm-maxval ppm1) (ppm-maxval ppm2))
	      (eq? (ppm-raw? ppm1) (ppm-raw? ppm2))
	      (= (pnm-width ppm1) (pnm-width ppm2))
	      (= (pnm-height ppm1) (pnm-height ppm2)))
  (panic
   "Arguments to PPM-THRESHOLDED-ABSOLUTE-DIFFERENCE are not matching PPMs"))
 (make-pbm (ppm-raw? ppm1)
	   (map-vector
	    (lambda (red1 green1 blue1 red2 green2 blue2)
	     (map-vector (lambda (red1 green1 blue1 red2 green2 blue2)
			  (> (sqrt (+ (sqr (- red1 red2))
				      (sqr (- green1 green2))
				      (sqr (- blue1 blue2))))
			     threshold))
			 red1 green1 blue1 red2 green2 blue2))
	    (ppm-red ppm1)
	    (ppm-green ppm1)
	    (ppm-blue ppm1)
	    (ppm-red ppm2)
	    (ppm-green ppm2)
	    (ppm-blue ppm2))))

;;; Point Images

(define (subsample-points points subsample-factor)
 (remove-if-not (lambda (point)
		 (and (zero? (remainder (x point) (x subsample-factor)))
		      (zero? (remainder (y point) (y subsample-factor)))))
		points))

(define (point< p q)
 (or (< (y p) (y q)) (and (= (y p) (y q)) (< (x p) (x q)))))

(define (normalize-points points) (sort points point< identity))

(define (pbm->points pbm)
 ;; Returns a list of the white (#T) points in the pbm image.
 (let ((points '())
       (height (pnm-height pbm))
       (width (pnm-width pbm))
       (bitmap (pbm-bitmap pbm)))
  (do ((y 0 (+ y 1))) ((>= y height))
   (do ((x 0 (+ x 1))) ((>= x width))
    (when (matrix-ref bitmap y x) (set! points (cons (vector x y) points)))))
  (normalize-points points)))

(define (points->pbm points)
 ;; Takes a list of what will be the white (#T) points in the pbm image.
 (let* ((height (+ (reduce max (map y points) 0) 1))
	(width (+ (reduce max (map x points) 0) 1))
	(bitmap (make-matrix height width #f)))
  (for-each (lambda (point) (matrix-set! bitmap (y point) (x point) #t))
	    points)
  (make-pbm #t bitmap)))

(define (points-proximity-clusterer points proximity-threshold)
 (setup-proximity-clusterer! proximity-threshold)
 ;; Clusters white (#T) pixels.
 (c-pbm-proximity-clusterer (points->pbm points) proximity-threshold))

;;; New Stuff

(define (pgm->image pgm)
 (map-vector (lambda (row) (map-vector exact->inexact row)) (pgm-grey pgm)))

(define (image->pgm image)
 (make-pgm
  #t
  *max-grey*
  (map-vector
   (lambda (row)
    (map-vector (lambda (element) (inexact->exact (floor element))) row))
   image)))

(define (image-filter-x! image filter)
 (let* ((ny (matrix-rows image))
	(nx (matrix-columns image))
	(nfilter (vector-length filter))
	(c (quotient nfilter 2))
	(row-copy (make-vector nx)))
  (do ((y 0 (+ y 1))) ((>= y ny))
   (do ((x 0 (+ x 1))) ((>= x nx))
    (vector-set! row-copy x (matrix-ref image y x)))
   (do ((x 0 (+ x 1))) ((>= x nx))
    (let ((s 0.0))
     (do ((fi 0 (+ fi 1))) ((>= fi nfilter))
      (let ((xi (+ (- x c) fi)))
       (when (and (>= xi 0) (< xi nx))
	(set! s (+ s (* (vector-ref row-copy xi) (vector-ref filter fi)))))))
     (matrix-set! image y x s))))))

(define (image-filter-y! image filter)
 (let* ((ny (matrix-rows image))
	(nx (matrix-columns image))
	(nfilter (vector-length filter))
	(c (quotient nfilter 2))
	(column-copy (make-vector ny)))
  (do ((x 0 (+ x 1))) ((>= x nx))
   (do ((y 0 (+ y 1))) ((>= y ny))
    (vector-set! column-copy y (matrix-ref image y x)))
   (do ((y 0 (+ y 1))) ((>= y ny))
    (let ((s 0.0))
     (do ((fi 0 (+ fi 1))) ((>= fi nfilter))
      (let ((yi (+ (- y c) fi)))
       (when (and (>= yi 0) (< yi ny))
	(set!
	 s (+ s (* (vector-ref column-copy yi) (vector-ref filter fi)))))))
     (matrix-set! image y x s))))))

(define (image-diff-x! image)
 (image-filter-x!
  image (vector (/ 12.0) (- (/ 2.0 3.0)) 0.0 (/ 2.0 3.0) (- (/ 12.0)))))

(define (image-diff-y! image)
 (image-filter-y!
  image (vector (/ 12.0) (- (/ 2.0 3.0)) 0.0 (/ 2.0 3.0) (- (/ 12.0)))))

(define (smooth-image! image sigma)
 (let* ((width 4.0)			; number of standard deviations to use
	(n (inexact->exact (ceiling (* width sigma))))
	(nfilter (+ (* 2 n) 1))
	(filter (make-vector nfilter))
	(k (/ 0.5 (sqr sigma)))
	(s 0.0))
  (do ((fi 0 (+ fi 1))) ((>= fi nfilter))
   (vector-set! filter fi (exp (- (* k (sqr (- fi n))))))
   (set! s (+ s (vector-ref filter fi))))
  ;; normalize filter so it sums to one
  (do ((fi 0 (+ fi 1))) ((>= fi nfilter))
   (vector-set! filter fi (/ (vector-ref filter fi) s)))
  (image-filter-x! image filter)
  (image-filter-y! image filter)))

(define (pgm-gaussian-smooth pgm sigma)
 (let ((image (pgm->image pgm)))
  (smooth-image! image sigma)
  (image->pgm image)))

;;; Interruptible usleep

;;; Constants

(define wall-fixed #x100000000)
(define wall-fix 0.0)
(define wall-bad #t)

;;; Procedures

(define (read-wall-clock)
 (let ((time (wall-clock)))
  (if (< time 0)
      (when wall-bad
       (set! wall-fix (+ wall-fixed wall-fix))
       (set! wall-bad #f))
      (set! wall-bad #t))
  (+ time wall-fix)))

(define (interruptible-usleep n)
 ;; needs work: This is hardwired to 100. It should be CLK_TCK. But we only
 ;;             expose CLOCKS_PER_SEC. So it goes.
 (let ((end (+ (read-wall-clock) (* (/ n 1000000.0) 100))))
  (let loop ()
   (usleep 10000)
   (unless (> (read-wall-clock) end) (loop)))))

;;; Meteor

;;; TTMTTD
;;;  1. way to startup just frame buffer, just frame buffer and viewfinder
;;;  2. (METEOR-SET-FRAME-FROM-PNM! i pnm)
;;;  3. (METEOR-SET-FRAMES-FROM-PNM-MOVIE! i j pnm-movie)
;;;  4. (METEOR-SET-FRAMES-FROM-VIDEO-FILE! i j pathname video-type)
;;;  5. make interchangeable between X and XShm
;;;  6. 24->32 bit so can use memcpy
;;;  7. double buffer, check for overrun
;;;  8. uncompressed single-file MBM, MGM, MPM file format
;;;       VIDEO-TYPE-ENUM
;;;       GET-VIDEO-WIDTH
;;;       GET-VIDEO-HEIGHT
;;;       READ-PPM-FROM-VIDEO-INPUT-PORT
;;;       READ-PIXMAP-FROM-VIDEO-INPUT-PORT
;;;       CLOSE-VIDEO-INPUT-PORT
;;;       OPEN-VIDEO-OUTPUT-FILE
;;;       WRITE-PNM-TO-VIDEO-OUTPUT-PORT
;;;       CLOSE-VIDEO-OUTPUT-PORT

;;; Constants

(define meteor-fmt-ntsc 256)
(define meteor-fmt-pal 512)
(define meteor-fmt-secam 1024)
(define meteor-fmt-automode 2048)

(define meteor-input-dev0 4096)
(define meteor-input-dev-rca 4096)
(define meteor-input-dev1 8192)
(define meteor-input-dev2 16384)
(define meteor-input-dev3 32768)
(define meteor-input-dev-rgb 40960)
(define meteor-input-dev-svideo	24576)

(define meteor-geo-rgb16 65536)
(define meteor-geo-rgb24 131072)
(define meteor-geo-yuv-packed 262144)
(define meteor-geo-yuv-planar 524288)
(define meteor-geo-yuv-planer 524288)
(define meteor-geo-unsigned 4194304)
(define meteor-geo-even-only 16777216)
(define meteor-geo-odd-only 33554432)
(define meteor-geo-field-mask 50331648)
(define meteor-geo-yuv-422 67108864)
(define meteor-geo-output-mask 68091904)
(define meteor-geo-yuv-12 268435456)
(define meteor-geo-yuv-9 1073741824)

;;; Variables

(define *meteor-height* #f)
(define *meteor-width* #f)

;;; Procedures

(define (get-motion? i) (not (zero? (get-motion-internal i))))

(define (meteor-startup height width nframes uninterlaced?)
 (set! *meteor-height* height)
 (set! *meteor-width* width)
 (meteor-startup-internal height width nframes (if uninterlaced? 1 0))
 #f)

(define (meteor-startup-offline height width nframes uninterlaced?)
 (set! *meteor-height* height)
 (set! *meteor-width* width)
 (meteor-startup-offline-internal height width nframes (if uninterlaced? 1 0))
 #f)

(define (meteor-startup-with-viewfinder height width nframes uninterlaced?)
 (unless (and
	  (= (visual-class (xdefaultvisual *display* *screen*)) truecolor)
	  (= (visual-red_mask (xdefaultvisual *display* *screen*)) 16711680)
	  (= (visual-green_mask (xdefaultvisual *display* *screen*)) 65280)
	  (= (visual-blue_mask (xdefaultvisual *display* *screen*)) 255)
	  (= (xdefaultdepth *display* *screen*) 24))
  (panic "Meteor can only handle 24-bit TRUECOLOR visuals"))
 (set! *meteor-height* height)
 (set! *meteor-width* width)
 (meteor-startup-with-viewfinder-internal
  height width nframes (if uninterlaced? 1 0)
  (cdr *display*) *screen* *display-pane*
  *display-pane-width* *display-pane-height*)
 #f)

(define (meteor-restartup height width nframes uninterlaced?)
 (set! *meteor-height* height)
 (set! *meteor-width* width)
 (meteor-restartup-internal height width nframes (if uninterlaced? 1 0))
 #f)

(define (meteor-turn-on-viewfinder! m)
 (meteor-turn-on-viewfinder-internal! m)
 #f)

(define (meteor-turn-on-capturing! i l n)
 (meteor-turn-on-capturing-internal! i l n)
 #f)

(define (meteor-turn-on-ellipse-tracker! i l n)
 (meteor-turn-on-ellipse-tracker-internal! i l n)
 #f)

(define (meteor-turn-on-ellipse-tracker-with-capturing! i l n)
 (meteor-turn-on-ellipse-tracker-with-capturing-internal! i l n)
 #f)

(define (meteor-turn-on-polygon-tracker! i l n h?)
 (meteor-turn-on-polygon-tracker-internal! i l n (if h? 1 0))
 #f)

(define (meteor-turn-on-polygon-tracker-with-capturing! i l n h?)
 (meteor-turn-on-polygon-tracker-with-capturing-internal! i l n (if h? 1 0))
 #f)

(define (meteor-turn-off!)
 (meteor-turn-off-internal!)
 #f)

(define (meteor-set-deinterlaced! deinterlaced?)
 (meteor-set-deinterlaced-internal! (if deinterlaced? 1 0))
 #f)

(define (meteor-frame->ppm i)
 (let ((red (make-matrix *meteor-height* *meteor-width*))
       (green (make-matrix *meteor-height* *meteor-width*))
       (blue (make-matrix *meteor-height* *meteor-width*)))
  (do ((y 0 (+ y 1))) ((>= y *meteor-height*))
   (do ((x 0 (+ x 1))) ((>= x *meteor-width*))
    (matrix-set! red y x (meteor-get-red i y x))
    (matrix-set! green y x (meteor-get-green i y x))
    (matrix-set! blue y x (meteor-get-blue i y x))))
  (make-ppm #t 255 red green blue)))

(define (meteor-code-frame->ppm i)
 (let ((red (make-matrix *meteor-height* *meteor-width*))
       (green (make-matrix *meteor-height* *meteor-width*))
       (blue (make-matrix *meteor-height* *meteor-width*)))
  (do ((y 0 (+ y 1))) ((>= y *meteor-height*))
   (do ((x 0 (+ x 1))) ((>= x *meteor-width*))
    (matrix-set! red y x (meteor-get-code-frame-red i y x))
    (matrix-set! green y x (meteor-get-code-frame-green i y x))
    (matrix-set! blue y x (meteor-get-code-frame-blue i y x))))
  (make-ppm #t 255 red green blue)))

(define (meteor-frames->ppm-movie m n)
 (map-n-vector (lambda (i) (meteor-frame->ppm (+ m i))) (- n m)))

(define (meteor-code-frames->ppm-movie m n)
 (map-n-vector (lambda (i) (meteor-code-frame->ppm (+ m i))) (- n m)))

(define (ximage->ppm ximage height width)
 ;; This is hardwired to raw and 8x8x8 RGB.
 (make-ppm
  #t
  255
  (map-n-vector
   (lambda (y)
    (map-n-vector (lambda (x) (quotient (xgetpixel ximage x y) 65536)) width))
   height)
  (map-n-vector
   (lambda (y)
    (map-n-vector
     (lambda (x) (quotient (remainder (xgetpixel ximage x y) 65536) 256))
     width))
   height)
  (map-n-vector
   (lambda (y)
    (map-n-vector (lambda (x) (remainder (xgetpixel ximage x y) 256)) width))
   height)))

(define (meteor-hack-track! m
			    n
			    o
			    colour-pixel-mass-threshold
			    colour-proximity-threshold
			    motion-threshold
			    motion-pixel-mass-threshold
			    motion-proximity-threshold)
 (meteor-start-hack-track! m)
 (for-each-n (lambda (i)
	      (meteor-hack-track-step! (+ m i)
				       (+ o i)
				       colour-pixel-mass-threshold
				       colour-proximity-threshold
				       motion-threshold
				       motion-pixel-mass-threshold
				       motion-proximity-threshold)
	      #f)
	     (- n m))
 (meteor-finish-hack-track!)
 #f)

(define (meteor-objects->pnm-movie m
				   n
				   o
				   colour-pixel-mass-threshold
				   colour-proximity-threshold
				   motion-threshold
				   motion-pixel-mass-threshold
				   motion-proximity-threshold)
 (meteor-hack-track! m
		     n
		     o
		     colour-pixel-mass-threshold
		     colour-proximity-threshold
		     motion-threshold
		     motion-pixel-mass-threshold
		     motion-proximity-threshold)
 (meteor-frames->ppm-movie o (+ o (- n m))))

(define (meteor-frame->pixmap i)
 (let* ((default-visual (xdefaultvisual *display* *screen*))
	(default-depth (xdefaultdepth *display* *screen*))
	(data (c-malloc (* 4 *meteor-width* *meteor-height*)))
	(ximage (cons 'ximagep
		      (c-xcreateimage
		       (cdr *display*) (cdr default-visual) default-depth
		       zpixmap 0 data *meteor-width* *meteor-height* 32 0)))
	(pixmap (xcreatepixmap
		 *display* *display-pane* *meteor-width* *meteor-height*
		 default-depth)))
  (meteor-copy-frame-to-ximage! i (cdr ximage))
  (xputimage
   *display* pixmap *color-gc* ximage 0 0 0 0 *meteor-width* *meteor-height*)
  (xdestroyimage ximage)
  pixmap))

(define (meteor-frames->pixmaps m n)
 (let* ((default-visual (xdefaultvisual *display* *screen*))
	(default-depth (xdefaultdepth *display* *screen*))
	(data (c-malloc (* 4 *meteor-width* *meteor-height*)))
	(ximage (cons 'ximagep
		      (c-xcreateimage
		       (cdr *display*) (cdr default-visual) default-depth
		       zpixmap 0 data *meteor-width* *meteor-height* 32 0)))
	(pixmaps (map-n-vector
		  (lambda (i)
		   (xcreatepixmap
		    *display* *display-pane* *meteor-width* *meteor-height*
		    default-depth))
		  (- n m))))
  (do ((i 0 (+ i 1))) ((>= i (- n m)))
   (meteor-copy-frame-to-ximage! (+ m i) (cdr ximage))
   (xputimage *display* (vector-ref pixmaps i) *color-gc* ximage 0 0 0 0
	      *meteor-width* *meteor-height*))
  (xdestroyimage ximage)
  pixmaps))

(define (meteor-motion->pixmap i)
 (let* ((default-visual (xdefaultvisual *display* *screen*))
	(default-depth (xdefaultdepth *display* *screen*))
	(data (c-malloc (* 4 *meteor-width* *meteor-height*)))
	(ximage (cons 'ximagep
		      (c-xcreateimage
		       (cdr *display*) (cdr default-visual) default-depth
		       zpixmap 0 data *meteor-width* *meteor-height* 32 0)))
	(pixmap (xcreatepixmap
		 *display* *display-pane* *meteor-width* *meteor-height*
		 default-depth)))
  (meteor-motion-detector! i (cdr ximage))
  (xputimage
   *display* pixmap *color-gc* ximage 0 0 0 0 *meteor-width* *meteor-height*)
  (xdestroyimage ximage)
  pixmap))

(define (meteor-motion->pixmaps m n)
 (let* ((default-visual (xdefaultvisual *display* *screen*))
	(default-depth (xdefaultdepth *display* *screen*))
	(data (c-malloc (* 4 *meteor-width* *meteor-height*)))
	(ximage (cons 'ximagep
		      (c-xcreateimage
		       (cdr *display*) (cdr default-visual) default-depth
		       zpixmap 0 data *meteor-width* *meteor-height* 32 0)))
	(pixmaps (map-n-vector
		  (lambda (i)
		   (xcreatepixmap
		    *display* *display-pane* *meteor-width* *meteor-height*
		    default-depth))
		  (- n m))))
  (do ((i 0 (+ i 1))) ((>= i (- n m)))
   ;; note: This is a kludge to deal with the fact that pairwise sequences
   ;;       are one shorter than the original.
   (meteor-motion-detector! (+ m (min i (- n 2))) (cdr ximage))
   (xputimage *display* (vector-ref pixmaps i) *color-gc* ximage 0 0 0 0
	      *meteor-width* *meteor-height*))
  (xdestroyimage ximage)
  pixmaps))

(define (meteor-objects->pixmaps m
				 n
				 o
				 colour-pixel-mass-threshold
				 colour-proximity-threshold
				 motion-threshold
				 motion-pixel-mass-threshold
				 motion-proximity-threshold)
 (meteor-hack-track! m
		     n
		     o
		     colour-pixel-mass-threshold
		     colour-proximity-threshold
		     motion-threshold
		     motion-pixel-mass-threshold
		     motion-proximity-threshold)
 (meteor-frames->pixmaps o (+ o (- n m))))

(define (meteor-show-frame i) (meteor-show-frame-internal i) #f)

(define (meteor-show-code-frame i) (meteor-show-code-frame-internal i) #f)

(define (meteor-show-frames m n)
 (do ((i m (+ i 1))) ((>= i n))
  (meteor-show-frame i)
  (usleep (inexact->exact (round (/ 1000000.0 *frame-rate*))))))

(define (meteor-set-hack-track! motion-threshold
				colour-pixel-mass-threshold
				colour-proximity-threshold
				motion-pixel-mass-threshold
				motion-proximity-threshold)
 (meteor-set-hack-track-internal! motion-threshold
				  colour-pixel-mass-threshold
				  colour-proximity-threshold
				  motion-pixel-mass-threshold
				  motion-proximity-threshold)
 #f)

(define (meteor-set-red-hack-track! hl hh sl sh vl vh)
 (meteor-set-red-hack-track-internal! hl hh sl sh vl vh)
 #f)

(define (meteor-set-green-hack-track! hl hh sl sh vl vh)
 (meteor-set-green-hack-track-internal! hl hh sl sh vl vh)
 #f)

(define (meteor-set-blue-hack-track! hl hh sl sh vl vh)
 (meteor-set-blue-hack-track-internal! hl hh sl sh vl vh)
 #f)

(define (meteor-set-yellow-hack-track! hl hh sl sh vl vh)
 (meteor-set-yellow-hack-track-internal! hl hh sl sh vl vh)
 #f)

(define (meteor-frames->video-file m n pathname video-type)
 (call-with-video-output-file
  pathname video-type (- n m)
  (lambda (video-output-port)
   (do ((i m (+ i 1))) ((>= i n))
    (write-pnm-to-video-output-port
     (meteor-frame->ppm i) video-output-port)))))

(define (meteor->pnm-movie before after height width nframes uninterlaced?)
 (let ((downsample-factor (/ 30.0 *frame-rate*)))
  (unless (< (abs (- downsample-factor (round downsample-factor))) 0.001)
   (panic "Frame rate does not yield an integral downsample factor"))
  (meteor-startup height width nframes (if uninterlaced? 1 0))
  (before)
  (meteor-capture-frames 0 (inexact->exact (round downsample-factor)) nframes)
  (after)
  (let ((pnm-movie (meteor-frames->ppm-movie 0 nframes)))
   (meteor-shutdown)
   pnm-movie)))

(define (meteor->video-file
	 before after height width nframes uninterlaced? pathname video-type)
 (let ((downsample-factor (/ 30.0 *frame-rate*)))
  (unless (< (abs (- downsample-factor (round downsample-factor))) 0.001)
   (panic "Frame rate does not yield an integral downsample factor"))
  (meteor-startup height width nframes (if uninterlaced? 1 0))
  (before)
  (meteor-capture-frames 0 (inexact->exact (round downsample-factor)) nframes)
  (after)
  (meteor-frames->video-file 0 nframes pathname video-type)
  (meteor-shutdown)
  #f))

(define (meteor-ppm->frame! i ppm)
 (unless (and (= *meteor-height* (pnm-height ppm))
	      (= *meteor-width* (pnm-width ppm)))
  (panic "Wrong height or width"))
 (let ((red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (do ((y 0 (+ y 1))) ((>= y *meteor-height*))
   (do ((x 0 (+ x 1))) ((>= x *meteor-width*))
    (meteor-set-red! i y x (matrix-ref red y x))
    (meteor-set-green! i y x (matrix-ref green y x))
    (meteor-set-blue! i y x (matrix-ref blue y x))))))

(define (get-ellipse-movie start end)
 (map-n-vector (lambda (frame)
		(map-n (lambda (k)
			(make-ellipse (get-ellipse-x (+ start frame) k)
				      (get-ellipse-y (+ start frame) k)
				      (get-ellipse-theta (+ start frame) k)
				      (get-ellipse-r (+ start frame) k)
				      (get-ellipse-s (+ start frame) k)))
		       (get-object-k (+ start frame))))
	       (- end start)))

(define (get-polygon-movie start finish)
 (map-n-vector
  (lambda (frame)
   (map-n-vector (lambda (k)
		  (map-n-vector (lambda (l)
				 (vector (get-polygon-x (+ start frame) k l)
					 (get-polygon-y (+ start frame) k l)))
				(get-polygon-l (+ start frame) k)))
		 (get-object-k (+ start frame))))
  (- finish start)))

(define (get-fudged-ellipse-movie end)
 (map-n-vector (lambda (frame)
		;; This fudge is because the moving objects don't appear until
		;; the second frame.
		(let ((frame (if (zero? frame) 1 frame)))
		 (map-n (lambda (k)
			 (make-ellipse (get-ellipse-x frame k)
				       (get-ellipse-y frame k)
				       (get-ellipse-theta frame k)
				       (get-ellipse-r frame k)
				       (get-ellipse-s frame k)))
			(get-object-k frame))))
	       end))

(define (get-fudged-polygon-movie finish)
 (map-n-vector
  (lambda (frame)
   ;; This fudge is because the moving objects don't appear until the second
   ;; frame.
   (let ((frame (if (zero? frame) 1 frame)))
    (map-n-vector (lambda (k)
		   (map-n-vector (lambda (l)
				  (vector (get-polygon-x frame k l)
					  (get-polygon-y frame k l)))
				 (get-polygon-l frame k)))
		  (get-object-k frame))))
  finish))

(define (merge-collinear-edges polygon epsilon)
 (if (< (vector-length polygon) 3)
     polygon
     (let loop ((i 0))
      (if (>= i (vector-length polygon))
	  polygon
	  (let ((p1 (vector-ref polygon i))
		(p2 (vector-ref polygon
				(modulo (+ i 1) (vector-length polygon))))
		(p3 (vector-ref polygon
				(modulo (+ i 2) (vector-length polygon)))))
	   (if (point-touches-line? p2 (make-line-segment p1 p3) epsilon)
	       (merge-collinear-edges
		(list->vector
		 (list-remove (vector->list polygon)
			      (modulo (+ i 1) (vector-length polygon))))
		epsilon)
	       (loop (+ i 1))))))))

(define (merge-close-vertices polygon epsilon)
 (let loop ((polygon polygon) (i 0))
  (cond
   ((or (< (vector-length polygon) 3) (>= i (vector-length polygon)))
    polygon)
   ((< (distance
	(vector-ref polygon i)
	(vector-ref polygon (modulo (+ i 1) (vector-length polygon))))
       epsilon)
    (loop (list->vector (list-remove (vector->list polygon) i)) 0))
   (else (loop polygon (+ i 1))))))

(define (convex-hull-of-polygon polygon)
 (merge-collinear-edges
  (merge-close-vertices (list->vector (convex-hull (vector->list polygon)))
			3.0)		;tolerance
  ;; tolerance
  0.0))

(define (convex-hull-of-polygons polygons)
 (map-vector convex-hull-of-polygon polygons))

(define (convex-hull-of-polygon-movie polygon-movie)
 (map-vector convex-hull-of-polygons polygon-movie))

(define (meteor-track-video-file pathname
				 video-type
				 uninterlaced?
				 capturing-code?
				 save-code?
				 tracking-ellipses?
				 tracking-polygons?
				 hull?
				 shutdown?
				 motion-threshold
				 colour-pixel-mass-threshold
				 colour-proximity-threshold
				 motion-pixel-mass-threshold
				 motion-proximity-threshold)
 (let* ((nframes (video-file-length pathname video-type))
	(ppm (video-file-frame->ppm pathname video-type 0))
	(height (pnm-height ppm))
	(width (pnm-width ppm)))
  (meteor-startup-offline height width nframes uninterlaced?)
  (call-with-video-input-file
   pathname video-type
   (lambda (video-input-port)
    (do ((i 0 (+ i 1))) ((>= i nframes))
     (meteor-ppm->frame!
      i (read-ppm-from-video-input-port video-input-port)))))
  (do ((i 0 (+ i 1))) ((>= i nframes))
   (meteor-offline-tracker i
			   height
			   width
			   (if uninterlaced? 1 0)
			   (if (or capturing-code? save-code?) 1 0)
			   (if tracking-ellipses? 1 0)
			   (if tracking-polygons? 1 0)
			   (if hull? 1 0)
			   motion-threshold
			   colour-pixel-mass-threshold
			   colour-proximity-threshold
			   motion-pixel-mass-threshold
			   motion-proximity-threshold))
  (when save-code?
   (call-with-video-output-file
    (if (has-extension? pathname)
	(default-extension
	 (string-append (strip-extension pathname) "-code")
	 (extension pathname))
	(string-append pathname "-code"))
    video-type
    nframes
    (lambda (video-output-port)
     (do ((i 0 (+ i 1))) ((>= i nframes))
      (write-pnm-to-video-output-port
       (meteor-code-frame->ppm i) video-output-port)))))
  (let ((results
	 (list (if tracking-ellipses? (get-fudged-ellipse-movie nframes) #f)
	       (if tracking-polygons? (get-fudged-polygon-movie nframes) #f))))
   (when shutdown? (meteor-shutdown-offline))
   results)))

;;; Tam V'Nishlam Shevah L'El Borei Olam
