(MODULE HLIB-POSTSCRIPT)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-postscript.sch")
(include "c-externals.sc")

;;; Parameters

(define *postscript-leader*
 "%!PS-Adobe-2.0 EPSF-1.2
%%DocumentFonts:
%%Pages: 1
%%BoundingBox: 0 0 320 240
%%EndComments
/flipmatrix where 	% only do it if it hasn't already been done
	{pop}
	{/flipmatrix [1 0 0 1 0 0] def
	 flipmatrix concat
	} ifelse

/localdict 30 dict def

/args { % value0 value1 ... [name0 name1 ...]
	dup length 1 sub	 % array len - 1
	localdict begin
	0 1 2 index
	{ % stack: ... names len-1 i
	  2 index 1 index get	 % get name_i
	  % stack: value0 value1 ... value_len-1 names len-1 i name_i
          % index of v_i will be 4 + (len-1 - i)
	  2 index 2 index sub 4 add index   % get v_i
	  def  % associate name and value in localdict
	  pop  % pop off index supplied by for
	} for
	{pop} repeat pop pop
} def

/arrowHeight 8 def
/arrowWidth 4 def
/brushWidth 1 def
/patternNone false def
/brushNone false def

/arrowheadpath {
newpath
0 0 moveto
arrowHeight neg arrowWidth 2 div lineto
arrowHeight neg arrowWidth 2 div neg lineto
closepath
} def

/arrowhead {
0 begin
transform originalCTM itransform
/taily exch def
/tailx exch def
transform originalCTM itransform
/tipy exch def
/tipx exch def
/dy tipy taily sub def
/dx tipx tailx sub def
/angle dx 0 ne dy 0 ne or { dy dx atan } { 90 } ifelse def
gsave
originalCTM setmatrix
tipx tipy translate
angle rotate
newpath
0 0 moveto
arrowHeight neg arrowWidth 2 div lineto
arrowHeight neg arrowWidth 2 div neg lineto
closepath
patternNone not {
originalCTM setmatrix
/padtip arrowHeight 2 exp 0.25 arrowWidth 2 exp mul add sqrt brushWidth mul
arrowWidth div def
/padtail brushWidth 2 div def
tipx tipy translate
angle rotate
padtip 0 translate
arrowHeight padtip add padtail add arrowHeight div dup scale
arrowheadpath
fill
} if
brushNone not {
originalCTM setmatrix
tipx tipy translate
angle rotate
arrowheadpath
[] 0 setdash
0 setgray
originalCTM setmatrix
stroke
} if
grestore
end
} dup 0 9 dict put def

/Line { [ /thick /arrowstart /arrowend /gray /dasharray /dashoffset
/xp /yp /xq /yq ] args
newpath
thick setlinewidth
gray setgray
dasharray dashoffset setdash
xp yp moveto
xq yq lineto
stroke
arrowstart 1 eq { xp yp xq yq arrowhead } if
arrowend 1 eq { xq yq xp yp arrowhead } if
end
} def

/Arc { [ /x /y /xrad /yrad /startangle /endangle /thick /gray ] args
newpath
thick setlinewidth
gray setgray
originalCTM setmatrix
x y translate
xrad yrad scale
0 0 1 startangle endangle arc
originalCTM setmatrix
stroke
end
} def

/FilledArc { [ /x /y /xrad /yrad /startangle /endangle /thick /gray ] args
newpath
thick setlinewidth
gray setgray
originalCTM setmatrix
x y translate
xrad yrad scale
0 0 1 startangle endangle arc
originalCTM setmatrix
eofill
stroke
end
} def

/CenteredString { [ /x /y /string ] args
x y moveto
gsave
1 1 scale
string dup stringwidth pop 2 div neg 0 rmoveto show
grestore
} def

/String { [ /x /y /string ] args
x y moveto
gsave
1 1 scale
string show
grestore
} def

[ 1 0 0 1 0 0 ] concat
/originalCTM matrix currentmatrix def
/Times-Roman findfont 10.0 scalefont setfont
")

(define *postscript-trailer*
 "
showpage

%%Trailer
")

;;; Procedures

(define (call-with-postscript-file pathname procedure)
 (call-with-output-file (default-extension pathname "ps")
  (lambda (ps)
   (display *postscript-leader* ps)
   (procedure ps)
   (display *postscript-trailer* ps))))

(define (postscript-draw-line-segment ps l thickness)
 (format ps "~a 0 0 0 [] 0 ~a ~a ~a ~a Line~%"
	 thickness (x (p l)) (- 240 (y (p l))) (x (q l)) (- 240 (y (q l)))))

(define (postscript-draw-arc
	 ps p x-radius y-radius start-angle end-angle thickness)
 (format ps "~a ~a ~a ~a ~a ~a ~a 0 Arc~%"
	 (x p) (- 240 (y p))
	 x-radius y-radius start-angle end-angle thickness))

(define (postscript-draw-circle ps p diameter)
 (postscript-draw-arc
  ps p (quotient diameter 2) (quotient diameter 2) 0 360 1))

(define (postscript-draw-filled-arc
	 ps p x-radius y-radius start-angle end-angle thickness)
 (format ps "~a ~a ~a ~a ~a ~a ~a 0 FilledArc~%"
	 (x p) (- 240 (y p))
	 x-radius y-radius start-angle end-angle thickness))

(define (postscript-draw-filled-circle ps p diameter)
 (postscript-draw-filled-arc
  ps p (quotient diameter 2) (quotient diameter 2) 0 360 1))

(define (postscript-draw-string ps p string)
 (format ps "~a ~a (~a) String~%" (x p) (- 240 (y p)) string))

(define (postscript-draw-centered-string ps p string)
 (format ps "~a ~a (~a) CenteredString~%" (x p) (- 240 (y p)) string))

(define (postscript-draw-ellipse ps ellipse thickness)
 (let* ((previous-x #f)
	(previous-y #f)
	(x0 (ellipse-x0 ellipse))
	(y0 (ellipse-y0 ellipse))
	(t0 (ellipse-t0 ellipse))
	(a (ellipse-a ellipse))
	(b (ellipse-b ellipse))
	(rxx (cos t0))
	(rxy (- (sin t0)))
	(ryx (- rxy))
	(ryy rxx))
  (for-each-n
   (lambda (i)
    (let* ((ellipse-x (* a (sin (degrees->radians (* 10 i)))))
	   (ellipse-y (* b (cos (degrees->radians (* 10 i)))))
	   (this-x (+ (* rxx ellipse-x) (* rxy ellipse-y) x0))
	   (this-y (+ (* ryx ellipse-x) (* ryy ellipse-y) y0)))
     (when previous-x
      (postscript-draw-line-segment
       ps
       (make-line-segment (vector this-x this-y)
			  (vector previous-x previous-y))
       thickness))
     (set! previous-x this-x)
     (set! previous-y this-y)))
   37)))

(define (postscript-draw-frame ps)
 (postscript-draw-line-segment ps (make-line-segment '#(0 0) '#(320 0)) 1)
 (postscript-draw-line-segment ps (make-line-segment '#(320 0) '#(320 240)) 1)
 (postscript-draw-line-segment ps (make-line-segment '#(320 240) '#(0 240)) 1)
 (postscript-draw-line-segment ps (make-line-segment '#(0 240) '#(0 0)) 1))

(define (postscript-draw-polygon ps p)
 (for-each-n
  (lambda (i)
   (postscript-draw-line-segment
    ps
    (make-line-segment (vector-ref p i)
		       (vector-ref p (modulo (+ i 1) (vector-length p))))
    1))
  (vector-length p)))

(define (postscript-draw-bar ps r l)
 (let* ((dx (- (x (p l)) (x (q l))))
	(dy (- (y (p l)) (y (q l))))
	(m (sqrt (+ (sqr dx) (sqr dy))))
	(dx (* 10.0 (/ dx m)))
	(dy (* 10.0 (/ dy m))))
  (postscript-draw-line-segment
   ps
   (make-line-segment (vector (- (x r) dx) (- (y r) dy))
		      (vector (+ (x r) dx) (+ (y r) dy)))
   3)))

(define (polygons->postscript polygons pathname)
 (call-with-postscript-file pathname
  (lambda (ps)
   (postscript-draw-frame ps)
   (for-each (lambda (polygon) (postscript-draw-polygon ps polygon))
	     polygons))))

(define (ellipse-movie->postscript pathname frame)
 (let ((ellipses (vector-ref (read-object-from-file
			      (replace-extension pathname "ellipse-movie"))
			     frame)))
  (call-with-postscript-file
    (format #f "~a-~a-track" (strip-extension pathname) frame)
   (lambda (ps)
    (postscript-draw-frame ps)
    (for-each (lambda (ellipse i) (postscript-draw-ellipse ps ellipse (+ i 1)))
	      ellipses
	      (enumerate (length ellipses)))))))

(define (spm-hull->postscript pathname frame)
 (let ((hulls (map convex-hull
		   (vector->list
		    (vector-ref (read-object-from-file
				 (replace-extension pathname "spm"))
				frame)))))
  (call-with-postscript-file
    (format #f "~a-~a-hull" (strip-extension pathname) frame)
   (lambda (ps)
    (postscript-draw-frame ps)
    (for-each
     (lambda (hull)
      (for-each
       (lambda (p q)
	(postscript-draw-line-segment ps (make-line-segment p q) 1))
       hull
       (cons (last hull) (but-last hull))))
     hulls)))))

(define (ground-location p)
 (let loop ((i 0))
  (if (>= i (- (vector-length p) 1))
      (vector (/ (+ (x (vector-ref p 0)) (x (vector-ref p 1))) 2)
	      (/ (+ (y (vector-ref p 0)) (y (vector-ref p 1))) 2))
      (let ((yg (/ (+ (y (vector-ref p i)) (y (vector-ref p (+ i 1)))) 2))
	    (xg (/ (+ (x (vector-ref p i)) (x (vector-ref p (+ i 1)))) 2)))
       (if (or (> (atan
		   (abs (- (y (vector-ref p i)) (y (vector-ref p (+ i 1)))))
		   (abs (- (x (vector-ref p i)) (x (vector-ref p (+ i 1))))))
		  (* 0.5 half-pi))
	       (< yg 10)
	       (> yg (- *height* 10))
	       (< xg 10)
	       (> xg (- *width* 10)))
	   (loop (+ i 1))
	   (vector xg yg))))))

(define (layer-id-location p)
 (let loop ((i 0))
  (if (>= i (vector-length p))
      (vector-ref p 0)
      (let ((yi (y (vector-ref p i)))
	    (xi (x (vector-ref p i))))
       (if (or (< yi 10)
	       (> yi (- *height* 10))
	       (< xi 10)
	       (> xi (- *width* 10)))
	   (loop (+ i 1))
	   (vector xi yi))))))

(define (kinematic-model->postscript polygons kinematic-model pathname)
 (call-with-postscript-file pathname
  (lambda (ps)
   (postscript-draw-frame ps)
   (for-each (lambda (polygon) (postscript-draw-polygon ps polygon)) polygons)
   (let* ((ns (enumerate (length polygons)))
	  (same-layer? (kinematic-model-same-layer? kinematic-model))
	  (joint-i? (kinematic-model-joint-i? kinematic-model))
	  (joint-j? (kinematic-model-joint-j? kinematic-model))
	  (joint-w? (kinematic-model-joint-w? kinematic-model))
	  (ground? (kinematic-model-ground? kinematic-model))
	  (layers
	   (equivalence-classesp
	    (lambda (pi pj)
	     (matrix-ref
	      same-layer? (positionq pi polygons) (positionq pj polygons)))
	    polygons))
	  (contacts (all-contacts polygons *epsilon2*)))
    (for-each (lambda (pi)
	       (let ((layer (number->string (positionp memq pi layers)))
		     (p (layer-id-location pi)))
		(postscript-draw-string ps (v+ p '#(3 3)) layer)))
	      polygons)
    (for-each (lambda (contact k)
	       (let ((p (contact-p contact))
		     (l1 (contact-l1 contact))
		     (l2 (contact-l2 contact)))
		(if (vector-ref joint-i? k)
		    (if (vector-ref joint-j? k)
			(if (vector-ref joint-w? k)
			    (postscript-draw-filled-circle ps p 10)
			    (postscript-draw-circle ps p 10))
			(if (vector-ref joint-w? k)
			    (begin (postscript-draw-filled-circle ps p 10)
				   (postscript-draw-bar ps p l2))
			    (begin (postscript-draw-circle ps p 10)
				   (postscript-draw-bar ps p l2))))
		    (if (vector-ref joint-j? k)
			(if (vector-ref joint-w? k)
			    (begin (postscript-draw-filled-circle ps p 10)
				   (postscript-draw-bar ps p l1))
			    (begin (postscript-draw-circle ps p 10)
				   (postscript-draw-bar ps p l1)))
			(if (vector-ref joint-w? k)
			    (begin (postscript-draw-filled-circle ps p 10)
				   (postscript-draw-bar ps p l1)
				   (postscript-draw-bar ps p l2))
			    #f)))))
	      contacts
	      (enumerate (length contacts)))
    (for-each (lambda (pi i)
	       (when (vector-ref ground? i)
		(let ((p (ground-location pi)))
		 (postscript-draw-line-segment
		  ps (make-line-segment p (v+ p '#(0 5))) 1)
		 (postscript-draw-line-segment
		  ps (make-line-segment (v+ p '#(-3 5)) (v+ p '#(3 5))) 1)
		 (postscript-draw-line-segment
		  ps (make-line-segment (v+ p '#(-2 7)) (v+ p '#(2 7))) 1)
		 (postscript-draw-line-segment
		  ps (make-line-segment (v+ p '#(-1 9)) (v+ p '#(1 9))) 1))))
	      polygons
	      ns)))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
