/* LaHaShem HaAretz U'Mloah */

#include <math.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#ifndef TRUE
#define TRUE (1==1)
#endif
#ifndef FALSE
#define FALSE (1!=1)
#endif
#ifndef PI
#define PI 3.141592653589793
#endif
#ifndef INFINITY
#define INFINITY (1.0/0.0)
#endif
#ifndef MIN
#define MIN(a, b) ((a)>(b)?(b):(a))
#endif
#ifndef MAX
#define MAX(a, b) ((a)>(b)?(a):(b))
#endif
#ifndef RGB_TO_GREY
#define RGB_TO_GREY(r, g, b) (0.299*(r)+0.587*(g)+0.114*(b))
#endif
#ifndef NEGATIVE_INFINITY
#define NEGATIVE_INFINITY (-1.0/0.0)
#endif

#define BLUE_INDEX 0
#define GREEN_INDEX 1
#define RED_INDEX 2

#define DELTA_SQUARED 1.0
#define THRESHOLD 1.0

/* global variables */

int gheight, gwidth;
int **gred = NULL, **ggreen, **gblue, **ggrey, **gprevious_grey;
float **gnfm, gdistance_squared;
int *gi, *gj, gn;
float **ggood;
float gx, gy, gtheta, gr, gs, gmu[3];
void (*pre_fit)(void);

/* prototypes */

void set_mu(int i, float x);
void set_height_width(int height, int width);
void set_rgb(int y, int x, int r, int g, int b);
void compute_nfm(void);
void c_set_ellipse(float x, float y, float theta, float r, float s);
void set_distance(float distance);
void set_cost_function(int cost_function);
int get_red(int y, int x);
int get_green(int y, int x);
int get_blue(int y, int x);
int get_grey(int y, int x);
int get_previous_grey(int y, int x);
float get_nfm(int y, int x);
int get_blurred_tnfm(int y, int x);
float get_maximum_nfm(void);
float get_x(void);
float get_y(void);
float get_theta(void);
float get_r(void);
float get_s(void);
void pre_fit1(void);
void pre_fit2(void);
float fit(float x, float y, float theta, float r, float s);
void step();
void free_all(void);

/* functions */

void set_mu(int i, float x)
{ if (i<0||i>=3) panic("set_mu: i out of bound");
  gmu[i] = x;}

void set_height_width(int height, int width)
{ int y, x;
  free_all();
  gwidth = width;
  gheight = height;
  gred = (int **)malloc(gheight*sizeof(int *));
  if (!gred) panic("allocation error 1 in set_height_width");
  for (y = 0; y<gheight; y++)
  { gred[y] = (int *)malloc(gwidth*sizeof(int));
    if (!gred[y]) panic("allocation error 2 in set_height_width");}
  ggreen = (int **)malloc(gheight*sizeof(int *));
  if (!ggreen) panic("allocation error 3 in set_height_width");
  for (y = 0; y<gheight; y++)
  { ggreen[y] = (int *)malloc(gwidth*sizeof(int));
    if (!ggreen[y]) panic("allocation error 4 in set_height_width");}
  gblue = (int **)malloc(gheight*sizeof(int *));
  if (!gblue) panic("allocation error 5 in set_height_width");
  for (y = 0; y<gheight; y++)
  { gblue[y] = (int *)malloc(gwidth*sizeof(int));
    if (!gblue[y]) panic("allocation error 6 in set_height_width");}
  ggrey = (int **)malloc(gheight*sizeof(int *));
  if (!ggrey) panic("allocation error 7 in set_height_width");
  for (y = 0; y<gheight; y++)
  { ggrey[y] = (int *)malloc(gwidth*sizeof(int));
    if (!ggrey[y]) panic("allocation error 8 in set_height_width");}
  gprevious_grey = (int **)malloc(gheight*sizeof(int *));
  if (!gprevious_grey) panic("allocation error 9 in set_height_width");
  for (y = 0; y<gheight; y++)
  { gprevious_grey[y] = (int *)malloc(gwidth*sizeof(int));
    if (!gprevious_grey[y]) panic("allocation error 10 in set_height_width");}
  gnfm = (float **)malloc(gheight*sizeof(float *));
  if (!gnfm) panic("allocation error 11 in set_height_width");
  for (y = 0; y<gheight; y++)
  { gnfm[y] = (float *)malloc(gwidth*sizeof(float));
    if (!gnfm[y]) panic("allocation error 12 in set_height_width");}
  for (y = 0; y<gheight; y++) gnfm[y][gwidth-1] = 0.0;
  for (x = 0; x<gwidth; x++) gnfm[gheight-1][x] = 0.0;
  gi = (int *)malloc(gheight*gwidth*sizeof(int));
  if (!gi) panic("allocation error 13 in set_height_width");
  gj = (int *)malloc(gheight*gwidth*sizeof(int));
  if (!gj) panic("allocation error 14 in set_height_width");
  ggood = (float **)malloc(gheight*sizeof(float *));
  if (!ggood) panic("allocation error 15 in set_height_width");
  for (y = 0; y<gheight; y++)
  { ggood[y] = (float *)malloc(gwidth*sizeof(float));
    if (!ggood[y]) panic("allocation error 16 in set_height_width");}}

void set_rgb(int y, int x, int r, int g, int b)
{ if (!gred) panic("set_rgb: haven't allocated storage");
  gred[y][x] = r;
  ggreen[y][x] = g;
  gblue[y][x] = b;
  gprevious_grey[y][x] = ggrey[y][x];
  ggrey[y][x] = RGB_TO_GREY(r, g, b);}

void compute_nfm(void)
{ int y, x;
  float ex, ey, et, l;
  if (!gred) panic("compute_nfm: haven't allocated storage");
  for (y = 0; y<gheight-1; y++)
  { for (x = 0; x<gwidth-1; x++)
    { ex = (gprevious_grey[y+1][x]+gprevious_grey[y+1][x+1]
	    +ggrey[y+1][x]+ggrey[y+1][x+1]
            -gprevious_grey[y][x]-gprevious_grey[y][x+1]
	    -ggrey[y][x]-ggrey[y][x+1])/4.0;
      ey = (gprevious_grey[y][x+1]+gprevious_grey[y+1][x+1]
	    +ggrey[y][x+1]+ggrey[y+1][x+1]
            -gprevious_grey[y][x]-gprevious_grey[y+1][x]
	    -ggrey[y][x]-ggrey[y+1][x])/4.0;
      et = (ggrey[y][x]+ggrey[y][x+1]
	    +ggrey[y+1][x]+ggrey[y+1][x+1]
            -gprevious_grey[y][x]-gprevious_grey[y][x+1]
	    -gprevious_grey[y+1][x]-gprevious_grey[y+1][x+1])/4.0;
      l = ex*ex+ey*ey;
      gnfm[y][x] = l>=DELTA_SQUARED?fabs(et/sqrt(l)):0.0;}}}

void c_set_ellipse(float x, float y, float theta, float r, float s)
{ gx = x;
  gy = y;
  gtheta = theta;
  gr = r;
  gs = s;}

void set_distance(float distance)
{ gdistance_squared = distance*distance;}

void set_cost_function(int cost_function)
{ switch (cost_function)
  { case 1:
    pre_fit = pre_fit1;
    return;
    case 2:
    pre_fit = pre_fit2;
    return;
    default:
    panic("Unrecognized cost function");}}

int get_red(int y, int x)
{ if (!gred) panic("get_red: haven't allocated storage");
  return gred[y][x];}

int get_green(int y, int x)
{ if (!gred) panic("get_green: haven't allocated storage");
  return ggreen[y][x];}

int get_blue(int y, int x)
{ if (!gred) panic("get_blue: haven't allocated storage");
  return gblue[y][x];}

int get_grey(int y, int x)
{ if (!gred) panic("get_grey: haven't allocated storage");
  return ggrey[y][x];}

int get_previous_grey(int y, int x)
{ if (!gred) panic("get_previous_grey: haven't allocated storage");
  return gprevious_grey[y][x];}

float get_nfm(int y, int x)
{ if (!gred) panic("get_nfm: haven't allocated storage");
  return gnfm[y][x];}

int get_blurred_tnfm(int y, int x)
{ if (!gred) panic("get_blurred_tnfm: haven't allocated storage");
  return gnfm[y][x]>=THRESHOLD||
	 gnfm[MAX(y-1,0)][x]>=THRESHOLD||
	 gnfm[MAX(y-1,0)][MAX(x-1,0)]>=THRESHOLD||
	 gnfm[y][MAX(x-1,0)]>=THRESHOLD||
	 gnfm[MIN(y+1,gheight-1)][MAX(x-1,0)]>=THRESHOLD||
	 gnfm[MIN(y+1,gheight-1)][x]>=THRESHOLD||
	 gnfm[MIN(y+1,gheight-1)][MIN(x+1,gwidth-1)]>=THRESHOLD||
	 gnfm[y][MIN(x+1,gwidth-1)]>=THRESHOLD||
	 gnfm[MAX(y-1,0)][MIN(x+1,gwidth-1)]>=THRESHOLD;}

float get_maximum_nfm(void)
{ float sintheta = sin(gtheta), costheta = cos(gtheta), t1, t2, rsmax;
  float max = NEGATIVE_INFINITY;
  int i, j, ihigh, jlow, jhigh, n;
  if (!gred) panic("get_maximum_nfm: haven't allocated storage");
  rsmax = MAX(gr, gs);
  ihigh = MIN(gheight, gy+rsmax);
  jlow = MAX(0, gx-rsmax);
  jhigh = MIN(gwidth, gx+rsmax);
  for (i = MAX(0, gy-rsmax); i<ihigh; i++)
  { for (j = jlow; j<jhigh; j++)
    { t1 = ((j-gx)*costheta+(i-gy)*sintheta)/gr;
      t2 = (-(j-gx)*sintheta+(i-gy)*costheta)/gs;
      if (t1*t1+t2*t2<=1.0&&gnfm[i][j]>=max) max = gnfm[i][j];}}
  return max;}

float get_x(void)
{ return gx;}

float get_y(void)
{ return gy;}

float get_theta(void)
{ return gtheta;}

float get_r(void)
{ return gr;}

float get_s(void)
{ return gs;}

void pre_fit1(void)
{ int i, j;
  if (!gred) panic("pre_fit1: haven't allocated storage");
  gn = 0;
  for (i = 0; i<gheight; i++)
  { for (j = 0; j<gwidth; j++)
    { if ((gred[i][j]-gmu[0])*(gred[i][j]-gmu[0])+
	  (ggreen[i][j]-gmu[1])*(ggreen[i][j]-gmu[1])+
	  (gblue[i][j]-gmu[2])*(gblue[i][j]-gmu[2])<=gdistance_squared)
      { gi[gn] = i;
	gj[gn] = j;
	gn++;
	ggood[i][j] = -1.0;}
      else ggood[i][j] = 1.0;}}}

void pre_fit2(void)
{ int i, j;
  if (!gred) panic("pre_fit2: haven't allocated storage");
  gn = 0;
  for (i = 0; i<gheight; i++)
  { for (j = 0; j<gwidth; j++)
    { if (gnfm[i][j]>=THRESHOLD||
	  gnfm[MAX(i-1,0)][j]>=THRESHOLD||
	  gnfm[MAX(i-1,0)][MAX(j-1,0)]>=THRESHOLD||
	  gnfm[i][MAX(j-1,0)]>=THRESHOLD||
	  gnfm[MIN(i+1,gheight-1)][MAX(j-1,0)]>=THRESHOLD||
	  gnfm[MIN(i+1,gheight-1)][j]>=THRESHOLD||
	  gnfm[MIN(i+1,gheight-1)][MIN(j+1,gwidth-1)]>=THRESHOLD||
	  gnfm[i][MIN(j+1,gwidth-1)]>=THRESHOLD||
	  gnfm[MAX(i-1,0)][MIN(j+1,gwidth-1)]>=THRESHOLD)
      { gi[gn] = i;
	gj[gn] = j;
	gn++;
	ggood[i][j] = -1.0;}
      else ggood[i][j] = 1.0;}}}

float fit(float x, float y, float theta, float r, float s)
{ float sintheta = sin(theta), costheta = cos(theta), t1, t2, rsmax;
  float good_pixels = 0.0, d, e = 0.0;
  int i, j, ihigh, jlow, jhigh, n;
  if (!gred) panic("fit: haven't allocated storage");
  if (r<=0.0||s<=0.0) return INFINITY;
  rsmax = MAX(r, s);
  ihigh = MIN(gheight, y+rsmax);
  jlow = MAX(0, x-rsmax);
  jhigh = MIN(gwidth, x+rsmax);
  for (i = MAX(0, y-rsmax); i<ihigh; i++)
  { for (j = jlow; j<jhigh; j++)
    { t1 = ((j-x)*costheta+(i-y)*sintheta)/r;
      t2 = (-(j-x)*sintheta+(i-y)*costheta)/s;
      if (t1*t1+t2*t2<=1.0) good_pixels += ggood[i][j];}}
  for (n = 0; n<gn; n++)
  { d = sqrt((gj[n]-x)*(gj[n]-x)+(gi[n]-y)*(gi[n]-y));
    if (d!=0.0) e += -1.0/d;}
  return e+good_pixels;}

void step(void)
{ float this, up, down;
  int flag = TRUE, steps = 100;
  if (!gred) panic("step: haven't allocated storage");
  pre_fit();
  this = fit(gx, gy, gtheta, gr, gs);
  while (flag&&steps-->0)
  { flag = FALSE;
    up = fit(gx+1.0, gy, gtheta, gr, gs);
    down = fit(gx-1.0, gy, gtheta, gr, gs);
    if (up<down)
    { if (up<this)
      { gx++;
	flag = TRUE;
	this = up;}}
    else
    { if (down<this)
      { gx--;
	flag = TRUE;
	this = down;}}
    up = fit(gx, gy+1.0, gtheta, gr, gs);
    down = fit(gx, gy-1.0, gtheta, gr, gs);
    if (up<down)
    { if (up<this)
      { gy++;
	flag = TRUE;
	this = up;}}
    else
    { if (down<this)
      { gy--;
	flag = TRUE;
	this = down;}}
    up = fit(gx, gy, gtheta+PI/180.0, gr, gs);
    down = fit(gx, gy, gtheta-PI/180.0, gr, gs);
    if (up<down)
    { if (up<this)
      { gtheta += PI/180.0;
	flag = TRUE;
	this = up;}}
    else
    { if (down<this)
      { gtheta -= PI/180.0;
	flag = TRUE;
	this = down;}}
    up = fit(gx, gy, gtheta, gr+1.0, gs);
    down = fit(gx, gy, gtheta, gr-1.0, gs);
    if (up<down)
    { if (up<this)
      { gr++;
	flag = TRUE;
	this = up;}}
    else
    { if (down<this)
      { gr--;
	flag = TRUE;
	this = down;}}
    up = fit(gx, gy, gtheta, gr, gs+1.0);
    down = fit(gx, gy, gtheta, gr, gs-1.0);
    if (up<down)
    { if (up<this)
      { gs++;
	flag = TRUE;
	this = up;}}
    else
    { if (down<this)
      { gs--;
	flag = TRUE;
	this = down;}}}}

void free_all(void)
{ int y;
  if (gred)
  { for (y = 0; y<gheight; y++) free(gred[y]);
    free(gred);
    gred = NULL;
    for (y = 0; y<gheight; y++) free(ggreen[y]);
    free(ggreen);
    ggreen = NULL;
    for (y = 0; y<gheight; y++) free(gblue[y]);
    free(gblue);
    gblue = NULL;
    for (y = 0; y<gheight; y++) free(ggrey[y]);
    free(ggrey);
    ggrey = NULL;
    for (y = 0; y<gheight; y++) free(gprevious_grey[y]);
    free(gprevious_grey);
    gprevious_grey = NULL;
    for (y = 0; y<gheight; y++) free(gnfm[y]);
    free(gnfm);
    gnfm = NULL;
    free(gi);
    free(gj);
    for (y = 0; y<gheight; y++) free(ggood[y]);
    free(ggood);
    ggood = NULL;}}

/* Tam V'Nishlam Shevah L'El Borei Olam */
