(MODULE HLIB-SDE)
;;; LaHaShem HaAretz U'Mloah

;;; needs work: SDE:TRACK prefixed to avoid name conflicts

(include "QobiScheme.sch")
(include "hlib-sde.sch")
(include "c-externals.sc")

;;; Structures

(define-structure target type mu)

;;; Procedures

(define (download-ppm! ppm)
 (let ((height (pnm-height ppm))
       (width (pnm-width ppm))
       (red (ppm-red ppm))
       (green (ppm-green ppm))
       (blue (ppm-blue ppm)))
  (set! *height* height)
  (set! *width* width)
  (do ((y 0 (+ y 1))) ((= y height))
   (do ((x 0 (+ x 1))) ((= x width))
    (set-rgb! y x
	      (matrix-ref red y x)
	      (matrix-ref green y x)
	      (matrix-ref blue y x))))
  (compute-nfm!)
  #f))

(define (set-ellipse! ellipse)
 (c-set-ellipse! (ellipse-x0 ellipse)
		 (ellipse-y0 ellipse)
		 (ellipse-t0 ellipse)
		 (ellipse-a ellipse)
		 (ellipse-b ellipse))
 #f)

(define (target-rgb-pixels ellipse)
 (set-ellipse! ellipse)
 (let* ((x (ellipse-x0 ellipse))
	(y (ellipse-y0 ellipse))
	(theta (ellipse-t0 ellipse))
	(r (ellipse-a ellipse))
	(s (ellipse-b ellipse))
	(sintheta (sin theta))
	(costheta (cos theta))
	(rsmax (max r s))
	(ilow (max 0 (- y rsmax)))
	(ihigh (min *height* (+ y rsmax)))
	(jlow (max 0 (- x rsmax)))
	(jhigh (min *width* (+ x rsmax)))
	(pixels '()))
  (do ((i ilow (+ i 1))) ((= i ihigh))
   (do ((j jlow (+ j 1))) ((= j jhigh))
    (when (<= (+ (sqr (/ (+ (* (- j x) costheta) (* (- i y) sintheta)) r))
		 (sqr (/ (- (* (- i y) costheta) (* (- j x) sintheta)) s)))
	      1)
     (set! pixels
	   (cons (vector (get-red i j) (get-green i j) (get-blue i j))
		 pixels)))))
  (list->vector pixels)))

(define (set-target! target)
 (let ((mu (target-mu target)))
  (case (target-type target)
   ((rgb)
    (set-cost-function! 1)
    (do ((i 0 (+ i 1))) ((= i (vector-length mu)))
     (set-mu! i (vector-ref mu i)))
    (let ((hue (vector-ref (rgb->hsv mu) 0)))
     (set-distance! (cond ((or (<= hue 30) (>= hue 330)) 40) ;red
			  ((<= 120 hue 180) 18) ;green
			  ((<= 210 hue 270) 30)	;blue
			  ((<= 30 hue 60) 45)	;yellow
			  (else 30)))))
   ((nfm) (set-cost-function! 2))
   (else (fuck-up)))
  #f))

(define (create-target ellipse cost-function)
 (case cost-function
  ((rgb) (make-target 'rgb (mean (target-rgb-pixels ellipse))))
  ((nfm) (make-target 'nfm #f))
  (else (fuck-up))))

(define (step ellipse target)
 (set-target! target)
 (set-ellipse! ellipse)
 (cond ((and (eq? (target-type target) 'nfm)
	     (not (= (ellipse-a ellipse) 5))
	     (not (= (ellipse-b ellipse) 5))
	     (<= (get-maximum-nfm) 20))
	ellipse)
       (else (step!)
	     (make-ellipse (get-x)
			   (get-y)
			   (get-theta)
			   (max (get-r) 5)
			   (max (get-s) 5)))))

(define (grow ellipse cost-function)
 (step ellipse (create-target ellipse cost-function)))

(define (grow* ellipse cost-function)
 (let loop ((ellipse ellipse))
  (let ((ellipse-prime (grow ellipse cost-function)))
   (if (equal? ellipse ellipse-prime) ellipse (loop ellipse-prime)))))

(define (sde:track ellipses cost-functions pathname video-type trace?)
 (let ((video-input-port (open-video-input-file pathname video-type)))
  (let ((ppm (read-ppm-from-video-input-port video-input-port)))
   (unless ppm (panic "Can't track empty movie"))
   (set-height-width! (pnm-height ppm) (pnm-width ppm))
   (download-ppm! ppm)
   (let loop ((ellipse-movie (list ellipses)) (i 0))
    (when trace? (format #t "~s~%" i))
    (let ((ppm (read-ppm-from-video-input-port video-input-port)))
     (cond (ppm (let ((targets (map create-target
				    (first ellipse-movie)
				    cost-functions)))
		 (download-ppm! ppm)
		 (loop (cons (map step (first ellipse-movie) targets)
			     ellipse-movie)
		       (+ i 1))))
	   (else (close-video-input-port video-input-port)
		 (list->vector (reverse ellipse-movie)))))))))

(define (blurred-tnfm-pbm)
 (let ((bitmap (make-matrix *height* *width*)))
  (do ((y 0 (+ y 1))) ((= y *height*))
   (do ((x 0 (+ x 1))) ((= x *width*))
    (matrix-set! bitmap y x (zero? (get-blurred-tnfm y x)))))
  (make-pbm #t bitmap)))

;;; Tam V'Nishlam Shevah L'El Borei Olam
