(MODULE HLIB-SPAN)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-span.sch")
(include "c-externals.sc")

;;; Things to explain in paper.
;;;  1. what INF and -INF really mean
;;;  2. {((-INF:INF)),((-INF:INF)],[(-INF:INF)),[(-INF:INF)]} denotes the set
;;;     of all intervals
;;;  3. resolution
;;;  4. Kleene +

;;; Structures

(define-structure extended-spanning-interval
 i j k l alpha? beta? gamma? delta? epsilon? zeta?)

;;; Parameters

(define *jair-aggregate-event-types*
 '(((pick-up x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (begin (and (not (diamond (attached? x y)))
		     (not (diamond (supports? x y)))
		     (supports? z y)
		     ;; The rest are common to both states.
		     (not (diamond (supported? x)))
		     (not (diamond (attached? y z)))
		     (not (diamond (supports? y x)))
		     (not (diamond (supports? y z)))
		     (not (diamond (supports? x z)))
		     (not (diamond (supports? z x))))
		(or (attached? x y) (attached? y z))
		(and (attached? x y)
		     (supports? x y)
		     (not (diamond (supports? z y)))
		     ;; The rest are common to both states.
		     (not (diamond (supported? x)))
		     (not (diamond (attached? y z)))
		     (not (diamond (supports? y x)))
		     (not (diamond (supports? y z)))
		     (not (diamond (supports? x z)))
		     (not (diamond (supports? z x)))))))
   ((put-down x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (begin (and (attached? x y)
		     (supports? x y)
		     (not (diamond (supports? z y)))
		     ;; The rest are common to both states.
		     (not (diamond (supported? x)))
		     (not (diamond (attached? y z)))
		     (not (diamond (supports? y x)))
		     (not (diamond (supports? y z)))
		     (not (diamond (supports? x z)))
		     (not (diamond (supports? z x))))
		(or (attached? x y) (attached? y z))
		(and (not (diamond (attached? x y)))
		     (not (diamond (supports? x y)))
		     (supports? z y)
		     ;; The rest are common to both states.
		     (not (diamond (supported? x)))
		     (not (diamond (attached? y z)))
		     (not (diamond (supports? y x)))
		     (not (diamond (supports? y z)))
		     (not (diamond (supports? x z)))
		     (not (diamond (supports? z x)))))))
   ((stack w x y z)
    (and (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (put-down w x y)
	 (supports? z y)
	 (not (attached? z y))))
   ((unstack w x y z)
    (and (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (pick-up w x y)
	 (supports? z y)
	 (not (attached? z y))))
   ((move w x y z)
    (and (not (diamond (equal? y z)))
	 (begin (pick-up w x y) (put-down w x z))))
   ((assemble w x y z) (and (<) (put-down w y z) (stack w x y z)))
   ((disassemble w x y z) (and (<) (unstack w x y z) (pick-up w y z)))
   ((pick-up-stack v w x y z) (begin (pick-up v w x) (stack v w y z)))
   ((unstack-put-down v w x y z) (begin (unstack v w x y) (put-down v w z)))
   ((unstack-stack u v w x y z) (begin (unstack u v w x) (stack u v y z)))))

(define *debugging-aggregate-event-types*
 '(((pick-up1 x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (and (not (diamond (attached? x y)))
	      (not (diamond (supports? x y)))
	      (supports? z y)
	      ;; The rest are common to both states.
	      (not (diamond (supported? x)))
	      (not (diamond (attached? y z)))
	      (not (diamond (supports? y x)))
	      (not (diamond (supports? y z)))
	      (not (diamond (supports? x z)))
	      (not (diamond (supports? z x))))))
   ((pick-up2 x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (or (attached? x y) (attached? y z))))
   ((pick-up3 x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (and (attached? x y)
	      (supports? x y)
	      (not (diamond (supports? z y)))
	      ;; The rest are common to both states.
	      (not (diamond (supported? x)))
	      (not (diamond (attached? y z)))
	      (not (diamond (supports? y x)))
	      (not (diamond (supports? y z)))
	      (not (diamond (supports? x z)))
	      (not (diamond (supports? z x))))))
   ((put-down1 x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (and (attached? x y)
	      (supports? x y)
	      (not (diamond (supports? z y)))
	      ;; The rest are common to both states.
	      (not (diamond (supported? x)))
	      (not (diamond (attached? y z)))
	      (not (diamond (supports? y x)))
	      (not (diamond (supports? y z)))
	      (not (diamond (supports? x z)))
	      (not (diamond (supports? z x))))))
   ((put-down2 x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (or (attached? x y) (attached? y z))))
   ((put-down3 x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (and (not (diamond (attached? x y)))
	      (not (diamond (supports? x y)))
	      (supports? z y)
	      ;; The rest are common to both states.
	      (not (diamond (supported? x)))
	      (not (diamond (attached? y z)))
	      (not (diamond (supports? y x)))
	      (not (diamond (supports? y z)))
	      (not (diamond (supports? x z)))
	      (not (diamond (supports? z x))))))
   ((stack1 w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (and (attached? w x)
	      (supports? w x)
	      (not (diamond (supports? y x)))
	      ;; The rest are common to both states.
	      (supports? z y)
	      (not (attached? z y))
	      (not (diamond (supported? w)))
	      (not (diamond (attached? x y)))
	      (not (diamond (supports? x w)))
	      (not (diamond (supports? x y)))
	      (not (diamond (supports? w y)))
	      (not (diamond (supports? y w))))))
   ((stack2 w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (or (attached? w x) (attached? x y))))
   ((stack3 w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (and (not (diamond (attached? w x)))
	      (not (diamond (supports? w x)))
	      (supports? y x)
	      ;; The rest are common to both states.
	      (supports? z y)
	      (not (attached? z y))
	      (not (diamond (supported? w)))
	      (not (diamond (attached? x y)))
	      (not (diamond (supports? x w)))
	      (not (diamond (supports? x y)))
	      (not (diamond (supports? w y)))
	      (not (diamond (supports? y w))))))
   ((unstack1 w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (and (not (diamond (attached? w x)))
	      (not (diamond (supports? w x)))
	      (supports? y x)
	      ;; The rest are common to both states.
	      (supports? z y)
	      (not (attached? z y))
	      (not (diamond (supported? w)))
	      (not (diamond (attached? x y)))
	      (not (diamond (supports? x w)))
	      (not (diamond (supports? x y)))
	      (not (diamond (supports? w y)))
	      (not (diamond (supports? y w))))))
   ((unstack2 w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (or (attached? w x) (attached? x y))))
   ((unstack3 w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (and (attached? w x)
	      (supports? w x)
	      (not (diamond (supports? y x)))
	      ;; The rest are common to both states.
	      (supports? z y)
	      (not (attached? z y))
	      (not (diamond (supported? w)))
	      (not (diamond (attached? x y)))
	      (not (diamond (supports? x w)))
	      (not (diamond (supports? x y)))
	      (not (diamond (supports? w y)))
	      (not (diamond (supports? y w))))))
   ((pick-up x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (or (begin (and (not (diamond (attached? x y)))
			 (not (diamond (supports? x y)))
			 (supports? z y)
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))
		    (or (attached? x y) (attached? y z))
		    (and (attached? x y)
			 (supports? x y)
			 (not (diamond (supports? z y)))
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x)))))
	     (begin (and (not (diamond (attached? x y)))
			 (not (diamond (supports? x y)))
			 (supports? z y)
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))
		    (and (attached? x y)
			 (supports? x y)
			 (not (diamond (supports? z y)))
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))))))
   ((put-down x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (or (begin (and (attached? x y)
			 (supports? x y)
			 (not (diamond (supports? z y)))
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))
		    (or (attached? x y) (attached? y z))
		    (and (not (diamond (attached? x y)))
			 (not (diamond (supports? x y)))
			 (supports? z y)
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x)))))
	     (begin (and (attached? x y)
			 (supports? x y)
			 (not (diamond (supports? z y)))
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))
		    (and (not (diamond (attached? x y)))
			 (not (diamond (supports? x y)))
			 (supports? z y)
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))))))
   ((stack w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (or (begin (and (attached? w x)
			 (supports? w x)
			 (not (diamond (supports? y x)))
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))
		    (or (attached? w x) (attached? x y))
		    (and (not (diamond (attached? w x)))
			 (not (diamond (supports? w x)))
			 (supports? y x)
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w)))))
	     (begin (and (attached? w x)
			 (supports? w x)
			 (not (diamond (supports? y x)))
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))
		    (and (not (diamond (attached? w x)))
			 (not (diamond (supports? w x)))
			 (supports? y x)
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))))))
   ((unstack w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (or (begin (and (not (diamond (attached? w x)))
			 (not (diamond (supports? w x)))
			 (supports? y x)
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))
		    (or (attached? w x) (attached? x y))
		    (and (attached? w x)
			 (supports? w x)
			 (not (diamond (supports? y x)))
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w)))))
	     (begin (and (not (diamond (attached? w x)))
			 (not (diamond (supports? w x)))
			 (supports? y x)
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))
		    (and (attached? w x)
			 (supports? w x)
			 (not (diamond (supports? y x)))
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))))))
   ((move w x y z)
    (and (not (diamond (equal? y z)))
	 (begin (pick-up w x y) (put-down w x z))))
   ((assemble w x y z) (and (<) (put-down w y z) (stack w x y z)))
   ((disassemble w x y z) (and (<) (unstack w x y z) (pick-up w y z)))))

(define *aggregate-event-types*
 '(((pick-up x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (or (begin (and (not (diamond (attached? x y)))
			 (not (diamond (supports? x y)))
			 (supports? z y)
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))
		    (or (attached? x y) (attached? y z))
		    (and (attached? x y)
			 (supports? x y)
			 (not (diamond (supports? z y)))
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x)))))
	     (begin (and (not (diamond (attached? x y)))
			 (not (diamond (supports? x y)))
			 (supports? z y)
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))
		    (and (attached? x y)
			 (supports? x y)
			 (not (diamond (supports? z y)))
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))))))
   ((put-down x y z)
    (and (not (diamond (equal? x y)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? y)
	 (not (diamond (attached? x z)))
	 (or (begin (and (attached? x y)
			 (supports? x y)
			 (not (diamond (supports? z y)))
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))
		    (or (attached? x y) (attached? y z))
		    (and (not (diamond (attached? x y)))
			 (not (diamond (supports? x y)))
			 (supports? z y)
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x)))))
	     (begin (and (attached? x y)
			 (supports? x y)
			 (not (diamond (supports? z y)))
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))
		    (and (not (diamond (attached? x y)))
			 (not (diamond (supports? x y)))
			 (supports? z y)
			 ;; The rest are common to both states.
			 (not (diamond (supported? x)))
			 (not (diamond (attached? y z)))
			 (not (diamond (supports? y x)))
			 (not (diamond (supports? y z)))
			 (not (diamond (supports? x z)))
			 (not (diamond (supports? z x))))))))
   ((stack w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (or (begin (and (attached? w x)
			 (supports? w x)
			 (not (diamond (supports? y x)))
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))
		    (or (attached? w x) (attached? x y))
		    (and (not (diamond (attached? w x)))
			 (not (diamond (supports? w x)))
			 (supports? y x)
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w)))))
	     (begin (and (attached? w x)
			 (supports? w x)
			 (not (diamond (supports? y x)))
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))
		    (and (not (diamond (attached? w x)))
			 (not (diamond (supports? w x)))
			 (supports? y x)
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))))))
   ((unstack w x y z)
    (and (not (diamond (equal? w x)))
	 (not (diamond (equal? y w)))
	 (not (diamond (equal? y x)))
	 (not (diamond (equal? z w)))
	 (not (diamond (equal? z x)))
	 (not (diamond (equal? z y)))
	 (supported? x)
	 (not (diamond (attached? w y)))
	 (or (begin (and (not (diamond (attached? w x)))
			 (not (diamond (supports? w x)))
			 (supports? y x)
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))
		    (or (attached? w x) (attached? x y))
		    (and (attached? w x)
			 (supports? w x)
			 (not (diamond (supports? y x)))
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w)))))
	     (begin (and (not (diamond (attached? w x)))
			 (not (diamond (supports? w x)))
			 (supports? y x)
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))
		    (and (attached? w x)
			 (supports? w x)
			 (not (diamond (supports? y x)))
			 ;; The rest are common to both states.
			 (supports? z y)
			 (not (attached? z y))
			 (not (diamond (supported? w)))
			 (not (diamond (attached? x y)))
			 (not (diamond (supports? x w)))
			 (not (diamond (supports? x y)))
			 (not (diamond (supports? w y)))
			 (not (diamond (supports? y w))))))))
   ((move w x y z)
    (and (not (diamond (equal? y z)))
	 (begin (pick-up w x y) (put-down w x z))))
   ((assemble w x y z) (and (<) (put-down w y z) (stack w x y z)))
   ((disassemble w x y z) (and (<) (unstack w x y z) (pick-up w y z)))))

;;; Variables

(define *polygon-movie* #f)
(define *kinematic-models* #f)
(define *contacts* #f)
(define *supports?* #f)
(define *contacts?* #f)
(define *attached?* #f)
(define *event-definition-set-name* #f)
(define *fuzz* #f)

;;; Procedures

(define (aggregate-event-types)
 (cond ((string=? *event-definition-set-name* "jair2001")
	*jair-aggregate-event-types*)
       ((string=? *event-definition-set-name* "debugging")
	*debugging-aggregate-event-types*)
       ((string=? *event-definition-set-name* "default")
	*aggregate-event-types*)
       (else (fuck-up))))

(define (extended-spanning-interval->string extended-spanning-interval)
 (define (-inf x) (if (= x minus-infinity) "-inf" x))
 (define (inf x) (if (= x infinity) "inf" x))
 (cond
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-j extended-spanning-interval)
	   (extended-spanning-interval-k extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(extended-spanning-interval-alpha? extended-spanning-interval)
	(extended-spanning-interval-beta? extended-spanning-interval)
	(extended-spanning-interval-gamma? extended-spanning-interval)
	(extended-spanning-interval-delta? extended-spanning-interval)
	(extended-spanning-interval-epsilon? extended-spanning-interval)
	(extended-spanning-interval-zeta? extended-spanning-interval))
   (format #f "{{~a}}"
	   (extended-spanning-interval-i extended-spanning-interval)))
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-j extended-spanning-interval))
	(= (extended-spanning-interval-k extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(extended-spanning-interval-gamma? extended-spanning-interval)
	(extended-spanning-interval-delta? extended-spanning-interval)
	(extended-spanning-interval-epsilon? extended-spanning-interval)
	(extended-spanning-interval-zeta? extended-spanning-interval))
   (format #f "~a{~a},{~a}~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-k extended-spanning-interval)
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-j extended-spanning-interval))
	(extended-spanning-interval-gamma? extended-spanning-interval)
	(extended-spanning-interval-delta? extended-spanning-interval))
   (format #f "~a{~a},~a~a,~a~a~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (extended-spanning-interval-i extended-spanning-interval)
	   (if (extended-spanning-interval-epsilon? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-k extended-spanning-interval))
	   (inf (extended-spanning-interval-l extended-spanning-interval))
	   (if (extended-spanning-interval-zeta? extended-spanning-interval)
	       "]"
	       ")")
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))
  ((and (= (extended-spanning-interval-k extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(extended-spanning-interval-epsilon? extended-spanning-interval)
	(extended-spanning-interval-zeta? extended-spanning-interval))
   (format #f "~a~a~a,~a~a,{~a}~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (if (extended-spanning-interval-gamma? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-i extended-spanning-interval))
	   (inf (extended-spanning-interval-j extended-spanning-interval))
	   (if (extended-spanning-interval-delta? extended-spanning-interval)
	       "]"
	       ")")
	   (extended-spanning-interval-k extended-spanning-interval)
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))
  ;; needs work: This is a special case to undo normalization. It needs to be
  ;;             made more general.
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-k extended-spanning-interval))
	(= (extended-spanning-interval-j extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(extended-spanning-interval-alpha? extended-spanning-interval)
	(not (extended-spanning-interval-beta? extended-spanning-interval))
	(extended-spanning-interval-gamma? extended-spanning-interval)
	(not (extended-spanning-interval-delta? extended-spanning-interval))
	(not (extended-spanning-interval-epsilon? extended-spanning-interval))
	(extended-spanning-interval-zeta? extended-spanning-interval))
   (format #f "[[~a:~a])"
	   (-inf (extended-spanning-interval-i extended-spanning-interval))
	   (inf (extended-spanning-interval-j extended-spanning-interval))))
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-k extended-spanning-interval))
	(= (extended-spanning-interval-j extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(eq? (extended-spanning-interval-gamma? extended-spanning-interval)
	     (extended-spanning-interval-epsilon? extended-spanning-interval))
	(eq? (extended-spanning-interval-delta? extended-spanning-interval)
	     (extended-spanning-interval-zeta? extended-spanning-interval)))
   (format #f "~a~a~a:~a~a~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (if (extended-spanning-interval-gamma? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-i extended-spanning-interval))
	   (inf (extended-spanning-interval-j extended-spanning-interval))
	   (if (extended-spanning-interval-delta? extended-spanning-interval)
	       "]"
	       ")")
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))
  (else
   (format #f "~a~a~a,~a~a,~a~a,~a~a~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (if (extended-spanning-interval-gamma? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-i extended-spanning-interval))
	   (inf (extended-spanning-interval-j extended-spanning-interval))
	   (if (extended-spanning-interval-delta? extended-spanning-interval)
	       "]"
	       ")")
	   (if (extended-spanning-interval-epsilon? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-k extended-spanning-interval))
	   (inf (extended-spanning-interval-l extended-spanning-interval))
	   (if (extended-spanning-interval-zeta? extended-spanning-interval)
	       "]"
	       ")")
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))))

(define (implies p q) (or (not p) q))

(define (extended-spanning-intervals i
				     j
				     k
				     l
				     alpha?
				     beta?
				     gamma?
				     delta?
				     epsilon?
				     zeta?)
 ;; This version is missing normalization conditions 5 and 9, and thus also
 ;; 12 and 8. But it contains the new conditions 13 and 14.
 (let ((j (min j l))
       (k (max k i)))
  (let ((gamma?
	 ;; Gamma is made false when i=-infinity.
	 (and gamma? (not (= i minus-infinity))))
	(delta?
	 ;; Delta is made false when either j=infinity or both j=l and either
	 ;; zeta, alpha, or beta is false.
	 (and delta?
	      (not (= j infinity))
	      (or (not (= j l)) (and zeta? alpha? beta?))))
	(epsilon?
	 ;; Epsilon is made false when either k=-infinity or both k=i and
	 ;; either gamma, beta, or alpha is false.
	 (and epsilon?
	      (not (= k minus-infinity))
	      (or (not (= k i)) (and gamma? beta? alpha?))))
	(zeta?
	 ;; Zeta is made false when i=infinity.
	 (and zeta? (not (= l infinity)))))
   (if (and (<= i j)
	    (<= k l)
	    (implies (= i j) (and gamma? delta?))
	    (implies (= k l) (and epsilon? zeta?))
	    (implies (= i l) (and alpha? beta?))
	    (not (= i infinity))
	    (not (= j minus-infinity))
	    (not (= k infinity))
	    (not (= l minus-infinity)))
       (list (make-extended-spanning-interval
	      i j k l alpha? beta? gamma? delta? epsilon? zeta?))
       '()))))

(define (spanning-intervals i j alpha? beta? gamma? delta?)
 (extended-spanning-intervals
  i j i j alpha? beta? gamma? delta? gamma? delta?))

(define (intervals i j alpha? beta?)
 (extended-spanning-intervals i i j j alpha? beta? #t #t #t #t))

(define (intersection-of is js)
 (reduce
  append
  (map (lambda (i)
	(reduce
	 append
	 (map (lambda (j)
	       (if (and (eq? (extended-spanning-interval-alpha? i)
			     (extended-spanning-interval-alpha? j))
			(eq? (extended-spanning-interval-beta? i)
			     (extended-spanning-interval-beta? j)))
		   (extended-spanning-intervals
		    (max (extended-spanning-interval-i i)
			 (extended-spanning-interval-i j))
		    (min (extended-spanning-interval-j i)
			 (extended-spanning-interval-j j))
		    (max (extended-spanning-interval-k i)
			 (extended-spanning-interval-k j))
		    (min (extended-spanning-interval-l i)
			 (extended-spanning-interval-l j))
		    (extended-spanning-interval-alpha? i)
		    (extended-spanning-interval-beta? i)
		    (cond ((> (extended-spanning-interval-i i)
			      (extended-spanning-interval-i j))
			   (extended-spanning-interval-gamma? i))
			  ((< (extended-spanning-interval-i i)
			      (extended-spanning-interval-i j))
			   (extended-spanning-interval-gamma? j))
			  (else (and (extended-spanning-interval-gamma? i)
				     (extended-spanning-interval-gamma? j))))
		    (cond ((< (extended-spanning-interval-j i)
			      (extended-spanning-interval-j j))
			   (extended-spanning-interval-delta? i))
			  ((> (extended-spanning-interval-j i)
			      (extended-spanning-interval-j j))
			   (extended-spanning-interval-delta? j))
			  (else (and (extended-spanning-interval-delta? i)
				     (extended-spanning-interval-delta? j))))
		    (cond ((> (extended-spanning-interval-k i)
			      (extended-spanning-interval-k j))
			   (extended-spanning-interval-epsilon? i))
			  ((< (extended-spanning-interval-k i)
			      (extended-spanning-interval-k j))
			   (extended-spanning-interval-epsilon? j))
			  (else (and (extended-spanning-interval-epsilon? i)
				     (extended-spanning-interval-epsilon? j))))
		    (cond ((< (extended-spanning-interval-l i)
			      (extended-spanning-interval-l j))
			   (extended-spanning-interval-zeta? i))
			  ((> (extended-spanning-interval-l i)
			      (extended-spanning-interval-l j))
			   (extended-spanning-interval-zeta? j))
			  (else (and (extended-spanning-interval-zeta? i)
				     (extended-spanning-interval-zeta? j)))))
		   '()))
	      js)
	 '()))
       is)
  '()))

(define (complement-of is)
 (if (null? is)
     (all-intervals)
     (intersection-of
      (let ((i (first is)))
       (append (extended-spanning-intervals
		minus-infinity
		infinity
		minus-infinity
		(extended-spanning-interval-k i)
		(extended-spanning-interval-alpha? i)
		(extended-spanning-interval-beta? i)
		#t
		#t
		#t
		(not (extended-spanning-interval-epsilon? i)))
	       (extended-spanning-intervals
		minus-infinity
		infinity
		(extended-spanning-interval-l i)
		infinity
		(extended-spanning-interval-alpha? i)
		(extended-spanning-interval-beta? i)
		#t
		#t
		(not (extended-spanning-interval-zeta? i))
		#t)
	       (extended-spanning-intervals
		minus-infinity
		(extended-spanning-interval-i i)
		minus-infinity
		infinity
		(extended-spanning-interval-alpha? i)
		(extended-spanning-interval-beta? i)
		#t
		(not (extended-spanning-interval-gamma? i))
		#t
		#t)
	       (extended-spanning-intervals
		(extended-spanning-interval-j i)
		infinity
		minus-infinity
		infinity
		(extended-spanning-interval-alpha? i)
		(extended-spanning-interval-beta? i)
		(not (extended-spanning-interval-delta? i))
		#t
		#t
		#t)
	       (extended-spanning-intervals
		minus-infinity
		infinity
		minus-infinity
		infinity
		(not (extended-spanning-interval-alpha? i))
		(extended-spanning-interval-beta? i)
		#t
		#t
		#t
		#t)
	       (extended-spanning-intervals
		minus-infinity
		infinity
		minus-infinity
		infinity
		(extended-spanning-interval-alpha? i)
		(not (extended-spanning-interval-beta? i))
		#t
		#t
		#t
		#t)
	       (extended-spanning-intervals
		minus-infinity
		infinity
		minus-infinity
		infinity
		(not (extended-spanning-interval-alpha? i))
		(not (extended-spanning-interval-beta? i))
		#t
		#t
		#t
		#t)))
      (complement-of (rest is)))))

(define (resolution-of i j)
 (append (if (and (eq? (extended-spanning-interval-alpha? i)
		       (extended-spanning-interval-alpha? j))
		  (eq? (extended-spanning-interval-beta? i)
		       (extended-spanning-interval-beta? j))
		  (<= (extended-spanning-interval-i i)
		      (extended-spanning-interval-j j))
		  (<= (extended-spanning-interval-i j)
		      (extended-spanning-interval-j i))
		  (implies (or (not (extended-spanning-interval-gamma? i))
			       (not (extended-spanning-interval-delta? j)))
			   (not (= (extended-spanning-interval-i i)
				   (extended-spanning-interval-j j))))
		  (implies (or (not (extended-spanning-interval-gamma? j))
			       (not (extended-spanning-interval-delta? i)))
			   (not (= (extended-spanning-interval-i j)
				   (extended-spanning-interval-j i)))))
	     (extended-spanning-intervals
	      (min (extended-spanning-interval-i i)
		   (extended-spanning-interval-i j))
	      (max (extended-spanning-interval-j i)
		   (extended-spanning-interval-j j))
	      (max (extended-spanning-interval-k i)
		   (extended-spanning-interval-k j))
	      (min (extended-spanning-interval-l i)
		   (extended-spanning-interval-l j))
	      (extended-spanning-interval-alpha? i)
	      (extended-spanning-interval-beta? i)
	      (cond ((< (extended-spanning-interval-i i)
			(extended-spanning-interval-i j))
		     (extended-spanning-interval-gamma? i))
		    ((> (extended-spanning-interval-i i)
			(extended-spanning-interval-i j))
		     (extended-spanning-interval-gamma? j))
		    (else (or (extended-spanning-interval-gamma? i)
			      (extended-spanning-interval-gamma? j))))
	      (cond ((> (extended-spanning-interval-j i)
			(extended-spanning-interval-j j))
		     (extended-spanning-interval-delta? i))
		    ((< (extended-spanning-interval-j i)
			(extended-spanning-interval-j j))
		     (extended-spanning-interval-delta? j))
		    (else (or (extended-spanning-interval-delta? i)
			      (extended-spanning-interval-delta? j))))
	      (cond ((> (extended-spanning-interval-k i)
			(extended-spanning-interval-k j))
		     (extended-spanning-interval-epsilon? i))
		    ((< (extended-spanning-interval-k i)
			(extended-spanning-interval-k j))
		     (extended-spanning-interval-epsilon? j))
		    (else (and (extended-spanning-interval-epsilon? i)
			       (extended-spanning-interval-epsilon? j))))
	      (cond ((< (extended-spanning-interval-l i)
			(extended-spanning-interval-l j))
		     (extended-spanning-interval-zeta? i))
		    ((> (extended-spanning-interval-l i)
			(extended-spanning-interval-l j))
		     (extended-spanning-interval-zeta? j))
		    (else (and (extended-spanning-interval-zeta? i)
			       (extended-spanning-interval-zeta? j)))))
	     '())
	 (if (and (eq? (extended-spanning-interval-alpha? i)
		       (extended-spanning-interval-alpha? j))
		  (eq? (extended-spanning-interval-beta? i)
		       (extended-spanning-interval-beta? j))
		  (<= (extended-spanning-interval-k i)
		      (extended-spanning-interval-l j))
		  (<= (extended-spanning-interval-k j)
		      (extended-spanning-interval-l i))
		  (implies (or (not (extended-spanning-interval-epsilon? i))
			       (not (extended-spanning-interval-zeta? j)))
			   (not (= (extended-spanning-interval-k i)
				   (extended-spanning-interval-l j))))
		  (implies (or (not (extended-spanning-interval-epsilon? j))
			       (not (extended-spanning-interval-zeta? i)))
			   (not (= (extended-spanning-interval-k j)
				   (extended-spanning-interval-l i)))))
	     (extended-spanning-intervals
	      (max (extended-spanning-interval-i i)
		   (extended-spanning-interval-i j))
	      (min (extended-spanning-interval-j i)
		   (extended-spanning-interval-j j))
	      (min (extended-spanning-interval-k i)
		   (extended-spanning-interval-k j))
	      (max (extended-spanning-interval-l i)
		   (extended-spanning-interval-l j))
	      (extended-spanning-interval-alpha? i)
	      (extended-spanning-interval-beta? i)
	      (cond ((> (extended-spanning-interval-i i)
			(extended-spanning-interval-i j))
		     (extended-spanning-interval-gamma? i))
		    ((< (extended-spanning-interval-i i)
			(extended-spanning-interval-i j))
		     (extended-spanning-interval-gamma? j))
		    (else (and (extended-spanning-interval-gamma? i)
			       (extended-spanning-interval-gamma? j))))
	      (cond ((< (extended-spanning-interval-j i)
			(extended-spanning-interval-j j))
		     (extended-spanning-interval-delta? i))
		    ((> (extended-spanning-interval-j i)
			(extended-spanning-interval-j j))
		     (extended-spanning-interval-delta? j))
		    (else (and (extended-spanning-interval-delta? i)
			       (extended-spanning-interval-delta? j))))
	      (cond ((< (extended-spanning-interval-k i)
			(extended-spanning-interval-k j))
		     (extended-spanning-interval-epsilon? i))
		    ((> (extended-spanning-interval-k i)
			(extended-spanning-interval-k j))
		     (extended-spanning-interval-epsilon? j))
		    (else (or (extended-spanning-interval-epsilon? i)
			      (extended-spanning-interval-epsilon? j))))
	      (cond ((> (extended-spanning-interval-l i)
			(extended-spanning-interval-l j))
		     (extended-spanning-interval-zeta? i))
		    ((< (extended-spanning-interval-l i)
			(extended-spanning-interval-l j))
		     (extended-spanning-interval-zeta? j))
		    (else (or (extended-spanning-interval-zeta? i)
			      (extended-spanning-interval-zeta? j)))))
	     '())))

(define (subsumes? j i)
 (and (eq? (extended-spanning-interval-alpha? i)
	   (extended-spanning-interval-alpha? j))
      (eq? (extended-spanning-interval-beta? i)
	   (extended-spanning-interval-beta? j))
      (>= (extended-spanning-interval-i i)
	  (extended-spanning-interval-i j))
      (<= (extended-spanning-interval-j i)
	  (extended-spanning-interval-j j))
      (>= (extended-spanning-interval-k i)
	  (extended-spanning-interval-k j))
      (<= (extended-spanning-interval-l i)
	  (extended-spanning-interval-l j))
      (implies (and (extended-spanning-interval-gamma? i)
		    (not (extended-spanning-interval-gamma? j)))
	       (not (= (extended-spanning-interval-i i)
		       (extended-spanning-interval-i j))))
      (implies (and (extended-spanning-interval-delta? i)
		    (not (extended-spanning-interval-delta? j)))
	       (not (= (extended-spanning-interval-j i)
		       (extended-spanning-interval-j j))))
      (implies (and (extended-spanning-interval-epsilon? i)
		    (not (extended-spanning-interval-epsilon? j)))
	       (not (= (extended-spanning-interval-k i)
		       (extended-spanning-interval-k j))))
      (implies (and (extended-spanning-interval-zeta? i)
		    (not (extended-spanning-interval-zeta? j)))
	       (not (= (extended-spanning-interval-l i)
		       (extended-spanning-interval-l j))))))

(define (span is js)
 (reduce
  append
  (map
   (lambda (i)
    (reduce
     append
     (map
      (lambda (j)
       (let* ((i1 (extended-spanning-interval-i i))
	      (j1 (extended-spanning-interval-j i))
	      (k1 (extended-spanning-interval-k i))
	      (l1 (extended-spanning-interval-l i))
	      (alpha1? (extended-spanning-interval-alpha? i))
	      (beta1? (extended-spanning-interval-beta? i))
	      (gamma1? (extended-spanning-interval-gamma? i))
	      (delta1? (extended-spanning-interval-delta? i))
	      (epsilon1? (extended-spanning-interval-epsilon? i))
	      (zeta1? (extended-spanning-interval-zeta? i))
	      (i2 (extended-spanning-interval-i j))
	      (j2 (extended-spanning-interval-j j))
	      (k2 (extended-spanning-interval-k j))
	      (l2 (extended-spanning-interval-l j))
	      (alpha2? (extended-spanning-interval-alpha? j))
	      (beta2? (extended-spanning-interval-beta? j))
	      (gamma2? (extended-spanning-interval-gamma? j))
	      (delta2? (extended-spanning-interval-delta? j))
	      (epsilon2? (extended-spanning-interval-epsilon? j))
	      (zeta2? (extended-spanning-interval-zeta? j))
	      (j-prime (min j1 j2))
	      (k-prime (max k1 k2))
	      (delta?-prime (or (and delta1? (<= j1 j2))
				(and delta2? (>= j1 j2))))
	      (epsilon?-prime (or (and epsilon1? (>= k1 k2))
				  (and epsilon2? (<= k1 k2)))))
	(append (extended-spanning-intervals
		 i1 j-prime k-prime l1
		 alpha1? beta1? gamma1? delta?-prime epsilon?-prime zeta1?)
		(extended-spanning-intervals
		 i1 j-prime k-prime l2
		 alpha1? beta2? gamma1? delta?-prime epsilon?-prime zeta2?)
		(extended-spanning-intervals
		 i2 j-prime k-prime l1
		 alpha2? beta1? gamma2? delta?-prime epsilon?-prime zeta1?)
		(extended-spanning-intervals
		 i2 j-prime k-prime l2
		 alpha2? beta2? gamma2? delta?-prime epsilon?-prime zeta2?))))
      js)
     '()))
   is)
  '()))

(define (diamond-allen r is)
 (reduce
  append
  (map
   (lambda (i)
    (let ((i1 (extended-spanning-interval-i i))
	  (j1 (extended-spanning-interval-j i))
	  (k1 (extended-spanning-interval-k i))
	  (l1 (extended-spanning-interval-l i))
	  (alpha1? (extended-spanning-interval-alpha? i))
	  (beta1? (extended-spanning-interval-beta? i))
	  (gamma1? (extended-spanning-interval-gamma? i))
	  (delta1? (extended-spanning-interval-delta? i))
	  (epsilon1? (extended-spanning-interval-epsilon? i))
	  (zeta1? (extended-spanning-interval-zeta? i)))
     (case r
      ((=) (list i))
      ((<) (append (extended-spanning-intervals
		    (- k1 *fuzz*) infinity minus-infinity infinity
		    #f #f (and (not beta1?) epsilon1?) #t #t #t)
		   (extended-spanning-intervals
		    (- k1 *fuzz*) infinity minus-infinity infinity
		    #f #t (and (not beta1?) epsilon1?) #t #t #t)
		   (extended-spanning-intervals
		    (- k1 *fuzz*) infinity minus-infinity infinity
		    #t #f #f #t #t #t)
		   (extended-spanning-intervals
		    (- k1 *fuzz*) infinity minus-infinity infinity
		    #t #t #f #t #t #t)))
      ((>) (append (extended-spanning-intervals
		    minus-infinity infinity minus-infinity (+ j1 *fuzz*)
		    #f #f #t #t #t (and (not alpha1?) delta1?))
		   (extended-spanning-intervals
		    minus-infinity infinity minus-infinity (+ j1 *fuzz*)
		    #f #t #t #t #t #f)
		   (extended-spanning-intervals
		    minus-infinity infinity minus-infinity (+ j1 *fuzz*)
		    #t #f #t #t #t (and (not alpha1?) delta1?))
		   (extended-spanning-intervals
		    minus-infinity infinity minus-infinity (+ j1 *fuzz*)
		    #t #t #t #t #t #f)))
      ((m) (append (extended-spanning-intervals
		    (- k1 *fuzz*) (+ l1 *fuzz*) minus-infinity infinity
		    (not beta1?) #f epsilon1? zeta1? #t #t)
		   (extended-spanning-intervals
		    (- k1 *fuzz*) (+ l1 *fuzz*)  minus-infinity infinity
		    (not beta1?) #t epsilon1? zeta1? #t #t)))
      ((mi) (append (extended-spanning-intervals
		     minus-infinity infinity (- i1 *fuzz*) (+ j1 *fuzz*)
		     #f (not alpha1?) #t #t gamma1? delta1?)
		    (extended-spanning-intervals
		     minus-infinity infinity (- i1 *fuzz*) (+ j1 *fuzz*)
		     #t (not alpha1?) #t #t gamma1? delta1?)))
      ((o) (append
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ l1 *fuzz*) (- k1 *fuzz*) infinity
	     #f #f (and alpha1? gamma1?) #f #f #t)
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ l1 *fuzz*) (- k1 *fuzz*) infinity
	     #t #f #f (and beta1? zeta1?) #f #t)
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ l1 *fuzz*) (- k1 *fuzz*) infinity
	     #f #t (and alpha1? gamma1?) #f (and (not beta1?) epsilon1?) #t)
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ l1 *fuzz*) (- k1 *fuzz*) infinity
	     #t #t #f (and beta1? zeta1?) (and (not beta1?) epsilon1?) #t)))
      ((oi) (append
	     (extended-spanning-intervals
	      minus-infinity (+ j1 *fuzz*) (- i1 *fuzz*) (+ l1 *fuzz*)
	      #f #f #t #f #f (and beta1? zeta1?))
	     (extended-spanning-intervals
	      minus-infinity (+ j1 *fuzz*) (- i1 *fuzz*) (+ l1 *fuzz*)
	      #t #f #t (and (not alpha1?) delta1?) #f (and beta1? zeta1?))
	     (extended-spanning-intervals
	      minus-infinity (+ j1 *fuzz*) (- i1 *fuzz*) (+ l1 *fuzz*)
	      #f #t #t #f (and alpha1? gamma1?) #f)
	     (extended-spanning-intervals
	      minus-infinity (+ j1 *fuzz*) (- i1 *fuzz*) (+ l1 *fuzz*)
	      #t #t #t (and (not alpha1?) delta1?) (and alpha1? gamma1?) #f)))
      ((s) (append
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ j1 *fuzz*) (- k1 *fuzz*) infinity
	     alpha1? #f gamma1? delta1? #f #t)
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ j1 *fuzz*) (- k1 *fuzz*) infinity
	     alpha1? #t gamma1? delta1? (and (not beta1?) epsilon1?) #t)))
      ((si) (append (extended-spanning-intervals
		     (- i1 *fuzz*) (+ j1 *fuzz*) minus-infinity (+ l1 *fuzz*)
		     alpha1? #f gamma1? delta1? #t (and beta1? zeta1?))
		    (extended-spanning-intervals
		     (- i1 *fuzz*) (+ j1 *fuzz*) minus-infinity (+ l1 *fuzz*)
		     alpha1? #t gamma1? delta1? #t #f)))
      ((f) (append
	    (extended-spanning-intervals
	     minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) (+ l1 *fuzz*)
	     #f beta1? #t #f epsilon1? zeta1?)
	    (extended-spanning-intervals
	     minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) (+ l1 *fuzz*)
	     #t beta1? #t (and (not alpha1?) delta1?) epsilon1? zeta1?)))
      ((fi) (append (extended-spanning-intervals
		     (- i1 *fuzz*) infinity (- k1 *fuzz*) (+ l1 *fuzz*)
		     #f beta1? (and alpha1? gamma1?) #t epsilon1? zeta1?)
		    (extended-spanning-intervals
		     (- i1 *fuzz*) infinity (- k1 *fuzz*) (+ l1 *fuzz*)
		     #t beta1? #f #t epsilon1? zeta1?)))
      ((d) (append (extended-spanning-intervals
		    minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) infinity
		    #f #f #t #f #f #t)
		   (extended-spanning-intervals
		    minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) infinity
		    #t #f #t (and (not alpha1?) delta1?) #f #t)
		   (extended-spanning-intervals
		    minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) infinity
		    #f #t #t #f (and (not beta1?) epsilon1?) #t)
		   (extended-spanning-intervals
		    minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) infinity
		    #t
		    #t
		    #t
		    (and (not alpha1?) delta1?)
		    (and (not beta1?) epsilon1?)
		    #t)))
      ((di) (append (extended-spanning-intervals
		     (- i1 *fuzz*) infinity minus-infinity (+ l1 *fuzz*)
		     #f #f (and alpha1? gamma1?) #t #t (and beta1? zeta1?))
		    (extended-spanning-intervals
		     (- i1 *fuzz*) infinity minus-infinity (+ l1 *fuzz*)
		     #f #t (and alpha1? gamma1?) #t #t #f)
		    (extended-spanning-intervals
		     (- i1 *fuzz*) infinity minus-infinity (+ l1 *fuzz*)
		     #t #f #f #t #t (and beta1? zeta1?))
		    (extended-spanning-intervals
		     (- i1 *fuzz*) infinity minus-infinity (+ l1 *fuzz*)
		     #t #t #f #t #t #f)))
      (else (fuck-up)))))
   is)
  '()))

(define (invert-allen r)
 (case r
  ((=) '=)
  ((<) '>)
  ((>) '<)
  ((m) 'mi)
  ((mi) 'm)
  ((o) 'oi)
  ((oi) 'o)
  ((s) 'si)
  ((si) 's)
  ((f) 'fi)
  ((fi) 'f)
  ((d) 'di)
  ((di) 'd)
  (else (fuck-up))))

(define (span-allen js r ks)
 (reduce
  append
  (map
   (lambda (j)
    (reduce
     append
     (map
      (lambda (k)
       (span
	(intersection-of (diamond-allen r (list j)) (list k))
	(intersection-of (diamond-allen (invert-allen r) (list k)) (list j))))
      ks)
     '()))
   js)
  '()))

(define (all-intervals)
 (append (spanning-intervals minus-infinity infinity #f #f #t #t)
	 (spanning-intervals minus-infinity infinity #f #t #t #t)
	 (spanning-intervals minus-infinity infinity #t #f #t #t)
	 (spanning-intervals minus-infinity infinity #t #t #t #t)))

(define (check-arguments e n variables)
 (unless (= (length (rest e)) n) (panic "Improper expression: ~s" e))
 (map (lambda (e)
       (when (symbol? e)
	(let ((variable (memq e variables)))
	 (unless variable (panic "Unbound symbol: ~s" e)))))
      (rest e)))

(define (check-syntax e variables)
 (unless (and (list? e) (symbol? (first e)))
  (panic "Improper event logic expression: ~s" e))
 (case (first e)
  ((not) (unless (= (length e) 2)
	  (panic "Improper event logic expression: ~s" e))
	 (check-syntax (second e) variables))
  ((or) (for-each (lambda (f) (check-syntax f variables)) (rest e)))
  ((every) (unless (and (= (length e) 3)
			(list? (second e))
			(every symbol? (second e))
			(not (some null? (second e))))
	    (panic "Improper event logic expression: ~s" e))
	   (check-syntax (third e) (append (second e) variables)))
  ((some) (unless (and (= (length e) 3)
		       (list? (second e))
		       (every symbol? (second e))
		       (not (some null? (second e))))
	   (panic "Improper event logic expression: ~s" e))	
	  (check-syntax (third e) (append (second e) variables)))
  ((and) (unless (or (and (>= (length e) 3)
			  (list? (second e))
			  (subsetq? (second e)
				    '(= < > m mi o oi s si f fi d di)))
		     (and (>= (length e) 2)
			  (or (not (list? (second e)))
			      (not (subsetq?
				    (second e)
				    '(= < > m mi o oi s si f fi d di))))))
	  (panic "Improper event logic expression: ~s" e))
	 (for-each (lambda (f) (check-syntax f variables))
		   (if (and (list? (second e))
			    (subsetq? (second e)
				      '(= < > m mi o oi s si f fi d di)))
		       (rest (rest e))
		       (rest e))))
  ((begin) (unless (>= (length e) 2)
	    (panic "Improper event logic expression: ~s" e))
	   (for-each (lambda (f) (check-syntax f variables)) (rest e)))
  ((diamond) (unless (or (and (= (length e) 3)
			      (list? (second e))
			      (subsetq? (second e)
					'(= < > m mi o oi s si f fi d di)))
			 (= (length e) 2))
	      (panic "Improper event logic expression: ~s" e))
	     (check-syntax (if (= (length e) 2) (second e) (third e))
			   variables))
  ((plus) (panic "Not yet implemented"))
  ((equal?) (check-arguments e 2 variables))
  ((supported?) (check-arguments e 1 variables))
  ((supports?) (check-arguments e 2 variables))
  ((contacts?) (check-arguments e 2 variables))
  ((attached?) (check-arguments e 2 variables))
  (else (let ((aggregate-event-type
	       (find-if (lambda (aggregate-event-type)
			 (eq? (first e) (first (first aggregate-event-type))))
			(aggregate-event-types))))
	 (unless aggregate-event-type
	  (panic "Improper event logic expression: ~s" e))
	 (check-arguments
	  e (length (rest (first aggregate-event-type))) variables)
	 (check-syntax (second aggregate-event-type)
		       (rest (first aggregate-event-type)))))))

(define (objects) (enumerate (length (vector-ref *polygon-movie* 0))))

(define (evaluate-arguments e bindings)
 (map (lambda (e) (if (symbol? e) (second (assq e bindings)) e)) (rest e)))

(define (precompute-contacts!)
 (set! *contacts*
       (map-vector (lambda (polygons) (all-contacts polygons *epsilon2*))
		   *polygon-movie*)))

(define (compute-supports? i x y polygons kinematic-model)
 (and (not (= x y))
      (not (vector-ref (kinematic-model-ground? kinematic-model) y))
      (let* ((original-contacts (vector-ref *contacts* i))
	     (attached?*
	      (reflexive-transitive-closure
	       (map-n-vector
		(lambda (x)
		 (map-n-vector
		  (lambda (y)
		   (some (lambda (joint-i? joint-j? joint-w? contact)
			  (and joint-i?
			       joint-j?
			       joint-w?
			       (or (and (= (contact-i contact) x)
					(= (contact-j contact) y))
				   (and (= (contact-i contact) y)
					(= (contact-j contact) x)))))
			 (vector->list
			  (kinematic-model-joint-i? kinematic-model))
			 (vector->list
			  (kinematic-model-joint-j? kinematic-model))
			 (vector->list
			  (kinematic-model-joint-w? kinematic-model))
			 original-contacts))
		  (length polygons)))
		(length polygons))))
	     (polygons (list-remove polygons x))
	     (contacts (map (lambda (contact)
			     (make-contact (contact-p contact)
					   (contact-l1 contact)
					   (contact-l2 contact)
					   (if (> (contact-i contact) x)
					       (- (contact-i contact) 1)
					       (contact-i contact))
					   (if (> (contact-j contact) x)
					       (- (contact-j contact) 1)
					       (contact-j contact))))
			    (remove-if (lambda (contact)
					(or (= (contact-i contact) x)
					    (= (contact-j contact) x)))
				       original-contacts))))
       (define (reduce v)
	(list->vector
	 (map (lambda (contact)
	       (vector-ref
		v
		(positionp
		 (lambda (contact1 contact2)
		  (and (equal? (contact-p contact1) (contact-p contact2))
		       (equal? (contact-l1 contact1) (contact-l1 contact2))
		       (equal? (contact-l2 contact1) (contact-l2 contact2))))
		 contact
		 original-contacts)))
	      contacts)))
       ;; See if scene is unstable when X is removed and all objects except for
       ;; Y and those rigidly attached directly or indirectly to Y are
       ;; grounded.
       (not (polygon-stable?
	     polygons
	     contacts
	     (make-kinematic-model
	      (map-vector
	       (lambda (row) (vector-remove row x))
	       (vector-remove (kinematic-model-same-layer? kinematic-model) x))
	      (reduce (kinematic-model-joint-i? kinematic-model))
	      (reduce (kinematic-model-joint-j? kinematic-model))
	      (reduce (kinematic-model-joint-w? kinematic-model))
	      (vector-remove
	       ;; Ground all objects except for Y and those rigidly attached
	       ;; directly or indirectly to Y.
	       (map-n-vector
		(lambda (z)
		 (or (vector-ref (kinematic-model-ground? kinematic-model) z)
		     (not (matrix-ref attached?* y z))))
		(vector-length (kinematic-model-ground? kinematic-model)))
	       x)))))))

(define (download-contacts! contacts)
 (set! *contacts* contacts)
 #t)

(define-service 'download-contacts! download-contacts!)

(define-service 'compute-supports? compute-supports?)

(define (precompute-supports?!)
 (deresult
  (map (lambda (server) (invoke-on server 'download-contacts! *contacts*))
       *servers*))
 (set! *supports?*
       (if (null? *servers*)
	   (list->vector
	    (map
	     (lambda (x)
	      (list->vector
	       (map
		(lambda (y)
		 (map-n-vector
		  (lambda (i)
		   (compute-supports?
		    i
		    x
		    y
		    (vector-ref *polygon-movie* i)
		    (first (vector-ref *kinematic-models* i))))
		  (vector-length *polygon-movie*)))
		(objects))))
	     (objects)))
	   (deresult
	    (list->vector
	     (map
	      (lambda (x)
	       (list->vector
		(map
		 (lambda (y)
		  (map-n-vector
		   (lambda (i)
		    (invoke 'compute-supports?
			    i
			    x
			    y
			    (vector-ref *polygon-movie* i)
			    (first (vector-ref *kinematic-models* i))))
		   (vector-length *polygon-movie*)))
		 (objects))))
	      (objects)))))))

(define (precompute-contacts?!)
 (set! *contacts?*
       (list->vector
	(map
	 (lambda (x)
	  (list->vector
	   (map
	    (lambda (y)
	     (map-n-vector
	      (lambda (i)
	       (let ((x (min x y))
		     (y (max x y)))
		(and (matrix-ref (kinematic-model-same-layer?
				  (first (vector-ref *kinematic-models* i)))
				 x
				 y)
		     (polygons-touch?
		      (list-ref (vector-ref *polygon-movie* i) x)
		      (list-ref (vector-ref *polygon-movie* i) y)
		      *epsilon1*))))
	      (vector-length *polygon-movie*)))
	    (objects))))
	 (objects)))))

(define (precompute-attached?!)
 (set! *attached?*
       (list->vector
	(map (lambda (x)
	      (list->vector
	       (map (lambda (y)
		     (map-n-vector
		      (lambda (i)
		       (some (lambda (joint-i? contact)
			      (and joint-i?
				   (or (and (= (contact-i contact) x)
					    (= (contact-j contact) y))
				       (and (= (contact-i contact) y)
					    (= (contact-j contact) x)))))
			     (vector->list
			      (kinematic-model-joint-i?
			       (first (vector-ref *kinematic-models* i))))
			     (vector-ref *contacts* i)))
		      (vector-length *polygon-movie*)))
		    (objects))))
	     (objects)))))

(define (precompute!)
 (precompute-contacts!)
 (precompute-supports?!)
 (precompute-contacts?!)
 (precompute-attached?!))

(define (supported? x i)
 (not (vector-ref
       (kinematic-model-ground? (first (vector-ref *kinematic-models* i))) x)))

(define (supports? x y i) (vector-ref (matrix-ref *supports?* x y) i))

(define (contacts? x y i) (vector-ref (matrix-ref *contacts?* x y) i))

(define (attached? x y i) (vector-ref (matrix-ref *attached?* x y) i))

(define (remove-subsumed-values subsumption-predicate set)
 (let ((values '()))
  (for-each
   (lambda (new-value)
    (unless (some (lambda (old-value)
		   (subsumption-predicate old-value new-value))
		  values)
     (set! values
	   (cons new-value
		 (remove-if (lambda (old-value)
			     (subsumption-predicate new-value old-value))
			    values)))))
   set)
  values))

(define (close-under-resolution-and-subsumption
	 resolution-function subsumption-predicate set)
 (let ((values '()))
  (let loop ()
   (unless (null? set)
    (let ((new-value (first set)))
     (set! set (rest set))
     (unless (some (lambda (old-value)
		    (subsumption-predicate old-value new-value))
		   values)
      (set! values
	    (remove-if
	     (lambda (old-value) (subsumption-predicate new-value old-value))
	     values))
      (for-each
       (lambda (old-value)
	(set! set (append (resolution-function new-value old-value) set)))
       values)
      (set! values (cons new-value values))))
    (loop)))
  values))

(define (make-irredundant is)
 (close-under-resolution-and-subsumption resolution-of subsumes? is))

(define (occurrences-of-some v e bindings)
 (if (null? v)
     (occurrences-of-internal e bindings)
     (make-irredundant
      (reduce append
	      (map (lambda (object)
		    (occurrences-of-some
		     (rest v) e (cons (list (first v) object) bindings)))
		   (objects))
	      '()))))

(define (occurrences-of-every-internal v o e bindings)
 (if (null? o)
     (all-intervals)
     (make-irredundant
      (intersection-of
       (occurrences-of-internal e (cons (list v (first o)) bindings))
       (occurrences-of-every-internal v (rest o) e bindings)))))

(define (occurrences-of-every v e bindings)
 (if (null? v)
     (occurrences-of-internal e bindings)
     (make-irredundant
      (intersection-of
       (occurrences-of-every-internal (first v) (objects) e bindings)
       (occurrences-of-every (rest v) e bindings)))))

(define (occurrences-of-and rs es bindings)
 (if (null? (rest es))
     (occurrences-of-internal (first es) bindings)
     (make-irredundant
      (reduce append
	      (map (lambda (r)
		    (span-allen (occurrences-of-internal (first es) bindings)
				r
				(occurrences-of-and rs (rest es) bindings)))
		   rs)
	      '()))))

(define (upward-pulse-filter l delta)
 (let loop ((l l) (count 0) (c '()))
  (cond
   ((null? l) (reverse (append (map-n (lambda (i) (> count delta)) count) c)))
   ((first l) (loop (rest l) (+ count 1) c))
   (else
    (loop (rest l)
	  0
	  (cons #f (append (map-n (lambda (i) (> count delta)) count) c)))))))

(define (downward-pulse-filter l delta)
 (let loop ((l l) (count 0) (c '()))
  (cond
   ((null? l) (reverse (append (map-n (lambda (i) (<= count delta)) count) c)))
   ((first l)
    (loop (rest l)
	  0
	  (cons #t (append (map-n (lambda (i) (<= count delta)) count) c))))
   (else (loop (rest l) (+ count 1) c)))))

(define (pulse-filter l delta)
 (let ((l-prime (downward-pulse-filter (upward-pulse-filter l delta) delta)))
  (if (equal? l-prime l) l (pulse-filter l-prime delta))))

(define (averaging-pulse-filter l delta)
 (let ((l (list->vector l)))
  (map-n (lambda (i)
	  (let ((count 0)
		(samples 0))
	   (do ((j (- i delta) (+ j 1))) ((> j (+ i delta)))
	    (when (and (>= j 0) (< j (vector-length l)))
	     (set! samples (+ samples 1))
	     (when (vector-ref l j) (set! count (+ count 1)))))
	   (>= (* 2 count) samples)))
	 (vector-length l))))

(define (recursive-averaging-pulse-filter l delta)
 (let loop ((l l))
  (let ((new (averaging-pulse-filter l delta)))
   (if (equal? l new) l (loop new)))))

(define (maximal-spanning-intervals l)
 (let loop ((l l) (i 0) (start #f) (c '()))
  (if (null? l)
      (if start
	  (append (spanning-intervals start i #t #f #t #t) c)
	  c)
      (if (first l)
	  (loop (rest l) (+ i 1) (or start i) c)
	  (loop (rest l)
		(+ i 1)
		#f
		(if start
		    (append (spanning-intervals start i #t #f #t #t) c)
		    c))))))

(define (maximal-spanning-intervals-that p)
 (maximal-spanning-intervals
  ;; tolerance
  (averaging-pulse-filter (map-n p (vector-length *polygon-movie*)) *delta*)))

(define (occurrences-of-internal e bindings)
 (make-irredundant
  (case (first e)
   ((not) (complement-of (occurrences-of-internal (second e) bindings)))
   ((or) (reduce
	  append
	  (map (lambda (e) (occurrences-of-internal e bindings)) (rest e))
	  '()))
   ((every) (occurrences-of-every (second e) (third e) bindings))
   ((some) (occurrences-of-some (second e) (third e) bindings))
   ((and) (if (and (list? (second e))
		   (subsetq? (second e) '(= < > m mi o oi s si f fi d di)))
	      (occurrences-of-and (second e) (rest (rest e)) bindings)
	      (occurrences-of-internal `(and (=) ,@(rest e)) bindings)))
   ((begin) (occurrences-of-internal `(and (m) ,@(rest e)) bindings))
   ((diamond) (if (= (length e) 2)
		  (occurrences-of-internal
		   `(diamond (= o oi s si f fi d di) ,(second e)) bindings)
		  (reduce append
			  (map (lambda (r)
				(diamond-allen
				 r
				 (occurrences-of-internal (third e) bindings)))
			       (second e))
			  '())))
   ((equal?) (let ((arguments (evaluate-arguments e bindings)))
	      (if (= (first arguments) (second arguments))
		  (all-intervals)
		  '())))
   ((supported?)
    (let ((arguments (evaluate-arguments e bindings)))
     (maximal-spanning-intervals-that
      (lambda (frame) (supported? (first arguments) frame)))))
   ((supports?)
    (let ((arguments (evaluate-arguments e bindings)))
     (maximal-spanning-intervals-that
      (lambda (frame)
       (supports? (first arguments) (second arguments) frame)))))
   ((contacts?)
    (let ((arguments (evaluate-arguments e bindings)))
     (maximal-spanning-intervals-that
      (lambda (frame)
       (contacts? (first arguments) (second arguments) frame)))))
   ((attached?)
    (let ((arguments (evaluate-arguments e bindings)))
     (maximal-spanning-intervals-that
      (lambda (frame)
       (attached? (first arguments) (second arguments) frame)))))
   (else
    (let* ((aggregate-event-type
	    (find-if (lambda (aggregate-event-type)
		      (eq? (first e) (first (first aggregate-event-type))))
		     (aggregate-event-types)))
	   (arguments (evaluate-arguments e bindings)))
     (occurrences-of-internal
      (second aggregate-event-type)
      (append (map list (rest (first aggregate-event-type)) arguments)
	      bindings)))))))

(define (occurrences-of e)
 (check-syntax e '())
 (occurrences-of-internal e '()))

;;; This kludge is because MAP still doesn't work for nondeterministic
;;; procedures.
;;; needs work: For some reason, the patch to scrt4 is not taking effect.
(define (my-map f l) (if (null? l) l (cons (f (first l)) (my-map f (rest l)))))

(define (an-event type)
 (if *fast?*
     (let loop ((objects (objects)) (arguments '()) (n (length (rest type))))
      (if (zero? n)
	  (cons (first type) (reverse arguments))
	  (let ((argument (a-member-of objects)))
	   (loop (removeq argument objects)
		 (cons argument arguments)
		 (- n 1)))))
     (cons (first type)
	   (my-map (lambda (x) (a-member-of (objects))) (rest type)))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
