(MODULE HLIB-SPEAK)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-speak.sch")
(include "c-externals.sc")

;;; From cdslib-signals

(define-structure signal
 ;; float vector or mapped file
 samples
 ;; inexact real
 prior-sample
 ;; inexact real
 posterior-sample
 ;; string
 info-string
 ;; exact integer
 min-index
 ;; exact integer
 max-index
 ;; exact integer
 shift
 ;; exact integer
 down-sample-factor
 ;; exact integer
 up-sample-factor
 ;; exact positive integer or #F
 sample-rate)

;;; Signal accessors

(define (signal-ref signal i)
 (cond ((< i (min-index signal)) (signal-prior-sample signal))
       ((>= i (max-index signal)) (signal-posterior-sample signal))
       (else (let ((i (- i (signal-shift signal))))
	      (if (zero? (remainder i (signal-up-sample-factor signal)))
		  (let ((i (* (quotient i (signal-up-sample-factor signal))
			      (signal-down-sample-factor signal)))
			(samples (signal-samples signal)))
		   (if (pair? (rest samples))
		       (mapped-au/fea-sd-file-ref samples i)
		       (float-vector-ref samples i)))
		  0.0)))))

;;; From cdslib-float-vectors

(define-c-external (float-vector-ref-internal pointer unsigned) float
 "float_vector_ref")

(define (float-vector-ref v i)
 (when (or (negative? i) (>= i (cdr v)))
  (panic "Index to FLOAT-VECTOR-REF is out of bounds: ~s" i))
 (float-vector-ref-internal (car v) i))

;;; From cdslib-byte-io
;;; needs work: This should all be done in C to speed it up. Or at least with
;;;             Scheme->C subprimitives.

(define (low-byte exact-integer)
 (let ((result (remainder exact-integer 256)))
  (if (negative? result) (+ result 256) result)))

(define (arithmetic-shift-byte-right exact-integer)
 (inexact->exact (floor (/ exact-integer 256))))

(define (double-byte->unsigned-integer byte1 byte0) (+ (* 256 byte1) byte0))

(define (double-byte->signed-integer byte1 byte0)
 (let ((result (+ (* 256 byte1) byte0)))
  (if (>= result 32768) (- result 65536) result)))

(define (quad-byte->unsigned-integer byte3 byte2 byte1 byte0)
 (+ (* 16777216 byte3) (* 65536 byte2) (* 256 byte1) byte0))

(define (quad-byte->signed-integer byte3 byte2 byte1 byte0)
 (panic "unimplemented"))

(define (read-unsigned-little-endian-double-byte input-port)
 (let ((byte1 (read-char input-port)))
  (if (eof-object? byte1)
      byte1
      (let ((byte0 (read-char input-port)))
       (when (eof-object? byte0) (panic "File is corrupt"))
       (double-byte->unsigned-integer (char->integer byte0)
				      (char->integer byte1))))))

(define (read-unsigned-big-endian-double-byte input-port)
 (let ((byte1 (read-char input-port)))
  (if (eof-object? byte1)
      byte1
      (let ((byte0 (read-char input-port)))
       (when (eof-object? byte0) (panic "File is corrupt"))
       (double-byte->unsigned-integer (char->integer byte1)
				      (char->integer byte0))))))

(define (read-signed-little-endian-double-byte input-port)
 (let ((byte1 (read-char input-port)))
  (if (eof-object? byte1)
      byte1
      (let ((byte0 (read-char input-port)))
       (when (eof-object? byte0) (panic "File is corrupt"))
       (double-byte->signed-integer (char->integer byte0)
				    (char->integer byte1))))))

(define (read-signed-big-endian-double-byte input-port)
 (let ((byte1 (read-char input-port)))
  (if (eof-object? byte1)
      byte1
      (let ((byte0 (read-char input-port)))
       (when (eof-object? byte0) (panic "File is corrupt"))
       (double-byte->signed-integer (char->integer byte1)
				    (char->integer byte0))))))

(define (read-unsigned-little-endian-quad-byte input-port)
 (let ((byte3 (read-char input-port)))
  (if (eof-object? byte3)
      byte3
      (let ((byte2 (read-char input-port)))
       (when (eof-object? byte2) (panic "File is corrupt"))
       (let ((byte1 (read-char input-port)))
	(when (eof-object? byte1) (panic "File is corrupt"))
	(let ((byte0 (read-char input-port)))
	 (when (eof-object? byte0) (panic "File is corrupt"))
	 (quad-byte->unsigned-integer (char->integer byte0)
				      (char->integer byte1)
				      (char->integer byte2)
				      (char->integer byte3))))))))

(define (read-unsigned-big-endian-quad-byte input-port)
 (let ((byte3 (read-char input-port)))
  (if (eof-object? byte3)
      byte3
      (let ((byte2 (read-char input-port)))
       (when (eof-object? byte2) (panic "File is corrupt"))
       (let ((byte1 (read-char input-port)))
	(when (eof-object? byte1) (panic "File is corrupt"))
	(let ((byte0 (read-char input-port)))
	 (when (eof-object? byte0) (panic "File is corrupt"))
	 (quad-byte->unsigned-integer (char->integer byte3)
				      (char->integer byte2)
				      (char->integer byte1)
				      (char->integer byte0))))))))

(define (read-signed-little-endian-quad-byte input-port)
 (let ((byte3 (read-char input-port)))
  (if (eof-object? byte3)
      byte3
      (let ((byte2 (read-char input-port)))
       (when (eof-object? byte2) (panic "File is corrupt"))
       (let ((byte1 (read-char input-port)))
	(when (eof-object? byte1) (panic "File is corrupt"))
	(let ((byte0 (read-char input-port)))
	 (when (eof-object? byte0) (panic "File is corrupt"))
	 (quad-byte->signed-integer (char->integer byte0)
				    (char->integer byte1)
				    (char->integer byte2)
				    (char->integer byte3))))))))

(define (read-signed-big-endian-quad-byte input-port)
 (let ((byte3 (read-char input-port)))
  (if (eof-object? byte3)
      byte3
      (let ((byte2 (read-char input-port)))
       (when (eof-object? byte2) (panic "File is corrupt"))
       (let ((byte1 (read-char input-port)))
	(when (eof-object? byte1) (panic "File is corrupt"))
	(let ((byte0 (read-char input-port)))
	 (when (eof-object? byte0) (panic "File is corrupt"))
	 (quad-byte->signed-integer (char->integer byte3)
				    (char->integer byte2)
				    (char->integer byte1)
				    (char->integer byte0))))))))

(define (write-little-endian-double-byte object output-port)
 (unless (and (exact? object) (integer? object))
  (panic "OBJECT is not an exact integer"))
 (write-char (integer->char (low-byte object)) output-port)
 (write-char (integer->char (low-byte (arithmetic-shift-byte-right object)))
	     output-port))

(define (write-big-endian-double-byte object output-port)
 (unless (and (exact? object) (integer? object))
  (panic "OBJECT is not an exact integer"))
 (write-char (integer->char (low-byte (arithmetic-shift-byte-right object)))
	     output-port)
 (write-char (integer->char (low-byte object)) output-port))

(define (write-little-endian-quad-byte object output-port)
 (unless (and (exact? object) (integer? object))
  (panic "OBJECT is not an exact integer"))
 (write-char (integer->char (low-byte object)) output-port)
 (write-char (integer->char (low-byte (arithmetic-shift-byte-right object)))
	     output-port)
 (write-char
  (integer->char
   (low-byte
    (arithmetic-shift-byte-right (arithmetic-shift-byte-right object))))
  output-port)
 (write-char
  (integer->char
   (low-byte
    (arithmetic-shift-byte-right
     (arithmetic-shift-byte-right (arithmetic-shift-byte-right object)))))
  output-port))

(define (write-big-endian-quad-byte object output-port)
 (unless (and (exact? object) (integer? object))
  (panic "OBJECT is not an exact integer"))
 (write-char
  (integer->char
   (low-byte
    (arithmetic-shift-byte-right
     (arithmetic-shift-byte-right (arithmetic-shift-byte-right object)))))
  output-port)
 (write-char
  (integer->char
   (low-byte
    (arithmetic-shift-byte-right (arithmetic-shift-byte-right object))))
  output-port)
 (write-char (integer->char (low-byte (arithmetic-shift-byte-right object)))
	     output-port)
 (write-char (integer->char (low-byte object)) output-port))

;;; From cdslib-generic

(define (min-index signal/spectrogram)
 (cond ((signal? signal/spectrogram) (signal-min-index signal/spectrogram))
       (else (panic "Argument to MIN-INDEX is not a signal"))))

(define (max-index signal/spectrogram)
 (cond ((signal? signal/spectrogram) (signal-max-index signal/spectrogram))
       (else (panic "Argument to MAX-INDEX is not a signal"))))

(define (sample-rate signal/spectrogram)
 (cond
  ((signal? signal/spectrogram) (signal-sample-rate signal/spectrogram))
  (else (panic "Argument to SAMPLE-RATE is not a signal"))))

;;; From cdslib-mapped-files

;;; The FPI does not check that an argument passed to an unsigned parameter
;;; is not negative. Thus we do that. But there is no upper bounds checking.

;;; FPI to Unix mmap

(define-c-external (c-file-length-internal pointer) unsigned "file_length")

(define (c-file-length pointer) (c-file-length-internal pointer))

(define-c-external (make-mapped-file-internal pointer unsigned) pointer
 "make_mapped_file")

(define (make-mapped-file pointer unsigned)
 (make-mapped-file-internal pointer unsigned))

(define-c-external (free-mapped-file-internal pointer unsigned) void
 "free_mapped_file")

(define (free-mapped-file pointer unsigned)
 (free-mapped-file-internal pointer unsigned)
 #f)

(define-c-external (mapped-file-ref-internal pointer unsigned) unsigned
 "mapped_file_ref")

(define (mapped-file-ref pointer unsigned)
 (mapped-file-ref-internal pointer unsigned))

;;; Mapped AU/FEA_SD files

(define (make-mapped-big-endian-fea-sd-file pathname)
 (let ((v (call-with-input-file pathname
	   (lambda (input-port)
	    (let* ((n (c-file-length (port->stdio-file input-port)))
		   (p (make-mapped-file (port->stdio-file input-port) n)))
	     (list p
		   n
		   double-byte->signed-integer
		   (quad-byte->unsigned-integer (mapped-file-ref p 8)
						(mapped-file-ref p 9)
						(mapped-file-ref p 10)
						(mapped-file-ref p 11))))))))
  (when-unreferenced
   v (lambda (v) (free-mapped-file (first v) (second v)) #f))
  v))

(define (make-mapped-little-endian-fea-sd-file pathname)
 (let ((v (call-with-input-file pathname
	   (lambda (input-port)
	    (let* ((n (c-file-length (port->stdio-file input-port)))
		   (p (make-mapped-file (port->stdio-file input-port) n)))
	     (list p
		   n
		   (lambda (byte1 byte0)
		    (double-byte->signed-integer byte0 byte1))
		   (quad-byte->unsigned-integer (mapped-file-ref p 11)
						(mapped-file-ref p 10)
						(mapped-file-ref p 9)
						(mapped-file-ref p 8))))))))
  (when-unreferenced
   v (lambda (v) (free-mapped-file (first v) (second v)) #f))
  v))

(define (mapped-au/fea-sd-file-ref v i)
 (when (negative? i)
  (panic "Index to MAPPED-AU/FEA-SD-FILE-REF is out of bounds: ~s" i))
 (exact->inexact ((third v)
		  (mapped-file-ref (first v) (+ (* 2 i) (fourth v)))
		  (mapped-file-ref (first v) (+ (* 2 i) (fourth v) 1)))))

;;; From cdslib-input

(define (big-endian-fea? pathname)
 ;; needs work: The calls READ-UNSIGNED-BIG-ENDIAN-QUAD-BYTE can fail if the
 ;;             file is corrupt.
 (call-with-input-file pathname
  (lambda (input-port)
   (and (not (eof-object? (read-unsigned-big-endian-quad-byte input-port)))
	(not (eof-object? (read-unsigned-big-endian-quad-byte input-port)))
	(not (eof-object? (read-unsigned-big-endian-quad-byte input-port)))
	(not (eof-object? (read-unsigned-big-endian-quad-byte input-port)))
	(equal? (read-unsigned-big-endian-quad-byte input-port) 27162)))))

(define (little-endian-fea? pathname)
 ;; needs work: The calls READ-LITTLE-BIG-ENDIAN-QUAD-BYTE can fail if the
 ;;             file is corrupt.
 (call-with-input-file pathname
  (lambda (input-port)
   (and (not (eof-object? (read-unsigned-little-endian-quad-byte input-port)))
	(not (eof-object? (read-unsigned-little-endian-quad-byte input-port)))
	(not (eof-object? (read-unsigned-little-endian-quad-byte input-port)))
	(not (eof-object? (read-unsigned-little-endian-quad-byte input-port)))
	(equal? (read-unsigned-little-endian-quad-byte input-port) 27162)))))

(define (mapped-file->signal pathname)
 (let ((pathname (default-extension pathname "sd")))
  (define (mapped-fea-sd-file->signal read-unsigned-quad-byte
				      make-mapped-fea-sd-file)
   ;; This is hardwired to read FEA-SD files with a single SHORT feature.
   ;; needs work: The calls to READ-UNSIGNED-QUAD-BYTE can fail if the file is
   ;;             corrupt.
   ;; needs work: We don't read the info-string.
   (call-with-input-file pathname
    (lambda (input-port)
     (let* ((machine-code (read-unsigned-quad-byte input-port))
	    (check-code (read-unsigned-quad-byte input-port))
	    (data-offset (read-unsigned-quad-byte input-port))
	    (record-size (read-unsigned-quad-byte input-port))
	    (check (read-unsigned-quad-byte input-port))
	    (edr (read-unsigned-quad-byte input-port))
	    (align-pad-size (read-unsigned-quad-byte input-port))
	    (foreign-header (read-unsigned-quad-byte input-port)))
      (unless (= check-code 3000)
       (panic "Can only read FEA-SD files with check code 3000"))
      (unless (= record-size 2)
       (panic "Can only read FEA-SD files with record size 2"))
      (unless (= check 27162) (fuck-up))
      (unless (zero? align-pad-size)
       (panic "Can only read FEA-SD files with align pad size 0"))
      (let* ((info-string "")
	     (sample-rate 22050)	;hardwired
	     ;; hardwired
	     (start-time (inexact->exact (round (* 0 sample-rate))))
	     (data-size
	      (- (c-file-length (port->stdio-file input-port)) data-offset))
	     (number-of-samples (quotient data-size 2)))
       (unless (even? data-size)
	(panic "Corrupt FEA-SD file; data size is odd"))
       (let ((signal (make-signal (make-mapped-fea-sd-file pathname)
				  0.0
				  0.0
				  info-string
				  start-time
				  (+ number-of-samples start-time)
				  start-time
				  1
				  1
				  sample-rate)))
	signal))))))
  (cond ((big-endian-fea? pathname)
	 (mapped-fea-sd-file->signal read-unsigned-big-endian-quad-byte
				     make-mapped-big-endian-fea-sd-file))
	((little-endian-fea? pathname)
	 (mapped-fea-sd-file->signal read-unsigned-little-endian-quad-byte
				     make-mapped-little-endian-fea-sd-file))
	(else (panic "Can only read FEA-SD files")))))

;;; From cdslib-output

(define-c-external (set-speed-and-format-internal pointer int) void
 "set_speed_and_format")

(define (set-speed-and-format pointer int)
 (set-speed-and-format-internal pointer int)
 #f)

(define (signal->/dev/dsp signal)
 (unless (sample-rate signal)
  (panic "Can't play signal with unknown sample rate"))
 (unless (let loop ((i (min-index signal)))
	  (or (not (negative? i))
	      (and (zero? (signal-ref signal i)) (loop (+ i 1)))))
  (panic "Can't write signals with non-zero prior content"))
 (unless (and (zero? (signal-posterior-sample signal))
	      (zero? (signal-prior-sample signal)))
  (panic "Can't write non-quiescent signals"))
 (call-with-output-file "/dev/dsp"
  (lambda (output-port)
   (set-speed-and-format (port->stdio-file output-port) (sample-rate signal))
   (do ((i 0 (+ i 1))) ((>= i (max-index signal)))
    (let ((sample (inexact->exact (round (signal-ref signal i)))))
     (unless (<= -32768 sample 32767) (panic "Sample out of range"))
     (write-little-endian-double-byte sample output-port))))))

;;; Variables

(define *spoken-word-cache* '())

;;; Procedures

(define (word->signal word)
 (let* ((word (format #f "~a/~a"
		      (case *language*
		       ((english) "english")
		       ((japanese) "japanese")
		       (else (fuck-up)))
		      word))
	(found (assoc word *spoken-word-cache*)))
  (if found
      (cdr found)
      (let ((signal (mapped-file->signal
		     (format #f "/home/qobi/howard/audio/~a.sd" word))))
       (set! *spoken-word-cache* (cons (cons word signal) *spoken-word-cache*))
       signal))))

(define (speak string)
 (for-each signal->/dev/dsp
	   (map (lambda (word) (word->signal word)) (fields string))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
