(MODULE HLIB-TRACK)
;;; LaHaShem HaAretz U'Mloah

;;; TTMTTD
;;; 1. Introduce Richard bug into cpt

(include "QobiScheme.sch")
(include "hlib-track.sch")
(include "c-externals.sc")

;;; Structures

(define-structure link data position predecessor successor synthetic?)

(define-structure coloured-ellipse colour ellipse)

;;; Variables

(define *max-x* 320)
(define *max-y* 240)
(define *max-t* half-pi)
;;; needs work: *max-a* and *max-b* aren't hard limits
(define *max-a* 120)
(define *max-b* 120)
(define *max-ap* 76800)

;;; Procedures

;;; Internal Correspondence

(define (make-empty-link data) (make-link data #f #f #f #f))

(define (coloured-ellipse->vector coloured-ellipse)
 (let ((ellipse (coloured-ellipse-ellipse coloured-ellipse)))
  (vector (coloured-ellipse-colour coloured-ellipse)
	  (ellipse-x0 ellipse)
	  (ellipse-y0 ellipse)
	  (ellipse-t0 ellipse)
	  (ellipse-a ellipse)
	  (ellipse-b ellipse))))

(define (weighted-squared-distance
	 v1 v2 weight-vector max-vector distance-functions)
 (/ (reduce-vector
     +
     (map-vector (lambda (v1 v2 weight max-value distance)
		  (* (sqr (/ (distance v1 v2) max-value)) weight))
		 v1 v2 weight-vector max-vector distance-functions)
     0.0)
    (reduce-vector + weight-vector 0.0)))

(define (weighted-coloured-ellipse-squared-distance
	 coloured-ellipse1 coloured-ellipse2 weight-vector)
 (weighted-squared-distance
  (coloured-ellipse->vector coloured-ellipse1)
  (coloured-ellipse->vector coloured-ellipse2)
  weight-vector
  (vector (/ *max-hue* 2) *max-x* *max-y* *max-t* *max-a* *max-b*)
  (vector (lambda (hue1 hue2)
	   (if (and hue1 hue2)
	       (separation hue1 hue2)
	       (if (or hue1 hue2) (/ *max-hue* 2) 0.0)))
	  -
	  -
	  (lambda (t1 t2)
	   (min (angle-separation t1 t2) (angle-separation t1 (- t2 pi))))
	  -
	  -)))

(define (weighted-ellipse-squared-distance
	 ellipse1 ellipse2
	 x-weight y-weight t-weight a-weight b-weight)
 (/ (+ (* (sqr (/ (- (ellipse-x0 ellipse1) (ellipse-x0 ellipse2))
		  *max-x*))
	  x-weight)
       (* (sqr (/ (- (ellipse-y0 ellipse1) (ellipse-y0 ellipse2))
		  *max-y*))
	  y-weight)
       (* (sqr (/ (min (angle-separation
			(ellipse-t0 ellipse1) (ellipse-t0 ellipse2))
		       (angle-separation
			(ellipse-t0 ellipse1) (- (ellipse-t0 ellipse2) pi)))
		  *max-t*))
	  t-weight)
       (* (sqr (/ (- (ellipse-a ellipse1) (ellipse-a ellipse2))
		  *max-a*))
	  a-weight)
       (* (sqr (/ (- (ellipse-b ellipse1) (ellipse-b ellipse2))
		  *max-b*))
	  b-weight))
    (+ x-weight y-weight t-weight a-weight b-weight)))

(define (weighted-polygon-squared-distance
	 polygon1 polygon2
	 x-weight y-weight a-weight)
 (/ (+ (* (sqr (/ (- (x (centroid polygon1)) (x (centroid polygon2))) *max-x*))
	  x-weight)
       (* (sqr (/ (- (y (centroid polygon1)) (y (centroid polygon2))) *max-y*))
	  y-weight)
       (* (sqr (/ (- (debugging-area-of-convex-polygon polygon1)
		     (debugging-area-of-convex-polygon polygon2))
		  *max-ap*))
	  a-weight))
    (+ x-weight y-weight a-weight)))

(define (centroid polygon)
 (mean (if (vector? polygon) polygon (list->vector polygon))))

(define (chain link)
 (if (link-successor link)
     (cons link (chain (link-successor link)))
     (list link)))

(define (correspondence sequence fill cost threshold)
 (let* ((cost (lambda (link1 link2)
	       (cost (link-data link1) (link-data link2))))
	(n (vector-length sequence))
	(sequence (map-vector (lambda (element) (map make-empty-link element))
			      sequence))
	(links (reduce-vector append sequence '())))
  (for-each-n (lambda (i)
	       (for-each (lambda (link) (set-link-position! link i))
			 (vector-ref sequence i)))
	      n)
  ;; For each link
  (for-each
   (lambda (link)
    (when (and
	   ;; that has no predecessor
	   (not (link-predecessor link))
	   ;; that isn't in the first frame
	   (not (zero? (link-position link)))
	   ;; such that the previous frame has some ellipses
	   (not (null? (vector-ref sequence (- (link-position link) 1)))))
     ;; Two links e1 and e2 are paired if the distance between e1 and e2 is
     ;; below a threshold and if e2 is the closest in its position to e1 and e1
     ;; is the closest in its position to e2.
     (let ((predecessor
	    (minp (lambda (predecessor1 predecessor2)
		   (< (cost predecessor1 link) (cost predecessor2 link)))
		  (vector-ref sequence (- (link-position link) 1)))))
      (when (and (< (cost predecessor link) threshold)
		 (every (lambda (sibling)
			 (>= (cost predecessor sibling)
			     (cost predecessor link)))
			(vector-ref sequence (link-position link))))
       ;; This symmetric criterion should mean that a greedy algorithm is
       ;; optimial.
       (when (link-successor predecessor) (fuck-up))
       (set-link-predecessor! link predecessor)
       (set-link-successor! predecessor link)))))
   links)
  (let ((chains (map chain (remove-if link-predecessor links))))
   (when #f				;debugging
    (format #t "Before merging~%")
    (for-each (lambda (chain)
	       (write (list (link-position (first chain))
			    (link-position (last chain))))
	       (newline))
	      chains)
    (newline))
   ;; This greedy algorithm for merging two chains can be suboptimal however.
   ;; Repeatedly merge two chains
   (let loop ()
    (let ((best-chain1 #f)
	  (best-chain2 #f))
     (for-effects
      (let ((chain1 (a-member-of chains))
	    (chain2 (a-member-of chains)))
       ;; if they are within fill distance from each other
       (when (and
	      (< (- (link-position (first chain2)) fill 2)
		 (link-position (last chain1))
		 ;; this allows chains to overlap
		 (+ (link-position (first chain2)) fill))
	      ;; needed because chains can now overlap
	      (< (link-position (last chain1))
		 (link-position (last chain2)))
	      ;; needed because chains can now overlap
	      (< (link-position (first chain1))
		 (link-position (first chain2))))
	;; among those that meet this criterion first merge the longest two
	;; and only if the cost is below a threshold
	(let ((this-cost (cost (last chain1) (first chain2))))
	 (when #f			;debugging
	  (write (list 'potential-merge
		       (list this-cost threshold)
		       (link-position (first chain1))
		       (link-position (last chain1))
		       (link-position (first chain2))
		       (link-position (last chain2))))
	  (newline))
	 (when (and (< this-cost threshold)
		    (or (not best-chain1)
			(> (+ (length chain1) (length chain2))
			   (+ (length best-chain1) (length best-chain2)))))
	  (set! best-chain1 chain1)
	  (set! best-chain2 chain2))))))
     (when best-chain1
      (when #f				;debugging
       (write (list 'merging
		    (link-position (first best-chain1))
		    (link-position (last best-chain1))
		    (link-position (first best-chain2))
		    (link-position (last best-chain2))))
       (newline))
      (set! chains
	    (cons
	     (append
	      best-chain1
	      (map-n (lambda (i) (first best-chain2))
		     (- (link-position (first best-chain2))
			(link-position (last best-chain1))
			1))
	      ;; needed because chains can now overlap
	      (list-tail best-chain2
			 (max 0
			      (+ 1
				 (- (link-position (last best-chain1))
				    (link-position (first best-chain2)))))))
	     (remq! best-chain2 (remq! best-chain1 chains))))
      (loop))))
   (when #f				;debugging
    (format #t "After merging~%")
    (for-each (lambda (chain)
	       (write (list (link-position (first chain))
			    (link-position (last chain))))
	       (newline))
	      chains)
    (newline))
   (let ((chains
	  (map (lambda (chain) (map link-data chain))
	       (removeq
		#f
		(map (lambda (chain)
		      (if (and (<= (link-position (first chain)) fill)
			       (>= (link-position (last chain))
				   (- n fill 1)))
			  (append (map-n (lambda (i) (first chain))
					 (link-position (first chain)))
				  chain
				  (map-n (lambda (i) (last chain))
					 (- n (link-position (last chain)) 1)))
			  #f))
		     chains)))))
    (when #f				;debugging
     (write (length chains))
     (newline)
     (newline))
    (if (null? chains)
	(make-vector n '())
	(map-vector vector->list
		    (transpose (list->vector (map list->vector chains)))))))))

(define (internal-correspondence ellipse-movie
				 fill
				 x-weight
				 y-weight
				 t-weight
				 a-weight
				 b-weight
				 threshold)
 (correspondence ellipse-movie
		 fill
		 (lambda (e1 e2)
		  (weighted-ellipse-squared-distance
		   e1 e2 x-weight y-weight t-weight a-weight b-weight))
		 threshold))

(define (polygon-correspondence polygon-movie
				fill
				x-weight
				y-weight
				a-weight
				threshold)
 (correspondence polygon-movie
		 fill
		 (lambda (p1 p2)
		  (weighted-polygon-squared-distance
		   p1 p2 x-weight y-weight a-weight))
		 threshold))

(define (colour-correspondence ellipse-movie
			       colour-movie
			       fill
			       weight-vector
			       threshold)
 (correspondence
  (map-vector
   (lambda (colours ellipses) (map make-coloured-ellipse colours ellipses))
   colour-movie ellipse-movie)
  fill
  (lambda (e1 e2)
   (weighted-coloured-ellipse-squared-distance e1 e2 weight-vector))
  threshold))

;;; Some Generic Procedures

(define (separation x y)
 (radians->degrees
  (angle-separation (degrees->radians x) (degrees->radians y))))

(define (hue-mu hues)
 (let* ((mu (radians->degrees (mean-angle (map degrees->radians hues))))
	(mu (inexact->exact (round (if (negative? mu) (+ mu 360.0) mu)))))
  (if (= mu 360) 0 mu)))

(define (mu x)
 (k*v (/ (length x)) (reduce v+ x (make-vector (vector-length (first x)) 0))))

(define (sigma x)
 (let ((mu (mu x)))
  (k*m (/ (length x))
       (reduce
	m+
	(map (lambda (x) (self-outer-product * (v- x mu))) x)
	(make-matrix (vector-length (first x)) (vector-length (first x)) 0)))))

(define (points->ellipse points)
 (let ((mu (mu points))
       (sigma (sigma points)))
  (create-ellipse (vector-ref mu 0)
		  (vector-ref mu 1)
		  (matrix-ref sigma 0 0)
		  (matrix-ref sigma 0 1)
		  (matrix-ref sigma 1 1))))

(define (pixels-with-similar-hue hue hues pixels cd-distance-threshold)
 (let ((index (positionv hue hues)))
  (remove-if-not
   (lambda (pixel)
    (and (<= (separation (pixel-hue pixel) hue) cd-distance-threshold)
	 (= index (classify (pixel-hue pixel) hues separation))))
   pixels)))

;;; Special-Purpose Clusterers

(define (c-histogram-clusterer
	 hues quanta pixel-mass-threshold hue-separation-threshold)
 ;; Hardwired to cluster angles in degrees
 (let* ((size (length hues))
	(pixel-hues (make-c-int-vector size))
	(bins (make-c-int-vector size))
	(new-hues (make-c-int-vector (+ (quantize *max-hue* quanta) 1))))
  (for-each (lambda (n hue)
	     (c-int-vector-set! pixel-hues n hue)
	     (c-int-vector-set! bins n 1)
	     #f)
	    (enumerate size)
	    hues)
  (let ((hue-number (c-histogram-clusterer-internal
		     pixel-hues size quanta pixel-mass-threshold
		     hue-separation-threshold bins new-hues)))
   (map-n (lambda (n) (c-int-vector-ref new-hues n)) hue-number))))

(define (hue-histogram-clusterer
	 hues quanta pixel-mass-threshold hue-separation-threshold)
 (histogram-clusterer hues
		      pixel-mass-threshold
		      hue-separation-threshold
		      hue-mu
		      (lambda (x) (quantize x quanta))
		      identity
		      separation))

(define (cd-histogram-clusterer
	 cds c1-quanta c2-quanta pixel-mass-threshold cd-distance-threshold)
 (histogram-clusterer cds
		      pixel-mass-threshold
		      cd-distance-threshold
		      mu
		      (lambda (x)
		       (vector (quantize (vector-ref x 0) c1-quanta)
			       (quantize (vector-ref x 1) c2-quanta)))
		      identity
		      distance))

;;; Some PNM operators

(define (subsample-vector v subsample)
 (list->vector (map-n (lambda (n) (vector-ref v (* n subsample)))
		      (quotient (vector-length v) subsample))))

(define (pixel->rgb pixel)
 (vector (pixel-red pixel) (pixel-green pixel) (pixel-blue pixel)))

(define (pixel->cd pixel)
 (let ((i (max (+ (pixel-red pixel) (pixel-green pixel) (pixel-blue pixel))
	       1)))
  (vector (inexact->exact (* *max-red* (/ (pixel-red pixel) i)))
	  (inexact->exact (* *max-green* (/ (pixel-green pixel) i))))))

(define (distance-clusterer cd-mus pixels)
 (let ((pixel-cd-mu-pairs
	(map (lambda (pixel)
	      (list pixel
		    (let ((best-cd-mu #f)
			  (best-distance infinity))
		     (for-each
		      (lambda (cd-mu)
		       (let ((distance
			      (distance-squared cd-mu (pixel->cd pixel))))
			(when (< distance best-distance)
			 (set! best-cd-mu cd-mu)
			 (set! best-distance distance))))
		      cd-mus)
		     best-cd-mu)))
	     pixels)))
  (map (lambda (cd-mu)
	(map first
	     (remove-if-not
	      (lambda (pixel-cd-mu-pair) (eq? (second pixel-cd-mu-pair) cd-mu))
	      pixel-cd-mu-pairs)))
       cd-mus)))

(define (background ppm left?)
 (map mu
      ;; needs work: This REMOVE-IF is here because of a bug in EM-CLUSTERER.
      (remove-if
       null?
       (em-clusterer
	(map-vector
	 rgb->cd (list->vector (subsample-ppm ppm 16 (if left? 16 0))))
	;; note: changed EMS-CONVERGENCE-TOLERANCE from 0.01 to 0.1
	'#(1.0 1.0) 1.0 0.01 0.1))))

(define (hand background background-threshold ppm left?)
 (map mu
      ;; needs work: This REMOVE-IF is here because of a bug in EM-CLUSTERER.
      (remove-if
       null?
       (em-clusterer
	(map-vector
	 rgb->cd
	 (list->vector
	  (subsample-ppm-that
	   (lambda (rgb)
	    (every
	     (lambda (background-cd)
	      (> (distance (rgb->cd rgb) background-cd) background-threshold))
	     background))
	   ppm
	   16
	   (if left? 16 0))))
	;; note: changed EMS-CONVERGENCE-TOLERANCE from 0.01 to 0.1
	'#(1.0 1.0) 1.0 0.01 0.1))))

(define (object background background-threshold ppm left?)
 (remove-duplicates
  (subsample-ppm-that
   (lambda (rgb)
    (every (lambda (background-cd)
	    (> (distance (rgb->cd rgb) background-cd) background-threshold))
	   background))
   ppm
   1
   (if left? 16 0))))

;;; Legacy Code

(define (old-normal-flow-magnitude pgm1 pgm2 epsilon sigma motion-sensitivity)
 ;; note: This code deliberately has the `Richard Bug.'
 (unless (and (pgm? pgm1)
	      (pgm? pgm2)
	      (eq? (pgm-raw? pgm1) (pgm-raw? pgm2))
	      (= (pnm-width pgm1) (pnm-width pgm2))
	      (= (pnm-height pgm1) (pnm-height pgm2)))
  (panic "Arguments to OLD-NORMAL-FLOW-MAGNITUDE are not matching PGMs"))
 (let* ((width (pnm-width pgm1))
	(height (pnm-height pgm1))
	(e1 (pgm-grey (pgm-smooth pgm1 sigma)))
	(e2 (pgm-grey (pgm-smooth pgm2 sigma)))
	(m (make-matrix height width 0))
	(maximum motion-sensitivity))
  (do ((i 0 (+ i 1))) ((= i (- height 1)))
   (do ((j 0 (+ j 1))) ((= j (- width 1)))
    (let* ((ex (/ (- (+ (matrix-ref e1 (+ i 1) j)
			(matrix-ref e1 (+ i 1) (+ j 1))
			(matrix-ref e2 (+ i 1) j)
			(matrix-ref e2 (+ i 1) (+ j 1)))
		     (+ (matrix-ref e1 i j)
			(matrix-ref e1 i (+ j 1))
			(matrix-ref e2 i j)
			(matrix-ref e2 i (+ j 1))))
		  4.0))
	   (ey (/ (- (+ (matrix-ref e1 i (+ j 1))
			(matrix-ref e1 (+ i 1) (+ j 1))
			(matrix-ref e2 i (+ j 1))
			(matrix-ref e2 (+ i 1) (+ j 1)))
		     (+ (matrix-ref e1 i j)
			(matrix-ref e1 (+ i 1) j)
			(matrix-ref e2 i j)
			(matrix-ref e2 (+ i 1) j)))
		  4.0))
	   (et (/ (- (+ (matrix-ref e2 i j)
			(matrix-ref e2 i (+ j 1))
			(matrix-ref e2 (+ i 1) j)
			(matrix-ref e2 (+ i 1) (+ j 1)))
		     (+ (matrix-ref e1 i j)
			(matrix-ref e1 i (+ j 1))
			(matrix-ref e1 (+ i 1) j)
			(matrix-ref e1 (+ i 1) (+ j 1))))
		  4.0))
	   (l (sqrt (+ (sqr ex) (sqr ey))))
	   (n (if (< l epsilon) 0.0 (abs et))))
     (set! maximum (max maximum n))
     (matrix-set! m i j n))))
  (do ((i 0 (+ i 1))) ((= i (- height 2)))
   (do ((j 0 (+ j 1))) ((= j (- width 2)))
    (matrix-set! m i j
		 (inexact->exact (* 255 (/ (matrix-ref m i j) maximum))))))
  (make-pgm (pgm-raw? pgm1) 255 m)))

;;; Generic Tracker Wrapper

(define (track pathname video-type track-movie? process-frame)
 (let* ((last-pgm #f)
	(last-moving-ellipses #f)
	(second-frame-moving-ellipses #f)
	(ellipse-movie '())
	(video-input-port (open-video-input-file pathname video-type))
	(video-output-port
	 (if track-movie?
	     (open-video-output-file
	      (string-append (strip-extension pathname) "-tracked")
	      video-type
	      5000)
	     #f)))
  (let loop ()
   (let ((ppm (read-ppm-from-video-input-port video-input-port)))
    (when ppm
     (let ((triple (process-frame ppm last-pgm last-moving-ellipses)))
      (when (and last-pgm (not second-frame-moving-ellipses))
       (set! second-frame-moving-ellipses (third triple)))
      (set! last-pgm (first triple))
      (set! ellipse-movie (cons (second triple) ellipse-movie))
      (set! last-moving-ellipses (third triple)))
     (when track-movie? (write-pnm-to-video-output-port ppm video-output-port))
     (loop))))
  (when track-movie? (close-video-output-port video-output-port))
  (close-video-input-port video-input-port)
  (let ((ellipse-movie (reverse ellipse-movie)))
   (list->vector
    (cons (append second-frame-moving-ellipses (first ellipse-movie))
	  (rest ellipse-movie))))))

;;; HST

(define (hst-process-frame ppm
			   last-pgm
			   last-moving-ellipses
			   saturation-threshold
			   value-threshold
			   quanta
			   colour-pixel-mass-threshold
			   hue-separation-threshold
			   colour-proximity-threshold
			   epsilon
			   sigma
			   motion-threshold
			   motion-sensitivity
			   motion-pixel-mass-threshold
			   motion-proximity-threshold
			   track-movie?
			   old?)
 (let* ((width (pnm-width ppm))
	(height (pnm-height ppm))
	(coloured-pbm
	 (ppm->coloured-pbm ppm saturation-threshold value-threshold))
	(pgm (ppm->pgm ppm))
	(coloured-pixels (pbm-pixels coloured-pbm ppm))
	(hues (c-histogram-clusterer (map pixel-hue coloured-pixels)
				     quanta
				     colour-pixel-mass-threshold
				     hue-separation-threshold))
	(hues (sort (remove-if
		     (lambda (hue)
		      (< (length
			  (pixels-with-similar-hue
			   hue hues coloured-pixels hue-separation-threshold))
			 colour-pixel-mass-threshold))
		     hues)
		    <
		    identity))
	(coloured-clusters
	 (remove-if
	  (lambda (cluster) (< (length cluster) colour-pixel-mass-threshold))
	  (reduce
	   append
	   (map (lambda (hue)
		 (let ((bitmap (make-matrix height width #f)))
		  (for-each
		   (lambda (p) (matrix-set! bitmap (pixel-y p) (pixel-x p) #t))
		   (pixels-with-similar-hue
		    hue hues coloured-pixels hue-separation-threshold))
		  (c-pbm-proximity-clusterer
		   (make-pbm #t bitmap) colour-proximity-threshold)))
		hues)
	   '())))
	(moving-pbm
	 (if last-pgm
	     (pbm-and (if old?
			  (pgm->pbm
			   (old-normal-flow-magnitude
			    last-pgm pgm epsilon sigma motion-sensitivity)
			   motion-threshold)
			  (c-threshold-normal-flow-magnitude
			   last-pgm pgm epsilon sigma motion-threshold))
		      (pbm-not coloured-pbm))
	     (pbm-constant width height #f)))
	(moving-pixels (pbm-pixels moving-pbm ppm))
	(moving-clusters
	 (remove-if
	  (lambda (cluster) (< (length cluster) motion-pixel-mass-threshold))
	  (c-pbm-proximity-clusterer moving-pbm motion-proximity-threshold)))
	(clusters (append coloured-clusters moving-clusters)))
  (when track-movie?
   (inject-pnm! (pgm-constant width height *max-grey*) ppm)
   (for-each (lambda (hue)
	      (overwrite-pixels-as!
	       (pixels-with-similar-hue
		hue hues coloured-pixels hue-separation-threshold)
	       ppm
	       (hsv->rgb (vector hue *max-saturation* *max-value*))))
	     hues)
   (overwrite-pixels-as! moving-pixels ppm (vector 0 0 0)))
  (if old?
      (list pgm
	    (map points->ellipse clusters)
	    (map points->ellipse moving-clusters))
      (list pgm
	    (append (map points->ellipse clusters)
		    (if (null? moving-clusters)
			(or last-moving-ellipses '())
			'()))
	    (or (and (null? moving-clusters) last-moving-ellipses)
		(map points->ellipse moving-clusters))))))

(define (hst pathname
	     video-type
	     saturation-threshold
	     value-threshold
	     quanta
	     colour-pixel-mass-threshold
	     hue-separation-threshold
	     colour-proximity-threshold
	     epsilon
	     sigma
	     motion-threshold
	     motion-sensitivity
	     motion-pixel-mass-threshold
	     motion-proximity-threshold
	     track-movie?
	     old?)
 (let ((frame-number 0)
       (number-of-frames (video-file-length pathname video-type)))
  (setup-proximity-clusterer!
   (max motion-proximity-threshold colour-proximity-threshold))
  (track pathname
	 video-type
	 track-movie?
	 (lambda (ppm last-pgm last-moving-ellipses)
	  (when *window*
	   (message (format #f "Frame ~s out of ~s (~s%)"
			    frame-number
			    number-of-frames
			    (inexact->exact
			     (/ (* frame-number 100) number-of-frames))))
	   (set! frame-number (+ frame-number 1)))
	  (hst-process-frame
	   ppm
	   last-pgm
	   last-moving-ellipses
	   saturation-threshold
	   value-threshold
	   quanta
	   colour-pixel-mass-threshold
	   hue-separation-threshold
	   colour-proximity-threshold
	   epsilon
	   sigma
	   motion-threshold
	   motion-sensitivity
	   motion-pixel-mass-threshold
	   motion-proximity-threshold
	   track-movie?
	   old?)))))

;;; CPT

(define (cpt-process-frame ppm
			   last-pgm
			   background-threshold
			   pixel-mass-threshold
			   proximity-threshold
			   epsilon
			   sigma
			   motion-threshold
			   c1-quanta
			   c2-quanta
			   cd-distance-threshold
			   track-movie?
			   left?
			   background)
 (let* ((width (pnm-width ppm))
	(height (pnm-height ppm))
	(pgm (ppm->pgm ppm))
	(pbm (pbm-or
	      (make-pbm
	       #t
	       (map-vector
		(lambda (red-row green-row blue-row)
		 (map-vector
		  (lambda (red green blue)
		   (every (lambda (background)
			   (>= (distance (rgb->cd (vector red green blue))
					 background)
			       background-threshold))
			  background))
		  red-row green-row blue-row))
		(ppm-red ppm) (ppm-green ppm) (ppm-blue ppm)))
	      (if last-pgm
		  (c-threshold-normal-flow-magnitude
		   last-pgm pgm epsilon sigma motion-threshold)
		  (pbm-constant width height #f))))
	(pbm (if left?
		 (pbm-and pbm (pbm-left-vertical-stripe width height 16))
		 pbm))
	(pixels (pbm-pixels pbm ppm))
	(cd-mus (map (lambda (cd-mu) (map-vector inexact->exact cd-mu))
		     (cd-histogram-clusterer (map pixel->cd pixels)
					     c1-quanta
					     c2-quanta
					     pixel-mass-threshold
					     cd-distance-threshold)))
	(clusters
	 (remove-if
	  (lambda (cluster) (< (length cluster) pixel-mass-threshold))
	  (reduce
	   append
	   (map (lambda (cd-mu)
		 (let ((bitmap (make-matrix height width #f)))
		  (for-each
		   (lambda (p)
		    (matrix-set! bitmap (pixel-y p) (pixel-x p) #t))
		   (similar-xs cd-mu cd-mus pixels cd-distance-threshold
			       pixel->cd distance))
		  (c-pbm-proximity-clusterer
		   (make-pbm #t bitmap) proximity-threshold)))
		cd-mus)
	   '()))))
  (when track-movie?
   (inject-pnm! (pgm-constant width height *max-grey*) ppm)
   (for-each (lambda (cd-mu)
	      (overwrite-pixels-as!
	       (similar-xs
		cd-mu cd-mus pixels cd-distance-threshold pixel->cd distance)
	       ppm
	       (vector (vector-ref cd-mu 0)
		       (vector-ref cd-mu 1)
		       (- (max *max-red* *max-green* *max-blue*)
			  (vector-ref cd-mu 0)
			  (vector-ref cd-mu 1)))))
	     cd-mus))
  (list pgm (map points->ellipse clusters) '())))

(define (cpt pathname
	     video-type
	     background-threshold
	     pixel-mass-threshold
	     proximity-threshold
	     epsilon
	     sigma
	     motion-threshold
	     c1-quanta
	     c2-quanta
	     cd-distance-threshold
	     track-movie?
	     left?
	     background)
 (let ((frame-number 0)
       (number-of-frames (video-file-length pathname video-type)))
  (setup-proximity-clusterer! proximity-threshold)
  (track pathname
	 video-type
	 track-movie?
	 (lambda (ppm last-pgm last-moving-ellipses)
	  (when *window*
	   (message (format #f "Frame ~s out of ~s (~s%)"
			    frame-number
			    number-of-frames
			    (inexact->exact
			     (/ (* frame-number 100) number-of-frames))))
	   (set! frame-number (+ frame-number 1)))
	  (cpt-process-frame ppm
			     last-pgm
			     background-threshold
			     pixel-mass-threshold
			     proximity-threshold
			     epsilon
			     sigma
			     motion-threshold
			     c1-quanta
			     c2-quanta
			     cd-distance-threshold
			     track-movie?
			     left?
			     background)))))

;;; CPT-HUE

(define (cpt-hue-process-frame ppm
			       last-pgm
			       background-threshold
			       pixel-mass-threshold
			       proximity-threshold
			       epsilon
			       sigma
			       motion-threshold
			       c1-quanta
			       c2-quanta
			       cd-distance-threshold
			       track-movie?
			       left?
			       background)
 (let* ((width (pnm-width ppm))
	(height (pnm-height ppm))
	(pgm (ppm->pgm ppm))
	(pbm (pbm-or
	      (make-pbm
	       #t
	       (map-vector
		(lambda (red-row green-row blue-row)
		 (map-vector
		  (lambda (red green blue)
		   (every (lambda (background)
			   (>= (distance (rgb->cd (vector red green blue))
					 background)
			       background-threshold))
			  background))
		  red-row green-row blue-row))
		(ppm-red ppm) (ppm-green ppm) (ppm-blue ppm)))
	      (if last-pgm
		  (c-threshold-normal-flow-magnitude
		   last-pgm pgm epsilon sigma motion-threshold)
		  (pbm-constant width height #f))))
	(pbm (if left?
		 (pbm-and pbm (pbm-left-vertical-stripe width height 16))
		 pbm))
	(pixels (pbm-pixels pbm ppm))
	(cd-mus (map (lambda (cd-mu) (map-vector inexact->exact cd-mu))
		     (cd-histogram-clusterer (map pixel->cd pixels)
					     c1-quanta
					     c2-quanta
					     pixel-mass-threshold
					     cd-distance-threshold)))
	(clusters
	 (remove-if
	  (lambda (cluster) (< (length cluster) pixel-mass-threshold))
	  (reduce
	   append
	   (map (lambda (cd-mu)
		 (let ((bitmap (make-matrix height width #f)))
		  (for-each
		   (lambda (p)
		    (matrix-set! bitmap (pixel-y p) (pixel-x p) #t))
		   (similar-xs cd-mu cd-mus pixels cd-distance-threshold
			       pixel->cd distance))
		  (c-pbm-proximity-clusterer
		   (make-pbm #t bitmap) proximity-threshold)))
		cd-mus)
	   '()))))
  (when track-movie?
   (inject-pnm! (pgm-constant width height *max-grey*) ppm)
   (for-each (lambda (cd-mu)
	      (overwrite-pixels-as!
	       (similar-xs
		cd-mu cd-mus pixels cd-distance-threshold pixel->cd distance)
	       ppm
	       (vector (vector-ref cd-mu 0)
		       (vector-ref cd-mu 1)
		       (- (max *max-red* *max-green* *max-blue*)
			  (vector-ref cd-mu 0)
			  (vector-ref cd-mu 1)))))
	     cd-mus))
  (let ((ellipses (map points->ellipse clusters))
	(hues (map
	       hue-mu
	       (map (lambda (cluster)
		     (map (lambda (point)
			   (hsv-hue
			    (rgb->hsv (ppm-rgb-ref ppm (y point) (x point)))))
			  cluster))
		    clusters))))
   (list pgm (map make-coloured-ellipse hues ellipses) '()))))

(define (cpt-hue pathname
		 video-type
		 background-threshold
		 pixel-mass-threshold
		 proximity-threshold
		 epsilon
		 sigma
		 motion-threshold
		 c1-quanta
		 c2-quanta
		 cd-distance-threshold
		 track-movie?
		 left?
		 background)
 (setup-proximity-clusterer! proximity-threshold)
 (let* ((frame-number 0)
	(number-of-frames (video-file-length pathname video-type))
	(coloured-ellipse-movie
	 (track
	  pathname
	  video-type
	  track-movie?
	  (lambda (ppm last-pgm last-moving-ellipses)
	   (when *window*
	    (message (format #f "Frame ~s out of ~s (~s%)"
			     frame-number
			     number-of-frames
			     (inexact->exact
			      (/ (* frame-number 100) number-of-frames))))
	    (set! frame-number (+ frame-number 1)))
	   (cpt-hue-process-frame ppm
				  last-pgm
				  background-threshold
				  pixel-mass-threshold
				  proximity-threshold
				  epsilon
				  sigma
				  motion-threshold
				  c1-quanta
				  c2-quanta
				  cd-distance-threshold
				  track-movie?
				  left?
				  background)))))
  (write-object-to-file (map-vector (lambda (frame)
				     (map coloured-ellipse-colour frame))
				    coloured-ellipse-movie)
			(replace-extension pathname "colour"))
  (map-vector (lambda (frame) (map coloured-ellipse-ellipse frame))
	      coloured-ellipse-movie)))

;;; CHMT

(define (chmt pathname
	      video-type
	      background-threshold
	      quanta
	      colour-pixel-mass-threshold
	      hue-separation-threshold
	      colour-proximity-threshold
	      motion-threshold
	      motion-pixel-mass-threshold
	      motion-proximity-threshold
	      track-movie?
	      left?
	      background)
 (let ((background-c (make-c-real-vector (length background)))
       (background-d (make-c-real-vector (length background))))
  (for-each (lambda (i cd)
	     (c-real-vector-set! background-c i (/ (vector-ref cd 0) 255))
	     (c-real-vector-set! background-d i (/ (vector-ref cd 1) 255)))
	    (enumerate (length background))
	    background)
  (let ((frames (c-chmt-internal
		 (replace-extension pathname video-type)
		 (replace-extension
		  (string-append (strip-extension pathname) "-tracked")
		  video-type)
		 (video-type-enum video-type)
		 (/ background-threshold 255)
		 quanta
		 colour-pixel-mass-threshold
		 hue-separation-threshold
		 colour-proximity-threshold
		 (* motion-threshold motion-threshold)
		 motion-pixel-mass-threshold
		 motion-proximity-threshold
		 (if track-movie? 1 0)
		 (if left? 1 0)
		 (length background)
		 background-c
		 background-d)))
   (free-c-vector! background-c)
   (free-c-vector! background-d)
   (list->vector (map-n (lambda (frame)
			 (map-n (lambda (k)
				 (make-ellipse (get-ellipse-x frame k)
					       (get-ellipse-y frame k)
					       (get-ellipse-theta frame k)
					       (get-ellipse-r frame k)
					       (get-ellipse-s frame k)))
				(get-object-k frame)))
			frames)))))

;;; HACK-TRACK

(define (hack-track pathname
		    video-type
		    colour-pixel-mass-threshold
		    colour-proximity-threshold
		    motion-threshold
		    motion-pixel-mass-threshold
		    motion-proximity-threshold
		    track-movie?)
 (let ((frames
	(let* ((ppm (video-file-frame->ppm pathname video-type 0))
	       (frames (video-file-length pathname video-type)))
	 (meteor-startup-offline
	  (pnm-height ppm)
	  (pnm-width ppm)
	  (* 2 frames)
	  #f)
	 frames)))
  (call-with-video-input-file
   pathname
   video-type
   (lambda (video-input-port)
    (let loop ((frame 0))
     (let ((ppm (read-ppm-from-video-input-port video-input-port)))
      (when ppm
       (let ((red (ppm-red ppm))
	     (green (ppm-green ppm))
	     (blue (ppm-blue ppm))
	     (height (pnm-height ppm))
	     (width (pnm-width ppm)))
	(do ((y 0 (+ y 1))) ((>= y height))
	 (do ((x 0 (+ x 1))) ((>= x width))
	  (meteor-set-red! frame y x (matrix-ref red y x))
	  (meteor-set-green! frame y x (matrix-ref green y x))
	  (meteor-set-blue! frame y x (matrix-ref blue y x)))))
       (loop (+ frame 1)))))))
  (meteor-hack-track! 0
		      frames
		      frames
		      colour-pixel-mass-threshold
		      colour-proximity-threshold
		      motion-threshold
		      motion-pixel-mass-threshold
		      motion-proximity-threshold)
  (when track-movie?
   (call-with-video-output-file
    (replace-extension (string-append (strip-extension pathname) "-tracked")
		       video-type)
    video-type
    frames
    (lambda (video-output-port)
     (for-each-n (lambda (i)
		  (write-pnm-to-video-output-port
		   (meteor-frame->ppm (+ frames i)) video-output-port))
		 frames))))
  (let ((ellipse-movie
	 (list->vector
	  (map-n (lambda (frame)
		  (map-n (lambda (k)
			  (make-ellipse (get-ellipse-x frame k)
					(get-ellipse-y frame k)
					(get-ellipse-theta frame k)
					(get-ellipse-r frame k)
					(get-ellipse-s frame k)))
			 (get-object-k frame)))
		 frames))))
   (meteor-shutdown-offline)
   ellipse-movie)))

;;; RFBT

(define (rfbt pathname
	      video-type
	      acceleration-vicinity
	      velocity-vicinity
	      position-vicinity
	      subsample
	      jerky
	      ratio
	      colour-pixel-mass-threshold
	      colour-proximity-threshold
	      motion-threshold
	      motion-pixel-mass-threshold
	      motion-proximity-threshold
	      hand-left
	      hand-top
	      hand-right
	      hand-bottom
	      track-movie?)
 (let ((frames (c-rfbt-internal
		(replace-extension pathname video-type)
		(replace-extension
		 (string-append (strip-extension pathname) "-tracked")
		 video-type)
		(video-type-enum video-type)
		acceleration-vicinity
		velocity-vicinity
		position-vicinity
		subsample
		jerky
		ratio
		colour-pixel-mass-threshold
		colour-proximity-threshold
		(* motion-threshold motion-threshold)
		motion-pixel-mass-threshold
		motion-proximity-threshold
		hand-left
		hand-top
		hand-right
		hand-bottom
		(if track-movie? 1 0))))
  (list->vector (map-n (lambda (frame)
			(map-n (lambda (k)
				(make-ellipse (get-ellipse-x frame k)
					      (get-ellipse-y frame k)
					      (get-ellipse-theta frame k)
					      (get-ellipse-r frame k)
					      (get-ellipse-s frame k)))
			       (get-object-k frame)))
		       frames))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
