(MODULE HLIB-TRAIN-CLASSIFY)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-train-classify.sch")
(include "c-externals.sc")

;;; Constants

(define sqrt-two-pi (sqrt (* 2.0 pi)))

(define log-two-pi (log (* 2.0 pi)))

;;; Structures

(define-structure match psi log-likelihood permutation)

(define-structure state-sequence states deltas)

;;; Variables

(define *delta* 1e-11)

(define *verbose?* #f)

;;; Procedures

(define (log-univariate-gaussian x mu sigma)
 (- (* -0.5 (sqr (/ (- x mu) sigma))) (log (* sigma sqrt-two-pi))))

(define (log-d-univariate-gaussian/d-x x mu sigma)
 (+ (log (/ (- mu x) (* sigma sigma sigma)))
    (log-univariate-gaussian x mu sigma)))

(define (lbessi0 x)
 (if (< (abs x) 3.75)
     (let ((y (sqr (/ x 3.75))))
      (log (+ 1.0
	      (* y
		 (+ 3.5156229
		    (* y
		       (+ 3.0899424
			  (* y
			     (+ 1.2067492
				(* y
				   (+ 0.2659732
				      (* y
					 (+ 0.360768e-1
					    (* y 0.45813e-2))))))))))))))
     (let* ((ax (abs x))
	    (y (/ 3.75 ax)))
      (+ (- ax (* 0.5 (log ax)))
	 (log
	  (+
	   0.39894228
	   (* y
	      (+
	       0.1328592e-1
	       (* y
		  (+
		   0.225319e-2
		   (* y
		      (+
		       -0.157565e-2
		       (* y
			  (+
			   0.916281e-2
			   (* y
			      (+
			       -0.2057706e-1
			       (* y
				  (+
				   0.2635537e-1
				   (* y
				      (+
				       -0.1647633e-1
				       (* y 0.392377e-2)))))))))))))))))))))

(define (log-von-mises x mean kappa)
 (- (* kappa (cos (- x mean))) log-two-pi (lbessi0 kappa)))

(define (log-d-von-mises/d-x x mean kappa)
 (+ (log (- (* kappa (sin (- x mean))))) (log-von-mises x mean kappa)))

(define (log-likelihood-of-feature-matrix psi feature-matrix)
 (let* ((uu (vector-length (psi-b psi)))
	(tt (matrix-rows feature-matrix))
	(ii (matrix-columns feature-matrix))
	(delta (make-matrix uu tt))
	(log-a (map-vector (lambda (row) (map-vector log row)) (psi-a psi)))
	(b (psi-b psi))
	(parameters (psi-parameters psi)))
  (for-each-n
   (lambda (t)
    (let ((feature-vector (vector-ref feature-matrix t)))
     (for-each-n
      (lambda (u)
       (let ((parameters (vector-ref parameters u)))
	(matrix-set!
	 delta u t
	 ;; needs work: Special version of sum that uses 0.0 instead of 0.
	 (+ (sum (lambda (i)
		  (let ((parameters (vector-ref parameters i)))
		   (case (first parameters)
		    ((continuous)
		     (log-univariate-gaussian (vector-ref feature-vector i)
					      (second parameters)
					      (third parameters)))
		    ((radial)
		     (log-von-mises (vector-ref feature-vector i)
				    (second parameters)
				    (third parameters)))
		    (else (fuck-up)))))
		 ii)
	    (if (zero? t)
		(log (vector-ref b u))
		(log-sum
		 (lambda (v)
		  (+ (matrix-ref log-a v u) (matrix-ref delta v (- t 1))))
		 uu))))))
      uu)))
   tt)
  (/ (log-sum (lambda (u) (matrix-ref delta u (- tt 1))) uu) ii)))

(define (d-log-likelihood-of-feature-matrix/d-feature-matrix
	 psi feature-matrix)
 (let* ((uu (vector-length (psi-b psi)))
	(tt (matrix-rows feature-matrix))
	(ii (matrix-columns feature-matrix))
	(delta (make-matrix uu tt))
	(d-delta/d-fti (make-matrix uu tt))
	(log-a (map-vector (lambda (row) (map-vector log row)) (psi-a psi)))
	(b (psi-b psi))
	(parameters (psi-parameters psi)))
  (for-each-n
   (lambda (t)
    (let ((feature-vector (vector-ref feature-matrix t)))
     (for-each-n
      (lambda (u)
       (let ((parameters (vector-ref parameters u)))
	(matrix-set!
	 delta u t
	 ;; needs work: Special version of sum that uses 0.0 instead of 0.
	 (+ (sum (lambda (i)
		  (let ((parameters (vector-ref parameters i)))
		   (case (first parameters)
		    ((continuous)
		     (log-univariate-gaussian (vector-ref feature-vector i)
					      (second parameters)
					      (third parameters)))
		    ((radial)
		     (log-von-mises (vector-ref feature-vector i)
				    (second parameters)
				    (third parameters)))
		    (else (fuck-up)))))
		 ii)
	    (if (zero? t)
		(log (vector-ref b u))
		(log-sum
		 (lambda (v)
		  (+ (matrix-ref log-a v u) (matrix-ref delta v (- t 1))))
		 uu))))))
      uu)))
   tt)
  (map-n-vector
   (lambda (t)
    (map-n-vector
     (lambda (i)
      (for-each-n
       (lambda (t-prime)
	(let ((feature-vector (vector-ref feature-matrix t-prime)))
	 (for-each-n
	  (lambda (u)
	   (let ((parameters (vector-ref parameters u)))
	    (matrix-set!
	     d-delta/d-fti u t-prime
	     (if (= t-prime t)
		 ;; needs work: Special version of sum that uses 0.0 instead
		 ;;             of 0.
		 (+ (sum (lambda (i-prime)
			  (let ((parameters (vector-ref parameters i-prime)))
			   (case (first parameters)
			    ((continuous)
			     (if (= i-prime i)
				 (log-d-univariate-gaussian/d-x
				  (vector-ref feature-vector i-prime)
				  (second parameters)
				  (third parameters))
				 (log-univariate-gaussian
				  (vector-ref feature-vector i-prime)
				  (second parameters)
				  (third parameters))))
			    ((radial)
			     (if (= i-prime i)
				 (log-d-von-mises/d-x
				  (vector-ref feature-vector i-prime)
				  (second parameters)
				  (third parameters))
				 (log-von-mises
				  (vector-ref feature-vector i-prime)
				  (second parameters)
				  (third parameters))))
			    (else (fuck-up)))))
			 ii)
		    (if (zero? t-prime)
			(log (vector-ref b u))
			(log-sum (lambda (v)
				  (+ (matrix-ref log-a v u)
				     (matrix-ref delta v (- t-prime 1))))
				 uu)))
		 (if (zero? t-prime)
		     0.0
		     ;; needs work: Special version of sum that uses 0.0
		     ;;             instead of 0.
		     (+ (sum (lambda (i-prime)
			      (let ((parameters
				     (vector-ref parameters i-prime)))
			       (case (first parameters)
				((continuous)
				 (log-univariate-gaussian
				  (vector-ref feature-vector i-prime)
				  (second parameters)
				  (third parameters)))
				((radial)
				 (log-von-mises
				  (vector-ref feature-vector i-prime)
				  (second parameters)
				  (third parameters)))
				(else (fuck-up)))))
			     ii)
			(log-sum
			 (lambda (v)
			  (+ (matrix-ref log-a v u)
			     (matrix-ref d-delta/d-fti v (- t-prime 1))))
			 uu)))))))
	  uu)))
       tt)
      (exp
       (/ (- (log-sum (lambda (u) (matrix-ref d-delta/d-fti u (- tt 1))) uu)
	     (log-sum (lambda (u) (matrix-ref delta u (- tt 1))) uu))
	  ii)))
     ii))
   tt)))

(define (numeric-d-log-likelihood-of-feature-matrix/d-feature-matrix
	 psi feature-matrix)		;debugging
 (let ((log-likelihood
	(log-likelihood-of-feature-matrix psi feature-matrix)))
  (let ((tt (matrix-rows feature-matrix))
	(ii (matrix-columns feature-matrix)))
   (map-n-vector
    (lambda (t)
     (map-n-vector
      (lambda (i)
       (/ (- (log-likelihood-of-feature-matrix
	      psi
	      (map-n-vector
	       (lambda (t-prime)
		(map-n-vector
		 (lambda (i-prime)
		  ;; needs work: can migrate MATRIX-REF
		  (if (and (= t-prime t) (= i-prime i))
		      (+ (matrix-ref feature-matrix t-prime i-prime) *delta*)
		      (matrix-ref feature-matrix t-prime i-prime)))
		 ii))
	       tt))
	     log-likelihood)
	  *delta*))
      ii))
    tt))))

(define (log-likelihood-of-ellipse-movie psi ellipse-movie)
 (log-likelihood-of-feature-matrix
  psi (ellipse-movie->feature-matrix ellipse-movie (psi-features psi))))

(define (e+ e0 e1)
 (make-ellipse (+ (ellipse-x0 e0) (ellipse-x0 e1))
	       (+ (ellipse-y0 e0) (ellipse-y0 e1))
	       (+ (ellipse-t0 e0) (ellipse-t0 e1))
	       (+ (ellipse-a e0) (ellipse-a e1))
	       (+ (ellipse-b e0) (ellipse-b e1))))

(define (sum-ellipse f n)
 (let loop ((n (- n 1))
	    ;; The constants are hardwired to be inexact for efficiency.
	    (c (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
  (if (negative? n) c (loop (- n 1) (e+ c (f n))))))

(define (k*e k e)
 (make-ellipse (* k (ellipse-x0 e))
	       (* k (ellipse-y0 e))
	       (* k (ellipse-t0 e))
	       (* k (ellipse-a e))
	       (* k (ellipse-b e))))

(define (d-log-likelihood-of-ellipse-movie/d-ellipse-movie psi ellipse-movie)
 (let* ((feature-matrix
	 (ellipse-movie->feature-matrix ellipse-movie (psi-features psi)))
	(tt (matrix-rows feature-matrix))
	(ii (matrix-columns feature-matrix))
	(kk (length (vector-ref ellipse-movie 0)))
	(d-log-likelihood/d-feature-matrix
	 ;; debugging: NUMERIC
	 (numeric-d-log-likelihood-of-feature-matrix/d-feature-matrix
	  psi feature-matrix))
	(d-feature-matrix/d-ellipse-movie
	 (d-ellipse-movie->feature-matrix/d-ellipse-movie
	  ellipse-movie (psi-features psi))))
  (map-n-vector
   (lambda (t)
    (map-n
     (lambda (k)
      (let ((d-feature-matrix/d-ellipse
	     ;; needs work: can migrate VECTOR-REF
	     (list-ref (vector-ref d-feature-matrix/d-ellipse-movie t) k)))
       ;; needs work: Special version of sum that uses 0.0 instead of 0.
       (sum-ellipse
	(lambda (t)
	 ;; needs work: Special version of sum that uses 0.0 instead of 0.
	 (sum-ellipse
	  (lambda (i)
	   ;; needs work: can migrate MATRIX-REF
	   (k*e (matrix-ref d-log-likelihood/d-feature-matrix t i)
		;; needs work: You know that the matrix has all zeros in all
		;;             rows except for t-2,t-1,t,t+1,t+2, where t is
		;;             the argument to this named procedure, so this
		;;             can be made more efficient.
		(matrix-ref d-feature-matrix/d-ellipse t i)))
	  ii))
	tt)))
     kk))
   ;; This is not TT because the feature matrix has two fewer rows than the
   ;; ellipse movie has frames.
   (vector-length ellipse-movie))))

(define (psi-ii psi) (vector-length (vector-ref (psi-parameters psi) 0)))

(define (psi-uu psi) (vector-length (psi-parameters psi)))

(define (psi->model psi)
 (let ((model (allocate-model (psi-ii psi) (psi-uu psi))))
  (do ((i 0 (+ i 1))) ((= i (psi-ii psi)))
   (case (first (matrix-ref (psi-parameters psi) 0 i))
    ((continuous) (set-model-feature-type-continuous! model i 100.0))
    ((radial) (set-model-feature-type-radial! model i))
    (else (panic "Unrecognized feature type"))))
  (do ((u 0 (+ u 1))) ((= u (psi-uu psi)))
   (set-model-b! model u (vector-ref (psi-b psi) u))
   (do ((v 0 (+ v 1))) ((= v (psi-uu psi)))
    (set-model-a! model u v (matrix-ref (psi-a psi) u v)))
   (do ((i 0 (+ i 1))) ((= i (psi-ii psi)))
    (let ((parameters (matrix-ref (psi-parameters psi) u i)))
     (case (first parameters)
      ((continuous)
       (set-model-parameter! model u i 0 (second parameters))
       (set-model-parameter! model u i 1 (third parameters)))
      ((radial)
       (set-model-parameter! model u i 0 (second parameters))
       (set-model-parameter! model u i 1 (third parameters)))
      (else (panic "Unrecognized feature type"))))))
  model))

(define (model->psi name features kk model)
 (make-psi name
	   features
	   kk
	   (let ((parameters (make-matrix (model-uu model) (model-ii model))))
	    (do ((u 0 (+ u 1))) ((= u (model-uu model)))
	     (do ((i 0 (+ i 1))) ((= i (model-ii model)))
	      (matrix-set! parameters
			   u
			   i
			   (list
			    (case (model-feature-type model i)
			     ((0) 'continuous)
			     ((1) 'radial)
			     (else (panic "Unrecognized feature type")))
			    (model-parameter model u i 0)
			    (model-parameter model u i 1)))))
	    parameters)
	   (let ((a (make-matrix (model-uu model) (model-uu model))))
	    (do ((u 0 (+ u 1))) ((= u (model-uu model)))
	     (do ((v 0 (+ v 1))) ((= v (model-uu model)))
	      (matrix-set! a u v (model-a model u v))))
	    a)
	   (let ((b (make-vector (model-uu model))))
	    (do ((u 0 (+ u 1))) ((= u (model-uu model)))
	     (vector-set! b u (model-b model u)))
	    b)))

(define (format-state-sequence sequence first-frame)
 (let ((string (string))
       (states (state-sequence-states sequence))
       (deltas (state-sequence-deltas sequence))
       (counter 0))
  (for-each-n (lambda (n)
	       (let* ((new-string
		       (format #f "~s -~s-> "
			       (vector-ref states n)
			       (+ (vector-ref deltas n) first-frame 2)))
		      (ns-length (string-length new-string)))
		(when (> (+ counter ns-length) 75)
		 (set! string (string-append string (format #f "~%")))
		 (set! counter 0))
		(set! string (string-append string new-string))
		(set! counter (+ counter ns-length))))
	      (vector-length deltas))
  (string-append
   string (format #f "~s" (vector-ref states (vector-length deltas))))))

(define (best-states tt model rmat)
 (let* ((best-state-sequence (best-state-sequence model rmat))
	(counter 0)
	(states (make-vector tt -1))
	(deltas (make-vector tt)))
  (when (= tt 0) (panic "BEST-STATES: movie has zero length"))
  (vector-set! states 0 (c-int-vector-ref best-state-sequence 0))
  (for-each-n (lambda (t)
	       (let ((state (c-int-vector-ref best-state-sequence t)))
		(unless (= (vector-ref states counter) state)
		 (vector-set! deltas counter t)
		 (set! counter (+ counter 1))
		 (vector-set! states counter state))))
	      tt)
  (let ((s (make-vector (+ counter 1)))
	(d (make-vector counter)))
   (vector-set! s counter (vector-ref states counter))
   (for-each-n (lambda (i)
		(vector-set! s i (vector-ref states i))
		(vector-set! d i (vector-ref deltas i)))
	       counter)
   (free-c-vector! best-state-sequence)
   (make-state-sequence s d))))

(define (best-state-sequence-of-ellipse-movie psi ellipse-movie)
 (unless (= (psi-kk psi) (ellipse-movie-kk ellipse-movie))
  (panic "Model and movie have different numbers of ellipses"))
 (let* ((rmat (ellipse-movie->rmat ellipse-movie (psi-features psi)))
	(model (psi->model psi))
	(state-sequence
	 (best-states (- (vector-length ellipse-movie) 2) model rmat)))
  (free-rmat rmat)
  (free-model model)
  state-sequence))

(define (c-log-likelihood-of-ellipse-movie psi ellipse-movie)
 (unless (= (psi-kk psi) (ellipse-movie-kk ellipse-movie))
  (panic "Model and movie have different numbers of ellipses"))
 (let* ((rmat (ellipse-movie->rmat ellipse-movie (psi-features psi)))
	(model (psi->model psi))
	(log-likelihood (model-log-likelihood model rmat)))
  (free-rmat rmat)
  (free-model model)
  log-likelihood))

(define (apply-permutation permutation ellipse-movie)
 (map-vector (lambda (ellipses)
	      (map (lambda (index) (list-ref ellipses index)) permutation))
	     ellipse-movie))

(define (apply-permutations permutations ellipse-movies)
 (if (null? permutations)
     '()
     (cons (apply-permutation (first permutations) (first ellipse-movies))
	   (apply-permutations (rest permutations) (rest ellipse-movies)))))

(define (a-permutation1 psi-kk ellipse-movie-kk)
 (unless (= psi-kk ellipse-movie-kk) (fail))
 (enumerate ellipse-movie-kk))

(define (a-permutation2 psi-kk ellipse-movie-kk)
 (unless (= psi-kk ellipse-movie-kk) (fail))
 (a-permutation-of (enumerate ellipse-movie-kk)))

(define (a-permutation3 psi-kk ellipse-movie-kk)
 (a-permutation-of (a-subset-of-size psi-kk (enumerate ellipse-movie-kk))))

(define (classify-ellipse-movie psis ellipse-movie generator)
 (let ((best-psi #f)
       (best-log-likelihood minus-infinity)
       (best-permutation #f))
  (for-each
   (lambda (psi)
    (for-effects
     (let* ((permutation
	     (generator (psi-kk psi) (ellipse-movie-kk ellipse-movie)))
	    (this-log-likelihood
	     (log-likelihood-of-ellipse-movie
	      psi (apply-permutation permutation ellipse-movie))))
      (when (> this-log-likelihood best-log-likelihood)
       (set! best-log-likelihood this-log-likelihood)
       (set! best-psi psi)
       (set! best-permutation permutation)))))
   psis)
  (unless best-psi (panic "No model has same number of ellipses as movie"))
  (make-match best-psi best-log-likelihood best-permutation)))

(define (rank-ellipse-movie psis ellipse-movie generator)
 (let ((matches '()))
  (for-each
   (lambda (psi)
    (let ((best-log-likelihood minus-infinity)
	  (best-permutation #f))
     (for-effects
      (let* ((permutation
	      (generator (psi-kk psi) (ellipse-movie-kk ellipse-movie)))
	     (this-log-likelihood
	      (log-likelihood-of-ellipse-movie
	       psi (apply-permutation permutation ellipse-movie))))
       (when (> this-log-likelihood best-log-likelihood)
	(set! best-log-likelihood this-log-likelihood)
	(set! best-permutation permutation))))
     (when best-permutation
      (set! matches
	    (cons (make-match psi best-log-likelihood best-permutation)
		  matches)))))
   psis)
  (when (null? matches)
   (panic "No model has same number of ellipses as movie"))
  (sort matches > match-log-likelihood)))

(define (rmats-log-likelihood model rmats)
 (reduce + (map (lambda (rmat) (model-log-likelihood model rmat)) rmats) 0.0))

(define (train-model-and-log-likelihood
	 ellipse-movies kk uu features upper-triangular?
	 random-a-and-b? min-epochs max-epochs training-epsilon nn)
 (when #f				;debugging
  (format #t "Entering TRAIN-MODEL-AND-LOG-LIKELIHOOD, L = ~s~%"
	  (length ellipse-movies)))
 (let* ((model (allocate-model (number-of-features kk features) uu))
	(best-model (allocate-model (number-of-features kk features) uu))
	(rmats (map (lambda (ellipse-movie)
		     (ellipse-movie->rmat ellipse-movie features))
		    ellipse-movies))
	(rmat-vector (allocate-rmat-vector (length rmats))))
  (do ((i 0 (+ i 1))) ((= i (number-of-features kk features)))
   (if (radial? kk features i)
       (set-model-feature-type-radial! model i)
       (set-model-feature-type-continuous! model i 100.0)))
  (do ((i 0 (+ i 1))) ((= i (number-of-features kk features)))
   (if (radial? kk features i)
       (set-model-feature-type-radial! best-model i)
       (set-model-feature-type-continuous! best-model i 100.0)))
  (for-each (lambda (rmat i) (rmat-vector-set! rmat-vector i rmat))
	    rmats
	    (enumerate (length rmats)))
  (let outer ((nn nn)
	      (best-model best-model)
	      (best minus-infinity)
	      (model model))
   (cond
    ((zero? nn)
     (when #f				;debugging
      (format #t "Exiting TRAIN-MODEL-AND-LOG-LIKELIHOOD, log L = ~s~%" best))
     (free-model model)
     (for-each free-rmat rmats)
     (free-rmat-vector rmat-vector)
     (list best-model best))
    (else
     (when #f (format #t "n = ~s~%" nn)) ;debugging
     (randomise-model! model upper-triangular? random-a-and-b?)
     (force-init-globals!)
     (let inner ((last (rmats-log-likelihood model rmats)) (epoch 1))
      (when #f (format #t "Last: ~s~%" last)) ;debugging
      (when (nan? last) (panic "Nan result")) ; debugging
      (cond
       ((and max-epochs (> epoch max-epochs))
	(when #f (format #t "Terminating pass, too many epochs~%")) ;debugging
	(outer (- nn 1) best-model best model))
       (else (update-model! model rmat-vector (length rmats))
	     (let ((this (rmats-log-likelihood model rmats)))
	      (cond
	       ((and (< (/ (- this last) (abs (* 2 (+ this last))))
			training-epsilon)
		     (>= epoch min-epochs))
		(when #f		;debugging
		 (format #t "Terminating pass, convergence, This = ~s~%" this))
		(if (> this best)
		    (outer (- nn 1) model this best-model)
		    (outer (- nn 1) best-model best model)))
	       (else (inner this (+ epoch 1)))))))))))))

(define (train-model ellipse-movies kk uu features upper-triangular?
		     random-a-and-b? min-epochs max-epochs training-epsilon nn)
 (let ((model-and-log-likelihood
	(train-model-and-log-likelihood
	 ellipse-movies kk uu features upper-triangular?
	 random-a-and-b? min-epochs max-epochs training-epsilon nn)))
  (first model-and-log-likelihood)))

(define (train-log-likelihood
	 ellipse-movies kk uu features upper-triangular?
	 random-a-and-b? min-epochs max-epochs training-epsilon nn)
 (let ((model-and-log-likelihood
	(train-model-and-log-likelihood
	 ellipse-movies kk uu features upper-triangular?
	 random-a-and-b? min-epochs max-epochs training-epsilon nn)))
  (free-model (first model-and-log-likelihood))
  (second model-and-log-likelihood)))

(define (external-correspondence1
	 ellipse-movies kk uu features upper-triangular?
	 random-a-and-b? min-epochs max-epochs training-epsilon nn generator)
 ;; This is the greedy algorithm.
 (let ((permutations '()))
  ;; needs work: This redundantly tries all permutations of the first movie.
  (for-each
   (lambda (ellipse-movie)
    (let ((best-log-likelihood minus-infinity)
	  (best-permutation #f))
     (for-effects
      (let* ((permutation (generator kk (ellipse-movie-kk ellipse-movie)))
	     (this-log-likelihood
	      (train-log-likelihood
	       (apply-permutations
		(append permutations (list permutation)) ellipse-movies)
	       kk uu features upper-triangular? random-a-and-b? min-epochs
	       max-epochs training-epsilon nn)))
       (when *verbose?* (format #t "Permutation~%"))
       (when (> this-log-likelihood best-log-likelihood)
	(set! best-log-likelihood this-log-likelihood)
	(set! best-permutation permutation))))
     (unless best-permutation (panic "Some movie has insufficient ellipses"))
     (set! permutations (append permutations (list best-permutation)))
     (when *verbose?* (format #t "Ellipse movie~%"))))
   ellipse-movies)
  permutations))

(define (external-correspondence1a
	 ellipse-movies kk uu features upper-triangular?
	 random-a-and-b? min-epochs max-epochs training-epsilon nn generator)
 ;; This is the quadratic greedy algorithm.
 (when (or (null? ellipse-movies) (null? (rest ellipse-movies)))
  (panic "There must be at least two ellipse movies"))
 (let ((permutations (map (lambda (ignore) #f) ellipse-movies))
       (ellipse-movies-to-be-placed ellipse-movies))
  (let ((best-log-likelihood minus-infinity)
	(best-permutation #f)
	(best-ellipse-movie1 #f)
	(best-ellipse-movie2 #f))
   (do ((ellipse-movies1 ellipse-movies (rest ellipse-movies1)))
     ((null? (rest ellipse-movies1)))
    (let ((ellipse-movie1 (first ellipse-movies1)))
     (for-each
      (lambda (ellipse-movie2)
       (for-effects
	(let* ((permutation (generator kk (ellipse-movie-kk ellipse-movie2)))
	       (this-log-likelihood
		(train-log-likelihood
		 (list ellipse-movie1
		       (apply-permutation permutation ellipse-movie2))
		 kk uu features upper-triangular? random-a-and-b? min-epochs
		 max-epochs training-epsilon nn)))
	 (when *verbose?* (format #t "Permutation~%"))
	 (when (> this-log-likelihood best-log-likelihood)
	  (set! best-log-likelihood this-log-likelihood)
	  (set! best-permutation permutation)
	  (set! best-ellipse-movie1 ellipse-movie1)
	  (set! best-ellipse-movie2 ellipse-movie2)))))
      (rest ellipse-movies1))))
   (unless best-permutation (panic "Some movie has insufficient ellipses"))
   (list-set! permutations
	      (positionq best-ellipse-movie1 ellipse-movies)
	      (enumerate (ellipse-movie-kk best-ellipse-movie1)))
   (list-set! permutations
	      (positionq best-ellipse-movie2 ellipse-movies)
	      best-permutation)
   (set! ellipse-movies-to-be-placed
	 (removeq best-ellipse-movie1
		  (removeq best-ellipse-movie2 ellipse-movies-to-be-placed)))
   (when *verbose?* (format #t "Ellipse movie~%")))
  (let loop ()
   (unless (null? ellipse-movies-to-be-placed)
    (let ((best-log-likelihood minus-infinity)
	  (best-permutation #f)
	  (best-ellipse-movie #f))
     (for-each
      (lambda (ellipse-movie)
       (for-effects
	(let* ((permutation (generator kk (ellipse-movie-kk ellipse-movie)))
	       (this-log-likelihood
		(train-log-likelihood
		 (removeq
		  #f
		  (map (lambda (permutation1 ellipse-movie1)
			(cond ((list? permutation1)
			       (apply-permutation permutation1 ellipse-movie1))
			      ((eq? ellipse-movie ellipse-movie1)
			       (apply-permutation permutation ellipse-movie))
			      (else #f)))
		       permutations
		       ellipse-movies))
		 kk uu features upper-triangular? random-a-and-b? min-epochs
		 max-epochs training-epsilon nn)))
	 (when *verbose?* (format #t "Permutation~%"))
	 (when (> this-log-likelihood best-log-likelihood)
	  (set! best-log-likelihood this-log-likelihood)
	  (set! best-permutation permutation)
	  (set! best-ellipse-movie ellipse-movie)))))
      ellipse-movies-to-be-placed)
     (unless best-permutation (panic "Some movie has insufficient ellipses"))
     (list-set! permutations
		(positionq best-ellipse-movie ellipse-movies)
		best-permutation)
     (set! ellipse-movies-to-be-placed
	   (removeq best-ellipse-movie ellipse-movies-to-be-placed))
     (when *verbose?* (format #t "Ellipse movie~%"))
     (loop))))
  permutations))

(define (external-correspondence2
	 ellipse-movies kk uu features upper-triangular?
	 random-a-and-b? min-epochs max-epochs training-epsilon nn generator)
 ;; This is the iterative refinement algorithm starting with a fixed initial
 ;; correspondence.
 (let* ((permutations
	 (map (lambda (ellipse-movie) (enumerate kk)) ellipse-movies))
	(current-log-likelihood
	 (train-log-likelihood
	  (apply-permutations permutations ellipse-movies)
	  kk uu features upper-triangular? random-a-and-b? min-epochs
	  max-epochs training-epsilon nn)))
  (let loop ()
   (let ((again? #f))
    (for-each-n
     (lambda (l)
      (when *verbose?* (format #t "l = ~s~%" l))
      (let ((permutation (list-ref permutations l))
	    (ellipse-movie (list-ref ellipse-movies l))
	    (best-log-likelihood minus-infinity)
	    (best-permutation #f))
       (for-effects
	(let* ((permutation (generator kk (ellipse-movie-kk ellipse-movie)))
	       (this-log-likelihood
		(train-log-likelihood
		 (apply-permutations
		  (append
		   (sublist permutations 0 l)
		   (list permutation)
		   (sublist permutations (+ l 1) (length permutations)))
		  ellipse-movies)
		 kk uu features upper-triangular? random-a-and-b? min-epochs
		 max-epochs training-epsilon nn)))
	 (when (> this-log-likelihood best-log-likelihood)
	  (set! best-log-likelihood this-log-likelihood)
	  (set! best-permutation permutation))))
       (when (> best-log-likelihood current-log-likelihood)
	(set! current-log-likelihood best-log-likelihood)
	(when *verbose?* (format #t "Best: ~s~%" best-log-likelihood))
	(when *verbose?* (write permutations) (newline))
	(unless (equal? best-permutation permutation)
	 (when *verbose?*
	  (format #t "~s -> ~s~%" permutation best-permutation))
	 (set! permutations
	       (append (sublist permutations 0 l)
		       (list best-permutation)
		       (sublist permutations (+ l 1) (length permutations))))
	 (when *verbose?* (write permutations) (newline))
	 (set! again? #t)))))
     (length permutations))
    (when again? (loop))))
  permutations))

(define (external-correspondence3
	 ellipse-movies kk uu features upper-triangular?
	 random-a-and-b? min-epochs max-epochs training-epsilon nn generator)
 ;; This is the greedy algorithm followed by the iterative refinement
 ;; algorithm.
 (let ((permutations '())
       (current-log-likelihood minus-infinity))
  ;; needs work: This redundantly tries all permutations of the first movie.
  (for-each
   (lambda (ellipse-movie)
    (let ((best-log-likelihood minus-infinity)
	  (best-permutation #f))
     (for-effects
      (let* ((permutation (generator kk (ellipse-movie-kk ellipse-movie)))
	     (this-log-likelihood
	      (train-log-likelihood
	       (apply-permutations
		(append permutations (list permutation)) ellipse-movies)
	       kk uu features upper-triangular? random-a-and-b? min-epochs
	       max-epochs training-epsilon nn)))
       (when *verbose?* (format #t "Permutation~%"))
       (when (> this-log-likelihood best-log-likelihood)
	(set! best-log-likelihood this-log-likelihood)
	(set! best-permutation permutation))))
     (unless best-permutation (panic "Some movie has insufficient ellipses"))
     (set! permutations (append permutations (list best-permutation)))
     (set! current-log-likelihood best-log-likelihood)
     (when *verbose?* (format #t "Ellipse movie~%"))))
   ellipse-movies)
  (let loop ()
   (let ((again? #f))
    (for-each-n
     (lambda (l)
      (when *verbose?* (format #t "l = ~s~%" l))
      (let ((permutation (list-ref permutations l))
	    (ellipse-movie (list-ref ellipse-movies l))
	    (best-log-likelihood minus-infinity)
	    (best-permutation #f))
       (for-effects
	(let* ((permutation (generator kk (ellipse-movie-kk ellipse-movie)))
	       (this-log-likelihood
		(train-log-likelihood
		 (apply-permutations
		  (append
		   (sublist permutations 0 l)
		   (list permutation)
		   (sublist permutations (+ l 1) (length permutations)))
		  ellipse-movies)
		 kk uu features upper-triangular? random-a-and-b? min-epochs
		 max-epochs training-epsilon nn)))
	 (when (> this-log-likelihood best-log-likelihood)
	  (set! best-log-likelihood this-log-likelihood)
	  (set! best-permutation permutation))))
       (when (> best-log-likelihood current-log-likelihood)
	(set! current-log-likelihood best-log-likelihood)
	(when *verbose?* (format #t "Best: ~s~%" best-log-likelihood))
	(when *verbose?* (write permutations) (newline))
	(unless (equal? best-permutation permutation)
	 (when *verbose?*
	  (format #t "~s -> ~s~%" permutation best-permutation))
	 (set! permutations
	       (append (sublist permutations 0 l)
		       (list best-permutation)
		       (sublist permutations (+ l 1) (length permutations))))
	 (when *verbose?* (write permutations) (newline))
	 (set! again? #t)))))
     (length permutations))
    (when again? (loop))))
  permutations))

(define (external-correspondence3a
	 ellipse-movies kk uu features upper-triangular?
	 random-a-and-b? min-epochs max-epochs training-epsilon nn generator)
 ;; This is the quadratic greedy algorithm followed by the iterative refinement
 ;; algorithm.
 (when (or (null? ellipse-movies) (null? (rest ellipse-movies)))
  (panic "There must be at least two ellipse movies"))
 (let ((permutations (map (lambda (ignore) #f) ellipse-movies))
       (ellipse-movies-to-be-placed ellipse-movies)
       (current-log-likelihood minus-infinity))
  (let ((best-log-likelihood minus-infinity)
	(best-permutation #f)
	(best-ellipse-movie1 #f)
	(best-ellipse-movie2 #f))
   (do ((ellipse-movies1 ellipse-movies (rest ellipse-movies1)))
     ((null? (rest ellipse-movies1)))
    (let ((ellipse-movie1 (first ellipse-movies1)))
     (for-each
      (lambda (ellipse-movie2)
       (for-effects
	(let* ((permutation (generator kk (ellipse-movie-kk ellipse-movie2)))
	       (this-log-likelihood
		(train-log-likelihood
		 (list ellipse-movie1
		       (apply-permutation permutation ellipse-movie2))
		 kk uu features upper-triangular? random-a-and-b? min-epochs
		 max-epochs training-epsilon nn)))
	 (when *verbose?* (format #t "Permutation~%"))
	 (when (> this-log-likelihood best-log-likelihood)
	  (set! best-log-likelihood this-log-likelihood)
	  (set! best-permutation permutation)
	  (set! best-ellipse-movie1 ellipse-movie1)
	  (set! best-ellipse-movie2 ellipse-movie2)))))
      (rest ellipse-movies1))))
   (unless best-permutation (panic "Some movie has insufficient ellipses"))
   (list-set! permutations
	      (positionq best-ellipse-movie1 ellipse-movies)
	      (enumerate (ellipse-movie-kk best-ellipse-movie1)))
   (list-set! permutations
	      (positionq best-ellipse-movie2 ellipse-movies)
	      best-permutation)
   (set! ellipse-movies-to-be-placed
	 (removeq best-ellipse-movie1
		  (removeq best-ellipse-movie2 ellipse-movies-to-be-placed)))
   (set! current-log-likelihood best-log-likelihood)
   (when *verbose?* (format #t "Ellipse movie~%")))
  (let loop ()
   (unless (null? ellipse-movies-to-be-placed)
    (let ((best-log-likelihood minus-infinity)
	  (best-permutation #f)
	  (best-ellipse-movie #f))
     (for-each
      (lambda (ellipse-movie)
       (for-effects
	(let* ((permutation (generator kk (ellipse-movie-kk ellipse-movie)))
	       (this-log-likelihood
		(train-log-likelihood
		 (removeq
		  #f
		  (map (lambda (permutation1 ellipse-movie1)
			(cond ((list? permutation1)
			       (apply-permutation permutation1 ellipse-movie1))
			      ((eq? ellipse-movie ellipse-movie1)
			       (apply-permutation permutation ellipse-movie))
			      (else #f)))
		       permutations
		       ellipse-movies))
		 kk uu features upper-triangular? random-a-and-b? min-epochs
		 max-epochs training-epsilon nn)))
	 (when *verbose?* (format #t "Permutation~%"))
	 (when (> this-log-likelihood best-log-likelihood)
	  (set! best-log-likelihood this-log-likelihood)
	  (set! best-permutation permutation)
	  (set! best-ellipse-movie ellipse-movie)))))
      ellipse-movies-to-be-placed)
     (unless best-permutation (panic "Some movie has insufficient ellipses"))
     (list-set! permutations
		(positionq best-ellipse-movie ellipse-movies)
		best-permutation)
     (set! ellipse-movies-to-be-placed
	   (removeq best-ellipse-movie ellipse-movies-to-be-placed))
     (set! current-log-likelihood best-log-likelihood)
     (when *verbose?* (format #t "Ellipse movie~%"))
     (loop))))
  (let loop ()
   (let ((again? #f))
    (for-each-n
     (lambda (l)
      (when *verbose?* (format #t "l = ~s~%" l))
      (let ((permutation (list-ref permutations l))
	    (ellipse-movie (list-ref ellipse-movies l))
	    (best-log-likelihood minus-infinity)
	    (best-permutation #f))
       (for-effects
	(let* ((permutation (generator kk (ellipse-movie-kk ellipse-movie)))
	       (this-log-likelihood
		(train-log-likelihood
		 (apply-permutations
		  (append
		   (sublist permutations 0 l)
		   (list permutation)
		   (sublist permutations (+ l 1) (length permutations)))
		  ellipse-movies)
		 kk uu features upper-triangular? random-a-and-b? min-epochs
		 max-epochs training-epsilon nn)))
	 (when (> this-log-likelihood best-log-likelihood)
	  (set! best-log-likelihood this-log-likelihood)
	  (set! best-permutation permutation))))
       (when (> best-log-likelihood current-log-likelihood)
	(set! current-log-likelihood best-log-likelihood)
	(when *verbose?* (format #t "Best: ~s~%" best-log-likelihood))
	(when *verbose?* (write permutations) (newline))
	(unless (equal? best-permutation permutation)
	 (when *verbose?*
	  (format #t "~s -> ~s~%" permutation best-permutation))
	 (set! permutations
	       (append (sublist permutations 0 l)
		       (list best-permutation)
		       (sublist permutations (+ l 1) (length permutations))))
	 (when *verbose?* (write permutations) (newline))
	 (set! again? #t)))))
     (length permutations))
    (when again? (loop))))
  permutations))

;;; Tam V'Nishlam Shevah L'El Borei Olam
