(MODULE HLIB-WBM)
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-wbm.sch")
(include "c-externals.sc")

;;; Structures

(define-structure psi name features kk parameters a b)

;;; Parameters

(define *full-features*
 '(magnitude-of-the-velocity-vector
   orientation-of-the-velocity-vector
   angular-velocity
   finite-difference-of-the-area
   finite-difference-of-the-eccentricity
   finite-difference-of-the-magnitude-of-the-velocity-vector
   finite-difference-of-the-orientation-of-the-velocity-vector
   finite-difference-of-the-angular-velocity
   second-finite-difference-of-the-area
   second-finite-difference-of-the-eccentricity
   distance-between-ellipse-centres
   orientation-of-the-vector-between-ellipse-centres
   difference-between-the-orientations-of-the-major-axes
   mouthful
   finite-difference-of-the-distance-between-ellipse-centres
   finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
   finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
   finite-difference-of-mouthful))

(define *partial-features*
 '(magnitude-of-the-velocity-vector
   orientation-of-the-velocity-vector
   finite-difference-of-the-magnitude-of-the-velocity-vector
   finite-difference-of-the-orientation-of-the-velocity-vector
   distance-between-ellipse-centres
   orientation-of-the-vector-between-ellipse-centres
   finite-difference-of-the-distance-between-ellipse-centres
   finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres))

;;; Procedures

(define (negate-ellipse ellipse)
 (make-ellipse (- (ellipse-x0 ellipse))
	       (- (ellipse-y0 ellipse))
	       (- (ellipse-t0 ellipse))
	       (- (ellipse-a ellipse))
	       (- (ellipse-b ellipse))))

(define (allocate-rmat ii tt) (allocate-rmat-internal ii tt))

(define (rmat-get rmat ii tt) (rmat-get-internal rmat ii tt))

(define (rmat-set! rmat ii tt value) (rmat-set!-internal rmat ii tt value) #f)

(define (free-rmat rmat) (free-rmat-internal rmat) #f)

(define (allocate-rmat-vector nn) (allocate-rmat-vector-internal nn))

(define (rmat-vector-set! rmat-vector n rmat)
 (rmat-vector-set!-internal rmat-vector n rmat)
 #f)

(define (free-rmat-vector rmat-vector)
 (free-rmat-vector-internal rmat-vector)
 #f)

(define (allocate-model ii uu) (allocate-model-internal ii uu))

(define (model-ii model) (model-ii-internal model))

(define (model-uu model) (model-uu-internal model))

(define (model-feature-type model i) (model-feature-type-internal model i))

(define (set-model-feature-type-continuous! model i value)
 (set-model-feature-type-continuous!-internal model i value)
 #f)

(define (set-model-feature-type-radial! model i)
 (set-model-feature-type-radial!-internal model i)
 #f)

(define (model-parameter model i u f) (model-parameter-internal model i u f))

(define (set-model-parameter! model i u f value)
 (set-model-parameter!-internal model i u f value)
 #f)

(define (model-a model u1 u2) (model-a-internal model u1 u2))

(define (set-model-a! model u1 u2 value)
 (set-model-a!-internal model u1 u2 value)
 #f)

(define (model-b model u) (model-b-internal model u))

(define (set-model-b! model u value) (set-model-b!-internal model u value) #f)

(define (randomise-model! model upper-triangular? random-a-and-b?)
 (randomise-model!-internal
  model (if upper-triangular? 1 0) (if random-a-and-b? 1 0))
 #f)

(define (print-model model) (print-model-internal model) #f)

(define (free-model model) (free-model-internal model) #f)

(define (model-log-likelihood model rmat)
 (model-log-likelihood-internal model rmat))

(define (best-state-sequence model rmat)
 (best-state-sequence-internal model rmat))

(define (update-model! model rmat-vector nn)
 (update-model!-internal model rmat-vector nn))

(define (number-of-features kk features)
 (let ((i 0))
  (for-each-n
   (lambda (k1)
    (when (memq 'magnitude-of-the-velocity-vector features)
     (set! i (+ i 1)))
    (when (memq 'orientation-of-the-velocity-vector features)
     (set! i (+ i 1)))
    (when (memq 'angular-velocity features)
     (set! i (+ i 1)))
    (when (memq 'finite-difference-of-the-area features)
     (set! i (+ i 1)))
    (when (memq 'finite-difference-of-the-eccentricity features)
     (set! i (+ i 1)))
    (when (memq 'finite-difference-of-the-magnitude-of-the-velocity-vector
		features)
     (set! i (+ i 1)))
    (when (memq 'finite-difference-of-the-orientation-of-the-velocity-vector
		features)
     (set! i (+ i 1)))
    (when (memq 'finite-difference-of-the-angular-velocity features)
     (set! i (+ i 1)))
    (when (memq 'second-finite-difference-of-the-area features)
     (set! i (+ i 1)))
    (when (memq 'second-finite-difference-of-the-eccentricity features)
     (set! i (+ i 1)))
    (for-each-n
     (lambda (k2)
      (when (memq 'distance-between-ellipse-centres features)
       (when (> k1 k2)
	(set! i (+ i 1))))
      (when (memq 'orientation-of-the-vector-between-ellipse-centres features)
       (when (> k1 k2)
	(set! i (+ i 1))))
      (when (memq 'difference-between-the-orientations-of-the-major-axes
		  features)
       (when (> k1 k2)
	(set! i (+ i 1))))
      (when (memq 'mouthful features)
       (unless (= k1 k2)
	(set! i (+ i 1))))
      (when (memq 'finite-difference-of-the-distance-between-ellipse-centres
		  features)
       (when (> k1 k2)
	(set! i (+ i 1))))
      (when (memq 'finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
		  features)
       (when (> k1 k2)
	(set! i (+ i 1))))
      (when (memq 'finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
		  features)
       (when (> k1 k2)
	(set! i (+ i 1))))
      (when (memq 'finite-difference-of-mouthful features)
       (unless (= k1 k2)
	(set! i (+ i 1)))))
     kk))
   kk)
  i))

(define (orientation u v)
 ;; This computes the orienation of a vector from U to V. It assumes image
 ;; coordinates, i.e. X increases from left to right but Y increases from
 ;; top to bottom. It returns an angle in [-pi, pi) where -pi and pi point to
 ;; the left and increasing angle denotes counterclockwise movement.
 (atan (- (y u) (y v)) (- (x v) (x u))))

(define (min-abs x y) (if (< (abs x) (abs y)) x y))

(define (symmetric-radial-distance theta phi)
 (min-abs (radial-distance theta phi) (radial-distance (+ theta pi) phi)))

(define (ellipse-movie-kk ellipse-movie)
 (when (zero? (vector-length ellipse-movie))
  (panic "Can't get number of ellipses from empty ellipse movie"))
 (let ((kk (length (vector-ref ellipse-movie 0))))
  (unless (every-vector (lambda (frame) (= (length frame) kk)) ellipse-movie)
   (panic "Frames don't all have same number of ellipses"))
  kk))

(define (magnitude-of-the-velocity-vector ellipse1a ellipse0a)
 ;; range = [0, infinity)
 ;; needs work: This shouldn't be modelled by a Guassian, perhaps a gamma.
 (distance (ellipse-center ellipse1a) (ellipse-center ellipse0a)))

(define (d-magnitude-of-the-velocity-vector/d-ellipse1a ellipse1a ellipse0a)
 (let ((s (magnitude-of-the-velocity-vector ellipse1a ellipse0a)))
  (make-ellipse (/ (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse0a)) s)
		(/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse0a)) s)
		0.0
		0.0
		0.0)))

(define (d-magnitude-of-the-velocity-vector/d-ellipse0a ellipse1a ellipse0a)
 (negate-ellipse
  (d-magnitude-of-the-velocity-vector/d-ellipse1a ellipse1a ellipse0a)))

(define (orientation-of-the-velocity-vector ellipse1a ellipse0a)
 ;; range = [-pi, pi)
 (orientation (ellipse-center ellipse1a) (ellipse-center ellipse0a)))

(define (d-orientation-of-the-velocity-vector/d-ellipse1a ellipse1a ellipse0a)
 ;; a kludge
 (if (= (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse0a))
			 (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse0a))
	  (* (sqr (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a))) s))
       (/ (* (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a)) s))
       0.0
       0.0
       0.0))))

(define (d-orientation-of-the-velocity-vector/d-ellipse0a ellipse1a ellipse0a)
 (negate-ellipse
  (d-orientation-of-the-velocity-vector/d-ellipse1a ellipse1a ellipse0a)))

(define (angular-velocity ellipse1a ellipse0a)
 ;; range = [-pi/2, pi/2)
 (symmetric-radial-distance (ellipse-t0 ellipse1a) (ellipse-t0 ellipse0a)))

(define (d-angular-velocity/d-ellipse1a ellipse1a ellipse0a)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 ;; Theta is -2.0 because we double the angular velocity so that its range is
 ;; [-pi, pi)
 (make-ellipse 0.0 0.0 -2.0 0.0 0.0))

(define (d-angular-velocity/d-ellipse0a ellipse1a ellipse0a)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 ;; Theta is 2.0 because we double the angular velocity so that its range is
 ;; [-pi, pi)
 (make-ellipse 0.0 0.0 2.0 0.0 0.0))

(define (finite-difference-of-the-area ellipse1a ellipse0a)
 ;; range = (-infinity, infinity)
 (- (ellipse-area ellipse0a) (ellipse-area ellipse1a)))

(define (d-finite-difference-of-the-area/d-ellipse1a ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (* minus-pi (ellipse-b ellipse1a))
	       (* minus-pi (ellipse-a ellipse1a))))

(define (d-finite-difference-of-the-area/d-ellipse0a ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (* pi (ellipse-b ellipse0a))
	       (* pi (ellipse-a ellipse0a))))

(define (finite-difference-of-the-eccentricity ellipse1a ellipse0a)
 ;; range = (-infinity, infinity)
 (- (ellipse-eccentricity ellipse0a) (ellipse-eccentricity ellipse1a)))

(define (d-finite-difference-of-the-eccentricity/d-ellipse1a
	 ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (- (/ (ellipse-b ellipse1a)))
	       (/ (ellipse-a ellipse1a) (sqr (ellipse-b ellipse1a)))))

(define (d-finite-difference-of-the-eccentricity/d-ellipse0a
	 ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (/ (ellipse-b ellipse0a))
	       (- (/ (ellipse-a ellipse0a) (sqr (ellipse-b ellipse0a))))))

(define (finite-difference-of-the-magnitude-of-the-velocity-vector
	 ellipse2a ellipse1a ellipse0a)
 ;; range = (-infinity, infinity)
 (- (magnitude-of-the-velocity-vector ellipse1a ellipse0a)
    (magnitude-of-the-velocity-vector ellipse2a ellipse1a)))

(define
 (d-finite-difference-of-the-magnitude-of-the-velocity-vector/d-ellipse2a
  ellipse2a ellipse1a ellipse0a)
 (let ((s (magnitude-of-the-velocity-vector ellipse2a ellipse1a)))
  (make-ellipse (/ (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse2a)) s)
		(/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse2a)) s)
		0.0
		0.0
		0.0)))

(define
 (d-finite-difference-of-the-magnitude-of-the-velocity-vector/d-ellipse1a
  ellipse2a ellipse1a ellipse0a)
 (let ((s (magnitude-of-the-velocity-vector ellipse2a ellipse1a))
       (t (magnitude-of-the-velocity-vector ellipse1a ellipse0a)))
  (make-ellipse (+ (/ (- (ellipse-x0 ellipse2a) (ellipse-x0 ellipse1a)) s)
		   (/ (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse0a)) t))
		(+ (/ (- (ellipse-y0 ellipse2a) (ellipse-y0 ellipse1a)) s)
		   (/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse0a)) t))
		0.0
		0.0
		0.0)))

(define
 (d-finite-difference-of-the-magnitude-of-the-velocity-vector/d-ellipse0a
  ellipse2a ellipse1a ellipse0a)
 (let ((s (magnitude-of-the-velocity-vector ellipse1a ellipse0a)))
  (make-ellipse (/ (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a)) s)
		(/ (- (ellipse-y0 ellipse0a) (ellipse-y0 ellipse1a)) s)
		0.0
		0.0
		0.0)))

(define (finite-difference-of-the-orientation-of-the-velocity-vector
	 ellipse2a ellipse1a ellipse0a)
 ;; range = [-pi, pi)
 (radial-distance
  (orientation-of-the-velocity-vector ellipse2a ellipse1a)
  (orientation-of-the-velocity-vector ellipse1a ellipse0a)))

(define
 (d-finite-difference-of-the-orientation-of-the-velocity-vector/d-ellipse2a
  ellipse2a ellipse1a ellipse0a)
 ;; a kludge
 (if (= (ellipse-x0 ellipse1a) (ellipse-x0 ellipse2a))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse2a) (ellipse-y0 ellipse1a))
			 (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse2a)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse2a))
	  (* (sqr (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse2a))) s))
       (/ (* (- (ellipse-x0 ellipse2a) (ellipse-x0 ellipse1a)) s))
       0.0
       0.0
       0.0))))

(define
 (d-finite-difference-of-the-orientation-of-the-velocity-vector/d-ellipse1a
  ellipse2a ellipse1a ellipse0a)
 ;; a kludge
 (if (or (= (ellipse-x0 ellipse1a) (ellipse-x0 ellipse2a))
	 (= (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a)))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse2a) (ellipse-y0 ellipse1a))
			 (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse2a))))))
	   (t (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse0a))
			 (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a)))))))
      (make-ellipse
       (+ (/ (- (ellipse-y0 ellipse2a) (ellipse-y0 ellipse1a))
	     (* (sqr (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse2a))) s))
	  (/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse0a))
	     (* (sqr (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a))) t)))
       (+ (/ (* (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse2a)) s))
	  (/ (* (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a)) t)))
       0.0
       0.0
       0.0))))

(define
 (d-finite-difference-of-the-orientation-of-the-velocity-vector/d-ellipse0a
  ellipse2a ellipse1a ellipse0a)
 ;; a kludge
 (if (= (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse0a))
			 (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse0a) (ellipse-y0 ellipse1a))
	  (* (sqr (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse1a))) s))
       (/ (* (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse0a)) s))
       0.0
       0.0
       0.0))))

(define (finite-difference-of-the-angular-velocity
	 ellipse2a ellipse1a ellipse0a)
 ;; range = [-pi, pi)
 (radial-distance (angular-velocity ellipse2a ellipse1a)
		  (angular-velocity ellipse1a ellipse0a)))

(define (d-finite-difference-of-the-angular-velocity/d-ellipse2a
	 ellipse2a ellipse1a ellipse0a)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 (make-ellipse 0.0 0.0 1.0 0.0 0.0))

(define (d-finite-difference-of-the-angular-velocity/d-ellipse1a
	 ellipse2a ellipse1a ellipse0a)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 (make-ellipse 0.0 0.0 -2.0 0.0 0.0))

(define (d-finite-difference-of-the-angular-velocity/d-ellipse0a
	 ellipse2a ellipse1a ellipse0a)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 (make-ellipse 0.0 0.0 1.0 0.0 0.0))

(define (second-finite-difference-of-the-area ellipse2a ellipse1a ellipse0a)
 ;; range = (-infinity, infinity)
 (- (finite-difference-of-the-area ellipse1a ellipse0a)
    (finite-difference-of-the-area ellipse2a ellipse1a)))

(define (d-second-finite-difference-of-the-area/d-ellipse2a
	 ellipse2a ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (* pi (ellipse-b ellipse2a))
	       (* pi (ellipse-a ellipse2a))))

(define (d-second-finite-difference-of-the-area/d-ellipse1a
	 ellipse2a ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (- (* two-pi (ellipse-b ellipse1a)))
	       (- (* two-pi (ellipse-a ellipse1a)))))

(define (d-second-finite-difference-of-the-area/d-ellipse0a
	 ellipse2a ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (* pi (ellipse-b ellipse0a))
	       (* pi (ellipse-a ellipse0a))))

(define (second-finite-difference-of-the-eccentricity
	 ellipse2a ellipse1a ellipse0a)
 ;; range = (-infinity, infinity)
 (- (finite-difference-of-the-eccentricity ellipse1a ellipse0a)
    (finite-difference-of-the-eccentricity ellipse2a ellipse1a)))

(define (d-second-finite-difference-of-the-eccentricity/d-ellipse2a
	 ellipse2a ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (/ (ellipse-b ellipse2a))
	       (- (/ (ellipse-a ellipse2a) (sqr (ellipse-b ellipse2a))))))

(define (d-second-finite-difference-of-the-eccentricity/d-ellipse1a
	 ellipse2a ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (/ -2.0 (ellipse-b ellipse1a))
	       (/ (* 2.0 (ellipse-a ellipse1a)) (sqr (ellipse-b ellipse1a)))))

(define (d-second-finite-difference-of-the-eccentricity/d-ellipse0a
	 ellipse2a ellipse1a ellipse0a)
 (make-ellipse 0.0
	       0.0
	       0.0
	       (/ (ellipse-b ellipse0a))
	       (- (/ (ellipse-a ellipse0a) (sqr (ellipse-b ellipse0a))))))

(define (distance-between-ellipse-centres ellipse0a ellipse0b)
 ;; range = [0, infinity)
 ;; needs work: This shouldn't be modelled by a Guassian, perhaps a gamma.
 (distance (ellipse-center ellipse0a) (ellipse-center ellipse0b)))

(define (new-distance-between-ellipse-centres ellipse0a ellipse0b)
 ;; debugging
 ;; range = [0, infinity)
 ;; needs work: This shouldn't be modelled by a Guassian, perhaps a gamma.
 (let* ((x0a (ellipse-x0 ellipse0a))
	(y0a (ellipse-y0 ellipse0a))
	(t0a (ellipse-t0 ellipse0a))
	(a0a (ellipse-a ellipse0a))
	(b0a (ellipse-b ellipse0a))
	(x0b (ellipse-x0 ellipse0b))
	(y0b (ellipse-y0 ellipse0b))
	(t0b (ellipse-t0 ellipse0b))
	(a0b (ellipse-a ellipse0b))
	(b0b (ellipse-b ellipse0b))
	(c0a (cos t0a))
	(s0a (sin t0a))
	(c0b (cos t0b))
	(s0b (sin t0b)))
  (sqrt (+ (*      x0a x0a)
	   (* -2.0 x0a x0b)
	   (*      x0b x0b)
	   (*      y0a y0a)
	   (* -2.0 y0a y0b)
	   (*      y0b y0b)
	   (*      b0a b0a)
	   (*      b0b b0b)
	   (*  2.0 a0a x0b c0a)
	   (*      a0b a0b c0b c0b)
	   (* -2.0 b0a y0a s0a)
	   (* -2.0 a0b x0b c0b)
	   (*      a0a a0a c0a c0a)
	   (* -2.0 b0b y0b s0b)
	   (* -2.0 a0a a0b c0a c0b)
	   (* -1.0 b0b b0b c0b c0b)
	   (* -1.0 b0a b0a c0a c0a)
	   (*  2.0 a0b x0a c0b)
	   (* -2.0 b0a b0b s0a s0b)
	   (*  2.0 b0b y0a s0b)
	   (* -2.0 a0a x0a c0a)
	   (*  2.0 b0a y0b s0a)))))

(define (newer-distance-between-ellipse-centres ellipse0a ellipse0b)
 ;; debugging
 ;; range = [0, infinity)
 ;; needs work: This shouldn't be modelled by a Guassian, perhaps a gamma.
 (define (powell p xi ftol func)
  (let* ((n (vector-length p))
	 (pt (make-vector n))
	 (ptt (make-vector n))
	 (xit (make-vector n))
	 (fret (func p)))
   (do ((j 0 (+ j 1))) ((= j n)) (vector-set! pt j (vector-ref p j)))
   (let loop ((iter 0))
    (let ((fp fret)
	  (ibig 0)
	  (del 0.0))
     (do ((i 0 (+ i 1))) ((= i n))
      (do ((j 0 (+ j 1))) ((= j n)) (vector-set! xit j (matrix-ref xi j i)))
      (let ((fptt fret))
       (set! fret (linmin p xit func))
       (when (> (abs (- fptt fret)) del)
	(set! del (abs (- fptt fret)))
	(set! ibig i))))
     (cond
      ((<= (abs (- fp fret)) ftol) fret)
      (else
       (when (= iter *itmax-powell*)
	(panic "Too many iterations in routine POWELL"))
       (do ((j 0 (+ j 1))) ((= j n))
	(vector-set! ptt j (* 2.0 (- (vector-ref p j) (vector-ref pt j))))
	(vector-set! xit j (- (vector-ref p j) (vector-ref pt j)))
	(vector-set! pt j (vector-ref p j)))
       (let ((fptt (func ptt)))
	(when (and
	       (< fptt fp)
	       (negative?
		(- (* 2.0 (+ (- fp (* 2.0 fret)) fptt) (sqr (- fp fret del)))
		   (* del (sqr (- fp fptt))))))
	 (set! fret (linmin p xit func))
	 (do ((j 0 (+ j 1))) ((= j n))
	  (matrix-set! xi j ibig (vector-ref xit j))))
	(loop (+ iter 1)))))))))
 (powell (vector 0.0 0.0)
	 (identity-matrix 2)
	 0.001
	 (lambda (x)
	  (let* ((ta (vector-ref x 0))
		 (tb (vector-ref x 1))
		 (xa (+ (ellipse-x0 ellipse0a)
			(* (ellipse-a ellipse0a)
			   (cos (+ ta (ellipse-t0 ellipse0a))))))
		 (ya (+ (ellipse-y0 ellipse0a)
			(* (ellipse-b ellipse0a)
			   (sin (+ ta (ellipse-t0 ellipse0a))))))
		 (xb (+ (ellipse-x0 ellipse0b)
			(* (ellipse-a ellipse0b)
			   (cos (+ tb (ellipse-t0 ellipse0b))))))
		 (yb (+ (ellipse-y0 ellipse0b)
			(* (ellipse-b ellipse0b)
			   (sin (+ tb (ellipse-t0 ellipse0b)))))))
	   (distance (vector xa ya) (vector xb yb))))))

(define (d-distance-between-ellipse-centres/d-ellipse0a ellipse0a ellipse0b)
 (let ((s (distance-between-ellipse-centres ellipse0a ellipse0b)))
  ;; a kludge
  (if (zero? s)
      (make-ellipse 0.0 0.0 0.0 0.0 0.0)
      (make-ellipse (/ (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)) s)
		    (/ (- (ellipse-y0 ellipse0a) (ellipse-y0 ellipse0b)) s)
		    0.0
		    0.0
		    0.0))))

(define (d-distance-between-ellipse-centres/d-ellipse0b ellipse0a ellipse0b)
 (negate-ellipse
  (d-distance-between-ellipse-centres/d-ellipse0a ellipse0a ellipse0b)))

(define (orientation-of-the-vector-between-ellipse-centres ellipse0a ellipse0b)
 ;; range = [-pi, pi)
 (orientation (ellipse-center ellipse0a) (ellipse-center ellipse0b)))

(define (d-orientation-of-the-vector-between-ellipse-centres/d-ellipse0a
	 ellipse0a ellipse0b)
 ;; a kludge
 (if (= (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse0b) (ellipse-y0 ellipse0a))
			 (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse0a) (ellipse-y0 ellipse0b))
	  (* (sqr (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))) s))
       (/ (* (- (ellipse-x0 ellipse0b) (ellipse-x0 ellipse0a)) s))
       0.0
       0.0
       0.0))))

(define (d-orientation-of-the-vector-between-ellipse-centres/d-ellipse0b
	 ellipse0a ellipse0b)
 (negate-ellipse
  (d-orientation-of-the-vector-between-ellipse-centres/d-ellipse0a
   ellipse0a ellipse0b)))

(define (difference-between-the-orientations-of-the-major-axes
	 ellipse0a ellipse0b)
 ;; range = [-pi/2, pi/2)
 (symmetric-radial-distance (ellipse-t0 ellipse0a) (ellipse-t0 ellipse0b)))

(define (d-difference-between-the-orientations-of-the-major-axes/d-ellipse0a
	 ellipse0a ellipse0b)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 ;; Theta is -2.0 because we double the difference between the orientations of
 ;; the major axes so that its range is [-pi, pi)
 (make-ellipse 0.0 0.0 -2.0 0.0 0.0))

(define (d-difference-between-the-orientations-of-the-major-axes/d-ellipse0b
	 ellipse0a ellipse0b)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 ;; Theta is 2.0 because we double the difference between the orientations of
 ;; the major axes so that its range is [-pi, pi)
 (make-ellipse 0.0 0.0 2.0 0.0 0.0))

(define (mouthful ellipse0a ellipse0b)
 ;; difference between the orientation of the major axis of the first ellipse
 ;; and the orientation of the vector between the centres of the ellipses.
 ;; This feature is not symmetric.
 ;; range = [-pi/2, pi/2)
 (symmetric-radial-distance
  (ellipse-t0 ellipse0a)
  (orientation (ellipse-center ellipse0a) (ellipse-center ellipse0b))))

(define (d-mouthful/d-ellipse0a ellipse0a ellipse0b)
 ;; a kludge
 (if (= (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     ;; needs work: This is not quite correct because it is missing MIN-ABS.
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse0b) (ellipse-y0 ellipse0a))
			 (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)))))))
      ;; The derivative is scaled by 2.0 because we double the mouthful so that
      ;; its range is [-pi, pi)
      (make-ellipse
       (/ (* 2.0 (- (ellipse-y0 ellipse0a) (ellipse-y0 ellipse0b)))
	  (* (sqr (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))) s))
       (/ 2.0 (* (- (ellipse-x0 ellipse0b) (ellipse-x0 ellipse0a)) s))
       -2.0
       0.0
       0.0))))

(define (d-mouthful/d-ellipse0b ellipse0a ellipse0b)
 ;; a kludge
 (if (= (ellipse-y0 ellipse0b) (ellipse-y0 ellipse0a))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     ;; needs work: This is not quite correct because it is missing MIN-ABS.
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse0b) (ellipse-y0 ellipse0a))
			 (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)))))))
      ;; The derivative is scaled by 2.0 because we double the mouthful so that
      ;; its range is [-pi, pi)
      (make-ellipse
       (/ (* 2.0 (- (ellipse-y0 ellipse0b) (ellipse-y0 ellipse0a)))
	  (* (sqr (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))) s))
       (/ 2.0 (* (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)) s))
       0.0
       0.0
       0.0))))

(define (finite-difference-of-the-distance-between-ellipse-centres
	 ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; range = (-infinity, infinity)
 (- (distance-between-ellipse-centres ellipse0a ellipse0b)
    (distance-between-ellipse-centres ellipse1a ellipse1b)))

(define
 (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse1a
  ellipse1a ellipse1b ellipse0a ellipse0b)
 (let ((s (distance-between-ellipse-centres ellipse1a ellipse1b)))
  ;; a kludge
  (if (zero? s)
      (make-ellipse 0.0 0.0 0.0 0.0 0.0)
      (make-ellipse (/ (- (ellipse-x0 ellipse1b) (ellipse-x0 ellipse1a)) s)
		    (/ (- (ellipse-y0 ellipse1b) (ellipse-y0 ellipse1a)) s)
		    0.0
		    0.0
		    0.0))))

(define
 (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse1b
  ellipse1a ellipse1b ellipse0a ellipse0b)
 (negate-ellipse
  (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse1a
   ellipse1a ellipse1b ellipse0a ellipse0b)))

(define
 (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse0a
  ellipse1a ellipse1b ellipse0a ellipse0b)
 (let ((s (distance-between-ellipse-centres ellipse0a ellipse0b)))
  ;; a kludge
  (if (zero? s)
      (make-ellipse 0.0 0.0 0.0 0.0 0.0)
      (make-ellipse (/ (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)) s)
		    (/ (- (ellipse-y0 ellipse0a) (ellipse-y0 ellipse0b)) s)
		    0.0
		    0.0
		    0.0))))

(define
 (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse0b
  ellipse1a ellipse1b ellipse0a ellipse0b)
 (negate-ellipse
  (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse0a
   ellipse1a ellipse1b ellipse0a ellipse0b)))

(define
 (finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
  ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; range = [-pi, pi)
 (radial-distance
  (orientation-of-the-vector-between-ellipse-centres ellipse1a ellipse1b)
  (orientation-of-the-vector-between-ellipse-centres ellipse0a ellipse0b)))

(define
 (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse1a
  ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; a kludge
 (if (= (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse1b) (ellipse-y0 ellipse1a))
			 (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse1b) (ellipse-y0 ellipse1a))
	  (* (sqr (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b))) s))
       (/ (* (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b)) s))
       0.0
       0.0
       0.0))))

(define
 (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse1b
  ellipse1a ellipse1b ellipse0a ellipse0b)
 (negate-ellipse
  (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse1a
   ellipse1a ellipse1b ellipse0a ellipse0b)))

(define
 (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse0a
  ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; a kludge
 (if (= (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse0b) (ellipse-y0 ellipse0a))
			 (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse0a) (ellipse-y0 ellipse0b))
	  (* (sqr (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))) s))
       (/ (* (- (ellipse-x0 ellipse0b) (ellipse-x0 ellipse0a)) s))
       0.0
       0.0
       0.0))))

(define
 (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse0b
  ellipse1a ellipse1b ellipse0a ellipse0b)
 (negate-ellipse
  (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse0a
   ellipse1a ellipse1b ellipse0a ellipse0b)))

(define
 (finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
  ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; range = [-pi, pi)
 (radial-distance
  (difference-between-the-orientations-of-the-major-axes ellipse1a ellipse1b)
  (difference-between-the-orientations-of-the-major-axes ellipse0a ellipse0b)))

(define
 (d-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes/d-ellipse1a
  ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 (make-ellipse 0.0 0.0 1.0 0.0 0.0))

(define
 (d-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes/d-ellipse1b
  ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 (make-ellipse 0.0 0.0 -1.0 0.0 0.0))

(define
 (d-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes/d-ellipse0a
  ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 (make-ellipse 0.0 0.0 -1.0 0.0 0.0))

(define
 (d-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes/d-ellipse0b
  ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; needs work: This is not quite correct because it is missing MIN-ABS.
 (make-ellipse 0.0 0.0 1.0 0.0 0.0))

(define (finite-difference-of-mouthful
	 ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; finite difference of the difference between the orientation of the major
 ;; axis of the first ellipse and the orientation of the vector between the
 ;; centres of the ellipses.  This feature is not symmetric.
 ;; range = [-pi, pi)
 (radial-distance
  (mouthful ellipse1a ellipse1b) (mouthful ellipse0a ellipse0b)))

(define (d-finite-difference-of-mouthful/d-ellipse1a
	 ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; a kludge
 (if (= (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     ;; needs work: This is not quite correct because it is missing MIN-ABS.
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse1b) (ellipse-y0 ellipse1a))
			 (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse1b) (ellipse-y0 ellipse1a))
	  (* (sqr (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b))) s))
       (/ (* (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b)) s))
       1.0
       0.0
       0.0))))

(define (d-finite-difference-of-mouthful/d-ellipse1b
	 ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; a kludge
 (if (= (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     ;; needs work: This is not quite correct because it is missing MIN-ABS.
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse1b) (ellipse-y0 ellipse1a))
			 (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse1a) (ellipse-y0 ellipse1b))
	  (* (sqr (- (ellipse-x0 ellipse1a) (ellipse-x0 ellipse1b))) s))
       (/ (* (- (ellipse-x0 ellipse1b) (ellipse-x0 ellipse1a)) s))
       0.0
       0.0
       0.0))))

(define (d-finite-difference-of-mouthful/d-ellipse0a
	 ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; a kludge
 (if (= (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     ;; needs work: This is not quite correct because it is missing MIN-ABS.
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse0b) (ellipse-y0 ellipse0a))
			 (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse0a) (ellipse-y0 ellipse0b))
	  (* (sqr (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))) s))
       (/ (* (- (ellipse-x0 ellipse0b) (ellipse-x0 ellipse0a)) s))
       -1.0
       0.0
       0.0))))

(define (d-finite-difference-of-mouthful/d-ellipse0b
	 ellipse1a ellipse1b ellipse0a ellipse0b)
 ;; a kludge
 (if (= (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))
     (make-ellipse 0.0 0.0 0.0 0.0 0.0)
     ;; needs work: This is not quite correct because it is missing MIN-ABS.
     (let ((s (+ 1.0
		 (sqr (/ (- (ellipse-y0 ellipse0b) (ellipse-y0 ellipse0a))
			 (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)))))))
      (make-ellipse
       (/ (- (ellipse-y0 ellipse0b) (ellipse-y0 ellipse0a))
	  (* (sqr (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b))) s))
       (/ (* (- (ellipse-x0 ellipse0a) (ellipse-x0 ellipse0b)) s))
       0.0
       0.0
       0.0))))

(define (ellipse-movie->feature-matrix ellipse-movie features)
 (let ((tt (vector-length ellipse-movie))
       (kk (ellipse-movie-kk ellipse-movie)))
  (unless (>= tt 3) (panic "Movie has fewer than three frames"))
  (let ((feature-matrix
	 (make-matrix (- tt 2) (number-of-features kk features))))
   (for-each-n
    (lambda (t)
     (let ((i 0)
	   (frame0 (vector-ref ellipse-movie (+ t 2)))
	   (frame1 (vector-ref ellipse-movie (+ t 1)))
	   (frame2 (vector-ref ellipse-movie t)))
      (for-each-n
       (lambda (k1)
	(let ((ellipse0a (list-ref frame0 k1))
	      (ellipse1a (list-ref frame1 k1))
	      (ellipse2a (list-ref frame2 k1)))
	 (when (memq 'magnitude-of-the-velocity-vector features)
	  (matrix-set!
	   feature-matrix t i
	   (magnitude-of-the-velocity-vector ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'orientation-of-the-velocity-vector features)
	  (matrix-set!
	   feature-matrix t i
	   (orientation-of-the-velocity-vector ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'angular-velocity features)
	  (matrix-set!
	   feature-matrix t i (* 2.0 (angular-velocity ellipse1a ellipse0a)))
	  (set! i (+ i 1)))
	 (when (memq 'finite-difference-of-the-area features)
	  (matrix-set!
	   feature-matrix t i
	   (finite-difference-of-the-area ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'finite-difference-of-the-eccentricity features)
	  (matrix-set!
	   feature-matrix t i
	   (finite-difference-of-the-eccentricity ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'finite-difference-of-the-magnitude-of-the-velocity-vector
		     features)
	  (matrix-set!
	   feature-matrix t i
	   (finite-difference-of-the-magnitude-of-the-velocity-vector
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq
		'finite-difference-of-the-orientation-of-the-velocity-vector
		features)
	  (matrix-set!
	   feature-matrix t i
	   (finite-difference-of-the-orientation-of-the-velocity-vector
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'finite-difference-of-the-angular-velocity features)
	  (matrix-set!
	   feature-matrix t i
	   (finite-difference-of-the-angular-velocity
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'second-finite-difference-of-the-area features)
	  (matrix-set!
	   feature-matrix t i
	   (second-finite-difference-of-the-area
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'second-finite-difference-of-the-eccentricity features)
	  (matrix-set!
	   feature-matrix t i
	   (second-finite-difference-of-the-eccentricity
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (for-each-n
	  (lambda (k2)
	   (let ((ellipse0b (list-ref frame0 k2))
		 (ellipse1b (list-ref frame1 k2))
		 (ellipse2b (list-ref frame2 k2)))
	    (when (memq 'distance-between-ellipse-centres features)
	     (when (> k1 k2)
	      (matrix-set!
	       feature-matrix t i
	       (distance-between-ellipse-centres ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'orientation-of-the-vector-between-ellipse-centres
			features)
	     (when (> k1 k2)
	      (matrix-set!
	       feature-matrix t i
	       (orientation-of-the-vector-between-ellipse-centres
		ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'difference-between-the-orientations-of-the-major-axes
			features)
	     (when (> k1 k2)
	      (matrix-set!
	       feature-matrix t i
	       (* 2.0 (difference-between-the-orientations-of-the-major-axes
		       ellipse0a ellipse0b)))
	      (set! i (+ i 1))))
	    (when (memq 'mouthful features)
	     (unless (= k1 k2)
	      (matrix-set!
	       feature-matrix t i (* 2.0 (mouthful ellipse0a ellipse0b)))
	      (set! i (+ i 1))))
	    (when (memq
		   'finite-difference-of-the-distance-between-ellipse-centres
		   features)
	     (when (> k1 k2)
	      (matrix-set!
	       feature-matrix t i
	       (finite-difference-of-the-distance-between-ellipse-centres
		ellipse1a ellipse1b ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
			features)
	     (when (> k1 k2)
	      (matrix-set!
	       feature-matrix t i
	       (finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
		ellipse1a ellipse1b ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
			features)
	     (when (> k1 k2)
	      (matrix-set!
	       feature-matrix t i
	       (finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
		ellipse1a ellipse1b ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'finite-difference-of-mouthful features)
	     (unless (= k1 k2)
	      (matrix-set!
	       feature-matrix t i
	       (finite-difference-of-mouthful
		ellipse1a ellipse1b ellipse0a ellipse0b))
	      (set! i (+ i 1))))))
	  kk)))
       kk)))
    (- tt 2))
   feature-matrix)))

(define (d-ellipse-movie->feature-matrix/d-ellipse-movie
	 ellipse-movie features)
 (let ((tt (vector-length ellipse-movie))
       (kk (ellipse-movie-kk ellipse-movie)))
  (unless (>= tt 3) (panic "Movie has fewer than three frames"))
  (map-n-vector
   (lambda (t)
    (map-n
     (lambda (k)
      (let ((feature-matrix
	     (make-matrix (- tt 2) (number-of-features kk features))))
       (for-each-n
	(lambda (t1)
	 (let ((i 0)
	       (frame0 (vector-ref ellipse-movie (+ t1 2)))
	       (frame1 (vector-ref ellipse-movie (+ t1 1)))
	       (frame2 (vector-ref ellipse-movie t1)))
	  (for-each-n
	   (lambda (k1)
	    (let ((ellipse0a (list-ref frame0 k1))
		  (ellipse1a (list-ref frame1 k1))
		  (ellipse2a (list-ref frame2 k1)))
	     (when (memq 'magnitude-of-the-velocity-vector features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-magnitude-of-the-velocity-vector/d-ellipse0a
			   ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-magnitude-of-the-velocity-vector/d-ellipse1a
			   ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (when (memq 'orientation-of-the-velocity-vector features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-orientation-of-the-velocity-vector/d-ellipse0a
			   ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-orientation-of-the-velocity-vector/d-ellipse1a
			   ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (when (memq 'angular-velocity features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-angular-velocity/d-ellipse0a ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-angular-velocity/d-ellipse1a ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (when (memq 'finite-difference-of-the-area features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-finite-difference-of-the-area/d-ellipse0a
			   ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-finite-difference-of-the-area/d-ellipse1a
			   ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (when (memq 'finite-difference-of-the-eccentricity features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-finite-difference-of-the-eccentricity/d-ellipse0a
			   ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-finite-difference-of-the-eccentricity/d-ellipse1a
			   ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (when (memq
		    'finite-difference-of-the-magnitude-of-the-velocity-vector
		    features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-finite-difference-of-the-magnitude-of-the-velocity-vector/d-ellipse0a
			   ellipse2a ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-finite-difference-of-the-magnitude-of-the-velocity-vector/d-ellipse1a
			   ellipse2a ellipse1a ellipse0a))
			 ((= t1 t)
			  (d-finite-difference-of-the-magnitude-of-the-velocity-vector/d-ellipse2a
			   ellipse2a ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (when (memq 'finite-difference-of-the-orientation-of-the-velocity-vector
			 features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-finite-difference-of-the-orientation-of-the-velocity-vector/d-ellipse0a
			   ellipse2a ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-finite-difference-of-the-orientation-of-the-velocity-vector/d-ellipse1a
			   ellipse2a ellipse1a ellipse0a))
			 ((= t1 t)
			  (d-finite-difference-of-the-orientation-of-the-velocity-vector/d-ellipse2a
			   ellipse2a ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (when (memq 'finite-difference-of-the-angular-velocity features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-finite-difference-of-the-angular-velocity/d-ellipse0a
			   ellipse2a ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-finite-difference-of-the-angular-velocity/d-ellipse1a
			   ellipse2a ellipse1a ellipse0a))
			 ((= t1 t)
			  (d-finite-difference-of-the-angular-velocity/d-ellipse2a
			   ellipse2a ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (when (memq 'second-finite-difference-of-the-area features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-second-finite-difference-of-the-area/d-ellipse0a
			   ellipse2a ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-second-finite-difference-of-the-area/d-ellipse1a
			   ellipse2a ellipse1a ellipse0a))
			 ((= t1 t)
			  (d-second-finite-difference-of-the-area/d-ellipse2a
			   ellipse2a ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (when (memq 'second-finite-difference-of-the-eccentricity
			 features)
	      (matrix-set!
	       feature-matrix t1 i
	       (if (= k1 k)
		   (cond ((= (+ t1 2) t)
			  (d-second-finite-difference-of-the-eccentricity/d-ellipse0a
			   ellipse2a ellipse1a ellipse0a))
			 ((= (+ t1 1) t)
			  (d-second-finite-difference-of-the-eccentricity/d-ellipse1a
			   ellipse2a ellipse1a ellipse0a))
			 ((= t1 t)
			  (d-second-finite-difference-of-the-eccentricity/d-ellipse2a
			   ellipse2a ellipse1a ellipse0a))
			 (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		   (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
	      (set! i (+ i 1)))
	     (for-each-n
	      (lambda (k2)
	       (let ((ellipse0b (list-ref frame0 k2))
		     (ellipse1b (list-ref frame1 k2))
		     (ellipse2b (list-ref frame2 k2)))
		(when (memq 'distance-between-ellipse-centres features)
		 (when (> k1 k2)
		  (matrix-set!
		   feature-matrix t1 i
		   (if (= (+ t1 2) t)
		       (cond ((= k1 k)
			      (d-distance-between-ellipse-centres/d-ellipse0a
			       ellipse0a ellipse0b))
			     ((= k2 k)
			      (d-distance-between-ellipse-centres/d-ellipse0b
			       ellipse0a ellipse0b))
			     (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		       (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		  (set! i (+ i 1))))
		(when (memq 'orientation-of-the-vector-between-ellipse-centres
			    features)
		 (when (> k1 k2)
		  (matrix-set!
		   feature-matrix t1 i
		   (if (= (+ t1 2) t)
		       (cond ((= k1 k)
			      (d-orientation-of-the-vector-between-ellipse-centres/d-ellipse0a
			       ellipse0a ellipse0b))
			     ((= k2 k)
			      (d-orientation-of-the-vector-between-ellipse-centres/d-ellipse0b
			       ellipse0a ellipse0b))
			     (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		       (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		  (set! i (+ i 1))))
		(when (memq
		       'difference-between-the-orientations-of-the-major-axes
		       features)
		 (when (> k1 k2)
		  (matrix-set!
		   feature-matrix t1 i
		   (if (= (+ t1 2) t)
		       (cond ((= k1 k)
			      (d-difference-between-the-orientations-of-the-major-axes/d-ellipse0a
			       ellipse0a ellipse0b))
			     ((= k2 k)
			      (d-difference-between-the-orientations-of-the-major-axes/d-ellipse0b
			       ellipse0a ellipse0b))
			     (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		       (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		  (set! i (+ i 1))))
		(when (memq 'mouthful features)
		 (unless (= k1 k2)
		  (matrix-set!
		   feature-matrix t1 i
		   (if (= (+ t1 2) t)
		       (cond ((= k1 k)
			      (d-mouthful/d-ellipse0a ellipse0a ellipse0b))
			     ((= k2 k)
			      (d-mouthful/d-ellipse0b ellipse0a ellipse0b))
			     (else (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		       (make-ellipse 0.0 0.0 0.0 0.0 0.0)))
		  (set! i (+ i 1))))
		(when (memq 'finite-difference-of-the-distance-between-ellipse-centres
			    features)
		 (when (> k1 k2)
		  (matrix-set!
		   feature-matrix t1 i
		   (cond
		    ((= (+ t1 2) t)
		     (cond ((= k1 k)
			    (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse0a
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   ((= k2 k)
			    (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse0b
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		    ((= (+ t1 1) t)
		     (cond ((= k1 k)
			    (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse1a
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   ((= k2 k)
			    (d-finite-difference-of-the-distance-between-ellipse-centres/d-ellipse1b
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		    (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		  (set! i (+ i 1))))
		(when (memq 'finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
			    features)
		 (when (> k1 k2)
		  (matrix-set!
		   feature-matrix t1 i
		   (cond
		    ((= (+ t1 2) t)
		     (cond ((= k1 k)
			    (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse0a
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   ((= k2 k)
			    (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse0b
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		    ((= (+ t1 1) t)
		     (cond ((= k1 k)
			    (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse1a
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   ((= k2 k)
			    (d-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres/d-ellipse1b
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		    (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		  (set! i (+ i 1))))
		(when (memq 'finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
			    features)
		 (when (> k1 k2)
		  (matrix-set!
		   feature-matrix t1 i
		   (cond
		    ((= (+ t1 2) t)
		     (cond ((= k1 k)
			    (d-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes/d-ellipse0a
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   ((= k2 k)
			    (d-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes/d-ellipse0b
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		    ((= (+ t1 1) t)
		     (cond ((= k1 k)
			    (d-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes/d-ellipse1a
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   ((= k2 k)
			    (d-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes/d-ellipse1b
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		    (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		  (set! i (+ i 1))))
		(when (memq 'finite-difference-of-mouthful features)
		 (unless (= k1 k2)
		  (matrix-set!
		   feature-matrix t1 i
		   (cond
		    ((= (+ t1 2) t)
		     (cond ((= k1 k)
			    (d-finite-difference-of-mouthful/d-ellipse0a
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   ((= k2 k)
			    (d-finite-difference-of-mouthful/d-ellipse0b
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		    ((= (+ t1 1) t)
		     (cond ((= k1 k)
			    (d-finite-difference-of-mouthful/d-ellipse1a
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   ((= k2 k)
			    (d-finite-difference-of-mouthful/d-ellipse1b
			     ellipse1a ellipse1b ellipse0a ellipse0b))
			   (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		    (else (make-ellipse 0.0 0.0 0.0 0.0 0.0))))
		  (set! i (+ i 1))))))
	      kk)))
	   kk)))
	(- tt 2))
       feature-matrix))
     kk))
   tt)))

(define (ellipse-movie->rmat ellipse-movie features)
 (let ((tt (vector-length ellipse-movie))
       (kk (ellipse-movie-kk ellipse-movie)))
  (unless (>= tt 3) (panic "Movie has fewer than three frames"))
  (let ((rmat (allocate-rmat (number-of-features kk features) (- tt 2))))
   (for-each-n
    (lambda (t)
     (let ((i 0)
	   (frame0 (vector-ref ellipse-movie (+ t 2)))
	   (frame1 (vector-ref ellipse-movie (+ t 1)))
	   (frame2 (vector-ref ellipse-movie t)))
      (for-each-n
       (lambda (k1)
	(let ((ellipse0a (list-ref frame0 k1))
	      (ellipse1a (list-ref frame1 k1))
	      (ellipse2a (list-ref frame2 k1)))
	 (when (memq 'magnitude-of-the-velocity-vector features)
	  (rmat-set!
	   rmat i t (magnitude-of-the-velocity-vector ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'orientation-of-the-velocity-vector features)
	  (rmat-set!
	   rmat i t (orientation-of-the-velocity-vector ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'angular-velocity features)
	  (rmat-set! rmat i t (* 2.0 (angular-velocity ellipse1a ellipse0a)))
	  (set! i (+ i 1)))
	 (when (memq 'finite-difference-of-the-area features)
	  (rmat-set!
	   rmat i t (finite-difference-of-the-area ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'finite-difference-of-the-eccentricity features)
	  (rmat-set!
	   rmat i t
	   (finite-difference-of-the-eccentricity ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'finite-difference-of-the-magnitude-of-the-velocity-vector
		     features)
	  (rmat-set!
	   rmat i t
	   (finite-difference-of-the-magnitude-of-the-velocity-vector
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq
		'finite-difference-of-the-orientation-of-the-velocity-vector
		features)
	  (rmat-set!
	   rmat i t
	   (finite-difference-of-the-orientation-of-the-velocity-vector
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'finite-difference-of-the-angular-velocity features)
	  (rmat-set!
	   rmat i t
	   (finite-difference-of-the-angular-velocity
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'second-finite-difference-of-the-area features)
	  (rmat-set!
	   rmat i t
	   (second-finite-difference-of-the-area
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (when (memq 'second-finite-difference-of-the-eccentricity features)
	  (rmat-set!
	   rmat i t
	   (second-finite-difference-of-the-eccentricity
	    ellipse2a ellipse1a ellipse0a))
	  (set! i (+ i 1)))
	 (for-each-n
	  (lambda (k2)
	   (let ((ellipse0b (list-ref frame0 k2))
		 (ellipse1b (list-ref frame1 k2))
		 (ellipse2b (list-ref frame2 k2)))
	    (when (memq 'distance-between-ellipse-centres features)
	     (when (> k1 k2)
	      (rmat-set!
	       rmat i t (distance-between-ellipse-centres ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'orientation-of-the-vector-between-ellipse-centres
			features)
	     (when (> k1 k2)
	      (rmat-set!
	       rmat i t
	       (orientation-of-the-vector-between-ellipse-centres
		ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'difference-between-the-orientations-of-the-major-axes
			features)
	     (when (> k1 k2)
	      (rmat-set!
	       rmat i t
	       (* 2.0 (difference-between-the-orientations-of-the-major-axes
		       ellipse0a ellipse0b)))
	      (set! i (+ i 1))))
	    (when (memq 'mouthful features)
	     (unless (= k1 k2)
	      (rmat-set! rmat i t (* 2.0 (mouthful ellipse0a ellipse0b)))
	      (set! i (+ i 1))))
	    (when (memq
		   'finite-difference-of-the-distance-between-ellipse-centres
		   features)
	     (when (> k1 k2)
	      (rmat-set!
	       rmat i t
	       (finite-difference-of-the-distance-between-ellipse-centres
		ellipse1a ellipse1b ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
			features)
	     (when (> k1 k2)
	      (rmat-set!
	       rmat i t
	       (finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
		ellipse1a ellipse1b ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
			features)
	     (when (> k1 k2)
	      (rmat-set!
	       rmat i t
	       (finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
		ellipse1a ellipse1b ellipse0a ellipse0b))
	      (set! i (+ i 1))))
	    (when (memq 'finite-difference-of-mouthful features)
	     (unless (= k1 k2)
	      (rmat-set!
	       rmat i t
	       (finite-difference-of-mouthful
		ellipse1a ellipse1b ellipse0a ellipse0b))
	      (set! i (+ i 1))))))
	  kk)))
       kk)))
    (- tt 2))
   rmat)))

(define (radial? kk features j)
 (call-with-current-continuation
  (lambda (return)
   (let ((i 0))
    (for-each-n
     (lambda (k1)
      (when (memq 'magnitude-of-the-velocity-vector features)
       (when (= i j) (return #f))
       (set! i (+ i 1)))
      (when (memq 'orientation-of-the-velocity-vector features)
       (when (= i j) (return #t))
       (set! i (+ i 1)))
      (when (memq 'angular-velocity features)
       (when (= i j) (return #t))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-area features)
       (when (= i j) (return #f))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-eccentricity features)
       (when (= i j) (return #f))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-magnitude-of-the-velocity-vector
		  features)
       (when (= i j) (return #f))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-orientation-of-the-velocity-vector
		  features)
       (when (= i j) (return #t))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-angular-velocity features)
       (when (= i j) (return #t))
       (set! i (+ i 1)))
      (when (memq 'second-finite-difference-of-the-area features)
       (when (= i j) (return #f))
       (set! i (+ i 1)))
      (when (memq 'second-finite-difference-of-the-eccentricity features)
       (when (= i j) (return #f))
       (set! i (+ i 1)))
      (for-each-n
       (lambda (k2)
	(when (memq 'distance-between-ellipse-centres features)
	 (when (> k1 k2)
	  (when (= i j) (return #f))
	  (set! i (+ i 1))))
	(when (memq 'orientation-of-the-vector-between-ellipse-centres
		    features)
	 (when (> k1 k2)
	  (when (= i j) (return #t))
	  (set! i (+ i 1))))
	(when (memq 'difference-between-the-orientations-of-the-major-axes
		    features)
	 (when (> k1 k2)
	  (when (= i j) (return #t))
	  (set! i (+ i 1))))
	(when (memq 'mouthful features)
	 (unless (= k1 k2)
	  (when (= i j) (return #t))
	  (set! i (+ i 1))))
	(when (memq 'finite-difference-of-the-distance-between-ellipse-centres
		    features)
	 (when (> k1 k2)
	  (when (= i j) (return #f))
	  (set! i (+ i 1))))
	(when (memq 'finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
		    features)
	 (when (> k1 k2)
	  (when (= i j) (return #t))
	  (set! i (+ i 1))))
	(when (memq 'finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
		    features)
	 (when (> k1 k2)
	  (when (= i j) (return #t))
	  (set! i (+ i 1))))
	(when (memq 'finite-difference-of-mouthful features)
	 (unless (= k1 k2)
	  (when (= i j) (return #t))
	  (set! i (+ i 1)))))
       kk))
     kk))
   (panic "Feature index out of bounds"))))

(define (feature-index feature kk features)
 (call-with-current-continuation
  (lambda (return)
   (let ((i 0))
    (for-each-n
     (lambda (k1)
      (when (memq 'magnitude-of-the-velocity-vector features)
       (when (equal? feature `(magnitude-of-the-velocity-vector ,k1))
	(return i))
       (set! i (+ i 1)))
      (when (memq 'orientation-of-the-velocity-vector features)
       (when (equal? feature `(orientation-of-the-velocity-vector ,k1))
	(return i))
       (set! i (+ i 1)))
      (when (memq 'angular-velocity features)
       (when (equal? feature `(angular-velocity ,k1)) (return i))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-area features)
       (when (equal? feature `(finite-difference-of-the-area ,k1)) (return i))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-eccentricity features)
       (when (equal? feature `(finite-difference-of-the-eccentricity ,k1))
	(return i))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-magnitude-of-the-velocity-vector
		  features)
       (when (equal?
	      feature
	      `(finite-difference-of-the-magnitude-of-the-velocity-vector ,k1))
	(return i))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-orientation-of-the-velocity-vector
		  features)
       (when (equal?
	      feature
	      `(finite-difference-of-the-orientation-of-the-velocity-vector
		,k1))
	(return i))
       (set! i (+ i 1)))
      (when (memq 'finite-difference-of-the-angular-velocity features)
       (when (equal? feature `(finite-difference-of-the-angular-velocity ,k1))
	(return i))
       (set! i (+ i 1)))
      (when (memq 'second-finite-difference-of-the-area features)
       (when (equal? feature `(second-finite-difference-of-the-area ,k1))
	(return i))
       (set! i (+ i 1)))
      (when (memq 'second-finite-difference-of-the-eccentricity features)
       (when (equal?
	      feature `(second-finite-difference-of-the-eccentricity ,k1))
	(return i))
       (set! i (+ i 1)))
      (for-each-n
       (lambda (k2)
	(when (memq 'distance-between-ellipse-centres features)
	 (when (> k1 k2)
	  (when (equal? feature `(distance-between-ellipse-centres ,k1 ,k2))
	   (return i))
	  (set! i (+ i 1))))
	(when (memq 'orientation-of-the-vector-between-ellipse-centres
		    features)
	 (when (> k1 k2)
	  (when (equal?
		 feature
		 `(orientation-of-the-vector-between-ellipse-centres ,k1 ,k2))
	   (return i))
	  (set! i (+ i 1))))
	(when (memq 'difference-between-the-orientations-of-the-major-axes
		    features)
	 (when (> k1 k2)
	  (when (equal?
		 feature
		 `(difference-between-the-orientations-of-the-major-axes
		   ,k1 ,k2))
	   (return i))
	  (set! i (+ i 1))))
	(when (memq 'mouthful features)
	 (unless (= k1 k2)
	  (when (equal? feature `(mouthful ,k1 ,k2)) (return i))
	  (set! i (+ i 1))))
	(when (memq 'finite-difference-of-the-distance-between-ellipse-centres
		    features)
	 (when (> k1 k2)
	  (when (equal?
		 feature
		 `(finite-difference-of-the-distance-between-ellipse-centres
		   ,k1 ,k2))
	   (return i))
	  (set! i (+ i 1))))
	(when (memq 'finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
		    features)
	 (when (> k1 k2)
	  (when (equal?
		 feature
		 `(finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
		   ,k1 ,k2))
	   (return i))
	  (set! i (+ i 1))))
	(when (memq 'finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
		    features)
	 (when (> k1 k2)
	  (when (equal?
		 feature
		 `(finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
		   ,k1 ,k2))
	   (return i))
	  (set! i (+ i 1))))
	(when (memq 'finite-difference-of-mouthful features)
	 (unless (= k1 k2)
	  (when (equal? feature `(finite-difference-of-mouthful ,k1 ,k2))
	   (return i))
	  (set! i (+ i 1)))))
       kk))
     kk))
   (panic "Feature not found"))))

(define (psi-mu-magnitude-of-the-velocity-vector psi u k)
 (second (matrix-ref (psi-parameters psi)
		     u
		     (feature-index `(magnitude-of-the-velocity-vector ,k)
				    (psi-kk psi)
				    (psi-features psi)))))

(define (psi-mean-orientation-of-the-velocity-vector psi u k)
 (second (matrix-ref (psi-parameters psi)
		     u
		     (feature-index `(orientation-of-the-velocity-vector ,k)
				    (psi-kk psi)
				    (psi-features psi)))))

(define (psi-mean-angular-velocity psi u k)
 (* 0.5 (second (matrix-ref (psi-parameters psi)
			    u
			    (feature-index `(angular-velocity ,k)
					   (psi-kk psi)
					   (psi-features psi))))))

(define (psi-mu-finite-difference-of-the-area psi u k)
 (second (matrix-ref (psi-parameters psi)
		     u
		     (feature-index `(finite-difference-of-the-area ,k)
				    (psi-kk psi)
				    (psi-features psi)))))

(define (psi-mu-finite-difference-of-the-eccentricity psi u k)
 (second (matrix-ref (psi-parameters psi)
		     u
		     (feature-index `(finite-difference-of-the-eccentricity ,k)
				    (psi-kk psi)
				    (psi-features psi)))))

(define (psi-mu-finite-difference-of-the-magnitude-of-the-velocity-vector
	 psi u k)
 (second
  (matrix-ref (psi-parameters psi)
	      u
	      (feature-index
	       `(finite-difference-of-the-magnitude-of-the-velocity-vector ,k)
	       (psi-kk psi)
	       (psi-features psi)))))

(define (psi-mean-finite-difference-of-the-orientation-of-the-velocity-vector
	 psi u k)
 (second (matrix-ref
	  (psi-parameters psi)
	  u
	  (feature-index
	   `(finite-difference-of-the-orientation-of-the-velocity-vector ,k)
	   (psi-kk psi)
	   (psi-features psi)))))

(define (psi-mean-finite-difference-of-the-angular-velocity psi u k)
 (second
  (matrix-ref (psi-parameters psi)
	      u
	      (feature-index `(finite-difference-of-the-angular-velocity ,k)
			     (psi-kk psi)
			     (psi-features psi)))))

(define (psi-mu-second-finite-difference-of-the-area psi u k)
 (second (matrix-ref (psi-parameters psi)
		     u
		     (feature-index `(second-finite-difference-of-the-area ,k)
				    (psi-kk psi)
				    (psi-features psi)))))

(define (psi-mu-second-finite-difference-of-the-eccentricity psi u k)
 (second
  (matrix-ref (psi-parameters psi)
	      u
	      (feature-index `(second-finite-difference-of-the-eccentricity ,k)
			     (psi-kk psi)
			     (psi-features psi)))))

(define (psi-mu-distance-between-ellipse-centres psi u k1 k2)
 (second
  (matrix-ref (psi-parameters psi)
	      u
	      (feature-index `(distance-between-ellipse-centres ,k1 ,k2)
			     (psi-kk psi)
			     (psi-features psi)))))

(define (psi-mean-orientation-of-the-vector-between-ellipse-centres
	 psi u k1 k2)
 (second
  (matrix-ref
   (psi-parameters psi)
   u
   (feature-index `(orientation-of-the-vector-between-ellipse-centres ,k1 ,k2)
		  (psi-kk psi)
		  (psi-features psi)))))

(define (psi-mean-difference-between-the-orientations-of-the-major-axes
	 psi u k1 k2)
 (* 0.5
    (second (matrix-ref
	     (psi-parameters psi)
	     u
	     (feature-index
	      `(difference-between-the-orientations-of-the-major-axes ,k1 ,k2)
	      (psi-kk psi)
	      (psi-features psi))))))

(define (psi-mean-mouthful psi u k1 k2)
 (* 0.5
    (second
     (matrix-ref
      (psi-parameters psi)
      u
      (feature-index `(mouthful ,k1 ,k2) (psi-kk psi) (psi-features psi))))))

(define (psi-mu-finite-difference-of-the-distance-between-ellipse-centres
	 psi u k1 k2)
 (second
  (matrix-ref
   (psi-parameters psi)
   u
   (feature-index
    `(finite-difference-of-the-distance-between-ellipse-centres ,k1 ,k2)
    (psi-kk psi)
    (psi-features psi)))))

(define (psi-mean-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
	 psi u k1 k2)
 (second
  (matrix-ref
   (psi-parameters psi)
   u
   (feature-index
    `(finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres ,k1 ,k2)
    (psi-kk psi)
    (psi-features psi)))))

(define (psi-mean-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
	 psi u k1 k2)
 (second
  (matrix-ref
   (psi-parameters psi)
   u
   (feature-index
    `(finite-difference-of-the-difference-between-the-orientations-of-the-major-axes ,k1 ,k2)
    (psi-kk psi)
    (psi-features psi)))))

(define (psi-mean-finite-difference-of-mouthful psi u k1 k2)
 (second (matrix-ref (psi-parameters psi)
		     u
		     (feature-index `(finite-difference-of-mouthful ,k1 ,k2)
				    (psi-kk psi)
				    (psi-features psi)))))

(define (psi-sigma-magnitude-of-the-velocity-vector psi u k)
 (third (matrix-ref (psi-parameters psi)
		    u
		    (feature-index `(magnitude-of-the-velocity-vector ,k)
				   (psi-kk psi)
				   (psi-features psi)))))

(define (psi-kappa-orientation-of-the-velocity-vector psi u k)
 (third (matrix-ref (psi-parameters psi)
		    u
		    (feature-index `(orientation-of-the-velocity-vector ,k)
				   (psi-kk psi)
				   (psi-features psi)))))

(define (psi-kappa-angular-velocity psi u k)
 (third (matrix-ref (psi-parameters psi)
		    u
		    (feature-index `(angular-velocity ,k)
				   (psi-kk psi)
				   (psi-features psi)))))

(define (psi-sigma-finite-difference-of-the-area psi u k)
 (third (matrix-ref (psi-parameters psi)
		    u
		    (feature-index `(finite-difference-of-the-area ,k)
				   (psi-kk psi)
				   (psi-features psi)))))

(define (psi-sigma-finite-difference-of-the-eccentricity psi u k)
 (third (matrix-ref (psi-parameters psi)
		    u
		    (feature-index `(finite-difference-of-the-eccentricity ,k)
				   (psi-kk psi)
				   (psi-features psi)))))

(define (psi-sigma-finite-difference-of-the-magnitude-of-the-velocity-vector
	 psi u k)
 (third
  (matrix-ref (psi-parameters psi)
	      u
	      (feature-index
	       `(finite-difference-of-the-magnitude-of-the-velocity-vector ,k)
	       (psi-kk psi)
	       (psi-features psi)))))

(define (psi-kappa-finite-difference-of-the-orientation-of-the-velocity-vector
	 psi u k)
 (third (matrix-ref
	 (psi-parameters psi)
	 u
	 (feature-index
	  `(finite-difference-of-the-orientation-of-the-velocity-vector ,k)
	  (psi-kk psi)
	  (psi-features psi)))))

(define (psi-kappa-finite-difference-of-the-angular-velocity psi u k)
 (third
  (matrix-ref (psi-parameters psi)
	      u
	      (feature-index `(finite-difference-of-the-angular-velocity ,k)
			     (psi-kk psi)
			     (psi-features psi)))))

(define (psi-sigma-second-finite-difference-of-the-area psi u k)
 (third (matrix-ref (psi-parameters psi)
		    u
		    (feature-index `(second-finite-difference-of-the-area ,k)
				   (psi-kk psi)
				   (psi-features psi)))))

(define (psi-sigma-second-finite-difference-of-the-eccentricity psi u k)
 (third
  (matrix-ref (psi-parameters psi)
	      u
	      (feature-index `(second-finite-difference-of-the-eccentricity ,k)
			     (psi-kk psi)
			     (psi-features psi)))))

(define (psi-sigma-distance-between-ellipse-centres psi u k1 k2)
 (third
  (matrix-ref (psi-parameters psi)
	      u
	      (feature-index `(distance-between-ellipse-centres ,k1 ,k2)
			     (psi-kk psi)
			     (psi-features psi)))))

(define (psi-kappa-orientation-of-the-vector-between-ellipse-centres
	 psi u k1 k2)
 (third
  (matrix-ref
   (psi-parameters psi)
   u
   (feature-index `(orientation-of-the-vector-between-ellipse-centres ,k1 ,k2)
		  (psi-kk psi)
		  (psi-features psi)))))

(define (psi-kappa-difference-between-the-orientations-of-the-major-axes
	 psi u k1 k2)
 (third
  (matrix-ref (psi-parameters psi)
	      u
	      (feature-index
	       `(difference-between-the-orientations-of-the-major-axes ,k1 ,k2)
	       (psi-kk psi)
	       (psi-features psi)))))

(define (psi-kappa-mouthful psi u k1 k2)
 (third (matrix-ref
	 (psi-parameters psi)
	 u
	 (feature-index `(mouthful ,k1 ,k2) (psi-kk psi) (psi-features psi)))))

(define (psi-sigma-finite-difference-of-the-distance-between-ellipse-centres
	 psi u k1 k2)
 (third
  (matrix-ref
   (psi-parameters psi)
   u
   (feature-index
    `(finite-difference-of-the-distance-between-ellipse-centres ,k1 ,k2)
    (psi-kk psi)
    (psi-features psi)))))

(define (psi-kappa-finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres
	 psi u k1 k2)
 (third
  (matrix-ref
   (psi-parameters psi)
   u
   (feature-index
    `(finite-difference-of-the-orientation-of-the-vector-between-ellipse-centres ,k1 ,k2)
    (psi-kk psi)
    (psi-features psi)))))

(define (psi-kappa-finite-difference-of-the-difference-between-the-orientations-of-the-major-axes
	 psi u k1 k2)
 (third
  (matrix-ref
   (psi-parameters psi)
   u
   (feature-index
    `(finite-difference-of-the-difference-between-the-orientations-of-the-major-axes ,k1 ,k2)
    (psi-kk psi)
    (psi-features psi)))))

(define (psi-kappa-finite-difference-of-mouthful psi u k1 k2)
 (third (matrix-ref (psi-parameters psi)
		    u
		    (feature-index `(finite-difference-of-mouthful ,k1 ,k2)
				   (psi-kk psi)
				   (psi-features psi)))))

(define (read-psis pathname)
 (map (lambda (psi)
       (if (boolean? (psi-features psi))
	   (make-psi
	    (psi-name psi)
	    (if (psi-features psi) *full-features* *partial-features*)
	    (psi-kk psi)
	    (psi-parameters psi)
	    (psi-a psi)
	    (psi-b psi))
	   psi))
      (read-object-from-file (replace-extension pathname "psis"))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
