(MODULE
 KGP
 (WITH
  QOBISCHEME
  XLIB
  HLIB-CHANGE-MOVIE
  HLIB-EM
  HLIB-FIND-LINES
  HLIB-HULL
  HLIB-INTERPRET
  HLIB-KINEMATIC
  HLIB-LOCATE
  HLIB-MERGE-LINES
  HLIB-MOST
  HLIB-POSTSCRIPT
  HLIB-RATIO-REGIONS
  HLIB-RILIB
  HLIB-SDE
  HLIB-SNAKE
  HLIB-SPAN
  HLIB-SPEAK
  HLIB-TRACK
  HLIB-TRAIN-CLASSIFY
  HLIB-WBM)
 (MAIN MAIN))
;;; LaHaShem HaAretz U'Mloah

;;; TTMTTD
;;; 1. The Same-Layer and Diff-Layer buttons can create a transient state where
;;;    the same-layer? relation is not transitive.
;;; 2. The Same-Layer button doesn't check that the polygons don't overlap.

(include "QobiScheme.sch")
(include "kgp.sch")
(include "c-externals.sc")

(set! *program* "kgp")
(set! *panic?* #t)

;;; Variables

(define *pathname* #f)
(define *mode* #f)
(define *polygons* #f)
(define *i* #f)
(define *point* #f)
(define *mark* #f)
(define *template* #f)
(define *frame* #f)

;;; Procedures

(define (substring? s1 s2)
 ;; Belongs in QobiScheme.
 (let ((n (string-length s1)))
  (some-n (lambda (i)
	   (every-n (lambda (j)
		     (char=? (string-ref s1 j) (string-ref s2 (+ j i))))
		    n))
	  (+ (- (string-length s2) n) 1))))

(define (draw-line-segment l)
 (xdrawline *display* *display-pane* *thin-gc*
	    (x (p l)) (y (p l)) (x (q l)) (y (q l))))

(define (draw-polygon p)
 (for-each-n
  (lambda (i)
   (draw-line-segment
    (make-line-segment (vector-ref p i)
		       (vector-ref p (modulo (+ i 1) (vector-length p))))))
  (vector-length p)))

(define (draw-bar r l)
 (let* ((dx (- (x (p l)) (x (q l))))
	(dy (- (y (p l)) (y (q l))))
	(m (sqrt (+ (sqr dx) (sqr dy))))
	(dx (* 10.0 (/ dx m)))
	(dy (* 10.0 (/ dy m))))
  (xdrawline *display* *display-pane* *medium-gc*
	     (- (x r) dx) (- (y r) dy) (+ (x r) dx) (+ (y r) dy))))

(define (draw-filled-circle p diameter)
 (xfillarc *display* *display-pane* *thin-gc*
	   (- (x p) (/ diameter 2))
	   (- (y p) (/ diameter 2))
	   diameter diameter 0 (* 360 64)))

(define (draw-unfilled-circle p diameter)
 (xdrawarc *display* *display-pane* *thin-gc*
	   (- (x p) (/ diameter 2))
	   (- (y p) (/ diameter 2))
	   diameter diameter 0 (* 360 64)))

(define (redraw!)
 (xclearwindow *display* *display-pane*)
 (for-each draw-polygon *polygons*)
 (unless (null? *kinematic-models*)
  (let* ((ns (enumerate (length *polygons*)))
	 (same-layer?
	  (kinematic-model-same-layer? (list-ref *kinematic-models* *i*)))
	 (joint-i?
	  (kinematic-model-joint-i? (list-ref *kinematic-models* *i*)))
	 (joint-j?
	  (kinematic-model-joint-j? (list-ref *kinematic-models* *i*)))
	 (joint-w?
	  (kinematic-model-joint-w? (list-ref *kinematic-models* *i*)))
	 (ground? (kinematic-model-ground? (list-ref *kinematic-models* *i*)))
	 (layers (equivalence-classesp
		  (lambda (pi pj)
		   (matrix-ref same-layer?
			       (positionq pi *polygons*)
			       (positionq pj *polygons*)))
		  *polygons*))
	 (contacts (all-contacts *polygons* *epsilon2*)))
   (for-each (lambda (pi)
	      (let ((layer (number->string (positionp memq pi layers)))
		    (p (layer-id-location pi)))
	       (xdrawstring *display* *display-pane* *roman-gc*
			    (x p) (y p) layer (string-length layer))))
	     *polygons*)
   (for-each
    (lambda (contact k)
     (let ((p (contact-p contact))
	   (l1 (contact-l1 contact))
	   (l2 (contact-l2 contact)))
      (if (vector-ref joint-i? k)
	  (if (vector-ref joint-j? k)
	      (if (vector-ref joint-w? k)
		  (draw-filled-circle p 10)
		  (draw-unfilled-circle p 10))
	      (if (vector-ref joint-w? k)
		  (begin (draw-filled-circle p 10)
			 (draw-bar p l2))
		  (begin (draw-unfilled-circle p 10)
			 (draw-bar p l2))))
	  (if (vector-ref joint-j? k)
	      (if (vector-ref joint-w? k)
		  (begin (draw-filled-circle p 10)
			 (draw-bar p l1))
		  (begin (draw-unfilled-circle p 10)
			 (draw-bar p l1)))
	      (if (vector-ref joint-w? k)
		  (begin (draw-filled-circle p 10)
			 (draw-bar p l1)
			 (draw-bar p l2))
		  #f)))))
    contacts
    (enumerate (length contacts)))
   (for-each
    (lambda (pi i)
     (when (vector-ref ground? i)
      (let ((p (ground-location pi)))
       (draw-line-segment (make-line-segment p (v+ p '#(0 5))))
       (draw-line-segment (make-line-segment (v+ p '#(-3 5)) (v+ p '#(3 5))))
       (draw-line-segment (make-line-segment (v+ p '#(-2 7)) (v+ p '#(2 7))))
       (draw-line-segment
	(make-line-segment (v+ p '#(-1 9)) (v+ p '#(1 9)))))))
    *polygons*
    ns)
   (unless (zero? (length *polygons*))
    (if (polygon-stable? *polygons* contacts (list-ref *kinematic-models* *i*))
	(xdrawstring
	 *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 25)
	 "Stable" (string-length "Stable"))
	(xdrawstring
	 *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 25)
	 "Unstable" (string-length "Unstable"))))))
 (when *pathname*
  (xdrawstring
   *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 5)
   *pathname* (string-length *pathname*))))

;;; Commands

(define (save-command)
 (message "")
 (unless *pathname*
  (message "You have not specified a pathname")
  (abort))
 (write-object-to-file (list *polygons* *kinematic-models*)
		       (default-extension *pathname* "kgp")))

(define (load-command)
 (message "")
 (unless *pathname*
  (message "You have not specified a pathname")
  (abort))
 (unless (can-open-file-for-input? (default-extension *pathname* "kgp"))
  (message "File does not exist")
  (abort))
 (let ((kgp (read-object-from-file (default-extension *pathname* "kgp"))))
  (set! *polygons* (first kgp))
  (set! *kinematic-models* (second kgp)))
 (set! *i* 0)
 (redraw-display-pane))

;;; Top Level

(define-application viewer 640 500 0 5 6
 (lambda ()
  (set! *polygons* '())
  (set! *kinematic-models* '())
  (set! *i* 0)
  (set! *mode* 'add-polygon)
  (set! *point* #f)
  (set! *mark* #f)
  (set! *frame* 0)
  (define-button 0 0 "Help" #f
   (lambda ()
    (message "")
    (help-command)))
  (define-button 1 0 "Abort" #f abort-command)
  (define-button 3 0 "Load" #f load-command)
  (define-button 4 0 "Save" #f save-command)
  (define-button 5 0 "Quit" #f quit)
  (define-button 2 1 "Same Layer" #f
   (lambda ()
    (when (null? *kinematic-models*) (message "No models") (abort))
    (unless (and *point* *mark*)
     (message "You need to specify two polygons")
     (abort))
    (let ((i (positionq *point* *polygons*))
	  (j (positionq *mark* *polygons*)))
     (cond ((< i j)
	    (matrix-set!
	     (kinematic-model-same-layer? (list-ref *kinematic-models* *i*))
	     i j #t))
	   ((< j i)
	    (matrix-set!
	     (kinematic-model-same-layer? (list-ref *kinematic-models* *i*))
	     j i #t))
	   (else (message "You need to specify two different polygons")
		 (abort))))
    (redraw-display-pane)))
  (define-button 3 1 "Diff Layer" #f
   (lambda ()
    (when (null? *kinematic-models*) (message "No models") (abort))
    (unless (and *point* *mark*)
     (message "You need to specify two polygons")
     (abort))
    (let ((i (positionq *point* *polygons*))
	  (j (positionq *mark* *polygons*)))
     (cond ((< i j)
	    (matrix-set!
	     (kinematic-model-same-layer? (list-ref *kinematic-models* *i*))
	     i j #f))
	   ((< j i)
	    (matrix-set!
	     (kinematic-model-same-layer? (list-ref *kinematic-models* *i*))
	     j i #f))
	   (else (message "You need to specify two different polygons")
		 (abort))))
    (redraw-display-pane)))
  (define-radio-buttons *mode* (lambda () (message ""))
   (2 0 coord "Coord")
   (0 1 add-polygon "Add Poly")
   (1 1 delete-polygon "Delete Poly")
   (4 1 ground "Ground")
   (5 1 not-ground "Not Ground")
   (0 2 rigid-i "Rigid i")
   (1 2 flex-i "Flex i")
   (2 2 rigid-j "Rigid j")
   (3 2 flex-j "Flex j")
   (4 2 rigid-w "Rigid w")
   (5 2 flex-w "Flex w"))
  (define-button 0 3 "Previous" #f
   (lambda ()
    (message "")
    (when (null? *kinematic-models*) (message "No models") (abort))
    (when (zero? *i*) (message "First model") (abort))
    (set! *i* (- *i* 1))
    (redraw-display-pane)))
  (define-button 1 3 "Next" #f
   (lambda ()
    (message "")
    (when (null? *kinematic-models*) (message "No models") (abort))
    (when (= *i* (- (length *kinematic-models*) 1))
     (message "Last model")
     (abort))
    (set! *i* (+ *i* 1))
    (redraw-display-pane)))
  (define-button 2 3 "Delete" #f
   (lambda ()
    (message "")
    (when (null? *kinematic-models*) (message "No models") (abort))
    (set! *kinematic-models* (list-remove *kinematic-models* *i*))
    (set! *i* (min *i* (- (length *kinematic-models*) 1)))
    (redraw-display-pane)))
  (define-button 3 3 "Add" #f
   (lambda ()
    (message "")
    (let ((n (length *polygons*))
	  (m (length (all-contacts *polygons* *epsilon2*))))
     (set! *kinematic-models*
	   (append *kinematic-models*
		   (list (make-kinematic-model (make-matrix n n #f)
					       (make-vector m #f)
					       (make-vector m #f)
					       (make-vector m #f)
					       (make-vector n #f))))))
    (set! *i* (- (length *kinematic-models*) 1))
    (redraw-display-pane)))
  (define-button 4 3 "Models" #f
   (lambda ()
    (message "")
    (set! *kinematic-models*
	  (simplified-minimal-polygon-kinematic-models *polygons*))
    (set! *i* 0)
    (redraw-display-pane)))
  (define-button 5 3 "Postscript" #f
   (lambda ()
    (let ((preferred-models (simplified-minimal-polygon-kinematic-models
			     *polygons*)))
     (for-each (lambda (preferred-model i)
		(kinematic-model->postscript
		 *polygons*
		 preferred-model
		 (format #f "~a-preferred-model-~s"
			 (strip-extension *pathname*) i)))
	       preferred-models
	       (enumerate (length preferred-models)))
     (let ((more-preferred-models
	    (cardinality-circumscription preferred-models)))
      (for-each (lambda (more-preferred-model i)
		 (kinematic-model->postscript
		  *polygons*
		  more-preferred-model
		  (format #f "~a-more-preferred-model-~s"
			  (strip-extension *pathname*) i)))
		more-preferred-models
		(enumerate (length more-preferred-models)))))))
  (define-button 0 4 "First Frame" #f
   (lambda ()
    (unless *template* (message "No template") (abort))
    (set! *frame* 0)
    (set! *pathname* (format #f *template* *frame*))
    (load-command)))
  (define-button 1 4 "Prev Frame" #f
   (lambda ()
    (unless *template* (message "No template") (abort))
    (when (zero? *frame*) (message "First frame") (abort))
    (set! *frame* (- *frame* 1))
    (set! *pathname* (format #f *template* *frame*))
    (load-command)))
  (define-button 2 4 "Next Frame" #f
   (lambda ()
    (unless *template* (message "No template") (abort))
    (unless (can-open-file-for-input?
	     (default-extension (format #f *template* (+ *frame* 1)) "kgp"))
     (message "Last frame")
     (abort))
    (set! *frame* (+ *frame* 1))
    (set! *pathname* (format #f *template* *frame*))
    (load-command)))
  (define-key (list (control #\x) (control #\c)) "Quit" quit)
  (define-key (list (control #\x) (control #\f)) "Load" load-command)
  (define-key (list (control #\x) (control #\s)) "Save" save-command)
  (define-key (control #\h) "Help"  help-command)
  (define-key (control #\g) "Abort" abort-command)
  (when (and *pathname* (not (substring? "~s" *pathname*))) (load-command)))
 (lambda () #f)
 (lambda () #f)
 (lambda ()
  (redraw!)
  (define-region 0 0 *display-pane-width* *display-pane-height*
   (lambda (x1 y1)
    (case *mode*
     ((coord)
      (message (format #f "~a ~a" x1 y1))
      (let ((p (find-if
		(lambda (polygon)
		 (point-touches-polygon? (vector x1 y1) polygon *epsilon1*))
		*polygons*)))
       (when p
	(set! *mark* *point*)
	(set! *point* p))))
     ((add-polygon)
      (message "")
      (set! *point* #f)
      (set! *mark* #f)
      (set! *polygons*
	    (append
	     *polygons*
	     (list
	      (list->vector
	       (convex-hull
		(let loop ((points
			    (list (vector (quantize x1 10) (quantize y1 10)))))
		 (let* ((result
			 (tracking-pointer
			  #t
			  #t
			  (lambda (x2 y2)
			   (when (>= (length points) 2)
			    (for-each
			     (lambda (p1 p2)
			      (xdrawline
			       *display* *display-pane* *thin-flipping-gc*
			       (x p1) (y p1) (x p2) (y p2)))
			     (but-last points)
			     (rest points)))
			   (xdrawline
			    *display* *display-pane* *thin-flipping-gc*
			    (x (last points))
			    (y (last points))
			    (quantize x2 10)
			    (quantize y2 10)))))
			(x2 (quantize (first result) 10))
			(y2 (quantize (second result) 10)))
		  (if (and (= x2 (x (first points))) (= y2 (y (first points))))
		      points
		      (loop (append points (list (vector x2 y2))))))))))))

      (let ((n (length *polygons*))
	    (m (length (all-contacts *polygons* *epsilon2*))))
       (for-each
	(lambda (kinematic-model)
	 (set-kinematic-model-same-layer?!
	  kinematic-model (make-matrix n n #f))
	 (set-kinematic-model-joint-i?! kinematic-model (make-vector m #f))
	 (set-kinematic-model-joint-j?! kinematic-model (make-vector m #f))
	 (set-kinematic-model-joint-w?! kinematic-model (make-vector m #f))
	 (set-kinematic-model-ground?! kinematic-model (make-vector n #f)))
	*kinematic-models*))
      (redraw-display-pane))
     ((delete-polygon)
      (message "")
      (set! *point* #f)
      (set! *mark* #f)
      (let ((i (position-if
		(lambda (polygon)
		 (point-touches-polygon? (vector x1 y1) polygon *epsilon1*))
		*polygons*)))
       (when i
	(set! *polygons* (list-remove *polygons* i))
	(let ((n (length *polygons*))
	      (m (length (all-contacts *polygons* *epsilon2*))))
	 (for-each
	  (lambda (kinematic-model)
	   (set-kinematic-model-same-layer?!
	    kinematic-model (make-matrix n n #f))
	   (set-kinematic-model-joint-i?! kinematic-model (make-vector m #f))
	   (set-kinematic-model-joint-j?! kinematic-model (make-vector m #f))
	   (set-kinematic-model-joint-w?! kinematic-model (make-vector m #f))
	   (set-kinematic-model-ground?! kinematic-model (make-vector n #f)))
	  *kinematic-models*))))
      (redraw-display-pane))
     ((ground)
      (message "")
      (when (null? *kinematic-models*) (message "No models") (abort))
      (for-each
       (lambda (pi i)
	(when (point-touches-polygon? (vector x1 y1) pi *epsilon1*)
	 (vector-set!
	  (kinematic-model-ground? (list-ref *kinematic-models* *i*)) i #t)))
       *polygons*
       (enumerate (length *polygons*)))
      (redraw-display-pane))
     ((not-ground)
      (message "")
      (when (null? *kinematic-models*) (message "No models") (abort))
      (for-each
       (lambda (pi i)
	(when (point-touches-polygon? (vector x1 y1) pi *epsilon1*)
	 (vector-set!
	  (kinematic-model-ground? (list-ref *kinematic-models* *i*)) i #f)))
       *polygons*
       (enumerate (length *polygons*)))
      (redraw-display-pane))
     ((rigid-i)
      (message "")
      (when (null? *kinematic-models*) (message "No models") (abort))
      (let ((contacts (all-contacts *polygons* *epsilon2*)))
       (for-each
	(lambda (contact i)
	 (when (< (distance (contact-p contact) (vector x1 y1)) 10)
	  (vector-set!
	   (kinematic-model-joint-i? (list-ref *kinematic-models* *i*))
	   i
	   #t)))
	contacts
	(enumerate (length contacts))))
      (redraw-display-pane))
     ((flex-i)
      (message "")
      (when (null? *kinematic-models*) (message "No models") (abort))
      (let ((contacts (all-contacts *polygons* *epsilon2*)))
       (for-each
	(lambda (contact i)
	 (when (< (distance (contact-p contact) (vector x1 y1)) 10)
	  (vector-set!
	   (kinematic-model-joint-i? (list-ref *kinematic-models* *i*))
	   i
	   #f)))
	contacts
	(enumerate (length contacts))))
      (redraw-display-pane))
     ((rigid-j)
      (message "")
      (when (null? *kinematic-models*) (message "No models") (abort))
      (let ((contacts (all-contacts *polygons* *epsilon2*)))
       (for-each
	(lambda (contact i)
	 (when (< (distance (contact-p contact) (vector x1 y1)) 10)
	  (vector-set!
	   (kinematic-model-joint-j? (list-ref *kinematic-models* *i*))
	   i
	   #t)))
	contacts
	(enumerate (length contacts))))
      (redraw-display-pane))
     ((flex-j)
      (message "")
      (when (null? *kinematic-models*) (message "No models") (abort))
      (let ((contacts (all-contacts *polygons* *epsilon2*)))
       (for-each
	(lambda (contact i)
	 (when (< (distance (contact-p contact) (vector x1 y1)) 10)
	  (vector-set!
	   (kinematic-model-joint-j? (list-ref *kinematic-models* *i*))
	   i
	   #f)))
	contacts
	(enumerate (length contacts))))
      (redraw-display-pane))
     ((rigid-w)
      (message "")
      (when (null? *kinematic-models*) (message "No models") (abort))
      (let ((contacts (all-contacts *polygons* *epsilon2*)))
       (for-each
	(lambda (contact i)
	 (when (< (distance (contact-p contact) (vector x1 y1)) 10)
	  (vector-set!
	   (kinematic-model-joint-w? (list-ref *kinematic-models* *i*))
	   i
	   #t)))
	contacts
	(enumerate (length contacts))))
      (redraw-display-pane))
     ((flex-w)
      (message "")
      (when (null? *kinematic-models*) (message "No models") (abort))
      (let ((contacts (all-contacts *polygons* *epsilon2*)))
       (for-each
	(lambda (contact i)
	 (when (< (distance (contact-p contact) (vector x1 y1)) 10)
	  (vector-set!
	   (kinematic-model-joint-w? (list-ref *kinematic-models* *i*))
	   i
	   #f)))
	contacts
	(enumerate (length contacts))))
      (redraw-display-pane))
     (else (fuck-up))))))
 (lambda ()
  (set! *pathname* (and (not (string=? *input* "")) (strip-extension *input*)))
  (when (and *pathname* (substring? "~s" *pathname*))
   (set! *template* *pathname*))
  (redraw-display-pane)))

(define-command (main
		 (at-most-one ("fast" fast?) ("cheap" cheap?))
		 (at-most-one ("window-position"
			       window-position?
			       (window-position-x "x" integer-argument 0)
			       (window-position-y "y" integer-argument 0)))
		 (optional (pathname "pathname" string-argument #f)))
 (set! *cache?* #f)
 (set! *fast?* (or fast? cheap?))
 (set! *cheap?* cheap?)
 (set! *window-position?* #f)
 (when window-position?
  (set! *window-position?* #t)
  (set! *window-position-x* window-position-x)
  (set! *window-position-y* window-position-y))
 (when (and pathname (string=? pathname "-")) (panic "pathname cannot be -"))
 (set! *pathname* (and pathname (strip-extension pathname)))
 (when (and *pathname* (substring? "~s" *pathname*))
  (set! *template* *pathname*))
 (when (and
	*pathname*
	(not (substring? "~s" *pathname*))
	(not (can-open-file-for-input? (default-extension *pathname* "kgp"))))
  (panic "File does not exist"))
 (viewer '()))

;;; Tam V'Nishlam Shevah L'El Borei Olam
