(MODULE
 LEONARD
 (WITH
  QOBISCHEME
  XLIB
  HLIB-CHANGE-MOVIE
  HLIB-EM
  HLIB-FIND-LINES
  HLIB-HULL
  HLIB-INTERPRET
  HLIB-KINEMATIC
  HLIB-LOCATE
  HLIB-MERGE-LINES
  HLIB-MOST
  HLIB-POSTSCRIPT
  HLIB-RATIO-REGIONS
  HLIB-RILIB
  HLIB-SDE
  HLIB-SNAKE
  HLIB-SPAN
  HLIB-SPEAK
  HLIB-TRACK
  HLIB-TRAIN-CLASSIFY
  HLIB-WBM)
 (MAIN MAIN))
;;; LaHaShem HaAretz U'Mloah

;;; needs work
;;;  1. meteor convex hull is buggy
;;;  2. Load command

;;; The following procedures have tolerances
;;;  POINT-TOUCHES-LINE?
;;;  POINT-TOUCHES-LINE-SEGMENT?
;;;  POINT-OF-NTERSECTION-OF-LINE-SEGMENTS
;;;  LINE-SEGMENTS-INTERSECT?
;;;  LINE-SEGMENTS-CROSS?
;;;  LINE-SEGMENTS-TOUCH?
;;;  POINT-TOUCHES-POLYGON?
;;;  CONTACTS
;;;  POLYGONS-CROSS?
;;;  POLYGONS-TOUCH?
;;;  ALL-CONTACTS
;;;  MERGE-COLLINEAR-EDGES

(include "QobiScheme.sch")
(include "leonard.sch")
(include "c-externals.sc")

(set! *program* "leonard")
(set! *panic?* #t)

;;; Parameters

(define *stretch* 10)

;;; Variables

(define *pathname* #f)
(define *t* #f)
(define *end* #f)
(define *start* #f)
(define *finish* #f)
(define *first-frame* #f)
(define *last-frame* #f)
(define *i* #f)
(define *mode* #f)
(define *point* #f)
(define *mark* #f)
(define *fill* #f)
(define *x-weight* #f)
(define *y-weight* #f)
(define *a-weight* #f)
(define *correspondence-threshold* #f)
(define *events* #f)
(define *red?* #f)
(define *green?* #f)
(define *blue?* #f)
(define *yellow?* #f)
(define *rhl* 13)
(define *rhh* 336)
(define *rsl* 36)
(define *rsh* 74)
(define *rvl* 35)
(define *rvh* 70)
(define *ghl* 122)
(define *ghh* 164)
(define *gsl* 29)
(define *gsh* 71)
(define *gvl* 22)
(define *gvh* 55)
(define *bhl* 192)
(define *bhh* 227)
(define *bsl* 32)
(define *bsh* 66)
(define *bvl* 27)
(define *bvh* 61)
(define *yhl* 50)
(define *yhh* 70)
(define *ysl* 40)
(define *ysh* 65)
(define *yvl* 55)
(define *yvh* 80)
(define *meteor-motion-threshold* 24)
(define *meteor-colour-pixel-mass-threshold* 60)
(define *meteor-colour-proximity-threshold* 4)
(define *meteor-motion-pixel-mass-threshold* 200)
(define *meteor-motion-proximity-threshold* 8)
(define *new-indexing?* #f)

;;; Procedures

(define (draw-line-segment l)
 (xdrawline *display* *display-pane* *thin-gc*
	    (x (p l)) (y (p l)) (x (q l)) (y (q l))))

(define (draw-polygon p)
 (for-each-n
  (lambda (i)
   (draw-line-segment
    (make-line-segment (vector-ref p i)
		       (vector-ref p (modulo (+ i 1) (vector-length p))))))
  (vector-length p)))

(define (draw-bar r l)
 (let* ((dx (- (x (p l)) (x (q l))))
	(dy (- (y (p l)) (y (q l))))
	(m (sqrt (+ (sqr dx) (sqr dy))))
	(dx (* 10.0 (/ dx m)))
	(dy (* 10.0 (/ dy m))))
  (xdrawline *display* *display-pane* *medium-gc*
	     (- (x r) dx) (- (y r) dy) (+ (x r) dx) (+ (y r) dy))))

(define (draw-filled-circle p diameter)
 (xfillarc *display* *display-pane* *thin-gc*
	   (- (x p) (/ diameter 2))
	   (- (y p) (/ diameter 2))
	   diameter diameter 0 (* 360 64)))

(define (draw-unfilled-circle p diameter)
 (xdrawarc *display* *display-pane* *thin-gc*
	   (- (x p) (/ diameter 2))
	   (- (y p) (/ diameter 2))
	   diameter diameter 0 (* 360 64)))

(define (show-polygons! t)
 (for-each draw-polygon (vector-ref *polygon-movie* t)))

(define (show-frame! t)
 (case *mode*
  ((movie) (meteor-show-frame (+ *start* t)))
  ((polygons)
   (xclearwindow *display* *display-pane*)
   (show-polygons! t))
  ((movie-and-polygons)
   (meteor-show-frame (+ *start* t))
   (xsetforeground *display* *thin-gc* (xcolor-pixel (second *background*)))
   (show-polygons! t)
   (xsetforeground *display* *thin-gc* (xcolor-pixel (second *foreground*))))
  ((code-and-polygons)
   (meteor-show-code-frame (+ *start* t))
   (show-polygons! t))
  (else (fuck-up))))

(define (redraw!)
 (when (or (eq? *mode* 'movie)
	   (eq? *mode* 'polygons)
	   (eq? *mode* 'movie-and-polygons)
	   (eq? *mode* 'code-and-polygons))
  (show-frame! *t*))
 (when (or (eq? *mode* 'polygons)
	   (eq? *mode* 'movie-and-polygons)
	   (eq? *mode* 'code-and-polygons))
  (when *kinematic-models*
   (unless (null? (vector-ref *kinematic-models* *t*))
    (let* ((polygons (vector-ref *polygon-movie* *t*))
	   (kinematic-model (list-ref (vector-ref *kinematic-models* *t*) *i*))
	   (ns (enumerate (length (vector-ref *polygon-movie* *t*))))
	   (same-layer? (kinematic-model-same-layer? kinematic-model))
	   (joint-i? (kinematic-model-joint-i? kinematic-model))
	   (joint-j? (kinematic-model-joint-j? kinematic-model))
	   (joint-w? (kinematic-model-joint-w? kinematic-model))
	   (ground? (kinematic-model-ground? kinematic-model))
	   (layers
	    (equivalence-classesp
	     (lambda (pi pj)
	      (matrix-ref
	       same-layer? (positionq pi polygons) (positionq pj polygons)))
	     polygons))
	   (contacts (all-contacts polygons *epsilon2*)))
     (when (eq? *mode* 'movie-and-polygons)
      (xsetforeground
       *display* *thin-gc* (xcolor-pixel (second *background*)))
      (xsetforeground
       *display* *medium-gc* (xcolor-pixel (second *background*)))
      (xsetforeground
       *display* *roman-gc* (xcolor-pixel (second *background*))))
     (for-each (lambda (pi)
		(let ((layer (number->string (positionp memq pi layers)))
		      (p (layer-id-location pi)))
		 (xdrawstring *display* *display-pane* *roman-gc*
			      (x p) (y p) layer (string-length layer))))
	       polygons)
     (for-each (lambda (contact k)
		(let ((p (contact-p contact))
		      (l1 (contact-l1 contact))
		      (l2 (contact-l2 contact)))
		 (if (vector-ref joint-i? k)
		     (if (vector-ref joint-j? k)
			 (if (vector-ref joint-w? k)
			     (draw-filled-circle p 10)
			     (draw-unfilled-circle p 10))
			 (if (vector-ref joint-w? k)
			     (begin (draw-filled-circle p 10)
				    (draw-bar p l2))
			     (begin (draw-unfilled-circle p 10)
				    (draw-bar p l2))))
		     (if (vector-ref joint-j? k)
			 (if (vector-ref joint-w? k)
			     (begin (draw-filled-circle p 10)
				    (draw-bar p l1))
			     (begin (draw-unfilled-circle p 10)
				    (draw-bar p l1)))
			 (if (vector-ref joint-w? k)
			     (begin (draw-filled-circle p 10)
				    (draw-bar p l1)
				    (draw-bar p l2))
			     #f)))))
	       contacts
	       (enumerate (length contacts)))
     (for-each
      (lambda (pi i)
       (when (vector-ref ground? i)
	(let ((p (ground-location pi)))
	 (draw-line-segment (make-line-segment p (v+ p '#(0 5))))
	 (draw-line-segment (make-line-segment (v+ p '#(-3 5)) (v+ p '#(3 5))))
	 (draw-line-segment (make-line-segment (v+ p '#(-2 7)) (v+ p '#(2 7))))
	 (draw-line-segment
	  (make-line-segment (v+ p '#(-1 9)) (v+ p '#(1 9)))))))
      polygons
      ns)
     (when (eq? *mode* 'movie-and-polygons)
      (xsetforeground *display* *thin-gc* (xcolor-pixel (second *foreground*)))
      (xsetforeground
       *display* *medium-gc* (xcolor-pixel (second *foreground*)))
      (xsetforeground
       *display* *roman-gc* (xcolor-pixel (second *foreground*))))
     (unless (zero? (length polygons))
      (if (polygon-stable? polygons contacts kinematic-model)
	  (xdrawstring
	   *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 25)
	   "Stable" (string-length "Stable"))
	  (xdrawstring
	   *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 25)
	   "Unstable" (string-length "Unstable"))))))))
 (for-each
  (lambda (i event)
   (xdrawstring *display* *display-pane* *roman-gc*
		(+ *width* 5) (+ (* (+ i 1) *roman-height*) 5)
		event (string-length event)))
  (enumerate (length *events*))
  *events*)
 (when *pathname*
  (xdrawstring
   *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 5)
   *pathname* (string-length *pathname*))))

(define (wait-until-motion)
 (let ((n 3)
       (m 2)
       (f (get-current-frame)))
  (let loop ()
   (when (< (get-current-frame) f) (set! f (get-current-frame)))
   (unless (and (>= (- (get-current-frame) f) n)
		(let loop ((i 0) (c 0))
		 (if (>= i n)
		     (>= c m)
		     (loop (+ i 1)
			   (if (get-motion? (- (get-current-frame) i))
			       (+ c 1)
			       c)))))
    (interruptible-usleep 33333)
    (when (abort?)
     (set! *mode* 'viewfinder)
     (redraw-buttons)
     (meteor-turn-on-viewfinder! 1)
     (abort))
    (loop)))))

(define (wait-until-no-motion)
 (let ((n 15)
       (m 2)
       (f (get-current-frame)))
  (let loop ()
   (when (< (get-current-frame) f) (set! f (get-current-frame)))
   (unless (and (>= (- (get-current-frame) f) n)
		(let loop ((i 0) (c 0))
		 (if (>= i n)
		     (< c m)
		     (loop (+ i 1)
			   (if (get-motion? (- (get-current-frame) i))
			       (+ c 1)
			       c)))))
    (interruptible-usleep 33333)
    (when (abort?)
     (set! *mode* 'viewfinder)
     (redraw-buttons)
     (meteor-turn-on-viewfinder! 1)
     (abort))
    (loop)))))

(define (wait-until-two-or-more-objects)
 (let ((n 3)
       (m 2)
       (f (get-current-frame)))
  (let loop ()
   (when (< (get-current-frame) f) (set! f (get-current-frame)))
   (unless (and (>= (- (get-current-frame) f) n)
		(let loop ((i 0) (c 0))
		 (if (>= i n)
		     (>= c m)
		     (loop (+ i 1)
			   (if (>= (get-object-k (- (get-current-frame) i)) 2)
			       (+ c 1)
			       c)))))
    (interruptible-usleep 33333)
    (when (abort?)
     (set! *mode* 'viewfinder)
     (redraw-buttons)
     (meteor-turn-on-viewfinder! 1)
     (abort))
    (loop)))))

(define (wait-until-less-than-two-objects)
 (let ((n 15)
       (m 2)
       (f (get-current-frame)))
  (let loop ()
   (when (< (get-current-frame) f) (set! f (get-current-frame)))
   (unless (and (>= (- (get-current-frame) f) n)
		(let loop ((i 0) (c 0))
		 (if (>= i n)
		     (< c m)
		     (loop (+ i 1)
			   (if (>= (get-object-k (- (get-current-frame) i)) 2)
			       (+ c 1)
			       c)))))
    (interruptible-usleep 33333)
    (when (abort?)
     (set! *mode* 'viewfinder)
     (redraw-buttons)
     (meteor-turn-on-viewfinder! 1)
     (abort))
    (loop)))))

(define (wait-until-six-or-more-objects)
 (let ((n 3)
       (m 2)
       (f (get-current-frame)))
  (let loop ()
   (when (< (get-current-frame) f) (set! f (get-current-frame)))
   (unless (and (>= (- (get-current-frame) f) n)
		(let loop ((i 0) (c 0))
		 (if (>= i n)
		     (>= c m)
		     (loop (+ i 1)
			   (if (>= (get-object-k (- (get-current-frame) i)) 6)
			       (+ c 1)
			       c)))))
    (interruptible-usleep 33333)
    (when (abort?)
     (set! *mode* 'viewfinder)
     (redraw-buttons)
     (meteor-turn-on-viewfinder! 1)
     (abort))
    (loop)))))

(define (wait-until-less-than-six-objects)
 (let ((n 15)
       (m 2)
       (f (get-current-frame)))
  (let loop ()
   (when (< (get-current-frame) f) (set! f (get-current-frame)))
   (unless (and (>= (- (get-current-frame) f) n)
		(let loop ((i 0) (c 0))
		 (if (>= i n)
		     (< c m)
		     (loop (+ i 1)
			   (if (>= (get-object-k (- (get-current-frame) i)) 6)
			       (+ c 1)
			       c)))))
    (interruptible-usleep 33333)
    (when (abort?)
     (set! *mode* 'viewfinder)
     (redraw-buttons)
     (meteor-turn-on-viewfinder! 1)
     (abort))
    (loop)))))

(define (wait-until-event)
 (let loop ()
  (wait-until-motion)
  (set! *start* (get-current-frame))
  (wait-until-no-motion)
  (set! *finish* (- (get-current-frame) (+ 15 3)))
  (if (< (- *finish* *start*) 11)
      (loop)
      (get-polygon-movie *start* *finish*))))

(define (debugging-wait-until-event)
 (let loop ()
  (wait-until-six-or-more-objects)
  (set! *start* (get-current-frame))
  (wait-until-less-than-six-objects)
  (set! *finish* (- (get-current-frame) (+ 15 3)))
  (if (< (- *finish* *start*) 11)
      (loop)
      (get-polygon-movie *start* *finish*))))

(define-service 'simplified-minimal-polygon-kinematic-models
 simplified-minimal-polygon-kinematic-models)

(define (download! polygon-movie
		   kinematic-models
		   contacts
		   supports?
		   contacts?
		   attached?)
 (set! *polygon-movie* polygon-movie)
 (set! *kinematic-models* kinematic-models)
 (set! *contacts* contacts)
 (set! *supports?* supports?)
 (set! *contacts?* contacts?)
 (set! *attached?* attached?)
 #t)

(define-service 'download! download!)

(define-service 'occurrences-of occurrences-of)

(define (set-parameters! fast? cheap? event-definition-set-name delta fuzz)
 (set! *fast?* fast?)
 (set! *cheap?* cheap?)
 (set! *event-definition-set-name* event-definition-set-name)
 (set! *delta* delta)
 (set! *fuzz* fuzz))

(define-service 'set-parameters! set-parameters!)

;;; Commands

(define (capture-command)
 (message "")
 (set! *kinematic-models* #f)
 (set! *events* '())
 (redraw-display-pane)
 (set! *mode* 'capture)
 (redraw-buttons)
 (meteor-turn-on-viewfinder! 1)
 (interruptible-usleep 5000000)
 (xbell *display* 100)
 (xflush *display*)
 (interruptible-usleep 300000)
 (xbell *display* 100)
 (xflush *display*)
 (meteor-turn-on-polygon-tracker-with-capturing! 0 1 5400 #t)
 (let ((polygon-movie (wait-until-event)))
  (meteor-turn-off!)
  (xbell *display* 100)
  (xflush *display*)
  (let ((polygon-movie (convex-hull-of-polygon-movie polygon-movie)))
   ;; SET! of *COLOURS* has to come before SET! of *POLYGON-MOVIE* because
   ;; POLYGON-CORRESPONDENCE may reorder the polygons in a frame so they no
   ;; longer correspond to the order used by GET-OBJECT-COLOUR.
   (set! *colours*
	 (map-n-vector
	  (lambda (t)
	   (map-n (lambda (i)
		   (list (matrix-ref polygon-movie t i)
			 (get-object-colour (+ *start* t) i)))
		  (vector-length (vector-ref polygon-movie t))))
	  (vector-length polygon-movie)))
   (set! *polygon-movie*
	 (map-vector
	  (lambda (frame) (map list->vector frame))
	  (polygon-correspondence
	   (map-vector (lambda (frame) (map vector->list (vector->list frame)))
		       polygon-movie)
	   *fill*
	   *x-weight*
	   *y-weight*
	   *a-weight*
	   *correspondence-threshold*)))
   (set! *colours*
	 (map-n-vector
	  (lambda (i)
	   (vector-ref
	    '#(moving red green blue yellow)
	    (mode
	     (removeq
	      #f
	      (map-n
	       (lambda (t)
		(if (assoc (list-ref (vector-ref *polygon-movie* t) i)
			   (vector-ref *colours* t))
		    (second (assoc (list-ref (vector-ref *polygon-movie* t) i)
				   (vector-ref *colours* t)))
		    #f))
	       (vector-length *polygon-movie*))))))
	  (length (vector-ref *polygon-movie* 0)))))
  (set! *t* 0)
  (set! *i* 0)
  (set! *end* (- (vector-length *polygon-movie*) 1))
  (set! *first-frame* 0)
  (set! *last-frame* *end*)
  (set! *mode* 'movie-and-polygons)
  (redraw-display-pane)
  (redraw-buttons)
  (write-object-to-file
   *polygon-movie* (default-extension (tmp "saved") "polygon-movie"))))

(define (play-command)
 (message "")
 (unless (or (eq? *mode* 'movie)
	     (eq? *mode* 'polygons)
	     (eq? *mode* 'movie-and-polygons)
	     (eq? *mode* 'code-and-polygons))
  (message "Can only do this in movie, polygons, or movie-and-polygons mode")
  (abort))
 (for-each-n
  (lambda (t)
   (show-frame! t)
   (xflush *display*)
   (usleep (inexact->exact (round (/ 1000000.0 *frame-rate*)))))
  (+ (- *last-frame* *first-frame*) 1))
 (redraw-display-pane))

(define (beginning-command)
 (message "")
 (unless (or (eq? *mode* 'movie)
	     (eq? *mode* 'polygons)
	     (eq? *mode* 'movie-and-polygons)
	     (eq? *mode* 'code-and-polygons))
  (message "Can only do this in movie, polygons, or movie-and-polygons mode")
  (abort))
 (unless (zero? *t*)
  (set! *t* 0)
  (set! *i* 0)
  (redraw-buttons)
  (redraw-display-pane)))

(define (decrement-t-command)
 (message "")
 (unless (or (eq? *mode* 'movie)
	     (eq? *mode* 'polygons)
	     (eq? *mode* 'movie-and-polygons)
	     (eq? *mode* 'code-and-polygons))
  (message "Can only do this in movie, polygons, or movie-and-polygons mode")
  (abort))
 (when (zero? *t*) (message "At beginning of movie") (abort))
 (set! *t* (- *t* 1))
 (set! *i* 0)
 (redraw-buttons)
 (redraw-display-pane))

(define (increment-t-command)
 (message "")
 (unless (or (eq? *mode* 'movie)
	     (eq? *mode* 'polygons)
	     (eq? *mode* 'movie-and-polygons)
	     (eq? *mode* 'code-and-polygons))
  (message "Can only do this in movie, polygons, or movie-and-polygons mode")
  (abort))
 (when (= *t* *end*) (message "At end of movie") (abort))
 (set! *t* (+ *t* 1))
 (set! *i* 0)
 (redraw-buttons)
 (redraw-display-pane))

(define (end-command)
 (message "")
 (unless (or (eq? *mode* 'movie)
	     (eq? *mode* 'polygons)
	     (eq? *mode* 'movie-and-polygons)
	     (eq? *mode* 'code-and-polygons))
  (message "Can only do this in movie, polygons, or movie-and-polygons mode")
  (abort))
 (unless (= *t* *end*)
  (set! *t* *end*)
  (set! *i* 0)
  (redraw-buttons)
  (redraw-display-pane)))

(define (save-command)
 (message "")
 (unless *pathname* (message "No pathname") (abort))
 (unless *polygon-movie* (message "You need to capture first") (abort))
 (meteor-frames->video-file (+ *start* *first-frame*)
			    (+ *start* *last-frame* 1)
			    (replace-extension *pathname* "Jpeg")
			    "Jpeg")
 (write-object-to-file
  (subvector *polygon-movie* *first-frame* (+ *last-frame* 1))
  (replace-extension *pathname* "polygon-movie")))

(define (models-command)
 (message "")
 (unless (or (eq? *mode* 'movie)
	     (eq? *mode* 'polygons)
	     (eq? *mode* 'movie-and-polygons)
	     (eq? *mode* 'code-and-polygons))
  (message
   "Can only do this in movie, polygons, or movie-and-polygons mode")
  (abort))
 (let ((t *t*))
  (cond
   ((null? *servers*)
    (set! *kinematic-models* (make-vector (+ *end* 1)))
    (set! *kinematic-models-cache* '())
    (for-each-n
     (lambda (t)
      (let* ((frame (vector-ref *polygon-movie* t))
	     (models
	      (forward-simplified-minimal-polygon-kinematic-models frame)))
       (vector-set! *kinematic-models* t models)
       (set! *t* t)
       (set! *i* 0)
       (redraw-buttons)
       (redraw-display-pane)
       (message (format #f "~s model~a ~s polygon~a ~s vertices"
			(length models)
			(if (= (length models) 1) "" "s")
			(length frame)
			(if (= (length frame) 1) "" "s")
			(reduce + (map vector-length frame) 0)))))
     (+ *end* 1))
    (message ""))
   (else (set! *kinematic-models*
	       (deresult
		(map-vector
		 (lambda (frame)
		  (invoke 'simplified-minimal-polygon-kinematic-models frame))
		 *polygon-movie*)))))
  (when (some-vector null? *kinematic-models*) (panic "No models"))
  (set! *kinematic-models* (circumscription *kinematic-models*))
  (set! *t* t)
  (set! *i* 0)
  (redraw-buttons)
  (redraw-display-pane)))

(define (track-command)
 (message "")
 (redraw-display-pane)
 (set! *mode* 'track)
 (redraw-buttons)
 (meteor-turn-off!)
 (meteor-turn-on-polygon-tracker! 0 1 5400 #t))

(define (track-and-view-command)
 (message "")
 (redraw-display-pane)
 (set! *mode* 'track-and-viewfinder)
 (redraw-buttons)
 (meteor-turn-on-viewfinder! 1)
 (meteor-turn-on-polygon-tracker! 0 1 5400 #t))

(define (viewfinder-command)
 (message "")
 (redraw-display-pane)
 (set! *mode* 'viewfinder)
 (redraw-buttons)
 (meteor-turn-on-viewfinder! 1))

(define (events-command)
 (message "")
 (unless (or (eq? *mode* 'movie)
	     (eq? *mode* 'polygons)
	     (eq? *mode* 'movie-and-polygons)
	     (eq? *mode* 'code-and-polygons))
  (message "Can only do this in movie, polygons, or movie-and-polygons mode")
  (abort))
 (unless *kinematic-models* (models-command))
 (precompute!)
 (set! *events* '())
 (deresult (map (lambda (server)
		 (invoke-on server
			    'download!
			    *polygon-movie*
			    *kinematic-models*
			    *contacts*
			    *supports?*
			    *contacts?*
			    *attached?*))
		*servers*))
 (let ((events '()))
  (for-effects
   (let ((e (an-event (first (a-member-of (aggregate-event-types))))))
    (set! events
	  (cons (list e
		      (if (null? *servers*)
			  (occurrences-of e)
			  (invoke 'occurrences-of e)))
		events))))
  (for-each
   (lambda (event)
    (let ((is (second event)))
     (unless (null? is)
      (set! *events* (cons (format-event (first event) is) *events*)))))
   (reverse (deresult events))))
 (set! *events* (cons "" *events*))
 (for-effects
  (let ((e (either (list 'supported? (a-member-of (objects)))
		   (list 'supports?
			 (a-member-of (objects))
			 (a-member-of (objects)))
		   (list 'contacts?
			 (a-member-of (objects))
			 (a-member-of (objects)))
		   (list 'attached?
			 (a-member-of (objects))
			 (a-member-of (objects))))))
   (when (and (or (eq? (first e) 'contacts?) (eq? (first e) 'attached?))
	      (> (second e) (third e)))
    (fail))
   (let ((is (occurrences-of e)))
    (unless (null? is) (set! *events* (cons (format-event e is) *events*))))))
 (set! *events* (reverse *events*))
 (redraw-display-pane))

(define (classify-command)
 (capture-command)
 (events-command))

;;; Top Level

(define-application viewer *width* *height* 0 6 6
 (lambda ()
  (set! *width* 320)
  (set! *height* 240)
  (set! *mode* 'viewfinder)
  (set! *polygon-movie* #f)
  (set! *kinematic-models* #f)
  (set! *events* '())
  (define-button 0 0 "Help" #f
   (lambda ()
    (meteor-turn-off!)
    (help-command)))
  (define-button 1 0 "Abort" #f abort-command)
  (define-button 2 0 "Previous"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   (lambda ()
    (message "")
    (unless (or (eq? *mode* 'movie)
		(eq? *mode* 'polygons)
		(eq? *mode* 'movie-and-polygons)
		(eq? *mode* 'code-and-polygons))
     (message
      "Can only do this in movie, polygons, or movie-and-polygons mode")
     (abort))
    (when (or (not *kinematic-models*)
	      (null? (vector-ref *kinematic-models* *t*)))
     (message "No models")
     (abort))
    (when (zero? *i*) (message "First model") (abort))
    (set! *i* (- *i* 1))
    (redraw-display-pane)))
  (define-button 3 0 "Next"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   (lambda ()
    (message "")
    (unless (or (eq? *mode* 'movie)
		(eq? *mode* 'polygons)
		(eq? *mode* 'movie-and-polygons)
		(eq? *mode* 'code-and-polygons))
     (message
      "Can only do this in movie, polygons, or movie-and-polygons mode")
     (abort))
    (when (or (not *kinematic-models*)
	      (null? (vector-ref *kinematic-models* *t*)))
     (message "No models")
     (abort))
    (when (= *i* (- (length (vector-ref *kinematic-models* *t*)) 1))
     (message "Last model")
     (abort))
    (set! *i* (+ *i* 1))
    (redraw-display-pane)))
  (define-button 4 0 "Models"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   models-command)
  (define-button 5 0 "Quit" #f quit)
  (define-button 0 1 "Track" (lambda () (eq? *mode* 'track)) track-command)
  (define-button 1 1 "Track&View"
   (lambda () (eq? *mode* 'track-and-viewfinder)) track-and-view-command)
  (define-button 2 1 "Viewfinder" (lambda () (eq? *mode* 'viewfinder))
   viewfinder-command)
  (define-button 3 1 "Movie" (lambda () (eq? *mode* 'movie))
   (lambda ()
    (message "")
    (unless *polygon-movie* (message "You need to capture first") (abort))
    (meteor-turn-off!)
    (set! *mode* 'movie)
    (redraw-buttons)
    (redraw-display-pane)))
  (define-button 4 1 "Polygons" (lambda () (eq? *mode* 'polygons))
   (lambda ()
    (message "")
    (unless *polygon-movie* (message "You need to capture first") (abort))
    (meteor-turn-off!)
    (set! *mode* 'polygons)
    (redraw-buttons)
    (redraw-display-pane)))
  (define-button 5 1 "Movie&Poly" (lambda () (eq? *mode* 'movie-and-polygons))
   (lambda ()
    (message "")
    (unless *polygon-movie* (message "You need to capture first") (abort))
    (meteor-turn-off!)
    (set! *mode* 'movie-and-polygons)
    (redraw-buttons)
    (redraw-display-pane)))
  (define-button 0 2 "Set Red" #f
   (lambda ()
    (message "")
    (unless (eq? *mode* 'viewfinder)
     (message "Can only do this in viewfinder mode")
     (abort))
    (unless *point* (message "No region") (abort))
    (let ((min-hue 0.0)
	  (max-hue 359.0)
	  (min-saturation 100.0)
	  (max-saturation 0.0)
	  (min-value 100.0)
	  (max-value 0.0))
     (do ((y1 (min (y *point*) (y *mark*)) (+ y1 1)))
       ((>= y1 (min (max (y *point*) (y *mark*)) *height*)))
      (do ((x1 (min (x *point*) (x *mark*)) (+ x1 1)))
	((>= x1 (min (max (x *point*) (x *mark*)) *width*)))
       (let ((hsv (rgb->hsv (vector (meteor-get-red 0 y1 x1)
				    (meteor-get-green 0 y1 x1)
				    (meteor-get-blue 0 y1 x1)))))
	(if (< (vector-ref hsv 0) 180.0)
	    (set! min-hue (max min-hue (vector-ref hsv 0)))
	    (set! max-hue (min max-hue (vector-ref hsv 0))))
	(set! min-saturation (min min-saturation (vector-ref hsv 1)))
	(set! max-saturation (max max-saturation (vector-ref hsv 1)))
	(set! min-value (min min-value (vector-ref hsv 2)))
	(set! max-value (max max-value (vector-ref hsv 2))))))
     (set! *rhl* (+ min-hue *stretch*))
     (set! *rhh* (- max-hue *stretch*))
     (set! *rsl* (- min-saturation *stretch*))
     (set! *rsh* (+ max-saturation *stretch*))
     (set! *rvl* (- min-value *stretch*))
     (set! *rvh* (+ max-value *stretch*))
     (set! *red?* #t)
     (meteor-set-red-hack-track!
      *rhl* *rhh* (if *red?* *rsl* 101) *rsh* *rvl* *rvh*)
     (redraw-buttons))))
  (define-button 1 2 "Set Green" #f
   (lambda ()
    (message "")
    (unless (eq? *mode* 'viewfinder)
     (message "Can only do this in viewfinder mode")
     (abort))
    (unless *point* (message "No region") (abort))
    (let ((min-hue 359.0)
	  (max-hue 0.0)
	  (min-saturation 100.0)
	  (max-saturation 0.0)
	  (min-value 100.0)
	  (max-value 0.0))
     (do ((y1 (min (y *point*) (y *mark*)) (+ y1 1)))
       ((>= y1 (min (max (y *point*) (y *mark*)) *height*)))
      (do ((x1 (min (x *point*) (x *mark*)) (+ x1 1)))
	((>= x1 (min (max (x *point*) (x *mark*)) *width*)))
       (let ((hsv (rgb->hsv (vector (meteor-get-red 0 y1 x1)
				    (meteor-get-green 0 y1 x1)
				    (meteor-get-blue 0 y1 x1)))))
	(set! min-hue (min min-hue (vector-ref hsv 0)))
	(set! max-hue (max max-hue (vector-ref hsv 0)))
	(set! min-saturation (min min-saturation (vector-ref hsv 1)))
	(set! max-saturation (max max-saturation (vector-ref hsv 1)))
	(set! min-value (min min-value (vector-ref hsv 2)))
	(set! max-value (max max-value (vector-ref hsv 2))))))
     (set! *ghl* (- min-hue *stretch*))
     (set! *ghh* (+ max-hue *stretch*))
     (set! *gsl* (- min-saturation *stretch*))
     (set! *gsh* (+ max-saturation *stretch*))
     (set! *gvl* (- min-value *stretch*))
     (set! *gvh* (+ max-value *stretch*))
     (set! *green?* #t)
     (meteor-set-green-hack-track!
      *ghl* *ghh* (if *green?* *gsl* 101) *gsh* *gvl* *gvh*)
     (redraw-buttons))))
  (define-button 2 2 "Set Blue" #f
   (lambda ()
    (message "")
    (unless (eq? *mode* 'viewfinder)
     (message "Can only do this in viewfinder mode")
     (abort))
    (unless *point* (message "No region") (abort))
    (let ((min-hue 359.0)
	  (max-hue 0.0)
	  (min-saturation 100.0)
	  (max-saturation 0.0)
	  (min-value 100.0)
	  (max-value 0.0))
     (do ((y1 (min (y *point*) (y *mark*)) (+ y1 1)))
       ((>= y1 (min (max (y *point*) (y *mark*)) *height*)))
      (do ((x1 (min (x *point*) (x *mark*)) (+ x1 1)))
	((>= x1 (min (max (x *point*) (x *mark*)) *width*)))
       (let ((hsv (rgb->hsv (vector (meteor-get-red 0 y1 x1)
				    (meteor-get-green 0 y1 x1)
				    (meteor-get-blue 0 y1 x1)))))
	(set! min-hue (min min-hue (vector-ref hsv 0)))
	(set! max-hue (max max-hue (vector-ref hsv 0)))
	(set! min-saturation (min min-saturation (vector-ref hsv 1)))
	(set! max-saturation (max max-saturation (vector-ref hsv 1)))
	(set! min-value (min min-value (vector-ref hsv 2)))
	(set! max-value (max max-value (vector-ref hsv 2))))))
     (set! *bhl* (- min-hue *stretch*))
     (set! *bhh* (+ max-hue *stretch*))
     (set! *bsl* (- min-saturation *stretch*))
     (set! *bsh* (+ max-saturation *stretch*))
     (set! *bvl* (- min-value *stretch*))
     (set! *bvh* (+ max-value *stretch*))
     (set! *blue?* #t)
     (meteor-set-blue-hack-track!
      *bhl* *bhh* (if *blue?* *bsl* 101) *bsh* *bvl* *bvh*)
     (redraw-buttons))))
  (define-button 3 2 "Set Yellow" #f
   (lambda ()
    (message "")
    (unless (eq? *mode* 'viewfinder)
     (message "Can only do this in viewfinder mode")
     (abort))
    (unless *point* (message "No region") (abort))
    (let ((min-hue 359.0)
	  (max-hue 0.0)
	  (min-saturation 100.0)
	  (max-saturation 0.0)
	  (min-value 100.0)
	  (max-value 0.0))
     (do ((y1 (min (y *point*) (y *mark*)) (+ y1 1)))
       ((>= y1 (min (max (y *point*) (y *mark*)) *height*)))
      (do ((x1 (min (x *point*) (x *mark*)) (+ x1 1)))
	((>= x1 (min (max (x *point*) (x *mark*)) *width*)))
       (let ((hsv (rgb->hsv (vector (meteor-get-red 0 y1 x1)
				    (meteor-get-green 0 y1 x1)
				    (meteor-get-blue 0 y1 x1)))))
	(set! min-hue (min min-hue (vector-ref hsv 0)))
	(set! max-hue (max max-hue (vector-ref hsv 0)))
	(set! min-saturation (min min-saturation (vector-ref hsv 1)))
	(set! max-saturation (max max-saturation (vector-ref hsv 1)))
	(set! min-value (min min-value (vector-ref hsv 2)))
	(set! max-value (max max-value (vector-ref hsv 2))))))
     (set! *yhl* (- min-hue *stretch*))
     (set! *yhh* (+ max-hue *stretch*))
     (set! *ysl* (- min-saturation *stretch*))
     (set! *ysh* (+ max-saturation *stretch*))
     (set! *yvl* (- min-value *stretch*))
     (set! *yvh* (+ max-value *stretch*))
     (set! *yellow?* #t)
     (meteor-set-yellow-hack-track!
      *yhl* *yhh* (if *yellow?* *ysl* 101) *ysh* *yvl* *yvh*)
     (redraw-buttons))))
  (define-button 4 2 "Save RGBY" #f
   (lambda ()
    (message "")
    (unless *pathname* (message "No pathname") (abort))
    (write-object-to-file
     (list (list *red?* *rhl* *rhh* *rsl* *rsh* *rvl* *rvh*)
	   (list *green?* *ghl* *ghh* *gsl* *gsh* *gvl* *gvh*)
	   (list *blue?* *bhl* *bhh* *bsl* *bsh* *bvl* *bvh*)
	   (list *yellow?* *yhl* *yhh* *ysl* *ysh* *yvl* *yvh*)
	   (list *meteor-motion-threshold*
		 *meteor-colour-pixel-mass-threshold*
		 *meteor-colour-proximity-threshold*
		 *meteor-motion-pixel-mass-threshold*
		 *meteor-motion-proximity-threshold*))
     (default-extension *pathname* "rgby"))))
  (define-button 5 2 "Code&Poly" (lambda () (eq? *mode* 'code-and-polygons))
   (lambda ()
    (message "")
    (unless *polygon-movie* (message "You need to capture first") (abort))
    (meteor-turn-off!)
    (set! *mode* 'code-and-polygons)
    (redraw-buttons)
    (redraw-display-pane)))
  (define-button 0 3 "Capture" (lambda () (eq? *mode* 'capture))
   capture-command)
  (define-button 1 3 "Set First"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   (lambda ()
    (message "")
    (unless (or (eq? *mode* 'movie)
		(eq? *mode* 'polygons)
		(eq? *mode* 'movie-and-polygons)
		(eq? *mode* 'code-and-polygons))
     (message
      "Can only do this in movie, polygons, or movie-and-polygons mode")
     (abort))
    (set! *first-frame* *t*)
    (set! *last-frame* (max *last-frame* *first-frame*))
    (redraw-buttons)))
  (define-button 2 3 "Play"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   play-command)
  (define-button 3 3 "Save" #f save-command)
  (define-button 4 3 "Set Last"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   (lambda ()
    (message "")
    (unless (or (eq? *mode* 'movie)
		(eq? *mode* 'polygons)
		(eq? *mode* 'movie-and-polygons)
		(eq? *mode* 'code-and-polygons))
     (message
      "Can only do this in movie, polygons, or movie-and-polygons mode")
     (abort))
    (set! *last-frame* *t*)
    (set! *first-frame* (min *first-frame* *last-frame*))
    (redraw-buttons)))
  (define-button 5 3 "Events"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   events-command)
  (define-button 0 4 "Beginning"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   beginning-command)
  (define-button 1 4
   (lambda ()
    (if (or (eq? *mode* 'movie)
	    (eq? *mode* 'polygons)
	    (eq? *mode* 'movie-and-polygons)
	    (eq? *mode* 'code-and-polygons))
	(format #f "First ~s" *first-frame*)
	"First"))
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   (lambda ()
    (message "")
    (unless (or (eq? *mode* 'movie)
		(eq? *mode* 'polygons)
		(eq? *mode* 'movie-and-polygons)
		(eq? *mode* 'code-and-polygons))
     (message
      "Can only do this in movie, polygons, or movie-and-polygons mode")
     (abort))
    (unless (= *t* *first-frame*)
     (set! *t* *first-frame*)
     (set! *i* 0)
     (redraw-buttons)
     (redraw-display-pane))))
  (define-button 2 4
   (lambda ()
    (if (or (eq? *mode* 'movie)
	    (eq? *mode* 'polygons)
	    (eq? *mode* 'movie-and-polygons)
	    (eq? *mode* 'code-and-polygons))
	(format #f "-T ~s" *t*)
	"-T"))
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   decrement-t-command)
  (define-button 3 4
   (lambda ()
    (if (or (eq? *mode* 'movie)
	    (eq? *mode* 'polygons)
	    (eq? *mode* 'movie-and-polygons)
	    (eq? *mode* 'code-and-polygons))
	(format #f "+T ~s" *t*)
	"+T"))
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   increment-t-command)
  (define-button 4 4
   (lambda ()
    (if (or (eq? *mode* 'movie)
	    (eq? *mode* 'polygons)
	    (eq? *mode* 'movie-and-polygons)
	    (eq? *mode* 'code-and-polygons))
	(format #f "Last ~s" *last-frame*)
	"Last"))
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   (lambda ()
    (message "")
    (unless (or (eq? *mode* 'movie)
		(eq? *mode* 'polygons)
		(eq? *mode* 'movie-and-polygons)
		(eq? *mode* 'code-and-polygons))
     (message
      "Can only do this in movie, polygons, or movie-and-polygons mode")
     (abort))
    (unless (= *t* *last-frame*)
     (set! *t* *last-frame*)
     (set! *i* 0)
     (redraw-buttons)
     (redraw-display-pane))))
  (define-button 5 4 "End"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   end-command)
  (define-toggle-button 0 5 "Red" *red?*
   (lambda ()
    (meteor-set-red-hack-track!
     *rhl* *rhh* (if *red?* *rsl* 101) *rsh* *rvl* *rvh*)))
  (define-toggle-button 1 5 "Green" *green?*
   (lambda ()
    (meteor-set-green-hack-track!
     *ghl* *ghh* (if *green?* *gsl* 101) *gsh* *gvl* *gvh*)))
  (define-toggle-button 2 5 "Blue" *blue?*
   (lambda ()
    (meteor-set-blue-hack-track!
     *bhl* *bhh* (if *blue?* *bsl* 101) *bsh* *bvl* *bvh*)))
  (define-toggle-button 3 5 "Yellow" *yellow?*
   (lambda ()
    (meteor-set-yellow-hack-track!
     *yhl* *yhh* (if *yellow?* *ysl* 101) *ysh* *yvl* *yvh*)))
  (define-button 4 5 "Postscript"
   (lambda ()
    (or (eq? *mode* 'movie)
	(eq? *mode* 'polygons)
	(eq? *mode* 'movie-and-polygons)
	(eq? *mode* 'code-and-polygons)))
   (lambda ()
    (message "")
    (unless (or (eq? *mode* 'movie)
		(eq? *mode* 'polygons)
		(eq? *mode* 'movie-and-polygons)
		(eq? *mode* 'code-and-polygons))
     (message
      "Can only do this in movie, polygons, or movie-and-polygons mode")
     (abort))
    (unless *pathname* (message "No pathname") (abort))
    (unless *kinematic-models* (models-command))
    (meteor-frames->video-file *start*
			       (+ *start* (vector-length *polygon-movie*))
			       (replace-extension *pathname* "Jpeg")
			       "Jpeg")
    (write-object-to-file
     *polygon-movie* (replace-extension *pathname* "polygon-movie"))
    (let ((preferred-models-movie
	   (map-vector forward-simplified-minimal-polygon-kinematic-models
		       *polygon-movie*)))
     (when (some-vector null? preferred-models-movie)
      (panic "No preferred models"))
     (let ((more-preferred-models-movie
	    (if *twice?*
		(map-vector
		 (lambda (preferred-models ground?)
		  (cardinality-circumscription
		   (remove-if-not
		    (lambda (preferred-model)
		     (equal? (kinematic-model-ground? preferred-model)
			     ground?))
		    preferred-models)))
		 preferred-models-movie
		 (map-vector
		  (lambda (preferred-models)
		   (kinematic-model-ground? (first preferred-models)))
		  (temporal-circumscription preferred-models-movie)))
		(map-vector cardinality-circumscription
			    preferred-models-movie))))
      (when (some-vector null? more-preferred-models-movie)
       (panic "No more-preferred models"))
      (for-each-n
       (lambda (i)
	(let ((frame (vector-ref *polygon-movie* i))
	      (preferred-models (vector-ref preferred-models-movie i))
	      (more-preferred-models
	       (vector-ref more-preferred-models-movie i))
	      (most-preferred-model (first (vector-ref *kinematic-models* i))))
	 ;; needs work: to convert frame-i-image.ppm to frame-i-image.ps
	 (write-pnm
	  (meteor-frame->ppm (+ *start* i))
	  (replace-extension
	   (format #f "~a-~s-image" (strip-extension *pathname*) i) "ppm"))
	 ;; needs work: to convert frame-i-code.ppm to frame-i-code.ps
	 (write-pnm
	  (meteor-code-frame->ppm (+ *start* i))
	  (replace-extension
	   (format #f "~a-~s-code" (strip-extension *pathname*) i) "ppm"))
	 (polygons->postscript
	  frame (format #f "~a-~s-polygons" (strip-extension *pathname*) i))
	 (write-object-to-file
	  (list frame preferred-models)
	  (replace-extension
	   (format #f "~a-~s-preferred-models" (strip-extension *pathname*) i)
	   "kgp"))
	 (for-each-indexed (lambda (preferred-model j)
			    (kinematic-model->postscript
			     frame
			     preferred-model
			     (format #f "~a-~s-preferred-model-~s"
				     (strip-extension *pathname*) i j)))
			   preferred-models)
	 (write-object-to-file
	  (list frame more-preferred-models)
	  (replace-extension
	   (format #f "~a-~s-more-preferred-models"
		   (strip-extension *pathname*) i)
	   "kgp"))
	 (if *new-indexing?*
	     (for-each
	      (lambda (more-preferred-model)
	       (kinematic-model->postscript
		frame
		more-preferred-model
		(format #f "~a-~s-more-preferred-model-~s"
			(strip-extension *pathname*)
			i
			(position more-preferred-model preferred-models))))
	      more-preferred-models)
	     (for-each-indexed
	      (lambda (more-preferred-model j)
	       (kinematic-model->postscript
		frame
		more-preferred-model
		(format #f "~a-~s-more-preferred-model-~s"
			(strip-extension *pathname*) i j)))
	      more-preferred-models))
	 (write-object-to-file
	  (list frame (list most-preferred-model))
	  (replace-extension
	   (format #f "~a-~s-most-preferred-model"
		   (strip-extension *pathname*) i)
	   "kgp"))
	 (if *new-indexing?*
	     (kinematic-model->postscript
	      frame
	      most-preferred-model
	      (format #f "~a-~s-most-preferred-model-~s"
		      (strip-extension *pathname*)
		      i
		      (position most-preferred-model preferred-models)))
	     (kinematic-model->postscript
	      frame
	      most-preferred-model
	      (format #f "~a-~s-most-preferred-model"
		      (strip-extension *pathname*) i)))
	 ;; needs work: to convert frame-i-code-and-polygons.ppm to
	 ;;             frame-i-code-and-polygons.ps
	 (write-pnm
	  (kinematic-model->ppm #t (+ *start* i) frame #f)
	  (replace-extension
	   (format #f "~a-~s-code-and-polygons" (strip-extension *pathname*) i)
	   "ppm"))
	 ;; needs work: to convert frame-i-code-and-most-preferred-model.ppm to
	 ;;             frame-i-code-and-most-preferred-model.ps
	 (write-pnm
	  (kinematic-model->ppm #t (+ *start* i) frame most-preferred-model)
	  (replace-extension
	   (format #f "~a-~s-code-and-most-preferred-model"
		   (strip-extension *pathname*) i)
	   "ppm"))
	 ;; needs work: to convert frame-i-image-and-polygons.ppm to
	 ;;             frame-i-image-and-polygons.ps
	 (write-pnm
	  (kinematic-model->ppm #f (+ *start* i) frame #f)
	  (replace-extension
	   (format #f "~a-~s-image-and-polygons"
		   (strip-extension *pathname*) i)
	   "ppm"))
	 ;; needs work: to convert frame-i-image-and-most-preferred-model.ppm
	 ;;             to frame-i-image-and-most-preferred-model.ps
	 (write-pnm
	  (kinematic-model->ppm #f (+ *start* i) frame most-preferred-model)
	  (replace-extension
	   (format #f "~a-~s-image-and-most-preferred-model"
		   (strip-extension *pathname*) i)
	   "ppm"))))
       (vector-length *polygon-movie*))))
    (write-file
     *events*
     (replace-extension
      (format #f "~a-events" (strip-extension *pathname*)) "text"))))
  (define-button 5 5 "Classify" #f classify-command)
  (define-key (list (control #\x) (control #\c)) "Quit" quit)
  (define-key (control #\h) "Help"
   (lambda ()
    (meteor-turn-off!)
    (help-command)))
  (define-key (control #\g) "Abort" abort-command)
  (define-key (control #\c) "Capture" capture-command)
  (define-key (meta #\f) "Increment T" increment-t-command)
  (define-key (meta #\b) "Decrement T" decrement-t-command)
  (define-key (control #\p) "Play" play-command)
  (define-key (list (control #\x) (control #\s)) "Save" save-command)
  (define-key (meta #\<) "Beginning" beginning-command)
  (define-key (meta #\>) "End" end-command)
  (define-key (meta #\t) "Track" track-command)
  (define-key (meta (control #\t)) "Track&View" track-and-view-command)
  (define-key (meta #\v) "Viewfinder" viewfinder-command)
  (define-key (meta #\e) "Events" events-command)
  (define-key (control #\s) "Classify" classify-command)
  (meteor-startup-with-viewfinder *height* *width* 1000 #t))
 (lambda () (meteor-turn-on-viewfinder! 1))
 (lambda () (meteor-shutdown))
 (lambda ()
  (define-region 0 0 *display-pane-width* *display-pane-height*
   (lambda (x1 y1)
    (message "")
    (unless (eq? *mode* 'viewfinder)
     (message "Can only do this in viewfinder mode")
     (abort))
    (meteor-turn-off!)
    (set! *polygon-movie* #f)
    (meteor-capture-frames 0 1 1)
    (set! x1 (min x1 (- *width* 1)))
    (set! y1 (min y1 (- *height* 1)))
    (let* ((result
	    (tracking-pointer
	     #f
	     #f
	     (lambda (x2 y2)
	      (set! x2 (min x2 (- *width* 1)))
	      (set! y2 (min y2 (- *height* 1)))
	      (redraw!)
	      (meteor-show-frame 0)
	      (xsetforeground
	       *display* *thin-gc* (xcolor-pixel (second *background*)))
	      (xdrawrectangle *display* *display-pane* *thin-gc*
			      x1 y1 (- x2 x1) (- y2 y1))
	      (xsetforeground
	       *display* *thin-gc* (xcolor-pixel (second *foreground*))))))
	   (x2 (min (first result) (- *width* 1)))
	   (y2 (min (second result) (- *height* 1))))
     (set! *point* (vector x1 y1))
     (set! *mark* (vector x2 y2))
     (meteor-turn-on-viewfinder! 1)
     (redraw-display-pane))))
  (xclearwindow *display* *display-pane*)
  (redraw!))
 (lambda ()
  (set! *pathname*
	(and (not (string=? *input* "")) (strip-extension *input*)))
  (redraw-display-pane)))

(define-command
 (command (at-most-one
	   ("servers" servers? (servers "pathname" string-argument *program*)))
	  (at-most-one ("fps" fps? (fps "#" real-argument 30.0)))
	  (at-most-one ("640x480" size640x480?)
		       ("800x600" size800x600?)
		       ("1024x480" size1024x480?)
		       ("1024x768" size1024x768?)
		       ("1280x1024" size1280x1024?)
		       ("1600x1024" size1600x1024?)
		       ("1600x1200" size1600x1200?))
	  (at-most-one ("window-position"
			window-position?
			(window-position-x "x" integer-argument 0)
			(window-position-y "y" integer-argument 0)))
	  (at-most-one ("rgby" rgby? (rgby "pathname" string-argument "rgby")))
	  (at-most-one ("fill" fill? (fill "#" integer-argument 5)))
	  (at-most-one ("x-weight" x-weight? (x-weight "#" real-argument 1.0)))
	  (at-most-one ("y-weight" y-weight? (y-weight "#" real-argument 1.0)))
	  (at-most-one ("a-weight" a-weight? (a-weight "#" real-argument 1.0)))
	  (at-most-one
	   ("correspondence-threshold"
	    correspondence-threshold?
	    (correspondence-threshold "#" real-argument 0.15)))
	  (at-most-one ("cache" cache?))
	  (at-most-one ("fast" fast?) ("cheap" cheap?))
	  (at-most-one ("twice" twice?))
	  (at-most-one ("new-temporal-circumscription-cost"
			new-temporal-circumscription-cost?))
	  (at-most-one ("new-indexing" new-indexing?))
	  (at-most-one
	   ("event-definition-set-name"
	    event-definition-set-name?
	    (event-definition-set-name "type" string-argument "default")))
	  (at-most-one ("delta" delta? (delta "#" integer-argument 2)))
	  (at-most-one ("fuzz" fuzz? (fuzz "#" integer-argument 0)))
	  (at-most-one ("jair2001" jair2001?)))
 (when jair2001?
  (set! fill? #f)
  (set! fill 5)
  (set! x-weight? #f)
  (set! x-weight 1.0)
  (set! y-weight? #f)
  (set! y-weight 1.0)
  (set! a-weight? #f)
  (set! a-weight 1.0)
  (set! correspondence-threshold? #f)
  (set! correspondence-threshold 0.15)
  (set! cache? #f)
  (set! fast? #f)
  (set! cheap? #f)
  (set! twice? #f)
  (set! new-temporal-circumscription-cost? #f)
  (set! new-indexing? #f)
  (set! event-definition-set-name "jair2001")
  (set! delta 2)
  (set! fuzz 0)
  (set! *jair2001-bug?* #t))
 (set! *cache?* cache?)
 (set! *fast?* (or fast? cheap?))
 (set! *cheap?* cheap?)
 (set! *twice?* twice?)
 (set! *new-temporal-circumscription-cost?* new-temporal-circumscription-cost?)
 (set! *new-indexing?* new-indexing?)
 (unless (member event-definition-set-name '("jair2001" "debugging" "default"))
  (panic "Unknown event-defintion--set name"))
 (set! *event-definition-set-name* event-definition-set-name)
 (set! *delta* delta)
 (set! *fuzz* fuzz)
 (set! *red?* #t)
 (set! *green?* #t)
 (set! *blue?* #t)
 (set! *yellow?* #t)
 (set! *frame-rate* fps)
 (let ((downsample-factor (/ 30.0 *frame-rate*)))
  (unless (< (abs (- downsample-factor (round downsample-factor))) 0.001)
   (panic "Frame rate does not yield an integral downsample factor")))
 (set! *window-position?* #f)
 (when window-position?
  (set! *window-position?* #t)
  (set! *window-position-x* window-position-x)
  (set! *window-position-y* window-position-y))
 (cond (size640x480? (set! *height* (- 480 250)) (set! *width* (- 640 10)))
       (size800x600? (set! *height* (- 600 250)) (set! *width* (- 800 10)))
       (size1024x480? (set! *height* (- 480 250)) (set! *width* (- 1024 10)))
       (size1024x768? (set! *height* (- 768 250)) (set! *width* (- 1024 10)))
       (size1280x1024? (set! *height* (- 1024 250)) (set! *width* (- 1280 10)))
       (size1600x1024? (set! *height* (- 1024 250)) (set! *width* (- 1600 10)))
       (size1600x1200? (set! *height* (- 1200 250)) (set! *width* (- 1600 10)))
       (else (set! *height* (- 1024 250)) (set! *width* (- 1600 10))))
 (when rgby?
  (let ((rgby (read-object-from-file (default-extension rgby "rgby"))))
   (set! *red?* (first (first rgby)))
   (set! *rhl* (second (first rgby)))
   (set! *rhh* (third (first rgby)))
   (set! *rsl* (fourth (first rgby)))
   (set! *rsh* (fifth (first rgby)))
   (set! *rvl* (sixth (first rgby)))
   (set! *rvh* (seventh (first rgby)))
   (set! *green?* (first (second rgby)))
   (set! *ghl* (second (second rgby)))
   (set! *ghh* (third (second rgby)))
   (set! *gsl* (fourth (second rgby)))
   (set! *gsh* (fifth (second rgby)))
   (set! *gvl* (sixth (second rgby)))
   (set! *gvh* (seventh (second rgby)))
   (set! *blue?* (first (third rgby)))
   (set! *bhl* (second (third rgby)))
   (set! *bhh* (third (third rgby)))
   (set! *bsl* (fourth (third rgby)))
   (set! *bsh* (fifth (third rgby)))
   (set! *bvl* (sixth (third rgby)))
   (set! *bvh* (seventh (third rgby)))
   (set! *yellow?* (first (fourth rgby)))
   (set! *yhl* (second (fourth rgby)))
   (set! *yhh* (third (fourth rgby)))
   (set! *ysl* (fourth (fourth rgby)))
   (set! *ysh* (fifth (fourth rgby)))
   (set! *yvl* (sixth (fourth rgby)))
   (set! *yvh* (seventh (fourth rgby)))
   (set! *meteor-motion-threshold* (first (fifth rgby)))
   (set! *meteor-colour-pixel-mass-threshold* (second (fifth rgby)))
   (set! *meteor-colour-proximity-threshold* (third (fifth rgby)))
   (set! *meteor-motion-pixel-mass-threshold* (fourth (fifth rgby)))
   (set! *meteor-motion-proximity-threshold* (fifth (fifth rgby)))
   (meteor-set-red-hack-track!
    *rhl* *rhh* (if *red?* *rsl* 101) *rsh* *rvl* *rvh*)
   (meteor-set-green-hack-track!
    *ghl* *ghh* (if *green?* *gsl* 101) *gsh* *gvl* *gvh*)
   (meteor-set-blue-hack-track!
    *bhl* *bhh* (if *blue?* *bsl* 101) *bsh* *bvl* *bvh*)
   (meteor-set-yellow-hack-track!
    *yhl* *yhh* (if *yellow?* *ysl* 101) *ysh* *yvl* *yvh*)
   (meteor-set-hack-track! *meteor-motion-threshold*
			   *meteor-colour-pixel-mass-threshold*
			   *meteor-colour-proximity-threshold*
			   *meteor-motion-pixel-mass-threshold*
			   *meteor-motion-proximity-threshold*)))
 (set! *fill* fill)
 (set! *x-weight* x-weight)
 (set! *y-weight* y-weight)
 (set! *a-weight* a-weight)
 (set! *correspondence-threshold* correspondence-threshold)
 (when servers?
  (for-each
   (lambda (line)
    (define-server (field-ref line 0) (string->number (field-ref line 1))))
   (read-file (default-extension servers "servers"))))
 (start-servers)
 (deresult
  (map
   (lambda (server)
    (invoke-on server 'set-parameters!
	       (or fast? cheap?) cheap? event-definition-set-name delta fuzz))
   *servers*))
 (viewer '())
 (shutdown-servers))

(define (main arguments)
 (cond ((and (= (length arguments) 3)
	     (string=? (second arguments) "-port")
	     (number? (string->number (third arguments)))
	     (integer? (string->number (third arguments)))
	     (exact? (string->number (third arguments))))
	(set-maximum-heap! (* 100 1024 1024))
	(server (string->number (third arguments)))
	#f)
       (else (command arguments))))

;;; Tam V'Nishlam Shevah L'El Borei Olam
