(MODULE
 PLAY-POLYGON-MOVIE
 (WITH
  QOBISCHEME
  XLIB
  HLIB-CHANGE-MOVIE
  HLIB-EM
  HLIB-FIND-LINES
  HLIB-HULL
  HLIB-INTERPRET
  HLIB-KINEMATIC
  HLIB-LOCATE
  HLIB-MERGE-LINES
  HLIB-MOST
  HLIB-POSTSCRIPT
  HLIB-RATIO-REGIONS
  HLIB-RILIB
  HLIB-SDE
  HLIB-SNAKE
  HLIB-SPAN
  HLIB-SPEAK
  HLIB-TRACK
  HLIB-TRAIN-CLASSIFY
  HLIB-WBM)
 (MAIN MAIN))
;;; LaHaShem HaAretz U'Mloah

;;; TTMTTD
;;;  1. Display linear and angular velocities and accelerations.

(include "QobiScheme.sch")
(include "play-polygon-movie.sch")
(include "c-externals.sc")

(set! *program* "play-polygon-movie")
(set! *panic?* #t)

;;; Variables

(define *directory* #f)
(define *pathname* #f)
(define *pathnames* #f)
(define *t* #f)
(define *end* #f)
(define *first-frame* #f)
(define *last-frame* #f)
(define *all?* #f)

;;; Procedures

(define (number-of-objects-in-frame movie t) (length (vector-ref movie t)))

(define (draw-polygon display display-pane gc polygon)
 (let ((polygon (vector->list polygon)))
  (for-each
   (lambda (p q) (xdrawline display display-pane gc (x p) (y p) (x q) (y q)))
   polygon
   (cons (last polygon) polygon))))

(define (draw-frame t)
 (xclearwindow *display* *display-pane*)
 (when *polygon-movie*
  (cond
   (*all?*
    (for-each
     (lambda (object) (draw-polygon *display* *display-pane* *thin-gc* object))
     (vector-ref *polygon-movie* t)))
   (else (draw-polygon *display* *display-pane* *thin-gc*
		       (list-ref (vector-ref *polygon-movie* t) *m*)))))
 (when *pathname*
  (xdrawstring
   *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 5)
   *pathname* (string-length *pathname*))))

(define (close-movie!) (set! *polygon-movie* #f))

(define (open-movie!)
 (when *pathname*
  (when (can-open-file-for-input?
	 (replace-extension
	  (string-append *directory* *pathname*) "polygon-movie"))
   (set! *polygon-movie*
	 (read-object-from-file
	  (replace-extension
	   (string-append *directory* *pathname*) "polygon-movie")))
   (set! *end* (- (vector-length *polygon-movie*) 1)))
  (set! *t* 0)
  (set! *m* 0)
  (set! *first-frame* 0)
  (set! *last-frame* *end*))
 (message "")
 (redraw-buttons)
 (redraw-display-pane))

(define (change-movie!) (close-movie!) (open-movie!))

;;; Commands

(define (increment-t-command)
 (message "")
 (unless *polygon-movie* (message "No movie") (abort))
 (when (= *t* *end*) (message "At end of movie") (abort))
 (set! *t* (+ *t* 1))
 (set! *m* (min *m* (- (number-of-objects-in-frame *polygon-movie* *t*) 1)))
 (redraw-buttons)
 (redraw-display-pane))

(define (decrement-t-command)
 (message "")
 (unless *polygon-movie* (message "No movie") (abort))
 (when (zero? *t*) (message "At beginning of movie") (abort))
 (set! *t* (- *t* 1))
 (set! *m* (min *m* (- (number-of-objects-in-frame *polygon-movie* *t*) 1)))
 (redraw-buttons)
 (redraw-display-pane))

(define (increment-m-command)
 (message "")
 (unless *polygon-movie* (message "No movie") (abort))
 (when (= *m* (- (number-of-objects-in-frame *polygon-movie* *t*) 1))
  (message "Last segment")
  (abort))
 (set! *m* (+ *m* 1))
 (set! *all?* #f)
 (redraw-buttons)
 (redraw-display-pane))

(define (decrement-m-command)
 (message "")
 (unless *polygon-movie* (message "No movie") (abort))
 (when (zero? *m*) (message "First segment") (abort))
 (set! *m* (- *m* 1))
 (set! *all?* #f)
 (redraw-buttons)
 (redraw-display-pane))

(define (play-command)
 (message "")
 (unless *polygon-movie* (message "No movie") (abort))
 (for-each-n (lambda (t)
	      (draw-frame t)
	      (xflush *display*)
	      (usleep (inexact->exact (round (/ 1000000.0 *frame-rate*)))))
	     *end*)
 (redraw-display-pane))

(define (beginning-command)
 (message "")
 (unless *polygon-movie* (message "No movie") (abort))
 (unless (zero? *t*)
  (set! *t* 0)
  (set! *m* (min *m* (- (number-of-objects-in-frame *polygon-movie* *t*) 1)))
  (redraw-buttons)
  (redraw-display-pane)))

(define (end-command)
 (message "")
 (unless *polygon-movie* (message "No movie") (abort))
 (unless (= *t* *end*)
  (set! *t* *end*)
  (set! *m* (min *m* (- (number-of-objects-in-frame *polygon-movie* *t*) 1)))
  (redraw-buttons)
  (redraw-display-pane)))

(define (reload-command)
 (message "")
 (unless *pathname* (message "You must specify a pathname") (abort))
 (change-movie!))

;;; Top Level

(define-application viewer 640 500 0 4 6
 (lambda ()
  (set! *directory* "")
  (set! *pathnames* '())
  (set! *polygon-movie* #f)
  (set! *all?* #t)
  (set! *m* 0)
  (open-movie!)
  (define-button 2 0 "Festival" #f
   (lambda ()
    (unless *pathname* (message "You must specify a pathname") (abort))
    (set! *pathnames*
	  (map strip-extension
	       (directory-list
		(replace-extension
		 (string-append *directory* *pathname*) "polygon-movie"))))
    (when (null? *pathnames*) (message "Empty festival") (abort))
    (set! *pathname* (first *pathnames*))
    (change-movie!)))
  (define-button 3 0 "Previous" #f
   (lambda ()
    (when (null? *pathnames*) (message "Empty festival") (abort))
    (set! *pathnames*
	  (cons (last *pathnames*) (reverse (rest (reverse *pathnames*)))))
    (set! *pathname* (first *pathnames*))
    (change-movie!)))
  (define-button 4 0 "Next" #f
   (lambda ()
    (when (null? *pathnames*) (message "Empty festival") (abort))
    (set! *pathnames* (append (rest *pathnames*) (list (first *pathnames*))))
    (set! *pathname* (first *pathnames*))
    (change-movie!)))
  (define-button 5 0 "Quit" #f quit)
  (define-button 0 1 "Directory" #f
   (lambda ()
    (message "")
    (unless *pathname* (message "You must specify a pathname") (abort))
    (close-movie!)
    (set! *directory*
	  (if (char=? (last (string->list *pathname*)) #\/)
	      *pathname*
	      (list->string (append (string->list *pathname*) (list #\/)))))
    (set! *pathname* #f)
    (redraw-display-pane)))
  (define-button 1 1 "Set First" #f
   (lambda ()
    (message "")
    (unless *polygon-movie* (message "No movie") (abort))
    (set! *first-frame* *t*)
    (set! *last-frame* (max *last-frame* *first-frame*))
    (redraw-buttons)))
  (define-button 2 1 "Play" #f play-command)
  (define-button 3 1 "Play Part" #f
   (lambda ()
    (message "")
    (unless *polygon-movie* (message "No movie") (abort))
    (for-each-n (lambda (t)
		 (draw-frame (+ *first-frame* t))
		 (xflush *display*)
		 (usleep (inexact->exact (round (/ 1000000.0 *frame-rate*)))))
		(+ (- *last-frame* *first-frame*) 1))
    (redraw-display-pane)))
  (define-button 4 1 "Set Last" #f
   (lambda ()
    (message "")
    (unless *polygon-movie* (message "No movie") (abort))
    (set! *last-frame* *t*)
    (set! *first-frame* (min *first-frame* *last-frame*))
    (redraw-buttons)))
  (define-button 0 2 "Beginning" #f beginning-command)
  (define-button 1 2
   (lambda ()
    (if *polygon-movie* (format #f "First ~s" *first-frame*) "First"))
   #f
   (lambda ()
    (message "")
    (unless *polygon-movie* (message "No movie") (abort))
    (unless (= *t* *first-frame*)
     (set! *t* *first-frame*)
     (set! *m*
	   (min *m* (- (number-of-objects-in-frame *polygon-movie* *t*) 1)))
     (redraw-buttons)
     (redraw-display-pane))))
  (define-button 2 2
   (lambda () (if *polygon-movie* (format #f "-T ~s" *t*) "-T"))
   #f
   decrement-t-command)
  (define-button 3 2
   (lambda () (if *polygon-movie* (format #f "+T ~s" *t*) "+T"))
   #f
   increment-t-command)
  (define-button 4 2
   (lambda () (if *polygon-movie* (format #f "Last ~s" *last-frame*) "Last"))
   #f
   (lambda ()
    (message "")
    (unless *polygon-movie* (message "No movie") (abort))
    (unless (= *t* *last-frame*)
     (set! *t* *last-frame*)
     (set! *m*
	   (min *m* (- (number-of-objects-in-frame *polygon-movie* *t*) 1)))
     (redraw-buttons)
     (redraw-display-pane))))
  (define-button 5 2 "End" #f end-command)
  (define-button 0 3
   (lambda () (if *polygon-movie* (format #f "-M ~s" *m*) "-M"))
   #f
   decrement-m-command)
  (define-toggle-button 1 3 "All" *all?* redraw-display-pane)
  (define-button 2 3
   (lambda () (if *polygon-movie* (format #f "+M ~s" *m*) "+M"))
   #f
   increment-m-command)
  (define-button 5 3 "Reload" #f reload-command)
  (define-key (meta #\f) "Increment T" increment-t-command)
  (define-key (meta #\b) "Decrement T" decrement-t-command)
  (define-key (control (meta #\f)) "Increment M" increment-m-command)
  (define-key (control (meta #\b)) "Decrement M" decrement-m-command)
  (define-key (control #\p) "Play" play-command)
  (define-key (list (control #\x) (control #\c)) "Quit" quit)
  (define-key (meta #\<) "Beginning" beginning-command)
  (define-key (meta #\>) "End" end-command)
  (define-key (control #\r) "Reload" reload-command))
 (lambda () #f)
 close-movie!
 (lambda () (draw-frame *t*))
 (lambda ()
  (set! *pathname* (and (not (string=? *input* "")) (strip-extension *input*)))
  (change-movie!)))

(define-command (main (at-most-one ("fps" fps? (fps "#" real-argument 30.0)))
		      (optional (pathname "pathname" string-argument #f)))
 (set! *frame-rate* fps)
 (set! *pathname* (and pathname (strip-extension pathname)))
 (viewer '()))

;;; Tam V'Nishlam Shevah L'El Borei Olam
