(MODULE
 PLAY-VIDEO
 (WITH
  QOBISCHEME
  XLIB
  HLIB-CHANGE-MOVIE
  HLIB-EM
  HLIB-FIND-LINES
  HLIB-HULL
  HLIB-INTERPRET
  HLIB-KINEMATIC
  HLIB-LOCATE
  HLIB-MERGE-LINES
  HLIB-MOST
  HLIB-POSTSCRIPT
  HLIB-RATIO-REGIONS
  HLIB-RILIB
  HLIB-SDE
  HLIB-SNAKE
  HLIB-SPAN
  HLIB-SPEAK
  HLIB-TRACK
  HLIB-TRAIN-CLASSIFY
  HLIB-WBM)
 (MAIN MAIN))
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "play-video.sch")
(include "c-externals.sc")

(set! *program* "play-video")
(set! *panic?* #t)

;;; Variables

(define *pathname* #f)
(define *pixmaps* #f)
(define *t* #f)
(define *end* #f)
(define *first-frame* #f)
(define *last-frame* #f)
(define *video-type* #f)

;;; Procedures

(define (video-type)
 (case *video-type*
  ((jpeg) "Jpeg")
  ((mpeg1) "Mpeg1")
  (else (fuck-up))))

(define (close-movie!)
 (when *pixmaps* (free-pixmaps *pixmaps*) (set! *pixmaps* #f)))

(define (open-movie!)
 (when *pathname*
  (when (can-open-file-for-input?
	 (replace-extension *pathname* (video-type)))
   (set! *pixmaps* (video-file->pixmaps
		    (replace-extension *pathname* (video-type)) (video-type)))
   (set! *end* (- (vector-length *pixmaps*) 1)))
  (set! *t* 0)
  (set! *first-frame* 0)
  (set! *last-frame* *end*))
 (message "")
 (redraw-buttons)
 (redraw-display-pane))

(define (change-movie!) (close-movie!) (open-movie!))

;;; Commands

(define (play-command)
 (message "")
 (unless *pixmaps* (message "No movie") (abort))
 (xclearwindow *display* *display-pane*)
 (when *pathname*
  (xdrawstring
   *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 5)
   *pathname* (string-length *pathname*)))
 (draw-pixmaps *pixmaps* 0 0)
 (redraw-display-pane))

(define (beginning-command)
 (message "")
 (unless *pixmaps* (message "No movie") (abort))
 (unless (zero? *t*)
  (set! *t* 0)
  (redraw-buttons)
  (redraw-display-pane)))

(define (decrement-t-command)
 (message "")
 (unless *pixmaps* (message "No movie") (abort))
 (when (zero? *t*) (message "At beginning of movie") (abort))
 (set! *t* (- *t* 1))
 (redraw-buttons)
 (redraw-display-pane))

(define (increment-t-command)
 (message "")
 (unless *pixmaps* (message "No movie") (abort))
 (when (= *t* *end*) (message "At end of movie") (abort))
 (set! *t* (+ *t* 1))
 (redraw-buttons)
 (redraw-display-pane))

(define (end-command)
 (message "")
 (unless *pixmaps* (message "No movie") (abort))
 (unless (= *t* *end*)
  (set! *t* *end*)
  (redraw-buttons)
  (redraw-display-pane)))

;;; Top Level

(define-application viewer 640 500 0 3 6
 (lambda ()
  (set! *video-type* 'jpeg)
  (open-movie!)
  (define-button 1 0 "Set First" #f
   (lambda ()
    (message "")
    (unless *pixmaps* (message "No movie") (abort))
    (set! *first-frame* *t*)
    (set! *last-frame* (max *last-frame* *first-frame*))
    (redraw-buttons)))
  (define-button 2 0 "Play" #f play-command)
  (define-button 3 0 "Play Part" #f
   (lambda ()
    (message "")
    (unless *pixmaps* (message "No movie") (abort))
    (xclearwindow *display* *display-pane*)
    (when *pathname*
     (xdrawstring
      *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 5)
      *pathname* (string-length *pathname*)))
    (for-each-n (lambda (t)
		 (draw-pixmap (vector-ref *pixmaps* (+ *first-frame* t)) 0 0)
		 (usleep (inexact->exact (round (/ 1000000.0 *frame-rate*)))))
		(+ (- *last-frame* *first-frame*) 1))
    (redraw-display-pane)))
  (define-button 4 0 "Set Last" #f
   (lambda ()
    (message "")
    (unless *pixmaps* (message "No movie") (abort))
    (set! *last-frame* *t*)
    (set! *first-frame* (min *first-frame* *last-frame*))
    (redraw-buttons)))
  (define-button 5 0 "Quit" #f quit)
  (define-button 0 1 "Beginning" #f beginning-command)
  (define-button 1 1
   (lambda () (if *pixmaps* (format #f "First ~s" *first-frame*) "First"))
   #f
   (lambda ()
    (message "")
    (unless *pixmaps* (message "No movie") (abort))
    (unless (= *t* *first-frame*)
     (set! *t* *first-frame*)
     (redraw-buttons)
     (redraw-display-pane))))
  (define-button 2 1
   (lambda () (if *pixmaps* (format #f "-T ~s" *t*) "-T"))
   #f
   decrement-t-command)
  (define-button 3 1
   (lambda () (if *pixmaps* (format #f "+T ~s" *t*) "+T"))
   #f
   increment-t-command)
  (define-button 4 1
   (lambda () (if *pixmaps* (format #f "Last ~s" *last-frame*) "Last"))
   #f
   (lambda ()
    (message "")
    (unless *pixmaps* (message "No movie") (abort))
    (unless (= *t* *last-frame*)
     (set! *t* *last-frame*)
     (redraw-buttons)
     (redraw-display-pane))))
  (define-button 5 1 "End" #f end-command)
  (define-radio-buttons *video-type* change-movie!
   (0 2 jpeg "Jpeg")
   (1 2 mpeg1 "Mpeg1"))
  (define-key (control #\p) "Play" play-command)
  (define-key (meta #\f) "Increment T" increment-t-command)
  (define-key (meta #\b) "Decrement T" decrement-t-command)
  (define-key (meta #\<) "Beginning" beginning-command)
  (define-key (meta #\>) "End" end-command)
  (define-key (list (control #\x) (control #\c)) "Quit" quit))
 (lambda () #f)
 close-movie!
 (lambda ()
  (when *pathname*
   (xdrawstring
    *display* *display-pane* *roman-gc* 5 (- *display-pane-height* 5)
    *pathname* (string-length *pathname*)))
  (when *pixmaps* (draw-pixmap (vector-ref *pixmaps* *t*) 0 0)))
 (lambda ()
  (set! *pathname* (and (not (string=? *input* "")) (strip-extension *input*)))
  (change-movie!)))

(define-command (main (at-most-one ("fps" fps? (fps "#" real-argument 30.0)))
		      (optional (pathname "pathname" string-argument #f)))
 (set! *frame-rate* fps)
 (set! *pathname* (and pathname (strip-extension pathname)))
 (viewer '()))

;;; Tam V'Nishlam Shevah L'El Borei Olam
