/* LaHaShem HaAretz U'Mloah */

/* Version 1.0: 5 July 1995
 * Version 1.1: 6 July 1995
 * - added set_version_printing(), print_version()
 * Version 1.2: 7 July 1995
 * - added __my_sqrt()
 * - added radial_distance()
 * Version 1.3: 7 August 1995
 * - changed version logging style
 * Version 1.4: 14 August 1995
 * - added free_matrix2d and __matrix2d
 * Version 1.5: 7 April 1996
 * - Added my_log since log(0.0) doesn't return -Inf on Alpha.
 * - Changed my_malloc() to check for zero argument because malloc(0) returns
 *   NULL on alpha.
 * Version 1.6: 30 April 1996
 * - Added my_atan2 since atan2(0.0, 0.0) returns NaNQ on Alpha.
 * Version 1.7: 20 October 1996
 * - Changed ifndef LINUX to ifdef SUN because SGI doesn't have matherr.
 * Version 1.8: 30 November 1996
 * - New HMM library.
 * Version 1.9: 1 May 1998
 * - Merged xil and hlib.
 */

/* TTMTTD :
   1.  A generalised matrix allocation and de-allocation function which takes
       as its first argument the dimensionality of the matrix.
   2.  Calculate sqrt(2*PI) and encode it in a #define */

/* Includes */

#include <math.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <string.h>
#include <memory.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <sched.h>
#include <netdb.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/extensions/XShm.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/resource.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <linux/soundcard.h>
#include "viff.h"
#include "vdefines.h"
#include "khoros.h"
#include "hlib-c.h"
#include "ioctl_meteor.h"

/* Type Definitions */

struct pair {Real sin, cos;};
enum predictor {NONE, ACCELERATION, VELOCITY, POSITION, ANYWHERE};
enum VideoType {JPEG, MPEG1};
typedef int TimeStamp;
typedef struct pict_image {
  unsigned char *luminance;              /* Luminance plane.   */
  unsigned char *Cr;                     /* Cr plane.          */
  unsigned char *Cb;                     /* Cb plane.          */
  unsigned char *display;                /* Display plane.     */
  int locked;                            /* Lock flag.         */
  TimeStamp show_time;                   /* Presentation time. */

#ifdef SH_MEM
  XShmSegmentInfo shminfo;               /* Segment info.  */
  XImage *ximage;                        /* Ximage struct. */
#endif

} PictImage;
typedef struct GoP {
  BOOLEAN drop_flag;                     /* Flag indicating dropped frame. */
  unsigned int tc_hours;                 /* Hour component of time code.   */
  unsigned int tc_minutes;               /* Minute component of time code. */
  unsigned int tc_seconds;               /* Second component of time code. */
  unsigned int tc_pictures;              /* Picture counter of time code.  */
  BOOLEAN closed_gop;                    /* Indicates no pred. vectors to
					    previous group of pictures.    */
  BOOLEAN broken_link;                   /* B frame unable to be decoded.  */
  char *ext_data;                        /* Extension data.                */
  char *user_data;                       /* User data.                     */
} GoP;
typedef struct pict {
  unsigned int temp_ref;                 /* Temporal reference.             */
  unsigned int code_type;                /* Frame type: P, B, I             */
  unsigned int vbv_delay;                /* Buffer delay.                   */
  BOOLEAN full_pel_forw_vector;          /* Forw. vectors specified in full
					    pixel values flag.              */
  unsigned int forw_r_size;              /* Used for vector decoding.       */
  unsigned int forw_f;                   /* Used for vector decoding.       */
  BOOLEAN full_pel_back_vector;          /* Back vectors specified in full
					    pixel values flag.              */
  unsigned int back_r_size;              /* Used in decoding.               */
  unsigned int back_f;                   /* Used in decoding.               */
  char *extra_info;                      /* Extra bit picture info.         */
  char *ext_data;                        /* Extension data.                 */
  char *user_data;                       /* User data.                      */
} Pict;
typedef struct slice {
  unsigned int vert_pos;                 /* Vertical position of slice. */
  unsigned int quant_scale;              /* Quantization scale.         */
  char *extra_info;                      /* Extra bit slice info.       */
} Slice;
typedef struct macroblock {
  int mb_address;                        /* Macroblock address.              */
  int past_mb_addr;                      /* Previous mblock address.         */
  int motion_h_forw_code;                /* Forw. horiz. motion vector code. */
  unsigned int motion_h_forw_r;          /* Used in decoding vectors.        */
  int motion_v_forw_code;                /* Forw. vert. motion vector code.  */
  unsigned int motion_v_forw_r;          /* Used in decdoinge vectors.       */
  int motion_h_back_code;                /* Back horiz. motion vector code.  */
  unsigned int motion_h_back_r;          /* Used in decoding vectors.        */
  int motion_v_back_code;                /* Back vert. motion vector code.   */
  unsigned int motion_v_back_r;          /* Used in decoding vectors.        */
  unsigned int cbp;                      /* Coded block pattern.             */
  BOOLEAN mb_intra;                      /* Intracoded mblock flag.          */
  BOOLEAN bpict_past_forw;               /* Past B frame forw. vector flag.  */
  BOOLEAN bpict_past_back;               /* Past B frame back vector flag.   */
  int past_intra_addr;                   /* Addr of last intracoded mblock.  */
  int recon_right_for_prev;              /* Past right forw. vector.         */
  int recon_down_for_prev;               /* Past down forw. vector.          */
  int recon_right_back_prev;             /* Past right back vector.          */
  int recon_down_back_prev;              /* Past down back vector.           */
} Macroblock;
typedef struct block {
  short int dct_recon[8][8];             /* Reconstructed dct coeff matrix. */
  short int dct_dc_y_past;               /* Past lum. dc dct coefficient.   */
  short int dct_dc_cr_past;              /* Past cr dc dct coefficient.     */
  short int dct_dc_cb_past;              /* Past cb dc dct coefficient.     */
} Block;
typedef struct vid_stream {
  unsigned int h_size;                         /* Horiz. size in pixels.     */
  unsigned int v_size;                         /* Vert. size in pixels.      */
  unsigned int mb_height;                      /* Vert. size in mblocks.     */
  unsigned int mb_width;                       /* Horiz. size in mblocks.    */
  unsigned char aspect_ratio;                  /* Code for aspect ratio.     */
  unsigned char picture_rate;                  /* Code for picture rate.     */
  unsigned int bit_rate;                       /* Bit rate.                  */
  unsigned int vbv_buffer_size;                /* Minimum buffer size.       */
  BOOLEAN const_param_flag;                    /* Contrained parameter flag. */
  unsigned char intra_quant_matrix[8][8];      /* Quantization matrix for
						  intracoded frames.         */
  unsigned char non_intra_quant_matrix[8][8];  /* Quanitization matrix for
						  non intracoded frames.     */
  char *ext_data;                              /* Extension data.            */
  char *user_data;                             /* User data.                 */
  GoP group;                                   /* Current group of pict.     */
  Pict picture;                                /* Current picture.           */
  Slice slice;                                 /* Current slice.             */
  Macroblock mblock;                           /* Current macroblock.        */
  Block block;                                 /* Current block.             */
  int state;                                   /* State of decoding.         */
  int bit_offset;                              /* Bit offset in stream.      */
  unsigned int *buffer;                        /* Pointer to next byte in
						  buffer.                    */
  int buf_length;                              /* Length of remaining buffer.*/
  unsigned int *buf_start;                     /* Pointer to buffer start.   */
  int max_buf_length;                          /* Max lenght of buffer.      */
  PictImage *past;                             /* Past predictive frame.     */
  PictImage *future;                           /* Future predictive frame.   */
  PictImage *current;                          /* Current frame.             */
  PictImage *ring[RING_BUF_SIZE];              /* Ring buffer of frames.     */
} VidStream;

/* Function Prototypes */

VidStream *open_video_input_file(char *pathname, enum VideoType video_type);
int read_video(VidStream *video_input_port);
int get_video_width(VidStream *video_input_port);
int get_video_height(VidStream *video_input_port);
unsigned int get_video_red(VidStream *video_input_port, int y, int x);
unsigned int get_video_green(VidStream *video_input_port, int y, int x);
unsigned int get_video_blue(VidStream *video_input_port, int y, int x);
void close_video_input_port(VidStream *video_input_port);
FILE *open_video_output_file(char *pathname,
			     enum VideoType video_type,
			     int numFrames);
void begin_video_frame(int width, int height, int maxVal);
void put_video_red(int y, int x, int value);
void put_video_green(int y, int x, int value);
void put_video_blue(int y, int x, int value);
void end_video_frame(void);
void close_video_output_port(FILE *video_output_port);
RMat *allocate_rmat_vector(size_t n);
RVec allocRVec(size_t x);
RMat allocRMat(size_t y, size_t x);
RMat3d allocRMat3d(size_t z, size_t y, size_t x);
void free_rmat_vector(RMat *v);
void freeRVec(RVec v);
void freeRMat(RMat m);
void freeRMat3d(RMat3d m);
Real *addRVec(Real *u, const Real *v, int size);
Real *copyRVec(Real *u, const Real *v, int size);
Real dotProdRVec(const Real *u, const Real *v, int size);
Real normaliseRVec(Real *v, int size);
Real rmat_get(RMat rmat, int i, int j);
void rmat_set(RMat rmat, int i, int j, Real x);
void rmat_vector_set(RMat *v, int i, RMat r);
Real *scaleRVec(Real *v, Real k, int size);
Real sumOfLogs(Real *v, int size);
Real sumRVec(Real *v, int size);
void set_trace(int n, int trace);
Real add_exp(Real e1, Real e2);
void clip_symmatrix(Real *xx, Real *xy, Real *yy, Real clip);
int distinct_reals(Real a, Real b);
Real dot_product(Real *u, Real *v, int size);
Real eigenvalue1(Real xx, Real xy, Real yx, Real yy);
Real eigenvalue2(Real xx, Real xy, Real yx, Real yy);
Real eigenvector_angle1(Real xx, Real xy, Real yx, Real yy);
Real eigenvector_angle2(Real xx, Real xy, Real yx, Real yy);
void eigenvectors(Real xx, Real xy, Real yx, Real yy, Real *exx, Real *exy,
		  Real *eyx, Real *eyy);
void free_matrix(Real **m, int nrl, int nrh, int ncl, int nch);
void free_matrix2d(Real **m, int n);
void free_matrix3d(Real ***a, int x, int y);
void free_matrix4d(Real ****a, int x, int y, int z);
void free_vector(Real *v, int nl, int nh);
int is_nan(Real x);
Real *map_real(Real (*f)(Real), Real *v, Real *rv, int size);
int max_int(int *v, int size);
#ifdef SUN
int matherr(struct exception *exc);
#endif
Real **__matrix(int nrl, int nrh, int ncl, int nch, int line,
		const char *file);
Real **__matrix2d(int x, int y, int line, const char *file);
Real ***__matrix3d(int x, int y, int z, int line, const char *file);
Real ****__matrix4d(int x, int y, int z, int w, int line, const char *file);
Real my_add_exp(Real e1, Real e2, int line, const char *file);
Real my_sqrt(Real x);
Real __my_sqrt(Real x, int line, const char *file);
Real my_exp(Real x);
void *my_malloc(int n, int line, const char *file);
Real my_log(Real x);
Real my_atan2(Real x, Real y);
Real normalise_rotation(Real rotation);
void panic(const char *error_text, ...);
void print_version(const char *version, const char *led, const char *file);
Real quadratic1(Real a, Real b, Real c);
Real quadratic2(Real a, Real b, Real c);
Real radial_distance(Real t, Real u);
void reportWarn(const char *error_text, ...);
void *_safe_malloc(size_t size, int line, char *file);
void set_version_printing(int boolean);
Real sqr_real(Real x);
Real sum_real(Real *v, int size);
void trace(int n, const char *error_text, ...);
Real *__vector(int nl, int nh, int line, const char *file);
Real log_normal(int n, Real *x, Real *mu, Real sqrt_big_sigma_determinant,
		Real **big_sigma_inverse);
Real log_normal_1d(Real x, Real mu, Real sigma);
Real normal(int n, Real *x, Real *mu, Real sqrt_big_sigma_determinant,
	    Real **big_sigma_inverse);
Real normal_1d(Real x, Real mu, Real sigma);
struct xvimage *create_image(unsigned int height, unsigned int width);
void viff_set_pixel(struct xvimage *viff,
		    unsigned int y, unsigned int x, unsigned int grey);
void detect_edges(float a1, float a2, int t1, int t2, int l, int w,
		  struct xvimage *viff);
void detect_line_segments(float a1, float a2, int t1, int t2, int l, int w,
			  struct xvimage *viff);
int edge_pixel(struct xvimage *viff, int y, int x);
int number_of_line_segments(void);
int line_segment_x1(int i);
int line_segment_y1(int i);
int line_segment_x2(int i);
int line_segment_y2(int i);
void free_xvimage(struct xvimage *viff);
int *make_c_int_vector(int l);
int c_int_vector_ref(int *v, int x);
void c_int_vector_set(int *v, int x, int i);
void free_c_int_vector(int *v);
int **make_c_int_matrix(int h, int w);
int c_int_matrix_ref(int **m, int y, int x);
void c_int_matrix_set(int **m, int y, int x, int i);
void free_c_int_matrix(int **m, int h);
float *make_c_float_vector(int l);
float c_float_vector_ref(float *v, int x);
void c_float_vector_set(float *v, int x, float i);
void free_c_float_vector(float *v);
float **make_c_float_matrix(int h, int w);
float c_float_matrix_ref(float **m, int y, int x);
void c_float_matrix_set(float **m, int y, int x, float i);
void free_c_float_matrix(float **m, int h);
Real *make_c_real_vector(int l);
Real c_real_vector_ref(Real *v, int x);
void c_real_vector_set(Real *v, int x, Real i);
void free_c_real_vector(Real *v);
Real **make_c_real_matrix(int h, int w);
Real c_real_matrix_ref(Real **m, int y, int x);
void c_real_matrix_set(Real **m, int y, int x, Real i);
void free_c_real_matrix(Real **m, int h);
void free_c_vector(void *v);
void free_c_matrix(void **m, int h);
int separation(int t, int u);
void setup_proximity_clusterer(int max);
int pbm_proximity_clusterer(int **m, int h, int w, int p,
			    int category, int parity, int start);
void threshold_normal_flow_magnitude(int width, int height,
				     const Real **e1,
				     const Real **e2,
				     int **t, Real epsilon,
				     int sigma, Real threshold);
int histogram_clusterer(const int *pixel_hues, int size, int quanta,
			int area_e, int hue_separation_e, int *bins,
			int *hues);
int get_first_frame(void);
int get_last_frame(void);
int get_current_frame(void);
int get_motion(int frame);
int get_object_k(int frame);
int get_object_colour(int frame, int k);
Real get_ellipse_x(int frame, int k);
Real get_ellipse_y(int frame, int k);
Real get_ellipse_theta(int frame, int k);
Real get_ellipse_r(int frame, int k);
Real get_ellipse_s(int frame, int k);
unsigned get_polygon_l(int frame, int k);
unsigned get_polygon_x(int frame, int k, int j);
unsigned get_polygon_y(int frame, int k, int j);
int chmt(char *pathname,
	 char *track_pathname,
	 enum VideoType video_type,
	 Real background_threshold,
	 int quanta,
	 int colour_pixel_mass_threshold,
	 int hue_separation_threshold,
	 int colour_proximity_threshold,
	 int motion_threshold,
	 int motion_pixel_mass_threshold,
	 int motion_proximity_threshold,
	 int track_movie,
	 int left,
	 int n,
	 Real *background_c,
	 Real *background_d);
void compute_autoclip_amount(char *pathname,
			     enum VideoType video_type,
			     int motion_threshold,
			     int motion_pixel_mass_threshold,
			     int motion_proximity_threshold);
void convolve(int predicted_x, int predicted_y,
	      int vicinity, enum predictor class,
	      int width, int height, int stamp_width, int stamp_height,
	      int subsample,
	      float **stamp_red,
	      float **stamp_green,
	      float **stamp_blue,
	      int **stamp_bitmap,
	      float **red, float **green, float **blue,
	      int *stamp_x, int *stamp_y, float *best,
	      enum predictor *predictor);
int rfbt(char *pathname,
	 char *track_pathname,
	 enum VideoType video_type,
	 int acceleration_vicinity,
	 int velocity_vicinity,
	 int position_vicinity,
	 int subsample,
	 Real jerky,
	 Real ratio,
	 int colour_pixel_mass_threshold,
	 int colour_proximity_threshold,
	 int motion_threshold,
	 int motion_pixel_mass_threshold,
	 int motion_proximity_threshold,
	 int hand_left,
	 int hand_top,
	 int hand_right,
	 int hand_bottom,
	 int track_movie);
int wall_clock(void);
void meteor_tracker(unsigned char *buffer,
		    int frame,
		    int height,
		    int width,
		    int uninterlaced,
		    int capturing,
                    int capturing_code,
		    int tracking_ellipses,
		    int tracking_polygons,
		    int hull,
		    int motion_threshold,
		    int colour_pixel_mass_threshold,
		    int colour_proximity_threshold,
		    int motion_pixel_mass_threshold,
		    int motion_proximity_threshold);
void meteor_handler(int signum);
void meteor_offline_tracker(int frame,
			    int height,
			    int width,
			    int uninterlaced,
			    int capturing_code,
			    int tracking_ellipses,
			    int tracking_polygons,
			    int hull,
			    int motion_threshold,
			    int colour_pixel_mass_threshold,
			    int colour_proximity_threshold,
			    int motion_pixel_mass_threshold,
			    int motion_proximity_threshold);
void meteor_draw_ellipse(float x0, float y0, float t0, float a, float b);
void meteor_set_real_time(void);
void meteor_set_nonreal_time(void);
void meteor_open(void);
void meteor_close(void);
void meteor_sigaction(void);
void meteor_set_geometry(int height, int width, int frames, int oformat,
			 int uninterlaced);
void meteor_set_video_format(int format);
void meteor_set_input_device(int device);
void meteor_get_frame_offset(struct meteor_frame_offset *offset);
void meteor_set_signal(int signal);
void meteor_block(void);
void meteor_unblock(void);
void meteor_capture(int cap);
void meteor_capframe(struct meteor_capframe *capframe);
void meteor_mmap(void);
void meteor_munmap(void);
void meteor_malloc(int nframes);
void meteor_page_in_frames(int iframes, int nframes);
void meteor_set_deinterlaced(int deinterlaced);
int meteor_get_red(int i, int y, int x);
int meteor_get_green(int i, int y, int x);
int meteor_get_blue(int i, int y, int x);
int meteor_get_pixel(int i, int y, int x);
void meteor_set_red(int i, int y, int x, int v);
void meteor_set_green(int i, int y, int x, int v);
void meteor_set_blue(int i, int y, int x, int v);
void meteor_set_pixel(int i, int y, int x, int v);
int meteor_get_code_frame_red(int i, int y, int x);
int meteor_get_code_frame_green(int i, int y, int x);
int meteor_get_code_frame_blue(int i, int y, int x);
void meteor_copy_frame_to_ximage(int i, XImage *ximage);
void meteor_motion_detector(int i, XImage *ximage);
void meteor_set_hack_track(int motion_threshold,
			   int colour_pixel_mass_threshold,
			   int colour_proximity_threshold,
			   int motion_pixel_mass_threshold,
			   int motion_proximity_threshold);
void meteor_set_red_hack_track(int hl, int hh,
			       int sl, int sh,
			       int vl, int vh);
void meteor_set_green_hack_track(int hl, int hh,
				 int sl, int sh,
				 int vl, int vh);
void meteor_set_blue_hack_track(int hl, int hh,
				int sl, int sh,
				int vl, int vh);
void meteor_set_yellow_hack_track(int hl, int hh,
				  int sl, int sh,
				  int vl, int vh);
void meteor_start_hack_track(int frame);
void meteor_hack_track_step(int frame,
			    int object_frame,
			    int colour_pixel_mass_threshold,
			    int colour_proximity_threshold,
			    int motion_threshold,
			    int motion_pixel_mass_threshold,
			    int motion_proximity_threshold);
void meteor_finish_hack_track(void);
void meteor_show_frame(int i);
void meteor_show_code_frame(int i);
void meteor_free(void);
void meteor_attach_viewfinder(Display *display, int screen, Window window,
			      int window_width, int window_height);
void meteor_detach_viewfinder(void);
void meteor_turn_on_viewfinder(int m);
void meteor_turn_on_capturing(int i, int l, int n);
void meteor_turn_on_ellipse_tracker(int i, int l, int n);
void meteor_turn_on_ellipse_tracker_with_capturing(int i, int l, int n);
void meteor_turn_on_polygon_tracker(int i, int l, int n, int h);
void meteor_turn_on_polygon_tracker_with_capturing(int i, int l, int n, int h);
void meteor_turn_off(void);
void meteor_capture_frames(int i, int l, int n);
void meteor_track_ellipses(int i, int l, int n);
void meteor_track_ellipses_and_capture_frames(int i, int l, int n);
void meteor_track_polygons(int i, int l, int n, int h);
void meteor_track_polygons_and_capture_frames(int i, int l, int n, int h);
void meteor_startup(int height, int width, int nframes, int uninterlaced);
void meteor_startup_offline(int height,
			    int width,
			    int nframes,
			    int uninterlaced);
void meteor_startup_with_viewfinder(int height, int width, int nframes,
				    int uninterlaced,
				    Display *display, int screen,
				    Window window,
				    int window_width, int window_height);
void meteor_restartup(int height, int width, int nframes, int uninterlaced);
void meteor_shutdown(void);
void meteor_shutdown_offline(void);
void send_frame(FILE *f, int i);
void receive_frame(FILE *f, int i);
void set_n_points(int n1, int n2);
void augmented_set_n_points(int n1, int n2,
			    int max_x_velocity, int max_y_velocity);
void set_point1(int i, int x, int y);
void set_point2(int i, int x, int y);
int xor_points_count(int dx, int dy);
void exhaustive_xor_velocity_points(int max_x_velocity, int max_y_velocity);
void hill_climbing_xor_velocity_points(int max_x_velocity, int max_y_velocity);
void xor_velocity_points(int max_x_velocity, int max_y_velocity);
void full_xor_velocity_points(int max_x_velocity, int max_y_velocity);
void clipped_xor_velocity_points(int max_x_velocity, int max_y_velocity,
				 int xl, int yl, int xh, int yh);
void augmented_model(int max_x_velocity, int max_y_velocity);
void sum_min_velocity_points(int max_x_velocity, int max_y_velocity);
void full_sum_min_velocity_points(int max_x_velocity, int max_y_velocity);
void clipped_sum_min_velocity_points(int max_x_velocity, int max_y_velocity,
				     int xl, int yl, int xh, int yh);
int get_x_velocity(void);
int get_y_velocity(void);
void free_points(void);
void augmented_free_points(int max_x_velocity, int max_y_velocity);
void edge_flow_set_n_points(int n1, int n2);
void edge_flow_set_point1(int i, int x, int y);
void edge_flow_set_point2(int i, int x, int y);
void edge_flow_set_height_width(int height, int width);
void edge_flow_set_rgb1(int x, int y, int red, int green, int blue);
void edge_flow_set_rgb2(int x, int y, int red, int green, int blue);
float edge_flow_absolute_difference(int x1, int y1, int x2, int y2,
				    int window_x, int window_y);
void edge_flow(int maximum_flow_magnitude, int window_x, int window_y);
float edge_flow_grey_difference(int x1, int y1, int x2, int y2,
				int window_x, int window_y);
void edge_flow_grey(int maximum_flow_magnitude, int window_x, int window_y);
int edge_flow_get(int i);
void edge_flow_free(void);
void set_speed_and_format(FILE *file, int speed);
float *make_float_vector(unsigned int length);
void free_float_vector(float *vector);
float float_vector_ref(float *vector, unsigned int i);
void float_vector_set(float *vector, unsigned int i, float x);
unsigned int file_length(FILE *file);
unsigned char *make_mapped_file(FILE *file, unsigned int length);
void free_mapped_file(unsigned char *mapped_file, unsigned int length);
unsigned int mapped_file_ref(unsigned char *mapped_file, unsigned int i);
void enqueue(int y, int x);
void relabel(int height,
	     int width,
	     int **mask,
	     int **c_right,
	     int **c_down,
	     int **w,
	     int v);
void preflow_push(int height,
		  int width,
		  int **mask,
		  int **c_right,
		  int **c_down,
		  int **w,
		  int v);
int **ratio_region(int height,
		   int width,
		   int **mask,
		   int **c_right,
		   int **c_down,
		   int **w,
		   int lg_max_v);

/* Global Variables */

int gtrace[TRACE_MAX_INDEX] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			       0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			       0, 0, 0, 0, 0};
struct xvimage *vect = NULL;
int **gpythag = NULL;
int gfirst_frame, glast_frame, gmotion[MAX_FRAMES], gobject_k[MAX_FRAMES];
int gobject_colour[MAX_FRAMES][MAX_OBJECTS];
Real gellipse_x[MAX_FRAMES][MAX_OBJECTS];
Real gellipse_y[MAX_FRAMES][MAX_OBJECTS];
Real gellipse_theta[MAX_FRAMES][MAX_OBJECTS];
Real gellipse_r[MAX_FRAMES][MAX_OBJECTS];
Real gellipse_s[MAX_FRAMES][MAX_OBJECTS];
unsigned short gpolygon_l[MAX_FRAMES][MAX_OBJECTS];
unsigned short gpolygon_x[MAX_FRAMES][MAX_OBJECTS][MAX_VERTICES];
unsigned short gpolygon_y[MAX_FRAMES][MAX_OBJECTS][MAX_VERTICES];
int meteor_real_time = FALSE;
int meteor_uid = -1;
int meteor_port;
int meteor_height, meteor_width;
int meteor_uninterlaced, meteor_deinterlaced;
int meteor_capturing = FALSE;
int meteor_tracking_ellipses = FALSE;
int meteor_tracking_polygons = FALSE;
int meteor_hull;
unsigned char *meteor_buffer;
struct meteor_frame_offset meteor_offset;
Display *meteor_display;
int meteor_screen;
Window meteor_window;
int meteor_window_width, meteor_window_height;
int meteor_i, meteor_n, meteor_j, meteor_m = 0, meteor_k, meteor_l = 0;
GC meteor_gc;
XShmSegmentInfo meteor_shminfo;
XImage *meteor_shmimage;
unsigned char *meteor_frames, *meteor_code_frames;
unsigned char *meteor_last_frame, **meteor_last_code, **meteor_conjure_code;
int meteor_starting, meteor_conjuring;
int meteor_motion_threshold = 24;
int meteor_colour_pixel_mass_threshold = 60;
int meteor_colour_proximity_threshold = 4;
int meteor_motion_pixel_mass_threshold = 200;
int meteor_motion_proximity_threshold = 8;
int meteor_rhl =  13, meteor_rhh = 336;
int meteor_rsl =  36, meteor_rsh =  74;
int meteor_rvl =  35, meteor_rvh =  70;
int meteor_ghl = 122, meteor_ghh = 164;
int meteor_gsl =  29, meteor_gsh =  71;
int meteor_gvl =  22, meteor_gvh =  55;
int meteor_bhl = 192, meteor_bhh = 227;
int meteor_bsl =  32, meteor_bsh =  66;
int meteor_bvl =  27, meteor_bvh =  61;
int meteor_yhl =  50, meteor_yhh =  70;
int meteor_ysl =  40, meteor_ysh =  65;
int meteor_yvl =  55, meteor_yvh =  80;
int meteor_first_frame, meteor_last_moving_valid, **meteor_last_moving;
int n_points1, n_points2, n_full_points1, n_full_points2;
int x_velocity, y_velocity;
struct point {int x; int y; int **a;};
struct point *points1, *points2, *full_points1, *full_points2;
int edge_flow_n_points1, edge_flow_n_points2, *edge_flow_map;
struct edge_flow_point {int x; int y;};
struct edge_flow_point *edge_flow_points1, *edge_flow_points2;
int edge_flow_height, edge_flow_width;
struct edge_flow_pixel {int red; int green; int blue;};
struct edge_flow_pixel **edge_flow_image1, **edge_flow_image2;

/* Function Definitions */

/* vector allocation and deallocation functions */

RMat *allocate_rmat_vector(size_t n)
{ RMat *v = (RMat *)safe_malloc(sizeof(RMat)*n);
  return v;}

RVec allocRVec(size_t x)
{ RVec v = (RVec)safe_malloc(sizeof(struct RVecStruct));
  VECTOR(v) = (Real *)safe_malloc(x*sizeof(Real));
  VLENGTH(v) = x;
  return v;}

RMat allocRMat(size_t y, size_t x)
{ int i;
  RMat m = (RMat)safe_malloc(sizeof(struct RMatStruct));
  MATRIX(m) = (Real **)safe_malloc(y*sizeof(Real *));
  for (i = 0; i<y; i++) MATRIX(m)[i] = (Real *)safe_malloc(x*sizeof(Real));
  ROWS(m) = y;
  COLUMNS(m) = x;
  return m;}

RMat3d allocRMat3d(size_t z, size_t y, size_t x)
{ int i, j;
  RMat3d m = (RMat3d)safe_malloc(sizeof(struct RMat3dStruct));
  MATRIX(m) = (Real ***)safe_malloc(z*sizeof(Real **));
  for (i = 0; i<z; i++)
  { MATRIX(m)[i] = (Real **)safe_malloc(y*sizeof(Real *));
    for (j = 0; j<y; j++)
    { MATRIX(m)[i][j] = (Real *)safe_malloc(x*sizeof(Real));}}
  m->z = z;
  m->y = y;
  m->x = x;
  return m;}

void free_rmat_vector(RMat *v) {free(v);}

void freeRVec(RVec v)
{ if (v!=NULL)
  { if (VECTOR(v)!=NULL) free(VECTOR(v));
    free(v);}}

void freeRMat(RMat m)
{ int i;
  if (m!=NULL)
  { if (MATRIX(m)!=NULL)
    { for (i = 0; i<ROWS(m); i++)
      { if (MATRIX(m)[i]!=NULL) free(MATRIX(m)[i]);}
      free(MATRIX(m));}
    free(m);}}

void freeRMat3d(RMat3d m)
{ int i, j;
  if (m!=NULL)
  { if (MATRIX(m)!=NULL)
    { for (i = 0; i<m->z; i++)
      { if (MATRIX(m)[i]!=NULL)
        { for (j = 0; j<m->y; j++)
          { if (MATRIX(m)[i][j]!=NULL) free(MATRIX(m)[i][j]);}
	  free(MATRIX(m)[i]);}}
      free(MATRIX(m));}
    free(m);}}

/* vector operations */

Real *addRVec(Real *u, const Real *v, int size)
{ for (size--; size>=0; size--) u[size] += v[size];
  return u;}

Real *copyRVec(Real *u, const Real *v, int size)
{ for (size--; size>=0; size--) u[size] = v[size];
  return u;}

Real dotProdRVec(const Real *u, const Real *v, int size)
{ Real x = 0.0;
  for (size--; size>=0; size--) x += u[size]*v[size];
  return x;}

/* normaliseRVec -
 * scales a vector so that the values in the vector lie in the range of
 * -1 to 1.  The scaled vector will only contain positive (or negative) values
 * if the original vector contained positive (or negative) values.  Returns
 * the maximum absolute value in the vector */

Real normaliseRVec(Real *v, int size)
{ int i;
  Real max = (size>0)?fabs(v[0]):0.0;
  for (i = 1; i<size; i++) if (fabs(v[i])>max) max = fabs(v[i]);
  if (max!=0.0) for (i = 0; i<size; i++) v[i] /= max;
  return max;}

Real rmat_get(RMat rmat, int i, int j)
{ if (i<0||i>=ROWS(rmat)) panic("i=%d out of bounds (0,%d)", i, ROWS(rmat));
  if (j<0||j>=COLUMNS(rmat))
  { panic("j=%d out of bounds (0,%d)", i, COLUMNS(rmat));}
  return MATRIX(rmat)[i][j];}

void rmat_set(RMat rmat, int i, int j, Real x)
{ if (i<0||i>=ROWS(rmat)) panic("i=%d out of bounds (0,%d)", i, ROWS(rmat));
  if (j<0||j>=COLUMNS(rmat))
  { panic("j=%d out of bounds (0,%d)", i, COLUMNS(rmat));}
  MATRIX(rmat)[i][j] = x;}

void rmat_vector_set(RMat *v, int i, RMat r) {v[i] = r;}

/* scaleRVec - v = k*v */
Real *scaleRVec(Real *v, Real k, int size)
{ for (size--; size>=0; size--) v[size] *= k;
  return v;}

/* sumOfLogs - x = log(exp(v[1]) + exp(v[2]) + ... + exp(v[size])) */

Real sumOfLogs(Real *v, int size)
{ Real x = NEGATIVE_INFINITY;
  for (size--; size>=0; size--) x = add_exp(x, v[size]);
  return x;}

Real sumRVec(Real *v, int size)
{ Real x = 0.0;
  for (size--; size>=0; size--) x += v[size];
  return x;}

/* debugging functions */

void set_trace(int n, int trace)
{ if (n<0||n>=TRACE_MAX_INDEX) panic("set_trace: Index out of range");
  gtrace[n] = trace;}

/* functions of general use */

/* needs work : maybe shouldn't check exact equality with NEGATIVE_INFINITY */
Real add_exp(Real e1, Real e2)
{ Real e_max = MAX(e1, e2), e_min = MIN(e1, e2);
  return (e_max==NEGATIVE_INFINITY)? NEGATIVE_INFINITY:
    (e_max-e_min>LOG_MATH_PRECISION)? e_max:
    my_log(1.0+my_exp(e_min-e_max))+e_max;}

void clip_symmatrix(Real *xx, Real *xy, Real *yy, Real clip)
{ Real exx, exy, eyx, eyy, v1, v2;
  v1 = MAX(eigenvalue1(*xx, *xy, *xy, *yy), clip);
  v2 = MAX(eigenvalue2(*xx, *xy, *xy, *yy), clip);
  eigenvectors(*xx, *xy, *xy, *yy, &exx, &exy, &eyx, &eyy);
  *xx = exx*exx*v1+eyx*eyx*v2;
  *xy = exy*exx*v1+eyy*eyx*v2;
  *yy = exy*exy*v1+eyy*eyy*v2;}

int distinct_reals(Real a, Real b)
{ return fabs(a-b)<MACH_EPS*(fabs(a)+fabs(b))?FALSE:TRUE;}

Real dot_product(Real *u, Real *v, int size)
{ int i;
  Real x = 0.0;
  for (i = 0; i<size; i++) x += u[i]*v[i];
  return x;}

Real eigenvalue1(Real xx, Real xy, Real yx, Real yy)
{ return quadratic1(1.0, -(yy+xx), xx*yy-xy*xy);}

Real eigenvalue2(Real xx, Real xy, Real yx, Real yy)
{ return quadratic2(1.0, -(yy+xx), xx*yy-xy*xy);}

Real eigenvector_angle1(Real xx, Real xy, Real yx, Real yy)
{ if (IS_ZERO(yx)&&IS_ZERO(xy))
  { if (yy>xx) return PI/2;
    else return 0.0;}
  else return my_atan2(yx, eigenvalue1(xx, xy, yx, yy)-yy);}

Real eigenvector_angle2(Real xx, Real xy, Real yx, Real yy)
{ if (IS_ZERO(yx)&&IS_ZERO(xy))
  { if (yy<=xx) return PI/2;
    else return 0.0;}
  else return my_atan2(yx, eigenvalue2(xx, xy, yx, yy)-yy);}

void eigenvectors(Real xx, Real xy, Real yx, Real yy, Real *exx, Real *exy,
		  Real *eyx, Real *eyy)
{ Real n;
  if (IS_ZERO(xy)&&IS_ZERO(yx))
  { *exx = yy>xx?0.0:1.0;
    *exy = yy>xx?1.0:0.0;
    *eyx = yy>xx?1.0:0.0;
    *eyy = yy>xx?0.0:1.0;}
  else
  { *exx = eigenvalue1(xx, xy, yx, yy)-yy;
    *eyx = eigenvalue2(xx, xy, yx, yy)-yy;
    n = sqrt(SQR(*exx)+yx*yx);
    *exx /= n;
    *exy = yx/n;
    n = sqrt(SQR(*eyx)+yx*yx);
    *eyx /= n;
    *eyy = yx/n;}}

void free_matrix(Real **m, int nrl, int nrh, int ncl, int nch)
{ int i;
  for(i = nrh; i>=nrl; i--) free((char*)(m[i]+ncl));
  free((char*)(m+nrl));}

void free_matrix2d(Real **m, int n)
{ int i;
  for (i = 0; i<n; i++) free(m[n]);
  free(m);}

void free_matrix3d(Real ***a, int x, int y)
{ int i, j;
  for (i = 0; i<x; i++)
  { for (j = 0; j<y; j++) free(a[i][j]);
    free(a[i]);}
  free(a);}

void free_matrix4d(Real ****a, int x, int y, int z)
{ int i;
  for (i = 0; i<x; i++) free_matrix3d(a[i], y, z);
  free(a);}

void free_vector(Real *v, int nl, int nh) {free((char*)(v+nl));}

int is_nan(Real x) {return !(x==x);}

Real *map_real(Real (*f)(Real), Real *v, Real *rv, int size)
{ int i;
  for (i = 0; i<size; i++) rv[i] = f(v[i]);
  return rv;}

int max_int(int *v, int size)
{ int max = v[size-1];
  for (size -= 2; size>=0; --size) if (v[size]>max) max = v[size];
  return max;}

#ifdef SUN
/* called when math exception is raised, see "man matherr" for details */
int matherr(struct exception *exc)
{ if (!is_nan(exc->retval))
  { if (gtrace[MATHERR])
    { trace(MATHERR, "matherr: type %d func %s arg1 %g arg2 %g retval %g\n",
	    exc->type, exc->name, exc->arg1, exc->arg2, exc->retval);}}
  else panic("matherr: type %d func %s arg1 %g arg2 %g retval %g\n", exc->type,
	     exc->name, exc->arg1, exc->arg2, exc->retval);
  /* the non-zero return values indicates that an exception shouldn't be
   * raised */
  return 1;}
#endif

Real **__matrix(int nrl, int nrh, int ncl, int nch, int line, const char *file)
{ int i;
  Real **m;
  m=(Real **) malloc((unsigned)(nrh-nrl+1)*sizeof(Real*));
  if (!m) panic("%s(%d): allocation failure 1 in matrix()", file, line);
  m -= nrl;
  for(i=nrl;i<=nrh;i++)
  { m[i]=(Real *) malloc((unsigned)(nch-ncl+1)*sizeof(Real));
    if (!m[i]) panic("%s(%d): allocation failure 2 in matrix()", file, line);
    m[i] -= ncl;}
  return m;}

Real **__matrix2d(int x, int y, int line, const char *file)
{ Real **m = (Real **)malloc(x*sizeof(Real *));
  int i;
  if (m==NULL) panic("matrix2d: out of memory, line = %d, file = %s",
		      line, file);
  for (i = 0; i<x; i++)
  { m[i] = (Real *)malloc(y*sizeof(Real));
    if (m[i]==NULL) panic("matrix2d: out of memory, line = %d, file = %s",
			  line, file);}
  return m;}

Real ***__matrix3d(int x, int y, int z, int line, const char *file)
{ int i, j;
  Real ***result;
  result = (Real ***)malloc(x*sizeof(Real **));
  if (!result)
  { panic("%s(%d) Matrix allocation failure %d*(Real **)", file, line, x);}
  for (i = 0; i<x; i++)
  { result[i] = (Real **)malloc(y*sizeof(Real *));
    if (!result[i])
    { panic("%s(%d) Matrix allocation failure %d*(Real *)", file, line, y);}
    for (j = 0; j<y; j++)
    { result[i][j] = (Real *)malloc(z*sizeof(Real));
      if (!result[i][j])
      { panic("%s(%d) Matrix allocation failure %d*(Real)", file, line, z);}}}
  return result;}

Real ****__matrix4d(int x, int y, int z, int w, int line, const char *file)
{ int i;
  Real ****result;
  result = (Real ****)malloc(x*sizeof(Real ***));
  if (!result)
  { panic("%s(%d) Matrix allocation failure %d*(Real ***)", file, line, x);}
  for (i = 0; i<x; i++)
  { result[i] = __matrix3d(y, z, w, line, file);}
  return result;}

Real my_add_exp(Real e1, Real e2, int line, const char *file)
{ Real e_max = MAX(e1, e2), e_min = MIN(e1, e2);
  Real factor = floor(e_min);
  Real rv = add_exp(e1, e2);
  if (is_nan(rv))
  { trace(MATHERR, "%s(%d) add_exp: Nan found e1=%g e2=%g e_max=%g e_min=%g\n",
	  file, line, e1, e2, e_max, e_min);
    panic("factor = %g, rv = %g\n", factor, rv);}
  return rv;}

Real my_sqrt(Real x)
{ x = (IS_ZERO(x)&&x<0.0)?0.0:x;
  return sqrt(x);}

Real __my_sqrt(Real x, int line, const char *file)
{ if (x<0.0) trace(MATHERR, "%s(%d) sqrt: arg1 = %g\n", file, line, x);
  x = (IS_ZERO(x)&&x<0.0)?0.0:x;
  return sqrt(x);}

Real my_exp(Real x)
{ if (x==NEGATIVE_INFINITY) return 0.0;
  return exp(x);}

void *my_malloc(int n, int line, const char *file)
{ static int total = 0;
  void *rv;
  if (n==0) panic("%s(%d) Call to malloc with 0 arg", file, line);
  total += n;
  trace(MALLOC, "%s(%d) Allocating %d bytes, %d total\n",
	file, line, n, total);
  rv =  malloc(n);
  if (rv==NULL) panic("%s(%d) Out of memory error", file, line);
  return rv;}

Real my_log(Real x)
{ if (x==0.0) return NEGATIVE_INFINITY;
  else return log(x);}

Real my_atan2(Real x, Real y)
{ if (x==0.0&&y==0.0) return 0.0;
  else if (x==0.0&&y==-1.0) return -PI;
  else return atan2(x, y);}

Real normalise_rotation(Real rotation)
{ if (rotation>PI) return normalise_rotation(rotation-2.0*PI);
  else if(rotation<=-PI) return normalise_rotation(rotation+2.0*PI);
  else return rotation;}

void panic(const char *error_text, ...)
{ va_list ap;
  char error_message[1024];
  va_start(ap, error_text);
  vsprintf(error_message, error_text, ap);
  fprintf(stderr, "panic: %s\n", error_message);
  exit(1);}

void print_version(const char *version, const char *led, const char *file)
{ trace(VERSION_PRINT, "%s: Version %s  Last Edit Date %s\n", file, version,
	led);}

Real quadratic1(Real a, Real b, Real c)
{ Real d = b*b-4*a*c;
  if (IS_ZERO(d)&&d<0) d = 0.0;
  return (-b+sqrt(d))/(2*a);}

Real quadratic2(Real a, Real b, Real c)
{ Real d = b*b-4*a*c;
  if (IS_ZERO(d)&&d<0) d = 0.0;
  return (-b-sqrt(d))/(2*a);}

Real radial_distance(Real t, Real u) {return normalise_rotation(u-t);}

void reportWarn(const char *error_text, ...)
{ va_list ap;
  char error_message[1024];
  va_start(ap, error_text);
  vsprintf(error_message, error_text, ap);
  fprintf(stderr, "WARNING: %s\n", error_message);
  return;}

void *_safe_malloc(size_t size, int line, char *file)
{ void *buffer;
#ifdef HLIBC_TRACE_SAFE_MALLOC
  fprintf(stderr, "%s(%d): allocating %d bytes\n", file, line, size);
#endif
  if (!(buffer = malloc(size)))
  { fprintf(stderr, "%s(%d): memory allocation failure. %ld bytes\n",
	    file, line, size);
   exit(1);}
  return buffer;}

void set_version_printing(int boolean) {gtrace[VERSION_PRINT] = boolean;}

Real sqr_real(Real x) {return x*x;}

Real sum_real(Real *v, int size)
{ int i;
  Real sum = 0.0;
  for (i = 0; i<size; i++) sum += v[i];
  return sum;}

void trace(int n, const char *error_text, ...)
{ va_list ap;
  va_start(ap, error_text);
  if (gtrace[n]) vfprintf(stderr, error_text, ap);
  va_end(ap);}

Real *__vector(int nl, int nh, int line, const char *file)
{ Real *v;
  v=(Real *)malloc((unsigned)(nh-nl+1)*sizeof(Real));
  if (!v) panic("%s(%d): allocation failure in vector()", file, line);
  return v-nl;}

/* PDFs */

Real log_normal(int n, Real *x, Real *mu, Real sqrt_big_sigma_determinant,
		Real **big_sigma_inverse)
{ Real r = 0.0;
  int i, j;
  for (i = 1; i<=n; i++)
  { for (j = 1; j<=n; j++)
    { r += big_sigma_inverse[i][j]*(x[i]-mu[i])*(x[j]-mu[j]);}}
  return -0.5*r-(my_log(pow(2.0*PI, n/2.0))+
		 my_log(sqrt_big_sigma_determinant));}

Real log_normal_1d(Real x, Real mu, Real sigma)
{ Real s = MAX(sigma, MINIMUM_SIGMA);
  if (sigma<MINIMUM_SIGMA)
  { trace(WARN,
	  "log_normal_1d: Sigma(%g) too small, replacing with minimum(%g)\n",
	  sigma,
	  MINIMUM_SIGMA);}
  return (-0.5*(x-mu)*(x-mu)/(s*s))-my_log(s*sqrt(2.0*PI));}

Real normal(int n, Real *x, Real *mu, Real sqrt_big_sigma_determinant,
	    Real **big_sigma_inverse)
{ Real r = 0.0;
  int i, j;
  for (i = 1; i<=n; i++)
  { for (j = 1; j<=n; j++)
    { r += big_sigma_inverse[i][j]*(x[i]-mu[i])*(x[j]-mu[j]);}}
  return my_exp(-0.5*r)/pow(2.0*PI, n/2.0)*sqrt_big_sigma_determinant;}

Real normal_1d(Real x, Real mu, Real sigma)
{ Real s = MAX(sigma, MINIMUM_SIGMA);
  if (sigma<MINIMUM_SIGMA)
  { trace(WARN,
	  "normal_1d: Sigma(%g) too small, replacing with minimum(%g)\n",
	  sigma,
	  MINIMUM_SIGMA);}
  return my_exp(-0.5*(x-mu)*(x-mu)/(s*s))/(s*sqrt(2.0*PI));}

struct xvimage *create_image(unsigned int height, unsigned int width)
{ return createimage(height, width, VFF_TYP_1_BYTE, 1, 1, "edges", 0, 0,
                     VFF_MS_NONE, VFF_MAPTYP_NONE, VFF_LOC_IMPLICIT, 0);}

void viff_set_pixel(struct xvimage *viff,
		    unsigned int y, unsigned int x, unsigned int grey)
{ unsigned char *array;
  array = (unsigned char *)viff->imagedata;
  array[PIXEL(x, y, (int)viff->col_size, (int)viff->row_size)] = grey;}

void detect_edges(float a1, float a2, int t1, int t2, int l, int w,
		  struct xvimage *viff)
{ unsigned int x, y;
  unsigned char *array;
  lvdrf(viff, a1, a2, t1, t2, l, w);
  array = (unsigned char *)viff->imagedata;
  for (y = 0; y<(int)viff->col_size; y++)
  { for (x = 0; x<(int)viff->row_size; x++)
    { array[PIXEL(x, y, (int)viff->col_size, (int)viff->row_size)] =
      array[PIXEL(x, y, (int)viff->col_size, (int)viff->row_size)]
      <128?0:255;}}}

void detect_line_segments(float a1, float a2, int t1, int t2, int l, int w,
			  struct xvimage *viff)
{ unsigned int x, y;
  unsigned char *array;
  lvdrf(viff, a1, a2, t1, t2, l, w);
  array = (unsigned char *)viff->imagedata;
  for (y = 0; y<(int)viff->col_size; y++)
  { for (x = 0; x<(int)viff->row_size; x++)
    { array[PIXEL(x, y, (int)viff->col_size, (int)viff->row_size)] =
      array[PIXEL(x, y, (int)viff->col_size, (int)viff->row_size)]
      <128?0:255;}}
  lvpolygon(viff, &vect, 0.6, 0.1, 10);}

int edge_pixel(struct xvimage *viff, int y, int x)
{ unsigned char *array;
  array = (unsigned char *)viff->imagedata;
  return array[PIXEL(x, y, (int)viff->col_size, (int)viff->row_size)];}

int number_of_line_segments(void) {return vect->col_size;}

int line_segment_x1(int i)
{ int *array;
  array = (int *)vect->imagedata;
  return array[BPIXEL(0, 0, i, (int)vect->col_size, (int)vect->row_size)];}

int line_segment_y1(int i)
{ int *array;
  array = (int *)vect->imagedata;
  return array[BPIXEL(1, 0, i, (int)vect->col_size, (int)vect->row_size)];}

int line_segment_x2(int i)
{ int *array;
  array = (int *)vect->imagedata;
  return array[BPIXEL(2, 0, i, (int)vect->col_size, (int)vect->row_size)];}

int line_segment_y2(int i)
{ int *array;
  array = (int *)vect->imagedata;
  return array[BPIXEL(3, 0, i, (int)vect->col_size, (int)vect->row_size)];}

void free_xvimage(struct xvimage *viff)
{ if (vect!=NULL)
  { free(vect->imagedata);
    free(vect);
    vect = NULL;}
  free(viff->imagedata);
  free(viff);}

int *make_c_int_vector(int l)
{ int *v;
  v = (int *)malloc(l*sizeof(int));
  if (v==NULL) panic("make_c_int_vector: out of memory.  size = %d", l);
  return v;}

int c_int_vector_ref(int *v, int x) {return v[x];}

void c_int_vector_set(int *v, int x, int i) {v[x] = i;}

void free_c_int_vector(int *v) {free(v);}

int **make_c_int_matrix(int h, int w)
{ int **m, i;
  m = (int **)malloc(h*sizeof(int *));
  if (m==NULL) panic("make_c_int_matrix: out of memory.  size = %d %d", h, w);
  for (i = 0; i<h; i++)
  { m[i] = (int *)malloc(w*sizeof(int));
    if (m[i]==NULL)
    { panic("make_c_int_matrix: out of memory.  size = %d %d", h, w);}}
  return m;}

int c_int_matrix_ref(int **m, int y, int x) {return m[y][x];}

void c_int_matrix_set(int **m, int y, int x, int i) {m[y][x] = i;}

void free_c_int_matrix(int **m, int h)
{ int i;
  for (i = 0; i<h; i++) free(m[i]);
  free(m);}

float *make_c_float_vector(int l)
{ float *v;
  v = (float *)malloc(l*sizeof(float));
  if (v==NULL) panic("make_c_float_vector: out of memory.  size = %d", l);
  return v;}

float c_float_vector_ref(float *v, int x) {return v[x];}

void c_float_vector_set(float *v, int x, float i) {v[x] = i;}

void free_c_float_vector(float *v) {free(v);}

float **make_c_float_matrix(int h, int w)
{ float **m;
  int i;
  m = (float **)malloc(h*sizeof(float *));
  if (m==NULL)
  { panic("make_c_float_matrix: out of memory.  size = %d %d", h, w);}
  for (i = 0; i<h; i++)
  { m[i] = (float *)malloc(w*sizeof(float));
    if (m[i]==NULL)
    { panic("make_c_float_matrix: out of memory.  size = %d %d", h, w);}}
  return m;}

float c_float_matrix_ref(float **m, int y, int x) {return m[y][x];}

void c_float_matrix_set(float **m, int y, int x, float i) {m[y][x] = i;}

void free_c_float_matrix(float **m, int h)
{ int i;
  for (i = 0; i<h; i++) free(m[i]);
  free(m);}

Real *make_c_real_vector(int l)
{ Real *v;
  v = (Real *)malloc(l*sizeof(Real));
  if (v==NULL) panic("make_c_real_vector: out of memory.  size = %d", l);
  return v;}

Real c_real_vector_ref(Real *v, int x) {return v[x];}

void c_real_vector_set(Real *v, int x, Real i) {v[x] = i;}

void free_c_real_vector(Real *v) {free(v);}

Real **make_c_real_matrix(int h, int w)
{ Real **m;
  int i;
  m = (Real **)malloc(h*sizeof(Real *));
  if (m==NULL) panic("make_c_real_matrix: out of memory.  size = %d %d", h, w);
  for (i = 0; i<h; i++)
  { m[i] = (Real *)malloc(w*sizeof(Real));
    if (m[i]==NULL)
    { panic("make_c_real_matrix: out of memory.  size = %d %d", h, w);}}
  return m;}

Real c_real_matrix_ref(Real **m, int y, int x) {return m[y][x];}

void c_real_matrix_set(Real **m, int y, int x, Real i) {m[y][x] = i;}

void free_c_real_matrix(Real **m, int h)
{ int i;
  for (i = 0; i<h; i++) free(m[i]);
  free(m);}

void free_c_vector(void *v) {free(v);}

void free_c_matrix(void **m, int h)
{ int i;
  for (i = 0; i<h; i++) free(m[i]);
  free(m);}

int separation(int t, int u)
{ Real max = MAX(t, u), min = MIN(t, u);
  if (max-min<180) return abs(max-min);
  else return abs(min-max+360);}

void setup_proximity_clusterer(int max)
{ int i, j;
  gpythag = (int **)malloc((max+1)*sizeof(int *));
  if (gpythag==NULL) panic("setup_proximity_clusterer: out of memory error");
  for (i = 0; i<=max; i++)
  { gpythag[i] = (int *)malloc((max+1)*sizeof(int));
    if (gpythag[i]==NULL)
    { panic("setup_proximity_clusterer: out of memory error");}}
  for (i = 0; i<=max; i++)
  { for (j = 0; j<=i; j++)
    { gpythag[i][j] = (int)rint(sqrt(i*i-j*j));}}}

int pbm_proximity_clusterer(int **m, int h, int w, int p,
			    int category, int parity, int start)
{ int i = start, x, y, x1, y1, x2, y2, q, again;
  for (y = 0; y<h; y++)
  { for (x = 0; x<w; x++)
    { if (m[y][x]==category)
      { m[y][x] = i;
	again = TRUE;
        while (again)
	{ again = FALSE;
          for (y1 = 0; y1<h; y1++)
	  { for (x1 = 0; x1<w; x1++)
	    { if (m[y1][x1]==i)
	      { for (y2 = MAX(0, y1-p); y2<MIN(h, y1+p+1); y2++)
		{ q = gpythag[p][abs(y2-y1)];
		  for (x2 = MAX(0, x1-q); x2<MIN(w, x1+q+1); x2++)
		  { if (m[y2][x2]==category)
		    { m[y2][x2] = i;
		      again = TRUE;}}}}}}}
	i += parity;}}}
  return i;}

void threshold_normal_flow_magnitude(int width, int height,
				     const Real **e1,
				     const Real **e2,
				     int **t, Real epsilon,
				     int sigma, Real threshold)
{ int i, j, x, y;
  Real ex, ey, et, l, **e1s, **e2s;
  e1s = (Real **)malloc(height*sizeof(Real *));
  if (!e1s) panic("allocation failure 1 in threshold_normal_flow_magnitude");
  for (i = 0; i<height; i++)
  { e1s[i] = (Real *)malloc(width*sizeof(Real));
    if (!e1s[i])
    { panic("allocation failure 2 in threshold_normal_flow_magnitude");}}
  e2s = (Real **)malloc(height*sizeof(Real *));
  if (!e2s) panic("allocation failure 3 in threshold_normal_flow_magnitude");
  for (i = 0; i<height; i++)
  { e2s[i] = (Real *)malloc(width*sizeof(Real));
    if (!e2s[i])
    { panic("allocation failure 4 in threshold_normal_flow_magnitude");}}
  epsilon *= epsilon;
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { e1s[y][x] = 0.0;
      e2s[y][x] = 0.0;}}
  for (y = sigma; y<height-sigma; y++)
  { for (x = sigma; x<width-sigma; x++)
    { for (i = y-sigma; i<=y+sigma; i++)
      { for (j = x-sigma; j<=x+sigma; j++)
	{ e1s[y][x] += e1[i][j];
	  e2s[y][x] += e2[i][j];}}
      /* This differs slightly from the Scheme code here in that e1s and e2s
         are not truncated. */
      e1s[y][x] /= (2*sigma+1)*(2*sigma+1);
      e2s[y][x] /= (2*sigma+1)*(2*sigma+1);}}
  for (i = 0; i<height-1; i++)
  { for (j = 0; j<width-1; j++)
    { ex = (e1s[i+1][j]+e1s[i+1][j+1]+e2s[i+1][j]+e2s[i+1][j+1]
            -e1s[i][j]-e1s[i][j+1]-e2s[i][j]-e2s[i][j+1])/4.0;
      ey = (e1s[i][j+1]+e1s[i+1][j+1]+e2s[i][j+1]+e2s[i+1][j+1]
            -e1s[i][j]-e1s[i+1][j]-e2s[i][j]-e2s[i+1][j])/4.0;
      et = (e2s[i][j]+e2s[i][j+1]+e2s[i+1][j]+e2s[i+1][j+1]
            -e1s[i][j]-e1s[i][j+1]-e1s[i+1][j]-e1s[i+1][j+1])/4.0;
      l = ex*ex+ey*ey;
      t[i][j] = (l>=epsilon)&&(fabs(et/sqrt(l))>=threshold);}}
  for (i = 0; i<height; i++) t[i][width-1] = FALSE;
  for (j = 0; j<width; j++) t[height-1][j] = FALSE;
  for (i = 0; i<height; i++)
  { free(e1s[i]);
    free(e2s[i]);}
  free(e1s);
  free(e2s);}

/* hard-wired to use hue values between 0 and 359 */

int histogram_clusterer(const int *pixel_hues, int size, int quanta,
			int area_e, int hue_separation_e, int *bins,
			int *hues)
{ int i, j, k, result, max_hues = MAX_HUE_VALUE/quanta+1;
  int *count, *map;
  struct pair *mean_hue;
  count = make_c_int_vector(MAX_HUE_VALUE/quanta+1);
  map = make_c_int_vector(MAX_HUE_VALUE/quanta+1);
  mean_hue = (struct pair *) malloc(max_hues*sizeof(struct pair));
  if (mean_hue==NULL) panic("mean_hue out of memory.  size = %d", max_hues);
  /* zero out the counter */
  for (i = 0; i<max_hues; i++) count[i] = 0;
  /* sort pixels into max_hue bins */
  for (i = 0; i<size; i++)
  { if (bins[i])
    { k = pixel_hues[i]/quanta;
      bins[i] = k;
      count[k]++;}
    else bins[i] = -1;}
  /* throw out bins with an insufficient number of pixels */
  /* and initialise mean_hue */
  for (i = 0; i<max_hues; i++)
  { map[i] = count[i]<area_e?-1:i;
    mean_hue[i].sin = mean_hue[i].cos = 0.0;}
  /* sum the sin, cos of remaining pixels and determine the mean hues */
  for (i = 0; i<size; i++)
  { if (bins[i]>=0&&(k = map[bins[i]])>=0)
    { mean_hue[k].sin += sin(DEG_TO_RAD(pixel_hues[i]));
      mean_hue[k].cos += cos(DEG_TO_RAD(pixel_hues[i]));}}
  for (j = 0; j<max_hues; j++)
  { if (map[j]>=0)
    { hues[j] = (int)rint(RAD_TO_DEG(my_atan2(mean_hue[j].sin,
                                              mean_hue[j].cos)));
      if (hues[j]<0) hues[j] += 360;}}
  /* collect clustered mean hues into bins */
  /* this could easily be translated into Scheme to replace
   * EQUIVALENCE-CLASSESP */
  for (i = 0; i<max_hues; i++)
  { if (map[i]>=0)
    { for (j = i+1; j<max_hues; j++)
      { if (map[j]>=0&&map[j]!=map[i]&&
	    (separation(hues[j], hues[i])<hue_separation_e))
	{ if (map[j]>map[i]) map[j] = map[i];
	  else
	  { for (k = 0; k<max_hues; k++)
	    { if (map[k]==map[i]) map[k] = map[j];}}}}}}
  /* sum sin's and cos's of hue bins and determine new mean hues */
  /* note: mean_hue[x].sin contains hues[x].sin*count[x] because of
   * the way mean_hue[x].sin was calculated, same with cos */
  for (i = 0; i<max_hues; i++)
  { k = map[i];
    if (k>=0&&k!=i)
    { mean_hue[k].sin += mean_hue[i].sin;
      mean_hue[k].cos += mean_hue[i].cos;}}
  /* note: the first occurences of new bins in map are guaranteed to be in
   * increasing order.  Trust me :) */
  for (i = 0, result = 0, k = -1; i<max_hues; i++)
  { j = map[i];
    if (j>=0&&j>k)
    { hues[result] = (int)rint(RAD_TO_DEG(my_atan2(mean_hue[j].sin,
						   mean_hue[j].cos)));
      if (hues[result]<0) hues[result] += 360;
      result++;
      k = j;}}
  free_c_int_vector(count);
  free_c_int_vector(map);
  free_c_vector(mean_hue);
  return result;}

int get_first_frame(void) {return gfirst_frame;}

int get_last_frame(void) {return glast_frame;}

int get_current_frame(void) {return meteor_i-1;}

int get_motion(int frame) {return gmotion[frame];}

int get_object_k(int frame) {return gobject_k[frame];}

int get_object_colour(int frame, int k) {return gobject_colour[frame][k];}

Real get_ellipse_x(int frame, int k) {return gellipse_x[frame][k];}

Real get_ellipse_y(int frame, int k) {return gellipse_y[frame][k];}

Real get_ellipse_theta(int frame, int k) {return gellipse_theta[frame][k];}

Real get_ellipse_r(int frame, int k) {return gellipse_r[frame][k];}

Real get_ellipse_s(int frame, int k) {return gellipse_s[frame][k];}

unsigned get_polygon_l(int frame, int k) {return gpolygon_l[frame][k];}

unsigned get_polygon_x(int frame, int k, int l)
{ return gpolygon_x[frame][k][l];}

unsigned get_polygon_y(int frame, int k, int l)
{ return gpolygon_y[frame][k][l];}

int chmt(char *pathname,
	 char *track_pathname,
	 enum VideoType video_type,
	 Real background_threshold,
	 int quanta,
	 int colour_pixel_mass_threshold,
	 int hue_separation_threshold,
	 int colour_proximity_threshold,
	 int motion_threshold,
	 int motion_pixel_mass_threshold,
	 int motion_proximity_threshold,
	 int track_movie,
	 int left,
	 int n,
	 Real *background_c,
	 Real *background_d)
{ int i, j, k, x, y, x1, y1, x2, y2, xl, yl, xh, yh, xp, yp, height, width;
  int q, frame = 0, count, size, result, last_moving_valid = FALSE;
  int max_hues = MAX_HUE_VALUE/quanta+1, no_moving_clusters, once;
  int *colour_pythag;
  int *motion_pythag;
  int new1, new2, new3;
  int **red, **green, **blue, **hue, **coloured;
  int **last_red, **last_green, **last_blue, **moving, **last_moving;
  int **coloured_cluster, **moving_cluster;
  int *pixel_hues, *bins, *hues;
  int *counter, *map;
  int **back_x, **back_y;
  /* These don't really need to be double precision */
  float r, g, b, c, d, saturation, value, m, rc, gc, bc, h;
  float mu_x, mu_y, sigma_xx, sigma_xy, sigma_yy;
  float fract;
  float float_background_threshold;
  float *float_background_c, *float_background_d;
  struct pair *mean_hue;
  VidStream *video_input_port;
  FILE *video_output_port;
  int tt; /* debugging */
  colour_pythag = make_c_int_vector(colour_proximity_threshold+1);
  motion_pythag = make_c_int_vector(motion_proximity_threshold+1);
  float_background_c = make_c_float_vector(n);
  float_background_d = make_c_float_vector(n);
  mean_hue = (struct pair *) malloc(max_hues*sizeof(struct pair));
  if (mean_hue==NULL) panic("mean_hue out of memory.  size = %d", max_hues);
  gfirst_frame = 0;
  /* Set up thresholds */
  float_background_threshold = background_threshold*background_threshold;
  /* Set up background */
  for (i = 0; i<n; i++)
  { float_background_c[i] = background_c[i];
    float_background_d[i] = background_d[i];}
  /* Set up proximity clusterer */
  for (i = 0; i<=colour_proximity_threshold; i++)
  { colour_pythag[i] = rint(sqrt(SQR(colour_proximity_threshold)-i*i));}
  for (i = 0; i<=motion_proximity_threshold; i++)
  { motion_pythag[i] = rint(sqrt(SQR(motion_proximity_threshold)-i*i));}
  /* Set up video ports */
  video_input_port = open_video_input_file(pathname, video_type);
  if (!video_input_port) panic("Can't open video input file");
  /* needs work: This won't work because you need to do read_video first */
  height = get_video_height(video_input_port);
  width = get_video_width(video_input_port);
  red = make_c_int_matrix(height, width);
  green = make_c_int_matrix(height, width);
  blue = make_c_int_matrix(height, width);
  hue = make_c_int_matrix(height, width);
  coloured = make_c_int_matrix(height, width);
  last_red = make_c_int_matrix(height, width);
  last_green = make_c_int_matrix(height, width);
  last_blue = make_c_int_matrix(height, width);
  moving = make_c_int_matrix(height, width);
  last_moving = make_c_int_matrix(height, width);
  coloured_cluster = make_c_int_matrix(height, width);
  moving_cluster = make_c_int_matrix(height, width);
  pixel_hues = make_c_int_vector(height*width);
  bins = make_c_int_vector(height*width);
  hues = make_c_int_vector(quanta*max_hues+1);
  counter = make_c_int_vector(max_hues);
  map = make_c_int_vector(max_hues);
  back_x = make_c_int_matrix(height, width);
  back_y = make_c_int_matrix(height, width);
  if (track_movie)
  { video_output_port =
    open_video_output_file(track_pathname, video_type, MAX_FRAMES);
    if (!video_output_port) panic("Can't open video output file");}
  while (TRUE)
  { tt = clock(); printf("Getting frame %d", frame); /* debugging */
    if (!read_video(video_input_port)) break;
    /* compute hue, chromaticity, and coloured */
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { red[y][x] = get_video_red(video_input_port, y, x);
	green[y][x] = get_video_green(video_input_port, y, x);
	blue[y][x] = get_video_blue(video_input_port, y, x);
	r = red[y][x]/255.0;
	g = green[y][x]/255.0;
	b = blue[y][x]/255.0;
	c = r/(r+g+b);
	d = g/(r+g+b);
	value = MAX(r, MAX(g, b));
	m = MIN(r, MIN(g, b));
	saturation = value==0.0?0.0:(value-m)/value;
	if (saturation==0.0) h = 0.0;
	else
	{ rc = (value-r)/(value-m);
	  gc = (value-g)/(value-m);
	  bc = (value-b)/(value-m);
	  h = (r==value?bc-gc:g==value?2.0+rc-bc:4.0+gc-rc)/6.0;
	  if (h<0.0) h += 1.0;}
	hue[y][x] = (MAX_HUE_VALUE+1)*h;
	if (left&&x<16) coloured[y][x] = FALSE;
	else
	{ coloured[y][x] = TRUE;
	  for (i = 0; i<n; i++)
	  { if ((c-float_background_c[i])*(c-float_background_c[i])+
		(d-float_background_d[i])*(d-float_background_d[i])<
		float_background_threshold)
	    { coloured[y][x] = FALSE;
	      break;}}}}}
    printf(" - %d\n", (clock()-tt)/1000); tt = clock(); /* debugging */
    printf("Histogram clusterer"); /* debugging */
    /* hard-wired to use hue values between 0 and 359 */
    /* Initialize pixel_hues */
    size = 0;
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { if (coloured[y][x]) pixel_hues[size++] = hue[y][x];}}
    /* Initialize bins */
    for (i = 0; i<height*width; i++) bins[i] = 1;
    /* zero out the counter */
    for (i = 0; i<max_hues; i++) counter[i] = 0;
    /* sort pixels into max_hue bins */
    for (i = 0; i<size; i++)
    { if (bins[i])
      { k = pixel_hues[i]/quanta;
	bins[i] = k;
	counter[k]++;}
    else bins[i] = -1;}
    /* throw out bins with an insufficient number of pixels
       and initialise mean_hue */
    for (i = 0; i<max_hues; i++)
    { map[i] = counter[i]<colour_pixel_mass_threshold?-1:i;
      mean_hue[i].sin = mean_hue[i].cos = 0.0;}
    /* sum the sin, cos of remaining pixels and determine the mean hues */
    for (i = 0; i<size; i++)
    { if (bins[i]>=0&&(k = map[bins[i]])>=0)
      { mean_hue[k].sin += sin(DEG_TO_RAD(pixel_hues[i]));
	mean_hue[k].cos += cos(DEG_TO_RAD(pixel_hues[i]));}}
    for (j = 0; j<max_hues; j++)
    { if (map[j]>=0)
      { hues[j] = (int)rint(RAD_TO_DEG(my_atan2(mean_hue[j].sin,
						mean_hue[j].cos)));
	if (hues[j]<0) hues[j] += 360;}}
    /* collect clustered mean hues into bins
       this could easily be translated into Scheme to replace
       EQUIVALENCE-CLASSESP */
    for (i = 0; i<max_hues; i++)
    { if (map[i]>=0)
      { for (j = i+1; j<max_hues; j++)
	{ if (map[j]>=0&&map[j]!=map[i]&&
	      (separation(hues[j], hues[i])<hue_separation_threshold))
	  { if (map[j]>map[i]) map[j] = map[i];
	  else
	  { for (k = 0; k<max_hues; k++)
	    { if (map[k]==map[i]) map[k] = map[j];}}}}}}
    /* sum sin's and cos's of hue bins and determine new mean hues
       note: mean_hue[x].sin contains hues[x].sin*counter[x] because of
       the way mean_hue[x].sin was calculated, same with cos */
    for (i = 0; i<max_hues; i++)
    { k = map[i];
      if (k>=0&&k!=i)
      { mean_hue[k].sin += mean_hue[i].sin;
	mean_hue[k].cos += mean_hue[i].cos;}}
    /* note: the first occurences of new bins in map are guaranteed to be
       in increasing order.  Trust me :) */
    for (i = 0, result = 0, k = -1; i<max_hues; i++)
    { j = map[i];
      if (j>=0&&j>k)
      { hues[result] = (int)rint(RAD_TO_DEG(my_atan2(mean_hue[j].sin,
						     mean_hue[j].cos)));
	if (hues[result]<0) hues[result] += 360;
	result++;
	k = j;}}
    /* Classify hue[][] according to hues[].
       This differs from the Scheme version in that a level of
       colour_pixel_mass_threshold filtering is missing here */
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { if (coloured[y][x])
	{ if (result==0) coloured[y][x] = FALSE;
	else
	{ j = hues[0];
	  for (i = 1; i<result; i++)
	  { if (separation(hue[y][x], hues[i])<separation(hue[y][x], j))
	    { j = hues[i];}}
	  hue[y][x] = j;}}}}
    printf(" - %d\n", (clock()-tt)/1000); tt = clock(); /* debugging */
    printf("Motion detection"); /* debugging */
    if (frame==0)
    { for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++)
	{ moving[y][x] = FALSE;
	  last_red[y][x] = red[y][x];
	  last_green[y][x] = green[y][x];
	  last_blue[y][x] = blue[y][x];}}}
    else
    { for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++)
	{ moving[y][x] =
	  /* note: These are integer multiplies */
	  (last_red[y][x]-red[y][x])*(last_red[y][x]-red[y][x])+
	  (last_green[y][x]-green[y][x])*(last_green[y][x]-green[y][x])+
	  (last_blue[y][x]-blue[y][x])*(last_blue[y][x]-blue[y][x])>
	  motion_threshold;
	  last_red[y][x] = red[y][x];
	  last_green[y][x] = green[y][x];
	  last_blue[y][x] = blue[y][x];}}}
    printf(" - %d\n", (clock()-tt)/1000); tt = clock(); /* debugging */
    printf("Colour proximity clusterer"); /* debugging */
    /* Cluster coloured objects by proximity, ignoring small clusters */
    k = 0;
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { coloured_cluster[y][x] = -1;}}
    for (i = 0; i<result; i++)
    { for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++)
	{ if (coloured[y][x]&&hue[y][x]==hues[i]&&coloured_cluster[y][x]==-1)
	  { coloured_cluster[y][x] = k;
	    count = 1;
	    back_y[y][x] = -1;
	    back_x[y][x] = -1;
	    yp = y;
	    xp = x;
	    yl = y;
	    xl = x;
	    yh = y;
	    xh = x;
	    while (yp!=-1)
	    { y1 = yp;
	      x1 = xp;
	      yp = back_y[y1][x1];
	      xp = back_x[y1][x1];
	      if (coloured_cluster[y1][x1]==k)
	      { for (y2 = MAX(0, y1-colour_proximity_threshold);
		     y2<MIN(height, y1+colour_proximity_threshold+1);
		     y2++)
		{ q = colour_pythag[abs(y2-y1)];
		  for (x2 = MAX(0, x1-q); x2<MIN(width, x1+q+1); x2++)
		  { if (coloured[y2][x2]&&
			hue[y2][x2]==hues[i]&&
			coloured_cluster[y2][x2]==-1)
		    { coloured_cluster[y2][x2] = k;
		      count++;
		      back_y[y2][x2] = yp;
		      back_x[y2][x2] = xp;
		      yp = y2;
		      xp = x2;
		      yl = MIN(yl, y2);
		      xl = MIN(xl, x2);
		      yh = MAX(yh, y2);
		      xh = MAX(xh, x2);}}}}}
	    if (count>=colour_pixel_mass_threshold) k++;
	    else
	    { for (y1 = yl; y1<=yh; y1++)
	      { for (x1 = xl; x1<=xh; x1++)
		if (coloured_cluster[y1][x1]==k)
		{ coloured_cluster[y1][x1] = -2;
		  coloured[y1][x1] = FALSE;}}}}}}}
    printf(" - %d\n", (clock()-tt)/1000); tt = clock(); /* debugging */
    printf("Motion proximity clusterer"); /* debugging */
    /* Cluster moving objects by proximity, ignoring small clusters */
    once = FALSE;
    again:
    no_moving_clusters = TRUE;
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++) moving_cluster[y][x] = -1;}
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { if (moving[y][x]&&moving_cluster[y][x]==-1)
	{ moving_cluster[y][x] = k;
	  back_y[y][x] = -1;
	  back_x[y][x] = -1;
	  yp = y;
	  xp = x;
	  yl = y;
	  xl = x;
	  yh = y;
	  xh = x;
	  count = 1;
	  while (yp!=-1)
	  { y1 = yp;
	    x1 = xp;
	    yp = back_y[y1][x1];
	    xp = back_x[y1][x1];
	    if (moving_cluster[y1][x1]==k)
	    { for (y2 = MAX(0, y1-motion_proximity_threshold);
		   y2<MIN(height, y1+motion_proximity_threshold+1);
		   y2++)
	      { q = motion_pythag[abs(y2-y1)];
		for (x2 = MAX(0, x1-q); x2<MIN(width, x1+q+1); x2++)
		{ if (moving[y2][x2]&&moving_cluster[y2][x2]==-1)
		  { moving_cluster[y2][x2] = k;
		    count++;
		    back_y[y2][x2] = yp;
		    back_x[y2][x2] = xp;
		    yp = y2;
		    xp = x2;
		    yl = MIN(yl, y2);
		    xl = MIN(xl, x2);
		    yh = MAX(yh, y2);
		    xh = MAX(xh, x2);}}}}}
	  if (count>=motion_pixel_mass_threshold)
	  { k++;
	    no_moving_clusters = FALSE;}
	  else
	  { for (y1 = yl; y1<=yh; y1++)
	    { for (x1 = xl; x1<=xh; x1++)
	      if (moving_cluster[y1][x1]==k)
	      { moving_cluster[y1][x1] = -2;
		moving[y1][x1] = FALSE;}}}}}}
    if (no_moving_clusters)
    { if (last_moving_valid)
      { if (!once)
	{ for (y = 0; y<height; y++)
	  { for (x = 0; x<width; x++) moving[y][x] = last_moving[y][x];}
	  printf(" conjuring"); /* debugging */
	  once = TRUE;
	  goto again;}
        else printf(" failed");}	/* debugging */
      else printf(" can't conjure");} /* debugging */
    else if (!once)
    { for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++) last_moving[y][x] = moving[y][x];}
      if (!last_moving_valid)
      { last_moving_valid = TRUE;
	gfirst_frame = frame;}
      glast_frame = frame;}
    printf(" - %d\n", (clock()-tt)/1000); tt = clock(); /* debugging */
    printf("Computing ellipses"); /* debugging */
    /* debugging */
    /*
       char fp[100]; FILE *f;
       sprintf(fp, "foo-%d.ppm", frame);
       f = fopen(fp, "w");
       fprintf(f, "P6\n320 240\n255\n");
       for (y = 0; y<height; y++)
       { for (x = 0; x<width; x++)
       { h = hue[y][x]/360.0;
       h *= 6.0;
       fract = h-floor(h);
       new1 = 0.0;
       new2 = 255.0*(1.0-fract);
       new3 = 255.0*fract;
       if ((int)h==0) {r = 255.0; g = new3; b = new1;}
       else if ((int)h==1) {r = new2; g = 255.0; b = new1;}
       else if ((int)h==2) {r = new1; g = 255.0; b = new3;}
       else if ((int)h==3) {r = new1; g = new2; b = 255.0;}
       else if ((int)h==4) {r = new3; g = new1; b = 255.0;}
       else if ((int)h==5) {r = 255.0; g = new1; b = new2;}
       else if ((int)h==6) {r = 255.0; g = new3; b = new1;}
       else panic("fuck up");
       putc(moving[y][x]?0:coloured[y][x]?(int)r:255, f);
       putc(moving[y][x]?0:coloured[y][x]?(int)g:255, f);
       putc(moving[y][x]?0:coloured[y][x]?(int)b:255, f);}}
       fclose(f);
       */
    if (k>MAX_OBJECTS) panic("Too many objects");
    /* Fit ellipses to clusters */
    gobject_k[frame] = k;
    for (i = 0; i<k; i++)
    { mu_x = 0.0;
      mu_y = 0.0;
      sigma_xx = 0.0;
      sigma_xy = 0.0;
      sigma_yy = 0.0;
      j = 0;
      for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++)
	if (coloured_cluster[y][x]==i||moving_cluster[y][x]==i)
	{ j++;
	  mu_x += x;
	  mu_y += y;}}
      mu_x /= j;
      mu_y /= j;
      for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++)
	if (coloured_cluster[y][x]==i||moving_cluster[y][x]==i)
	{ sigma_xx += (x-mu_x)*(x-mu_x);
	  sigma_xy += (x-mu_x)*(y-mu_y);
	  sigma_yy += (y-mu_y)*(y-mu_y);}}
      sigma_xx /= j;
      sigma_xy /= j;
      sigma_yy /= j;
      gellipse_x[frame][i] = mu_x;
      gellipse_y[frame][i] = mu_y;
      gellipse_theta[frame][i] = eigenvector_angle1(sigma_xx,
						    sigma_xy,
						    sigma_xy,
						    sigma_yy);
      gellipse_r[frame][i] = sqrt(2.0*eigenvalue1(sigma_xx,
						  sigma_xy,
						  sigma_xy,
						  sigma_yy));
      gellipse_s[frame][i] = sqrt(2.0*eigenvalue2(sigma_xx,
						  sigma_xy,
						  sigma_xy,
						  sigma_yy));}
    printf(" - %d\n", (clock()-tt)/1000); tt = clock(); /* debugging */
    if (track_movie)
    { begin_video_frame(width, height, 255);
      for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++)
	{ h = hue[y][x]/360.0;
	  h *= 6.0;
	  fract = h-floor(h);
	  new1 = 0.0;
	  new2 = 255.0*(1.0-fract);
	  new3 = 255.0*fract;
	  if ((int)h==0) {r = 255.0; g = new3; b = new1;}
	  else if ((int)h==1) {r = new2; g = 255.0; b = new1;}
	  else if ((int)h==2) {r = new1; g = 255.0; b = new3;}
	  else if ((int)h==3) {r = new1; g = new2; b = 255.0;}
	  else if ((int)h==4) {r = new3; g = new1; b = 255.0;}
	  else if ((int)h==5) {r = 255.0; g = new1; b = new2;}
	  else if ((int)h==6) {r = 255.0; g = new3; b = new1;}
	  else panic("fuck up");
	  if (TRUE)
	  { put_video_red(y, x, moving[y][x]?0:coloured[y][x]?r:255);
	    put_video_green(y, x, moving[y][x]?0:coloured[y][x]?g:255);
	    put_video_blue(y, x, moving[y][x]?0:coloured[y][x]?b:255);}
	  else
	  { put_video_red(y, x, coloured[y][x]?r:moving[y][x]?0:255);
	    put_video_green(y, x, coloured[y][x]?g:moving[y][x]?0:255);
	    put_video_blue(y, x, coloured[y][x]?b:moving[y][x]?0:255);}}}
      end_video_frame();}
    frame++;
    if (frame>=MAX_FRAMES) panic("Too many frames");}
  if (track_movie) close_video_output_port(video_output_port);
  close_video_input_port(video_input_port);
  free_c_int_vector(colour_pythag);
  free_c_int_vector(motion_pythag);
  free_c_float_vector(float_background_c);
  free_c_float_vector(float_background_d);
  free_c_vector(mean_hue);
  free_c_int_matrix(red, height);
  free_c_int_matrix(green, height);
  free_c_int_matrix(blue, height);
  free_c_int_matrix(hue, height);
  free_c_int_matrix(coloured, height);
  free_c_int_matrix(last_red, height);
  free_c_int_matrix(last_green, height);
  free_c_int_matrix(last_blue, height);
  free_c_int_matrix(moving, height);
  free_c_int_matrix(last_moving, height);
  free_c_int_matrix(coloured_cluster, height);
  free_c_int_matrix(moving_cluster, height);
  free_c_int_vector(pixel_hues);
  free_c_int_vector(bins);
  free_c_int_vector(hues);
  free_c_int_vector(counter);
  free_c_int_vector(map);
  free_c_int_matrix(back_x, height);
  free_c_int_matrix(back_y, height);
  return frame;}

void compute_autoclip_amount(char *pathname,
			     enum VideoType video_type,
			     int motion_threshold,
			     int motion_pixel_mass_threshold,
			     int motion_proximity_threshold)
{ int i, k, x, y, x1, y1, x2, y2, xl, yl, xh, yh, xp, yp, height, width;
  int q, frame = 0, count, last_moving_valid = FALSE;
  int no_moving_clusters, once;
  int *motion_pythag;
  int **red, **green, **blue;
  int **last_red, **last_green, **last_blue;
  int **moving, **last_moving, **moving_cluster;
  int **back_x, **back_y;
  VidStream *video_input_port;
  motion_pythag = make_c_int_vector(motion_proximity_threshold+1);
  gfirst_frame = 0;
  /* Set up thresholds */
  motion_threshold *= motion_threshold;
  /* Set up proximity clusterer */
  for (i = 0; i<=motion_proximity_threshold; i++)
  { motion_pythag[i] = rint(sqrt(SQR(motion_proximity_threshold)-i*i));}
  /* Set up video port */
  video_input_port = open_video_input_file(pathname, video_type);
  if (!video_input_port) panic("Can't open video input file");
  /* needs work: This won't work because you need to do read_video first */
  height = get_video_height(video_input_port);
  width = get_video_width(video_input_port);
  red = make_c_int_matrix(height, width);
  green = make_c_int_matrix(height, width);
  blue = make_c_int_matrix(height, width);
  last_red = make_c_int_matrix(height, width);
  last_green = make_c_int_matrix(height, width);
  last_blue = make_c_int_matrix(height, width);
  moving = make_c_int_matrix(height, width);
  last_moving = make_c_int_matrix(height, width);
  moving_cluster = make_c_int_matrix(height, width);
  back_x = make_c_int_matrix(height, width);
  back_y = make_c_int_matrix(height, width);
  while (read_video(video_input_port))
  /* compute hue and chromaticity */
  { for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { red[y][x] = get_video_red(video_input_port, y, x);
	green[y][x] = get_video_green(video_input_port, y, x);
	blue[y][x] = get_video_blue(video_input_port, y, x);}}
    if (frame==0)
    { for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++)
	{ moving[y][x] = FALSE;
	  last_red[y][x] = red[y][x];
	  last_green[y][x] = green[y][x];
	  last_blue[y][x] = blue[y][x];}}}
    else
    { for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++)
	{ moving[y][x] =
	  (last_red[y][x]-red[y][x])*(last_red[y][x]-red[y][x])+
	  (last_green[y][x]-green[y][x])*(last_green[y][x]-green[y][x])+
	  (last_blue[y][x]-blue[y][x])*(last_blue[y][x]-blue[y][x])>
	  motion_threshold;
	  last_red[y][x] = red[y][x];
	  last_green[y][x] = green[y][x];
	  last_blue[y][x] = blue[y][x];}}}
    /* Cluster moving objects by proximity, ignoring small clusters */
    k = 0;
    once = FALSE;
    again:
    no_moving_clusters = TRUE;
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++) moving_cluster[y][x] = -1;}
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { if (moving[y][x]&&moving_cluster[y][x]==-1)
	{ moving_cluster[y][x] = k;
	  back_y[y][x] = -1;
	  back_x[y][x] = -1;
	  yp = y;
	  xp = x;
	  yl = y;
	  xl = x;
	  yh = y;
	  xh = x;
	  count = 1;
	  while (yp!=-1)
	  { y1 = yp;
	    x1 = xp;
	    yp = back_y[y1][x1];
	    xp = back_x[y1][x1];
	    if (moving_cluster[y1][x1]==k)
	    { for (y2 = MAX(0, y1-motion_proximity_threshold);
		   y2<MIN(height, y1+motion_proximity_threshold+1);
		   y2++)
	      { q = motion_pythag[abs(y2-y1)];
		for (x2 = MAX(0, x1-q); x2<MIN(width, x1+q+1); x2++)
		{ if (moving[y2][x2]&&moving_cluster[y2][x2]==-1)
		  { moving_cluster[y2][x2] = k;
		    count++;
		    back_y[y2][x2] = yp;
		    back_x[y2][x2] = xp;
		    yp = y2;
		    xp = x2;
		    yl = MIN(yl, y2);
		    xl = MIN(xl, x2);
		    yh = MAX(yh, y2);
		    xh = MAX(xh, x2);}}}}}
	  if (count>=motion_pixel_mass_threshold)
	  { k++;
	    no_moving_clusters = FALSE;}
	  else
	  { for (y1 = yl; y1<=yh; y1++)
	    { for (x1 = xl; x1<=xh; x1++)
	      if (moving_cluster[y1][x1]==k)
	      { moving_cluster[y1][x1] = -2;
		moving[y1][x1] = FALSE;}}}}}}
    if (no_moving_clusters)
    { if (last_moving_valid)
      { if (!once)
	{ for (y = 0; y<height; y++)
	  { for (x = 0; x<width; x++) moving[y][x] = last_moving[y][x];}
	  once = TRUE;
	  goto again;}}}
    else if (!once)
    { for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++) last_moving[y][x] = moving[y][x];}
      if (!last_moving_valid)
      { last_moving_valid = TRUE;
	gfirst_frame = frame;}
      glast_frame = frame;}
    frame++;}
  close_video_input_port(video_input_port);
  free_c_vector(motion_pythag);
  free_c_int_matrix(red, height);
  free_c_int_matrix(green, height);
  free_c_int_matrix(blue, height);
  free_c_int_matrix(last_red, height);
  free_c_int_matrix(last_green, height);
  free_c_int_matrix(last_blue, height);
  free_c_int_matrix(moving, height);
  free_c_int_matrix(last_moving, height);
  free_c_int_matrix(moving_cluster, height);
  free_c_int_matrix(back_x, height);
  free_c_int_matrix(back_y, height);}

void convolve(int predicted_x, int predicted_y,
	      int vicinity, enum predictor class,
	      int width, int height, int stamp_width, int stamp_height,
	      int subsample,
	      float **stamp_red,
	      float **stamp_green,
	      float **stamp_blue,
	      int **stamp_bitmap,
	      float **red, float **green, float **blue,
	      int *stamp_x, int *stamp_y, float *best,
	      enum predictor *predictor)
{ int x, y, this_x, this_y, j;
  float sum, r, g, b;
  for (this_y = predicted_y-vicinity; this_y<=predicted_y+vicinity; this_y++)
  { for (this_x = predicted_x-vicinity; this_x<=predicted_x+vicinity; this_x++)
    { sum = 0.0;
      j = 0;
      for (y = MAX(0, -this_y);
	   y<MIN(stamp_height, height-this_y);
	   y += subsample)
      { for (x = MAX(0, -this_x);
	     x<MIN(stamp_width, width-this_x);
	     x += subsample)
	{ if (stamp_bitmap[y][x])
	  { r = red[this_y+y][this_x+x]-stamp_red[y][x];
	    g = green[this_y+y][this_x+x]-stamp_green[y][x];
	    b = blue[this_y+y][this_x+x]-stamp_blue[y][x];
	    sum += r*r+g*g+b*b;
	    j++;}}}
      if (j!=0)
      { sum /= j;
	if (sum<*best)
	{ *stamp_x = this_x;
	  *stamp_y = this_y;
	  *best = sum;
	  *predictor = class;}}}}}

int rfbt(char *pathname,
	 char *track_pathname,
	 enum VideoType video_type,
	 int acceleration_vicinity,
	 int velocity_vicinity,
	 int position_vicinity,
	 int subsample,
	 Real jerky,
	 Real ratio,
	 int colour_pixel_mass_threshold,
	 int colour_proximity_threshold,
	 int motion_threshold,
	 int motion_pixel_mass_threshold,
	 int motion_proximity_threshold,
	 int hand_left,
	 int hand_top,
	 int hand_right,
	 int hand_bottom,
	 int track_movie)
{ int i, j, k = 0, k1 = 0, k2 = 0, l1, l2, height, width;
  int x, y, x1, y1, x2, y2, xl, yl, xh, yh, xp, yp;
  int min_x, min_y, max_x, max_y;
  int q, count;
  float r, g, b, h, s, v, m, rc, gc, bc;
  float object_x[MAX_OBJECTS], object_y[MAX_OBJECTS];
  float object_x1[MAX_OBJECTS], object_y1[MAX_OBJECTS];
  float object_x2[MAX_OBJECTS], object_y2[MAX_OBJECTS];
  Real object_sigma_xx[MAX_OBJECTS];
  Real object_sigma_xy[MAX_OBJECTS];
  Real object_sigma_yy[MAX_OBJECTS];
  float object_vx[MAX_OBJECTS], object_vy[MAX_OBJECTS];
  float object_ax[MAX_OBJECTS], object_ay[MAX_OBJECTS];
  float predicted_x, predicted_y;
  float determinant, sigma_inverse_xx, sigma_inverse_xy, sigma_inverse_yy;
  float variance_enhancer = .30;
  int stamp_x[MAX_OBJECTS], stamp_y[MAX_OBJECTS];
  int stamp_x1[MAX_OBJECTS], stamp_y1[MAX_OBJECTS];
  int stamp_x2[MAX_OBJECTS], stamp_y2[MAX_OBJECTS];
  int stamp_width[MAX_OBJECTS], stamp_height[MAX_OBJECTS];
  int stamp_vx[MAX_OBJECTS], stamp_vy[MAX_OBJECTS];
  int stamp_ax[MAX_OBJECTS], stamp_ay[MAX_OBJECTS];
  int stamp_coloured[MAX_OBJECTS];
  int frame = 0;
  enum predictor predictor, predictor_tried;
  float best[MAX_OBJECTS], last_best[MAX_OBJECTS], last_last_best[MAX_OBJECTS];
  int *colour_pythag;
  int *motion_pythag;
  float mu_x, mu_y, sigma_xx, sigma_xy, sigma_yy;
  VidStream *video_input_port;
  FILE *video_output_port;
  float **stamp_red[MAX_OBJECTS];
  float **stamp_green[MAX_OBJECTS];
  float **stamp_blue[MAX_OBJECTS];
  int **stamp_bitmap[MAX_OBJECTS];
  float **red, **green, **blue;
  float **last_red, **last_green, **last_blue;
  float **hue, **saturation, **value;
  float **last_hue, **last_saturation, **last_value;
  int **colour, **moving, **last_moving, **coloured_cluster, **moving_cluster;
  int **back_x, **back_y;
  colour_pythag = make_c_int_vector(colour_proximity_threshold+1);
  motion_pythag = make_c_int_vector(motion_proximity_threshold+1);
  /* Set up proximity clusterer */
  for (i = 0; i<=colour_proximity_threshold; i++)
  { colour_pythag[i] = rint(sqrt(SQR(colour_proximity_threshold)-i*i));}
  for (i = 0; i<=motion_proximity_threshold; i++)
  { motion_pythag[i] = rint(sqrt(SQR(motion_proximity_threshold)-i*i));}
  /* Set up video ports */
  video_input_port = open_video_input_file(pathname, video_type);
  if (!video_input_port) panic("Can't open video input file");
  /* needs work: This won't work because you need to do read_video first */
  height = get_video_height(video_input_port);
  width = get_video_width(video_input_port);
  for (i = 0; i<MAX_OBJECTS; i++)
  { stamp_red[i] = make_c_float_matrix(height, width);
    stamp_green[i] = make_c_float_matrix(height, width);
    stamp_blue[i] = make_c_float_matrix(height, width);
    stamp_bitmap[i] = make_c_int_matrix(height, width);}
  red = make_c_float_matrix(height, width);
  green = make_c_float_matrix(height, width);
  blue = make_c_float_matrix(height, width);
  last_red = make_c_float_matrix(height, width);
  last_green = make_c_float_matrix(height, width);
  last_blue = make_c_float_matrix(height, width);
  hue = make_c_float_matrix(height, width);
  saturation = make_c_float_matrix(height, width);
  value = make_c_float_matrix(height, width);
  last_hue = make_c_float_matrix(height, width);
  last_saturation = make_c_float_matrix(height, width);
  last_value = make_c_float_matrix(height, width);
  colour = make_c_int_matrix(height, width);
  moving = make_c_int_matrix(height, width);
  last_moving = make_c_int_matrix(height, width);
  coloured_cluster = make_c_int_matrix(height, width);
  moving_cluster = make_c_int_matrix(height, width);
  back_x = make_c_int_matrix(height, width);
  back_y = make_c_int_matrix(height, width);
  if (track_movie)
  { video_output_port =
    open_video_output_file(track_pathname, video_type, MAX_FRAMES);
    if (!video_output_port) panic("Can't open video output file");}
  while (read_video(video_input_port))
  /* Compute colour and moving for each frame */
  { for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { last_red[y][x] = red[y][x];
	last_green[y][x] = green[y][x];
	last_blue[y][x] = blue[y][x];
	last_hue[y][x] = hue[y][x];
	last_saturation[y][x] = saturation[y][x];
	last_value[y][x] = value[y][x];
	red[y][x] = get_video_red(video_input_port, y, x);
	green[y][x] = get_video_green(video_input_port, y, x);
	blue[y][x] = get_video_blue(video_input_port, y, x);
	r = red[y][x]/255.0;
	g = green[y][x]/255.0;
	b = blue[y][x]/255.0;
	v = MAX(r, MAX(g, b));
	m = MIN(r, MIN(g, b));
	s = v==0.0?0.0:(v-m)/v;
	if (s==0.0) h = 0.0;
	else
	{ rc = (v-r)/(v-m);
	  gc = (v-g)/(v-m);
	  bc = (v-b)/(v-m);
	  h = (r==v?bc-gc:g==v?2.0+rc-bc:4.0+gc-rc)/6.0;
	  if (h<0.0) h += 1.0;}
	h = (MAX_HUE_VALUE+1)*h;
	hue[y][x] = h;
	saturation[y][x] = s;
	value[y][x] = v;
	colour[y][x] = ((h>=RED_BLOCK_HUE_HIGH||h<=RED_BLOCK_HUE_LOW)&&
			s>=RED_BLOCK_SATURATION_LOW)?0:-1;
	moving[y][x] =
	(last_red[y][x]-red[y][x])*(last_red[y][x]-red[y][x])+
	(last_green[y][x]-green[y][x])*(last_green[y][x]-green[y][x])+
	(last_blue[y][x]-blue[y][x])*(last_blue[y][x]-blue[y][x])>
	motion_threshold;}}
    /* Cluster coloured objects by proximity, ignoring small clusters */
    if (frame==1)
    { k = 0;
      for (i = 0; i<1; i++)	/* note: n is fixed to be 1 here */
      { for (y = 0; y<height; y++)
	{ for (x = 0; x<width; x++) coloured_cluster[y][x] = -1;}
	for (y = 0; y<height; y++)
	{ for (x = 0; x<width; x++)
	  { if (colour[y][x]==i&&coloured_cluster[y][x]==-1)
	    { coloured_cluster[y][x] = k;
	      count = 1;
	      back_y[y][x] = -1;
	      back_x[y][x] = -1;
	      yp = y;
	      xp = x;
	      yl = y;
	      xl = x;
	      yh = y;
	      xh = x;
	      while (yp!=-1)
	      { y1 = yp;
		x1 = xp;
		yp = back_y[y1][x1];
		xp = back_x[y1][x1];
		if (coloured_cluster[y1][x1]==k)
		{ for (y2 = MAX(0, y1-colour_proximity_threshold);
		       y2<MIN(height, y1+colour_proximity_threshold+1);
		       y2++)
		  { q = colour_pythag[abs(y2-y1)];
		    for (x2 = MAX(0, x1-q); x2<MIN(width, x1+q+1); x2++)
		    { if (colour[y2][x2]==i&&coloured_cluster[y2][x2]==-1)
		      { coloured_cluster[y2][x2] = k;
			count++;
			back_y[y2][x2] = yp;
			back_x[y2][x2] = xp;
			yp = y2;
			xp = x2;
			yl = MIN(yl, y2);
			xl = MIN(xl, x2);
			yh = MAX(yh, y2);
			xh = MAX(xh, x2);}}}}}
	      if (count>=colour_pixel_mass_threshold) k++;
	      else
	      { for (y1 = yl; y1<=yh; y1++)
		{ for (x1 = xl; x1<=xh; x1++)
		  if (coloured_cluster[y1][x1]==k)
		  { coloured_cluster[y1][x1] = -2;
		    /* note: this can affect the stamp tracker */
		    colour[y1][x1] = -1;}}}}}}}
      k1 = k;}
    /* Cluster moving objects by proximity, ignoring small clusters */
    if (frame==1)
    { for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++) moving_cluster[y][x] = -1;}
      for (y = 0; y<height; y++)
      { for (x = 0; x<width; x++)
	{ if (moving[y][x]&&moving_cluster[y][x]==-1)
	  { moving_cluster[y][x] = k;
	    back_y[y][x] = -1;
	    back_x[y][x] = -1;
	    yp = y;
	    xp = x;
	    yl = y;
	    xl = x;
	    yh = y;
	    xh = x;
	    count = 1;
	    while (yp!=-1)
	    { y1 = yp;
	      x1 = xp;
	      yp = back_y[y1][x1];
	      xp = back_x[y1][x1];
	      if (moving_cluster[y1][x1]==k)
	      { for (y2 = MAX(0, y1-motion_proximity_threshold);
		     y2<MIN(height, y1+motion_proximity_threshold+1);
		     y2++)
		{ q = motion_pythag[abs(y2-y1)];
		  for (x2 = MAX(0, x1-q); x2<MIN(width, x1+q+1); x2++)
		  { if (moving[y2][x2]&&moving_cluster[y2][x2]==-1)
		    { moving_cluster[y2][x2] = k;
		      count++;
		      back_y[y2][x2] = yp;
		      back_x[y2][x2] = xp;
		      yp = y2;
		      xp = x2;
		      yl = MIN(yl, y2);
		      xl = MIN(xl, x2);
		      yh = MAX(yh, y2);
		      xh = MAX(xh, x2);}}}}}
	    if (count>=motion_pixel_mass_threshold) k++;
	    else
	    { for (y1 = yl; y1<=yh; y1++)
	      { for (x1 = xl; x1<=xh; x1++)
		if (moving_cluster[y1][x1]==k)
		{ moving_cluster[y1][x1] = -2;
		  /* note: this can affect the stamp tracker */
		  moving[y1][x1] = FALSE;}}}}}}
      k2 = k-k1;}
    if (k>MAX_OBJECTS) panic("Too many objects");
    /* needs work: assumes already clipped */
    if (frame==1)
    /* Initialize stamps */
    { l1 = k1;
      for (i = 0; i<l1; i++)
      { min_x = width;
	min_y = height;
	max_x = 0;
	max_y = 0;
	count = 0;
	for (y = 0; y<height; y++)
	{ for (x = 0; x<width; x++)
	  if (coloured_cluster[y][x]==i||moving_cluster[y][x]==i)
	  { count++;
	    min_x = MIN(min_x, x);
	    min_y = MIN(min_y, y);
	    max_x = MAX(max_x, x);
	    max_y = MAX(max_y, y);}}
	stamp_x[i] = min_x;
	stamp_y[i] = min_y;
	stamp_width[i] = max_x-min_x;
	stamp_height[i] = max_y-min_y;
	stamp_coloured[i] = i<k1;
	for (y = 0; y<stamp_height[i]; y++)
	{ for (x = 0; x<stamp_width[i]; x++)
	  { stamp_bitmap[i][y][x] =
	    (stamp_coloured[i]?
	     coloured_cluster[y+stamp_y[i]][x+stamp_x[i]]==i:
	     moving_cluster[y+stamp_y[i]][x+stamp_x[i]]==i&&
	     colour[y+stamp_y[i]][x+stamp_x[i]]==-1)?PERSISTENCE:0;}}
	stamp_x1[i] = min_x;
	stamp_y1[i] = min_y;
	stamp_x2[i] = min_x;
	stamp_y2[i] = min_y;
	stamp_vx[i] = 0;
	stamp_vy[i] = 0;
	stamp_ax[i] = 0;
	stamp_ay[i] = 0;
	printf("%3d %d %4d %4d %3d %3d\n",
	       frame, i,
	       stamp_x[i], stamp_y[i], stamp_width[i], stamp_height[i]);}
      for (i = 0; i<l1; i++)
      { for (y = MAX(0, -stamp_y[i]);
	     y<MIN(stamp_height[i], height-stamp_y[i]);
	     y++)
	{ for (x = MAX(0, -stamp_x[i]);
	       x<MIN(stamp_width[i], width-stamp_x[i]);
	       x++)
	  { stamp_red[i][y][x] = red[y+stamp_y[i]][x+stamp_x[i]];
	    stamp_green[i][y][x] = green[y+stamp_y[i]][x+stamp_x[i]];
	    stamp_blue[i][y][x] = blue[y+stamp_y[i]][x+stamp_x[i]];}}}
      /* Initialize objects */
      l2 = k2;
      for (i = 0, j = k1; i<l2; i++, j++)
      { mu_x = 0.0;
	mu_y = 0.0;
	sigma_xx = 0.0;
	sigma_xy = 0.0;
	sigma_yy = 0.0;
	count = 0;
	for (y = 0; y<height; y++)
	{ for (x = 0; x<width; x++)
	  { if (moving_cluster[y][x]==j)
	    { mu_x += x;
	      mu_y += y;
	      count++;}}}
	mu_x /= count;
	mu_y /= count;
	for (y = 0; y<height; y++)
	{ for (x = 0; x<width; x++)
	  { if (moving_cluster[y][x]==j)
	    { sigma_xx += (x-mu_x)*(x-mu_x);
	      sigma_xy += (x-mu_x)*(y-mu_y);
	      sigma_yy += (y-mu_y)*(y-mu_y);}}}
	sigma_xx /= count;
	sigma_xy /= count;
	sigma_yy /= count;
	object_x[i] = mu_x;
	object_y[i] = mu_y;
	object_sigma_xx[i] = sigma_xx;
	object_sigma_xy[i] = sigma_xy;
	object_sigma_yy[i] = sigma_yy;
	object_x1[i] = mu_x;
	object_y1[i] = mu_y;
	object_x2[i] = mu_x;
	object_y2[i] = mu_y;
	object_vx[i] = 0.0;
	object_vy[i] = 0.0;
	object_ax[i] = 0.0;
	object_ay[i] = 0.0;
	printf("%3d %d %4.0f %4.0f\n", frame, j, object_x[i], object_y[i]);}
      printf("\n");}
    /* needs work: assumes already clipped */
    else if (frame>1)
    /* Search for next stamp position */
    { for (i = 0; i<l1; i++)
      { stamp_x2[i] = stamp_x1[i];
	stamp_y2[i] = stamp_y1[i];
	stamp_x1[i] = stamp_x[i];
	stamp_y1[i] = stamp_y[i];
	last_last_best[i] = last_best[i];
	last_best[i] = best[i];
	best[i] = INFINITY;
	predictor = NONE;
	predictor_tried = ACCELERATION;
	convolve(stamp_x1[i]+stamp_vx[i]+0.5*stamp_ax[i],
		 stamp_y1[i]+stamp_vy[i]+0.5*stamp_ay[i],
		 acceleration_vicinity, ACCELERATION,
		 width, height, stamp_width[i], stamp_height[i],
		 subsample,
		 stamp_red[i], stamp_green[i], stamp_blue[i],
		 stamp_bitmap[i],
		 red, green, blue,
		 &stamp_x[i], &stamp_y[i], &best[i], &predictor);
	if (frame>=4&&best[i]-2*last_best[i]+last_last_best[i]>jerky)
	{ predictor_tried = VELOCITY;
	  convolve(stamp_x1[i]+stamp_vx[i], stamp_y1[i]+stamp_vy[i],
		   velocity_vicinity, VELOCITY,
		   width, height, stamp_width[i], stamp_height[i],
		   subsample,
		   stamp_red[i], stamp_green[i], stamp_blue[i],
		   stamp_bitmap[i],
		   red, green, blue,
		   &stamp_x[i], &stamp_y[i], &best[i], &predictor);
	  if (frame>=4&&best[i]-2*last_best[i]+last_last_best[i]>jerky)
	  { predictor_tried = POSITION;
	    convolve(stamp_x1[i], stamp_y1[i],
		     position_vicinity, POSITION,
		     width, height, stamp_width[i], stamp_height[i],
		     subsample,
		     stamp_red[i], stamp_green[i], stamp_blue[i],
		     stamp_bitmap[i],
		     red, green, blue,
		     &stamp_x[i], &stamp_y[i], &best[i], &predictor);}}
	
	/* Update predictor */
	stamp_vx[i] = stamp_x[i]-stamp_x1[i];
	stamp_vy[i] = stamp_y[i]-stamp_y1[i];
	stamp_ax[i] = stamp_x[i]-2*stamp_x1[i]+stamp_x2[i];
	stamp_ay[i] = stamp_y[i]-2*stamp_y1[i]+stamp_y2[i];
	printf("%3d %d %4d %4d %3d %3d %4d %4d %4d %4d %5.0f %5.0f %5.0f",
	       frame, i,
	       stamp_x[i], stamp_y[i], stamp_width[i], stamp_height[i],
	       stamp_vx[i], stamp_vy[i], stamp_ax[i], stamp_ay[i],
	       best[i],
	       frame>=3?best[i]-last_best[i]:0.0,
	       frame>=4?best[i]-2*last_best[i]+last_last_best[i]:0.0);
	printf(" %s %s",
	       predictor_tried==ACCELERATION?"acc":
	       predictor_tried==VELOCITY?"vel":
	       predictor_tried==POSITION?"pos":
	       predictor_tried==ANYWHERE?"any":"non",
	       predictor==ACCELERATION?"acc":
	       predictor==VELOCITY?"vel":
	       predictor==POSITION?"pos":
	       predictor==ANYWHERE?"any":"non");
	if (frame>=4&&best[i]-2*last_best[i]+last_last_best[i]>jerky)
	printf("*\n");
	else printf("\n");}
      /* note: an IIR filter */
      for (i = 0; i<l1; i++)
      { count = 0;
	for (y = MAX(0, -stamp_y[i]);
	     y<MIN(stamp_height[i], height-stamp_y[i]);
	     y++)
	{ for (x = MAX(0, -stamp_x[i]);
	       x<MIN(stamp_width[i], width-stamp_x[i]);
	       x++)
	  { stamp_red[i][y][x] = ratio*stamp_red[i][y][x]+
	    (1-ratio)*red[y+stamp_y[i]][x+stamp_x[i]];
	    stamp_green[i][y][x] = ratio*stamp_green[i][y][x]+
	    (1-ratio)*green[y+stamp_y[i]][x+stamp_x[i]];
	    stamp_blue[i][y][x] = ratio*stamp_blue[i][y][x]+
	    (1-ratio)*blue[y+stamp_y[i]][x+stamp_x[i]];
	    count += moving[y+stamp_y[i]][x+stamp_x[i]];}}
	if (count>motion_pixel_mass_threshold)
	{ for (y = MAX(0, -stamp_y[i]);
	       y<MIN(stamp_height[i], height-stamp_y[i]);
	       y++)
	  { for (x = MAX(0, -stamp_x[i]);
		 x<MIN(stamp_width[i], width-stamp_x[i]);
		 x++)
	    { stamp_bitmap[i][y][x] =
	      (stamp_coloured[i]?
	       colour[y+stamp_y[i]][x+stamp_x[i]]!=-1:
	       moving[y+stamp_y[i]][x+stamp_x[i]]&&
	       colour[y+stamp_y[i]][x+stamp_x[i]]==-1)?
	       PERSISTENCE:MAX(0, stamp_bitmap[i][y][x]-1);}}}}
      /* Search for next object position */
      for (i = 0, j = k1; i<l2; i++, j++)
      { predicted_x = object_x[i]+object_vx[i]+0.5*object_ax[i];
	predicted_y = object_y[i]+object_vy[i]+0.5*object_ay[i];
	determinant = (object_sigma_xx[i]*object_sigma_yy[i]-
		       object_sigma_xy[i]*object_sigma_xy[i]);
	sigma_inverse_xx = object_sigma_yy[i]/determinant;
	sigma_inverse_xy = -object_sigma_xy[i]/determinant;
	sigma_inverse_yy = object_sigma_xx[i]/determinant;
	mu_x = 0.0;
	mu_y = 0.0;
	sigma_xx = 0.0;
	sigma_xy = 0.0;
	sigma_yy = 0.0;
	count = 0;
	for (y = 0; y<height; y++)
	{ for (x = 0; x<width; x++)
	  { if (moving[y][x]&&
		(x-predicted_x)*sigma_inverse_xx*(x-predicted_x)+
		(x-predicted_x)*sigma_inverse_xy*(y-predicted_y)+
		(y-predicted_y)*sigma_inverse_xy*(x-predicted_x)+
		(y-predicted_y)*sigma_inverse_yy*(y-predicted_y)<=4.0)
	    { mu_x += x;
	      mu_y += y;
	      count++;}}}
	mu_x /= count;
	mu_y /= count;
	for (y = 0; y<height; y++)
	{ for (x = 0; x<width; x++)
	  { if (moving[y][x]&&
		(x-predicted_x)*sigma_inverse_xx*(x-predicted_x)+
		(x-predicted_x)*sigma_inverse_xy*(y-predicted_y)+
		(y-predicted_y)*sigma_inverse_xy*(x-predicted_x)+
		(y-predicted_y)*sigma_inverse_yy*(y-predicted_y)<=4.0)
	    { sigma_xx += (x-mu_x)*(x-mu_x);
	      sigma_xy += (x-mu_x)*(y-mu_y);
	      sigma_yy += (y-mu_y)*(y-mu_y);}}}
	sigma_xx /= count;
	sigma_xy /= count;
	sigma_yy /= count;
	
	/* Update predictor */
	object_x2[i] = object_x1[i];
	object_y2[i] = object_y1[i];
	object_x1[i] = object_x[i];
	object_y1[i] = object_y[i];
	if (count/
	    (4*PI*
	     sqrt(eigenvalue1(sigma_xx, sigma_xy, sigma_xy, sigma_yy))*
	     sqrt(eigenvalue2(sigma_xx, sigma_xy, sigma_xy, sigma_yy)))
	    >0.1)
	{ object_vx[i] = object_x[i]-object_x1[i];
	  object_vy[i] = object_y[i]-object_y1[i];
	  object_ax[i] = object_x[i]-2*object_x1[i]+object_x2[i];
	  object_ay[i] = object_y[i]-2*object_y1[i]+object_y2[i];
	  object_x[i] = mu_x;
	  object_y[i] = mu_y;
	  object_sigma_xx[i] = sigma_xx*(1.0+variance_enhancer);
	  object_sigma_xy[i] = sigma_xy*(1.0+variance_enhancer);
	  object_sigma_yy[i] = sigma_yy*(1.0+variance_enhancer);
	  clip_symmatrix(&object_sigma_xx[i],
			 &object_sigma_xy[i],
			 &object_sigma_yy[i],
			 400.0);}
	printf("%3d %d %4.0f %4.0f %3.0f %3.0f %4.0f %4.0f %4.0f %4.0f\n",
	       frame, j, object_x[i], object_y[i],
	       2.0*sqrt(eigenvalue1(object_sigma_xx[i],
				    object_sigma_xy[i],
				    object_sigma_xy[i],
				    object_sigma_yy[i])),
	       2.0*sqrt(eigenvalue2(object_sigma_xx[i],
				    object_sigma_xy[i],
				    object_sigma_xy[i],
				    object_sigma_yy[i])),
	       object_vx[i], object_vy[i], object_ax[i], object_ay[i]);}
      printf("\n");}
    /* Fit ellipses to stamps and objects */
    if (frame>0)
    { gobject_k[frame] = l1+l2;
      for (i = 0; i<l1; i++)
      { mu_x = 0.0;
	mu_y = 0.0;
	sigma_xx = 0.0;
	sigma_xy = 0.0;
	sigma_yy = 0.0;
	count = 0;
	for (y = 0; y<stamp_height[i]; y++)
	{ for (x = 0; x<stamp_width[i]; x++)
	  if (stamp_bitmap[i][y][x])
	  { count++;
	    mu_x += x+stamp_x[i];
	    mu_y += y+stamp_y[i];}}
	mu_x /= count;
	mu_y /= count;
	for (y = 0; y<stamp_height[i]; y++)
	{ for (x = 0; x<stamp_width[i]; x++)
	  if (stamp_bitmap[i][y][x])
	  { sigma_xx += (x+stamp_x[i]-mu_x)*(x+stamp_x[i]-mu_x);
	    sigma_xy += (x+stamp_x[i]-mu_x)*(y+stamp_y[i]-mu_y);
	    sigma_yy += (y+stamp_y[i]-mu_y)*(y+stamp_y[i]-mu_y);}}
	sigma_xx /= count;
	sigma_xy /= count;
	sigma_yy /= count;
	gellipse_x[frame][i] = mu_x;
	gellipse_y[frame][i] = mu_y;
	gellipse_theta[frame][i] = eigenvector_angle1(sigma_xx,
						      sigma_xy,
						      sigma_xy,
						      sigma_yy);
	gellipse_r[frame][i] = sqrt(2.0*eigenvalue1(sigma_xx,
						    sigma_xy,
						    sigma_xy,
						    sigma_yy));
	gellipse_s[frame][i] = sqrt(2.0*eigenvalue2(sigma_xx,
						    sigma_xy,
						    sigma_xy,
						    sigma_yy));}
      for (i = 0, j = l1; i<l2; i++, j++)
      { gellipse_x[frame][j] = object_x[i];
	gellipse_y[frame][j] = object_y[i];
	gellipse_theta[frame][j] = eigenvector_angle1(object_sigma_xx[i],
						      object_sigma_xy[i],
						      object_sigma_xy[i],
						      object_sigma_yy[i]);
	gellipse_r[frame][j] = sqrt(2.0*eigenvalue1(object_sigma_xx[i],
						    object_sigma_xy[i],
						    object_sigma_xy[i],
						    object_sigma_yy[i]));
	gellipse_s[frame][j] = sqrt(2.0*eigenvalue2(object_sigma_xx[i],
						    object_sigma_xy[i],
						    object_sigma_xy[i],
						    object_sigma_yy[i]));}}
    /* Produce track movie */
    if (track_movie)
    { begin_video_frame(width, height, 255);
      switch (1)		/* select type of track movie */
      { case 0:
	for (y = 0; y<height; y++)
	{ for (x = 0; x<width; x++)
	  { put_video_red(y, x,
			  colour[y][x]!=-1?red[y][x]:moving[y][x]?0:255);
	    put_video_green(y, x,
			    colour[y][x]!=-1?green[y][x]:moving[y][x]?0:255);
	    put_video_blue(y, x,
			   colour[y][x]!=-1?blue[y][x]:moving[y][x]?0:255);}}
	break;
	case 1:
	for (y = 0; y<height; y++)
	{ for (x = 0; x<width; x++)
	  { put_video_red(y, x, red[y][x]);
	    put_video_green(y, x, green[y][x]);
	    put_video_blue(y, x, blue[y][x]);}}
	if (frame>0)
	{ for (i = 0; i<l1; i++)
	  { for (y = MAX(0, -stamp_y[i]);
		 y<MIN(stamp_height[i], height-stamp_y[i]);
		 y++)
	    { for (x = MAX(0, -stamp_x[i]);
		   x<MIN(stamp_width[i], width-stamp_x[i]);
		   x++)
	      { if (stamp_bitmap[i][y][x])
		{ put_video_red(y+stamp_y[i], x+stamp_x[i],
				stamp_coloured[i]?255:255);
		  put_video_green(y+stamp_y[i], x+stamp_x[i],
				  stamp_coloured[i]?0:255);
		  put_video_blue(y+stamp_y[i], x+stamp_x[i],
				 stamp_coloured[i]?0:255);}}}
	    for (y = 0; y<stamp_height[i]; y++)
	    { put_video_red(MIN(MAX(stamp_y[i]+y, 0), height-1),
			    MIN(MAX(stamp_x[i], 0), width-1),
			    255);
	      put_video_green(MIN(MAX(stamp_y[i]+y, 0), height-1),
			      MIN(MAX(stamp_x[i], 0), width-1),
			      255);
	      put_video_blue(MIN(MAX(stamp_y[i]+y, 0), height-1),
			     MIN(MAX(stamp_x[i], 0), width-1),
			     255);
	      put_video_red(MIN(MAX(stamp_y[i]+y, 0), height-1),
			    MIN(MAX(stamp_x[i]+stamp_width[i]-1, 0),
				width-1),
			    255);
	      put_video_green(MIN(MAX(stamp_y[i]+y, 0), height-1),
			      MIN(MAX(stamp_x[i]+stamp_width[i]-1, 0),
				  width-1),
			      255);
	      put_video_blue(MIN(MAX(stamp_y[i]+y, 0), height-1),
			     MIN(MAX(stamp_x[i]+stamp_width[i]-1, 0),
				 width-1),
			     255);}
	    for (x = 0; x<stamp_width[i]; x++)
	    { put_video_red(MIN(MAX(stamp_y[i], 0), height-1),
			    MIN(MAX(stamp_x[i]+x, 0), width-1),
			    255);
	      put_video_green(MIN(MAX(stamp_y[i], 0), height-1),
			      MIN(MAX(stamp_x[i]+x, 0), width-1),
			      255);
	      put_video_blue(MIN(MAX(stamp_y[i], 0), height-1),
			     MIN(MAX(stamp_x[i]+x, 0), width-1),
			     255);
	      put_video_red(MIN(MAX(stamp_y[i]+stamp_height[i]-1, 0),
				height-1),
			    MIN(MAX(stamp_x[i]+x, 0), width-1),
			    255);
	      put_video_green(MIN(MAX(stamp_y[i]+stamp_height[i]-1, 0),
				  height-1),
			      MIN(MAX(stamp_x[i]+x, 0), width-1),
			      255);
	      put_video_blue(MIN(MAX(stamp_y[i]+stamp_height[i]-1, 0),
				 height-1),
			     MIN(MAX(stamp_x[i]+x, 0), width-1),
			     255);}}
	  for (y = 0; y<height; y++)
	  { for (x = 0; x<width; x++)
	    { if (moving[y][x])
	      { put_video_red(y, x, 255);
		put_video_green(y, x, 255);
		put_video_blue(y, x, 255);}}}}
	break;}
      end_video_frame();}
    frame++;
    if (frame>=MAX_FRAMES) panic("Too many frames");}
  /* A kludge since this doesn't track the first frame */
  gobject_k[0] = gobject_k[1];
  for (i = 0; i<l1+l2; i++)
  { gellipse_x[0][i] = gellipse_x[1][i];
    gellipse_y[0][i] = gellipse_y[1][i];
    gellipse_theta[0][i] = gellipse_theta[1][i];
    gellipse_r[0][i] = gellipse_r[1][i];
    gellipse_s[0][i] = gellipse_s[1][i];}
  if (track_movie) close_video_output_port(video_output_port);
  close_video_input_port(video_input_port);
  gfirst_frame = 0;		/* needs work */
  glast_frame = frame-1;	/* needs work */
  free_c_int_vector(colour_pythag);
  free_c_int_vector(motion_pythag);
  for (i = 0; i<MAX_OBJECTS; i++)
  { free_c_float_matrix(stamp_red[i], height);
    free_c_float_matrix(stamp_green[i], height);
    free_c_float_matrix(stamp_blue[i], height);
    free_c_int_matrix(stamp_bitmap[i], height);}
  free_c_float_matrix(red, height);
  free_c_float_matrix(green, height);
  free_c_float_matrix(blue, height);
  free_c_float_matrix(last_red, height);
  free_c_float_matrix(last_green, height);
  free_c_float_matrix(last_blue, height);
  free_c_float_matrix(hue, height);
  free_c_float_matrix(saturation, height);
  free_c_float_matrix(value, height);
  free_c_float_matrix(last_hue, height);
  free_c_float_matrix(last_saturation, height);
  free_c_float_matrix(last_value, height);
  free_c_int_matrix(colour, height);
  free_c_int_matrix(moving, height);
  free_c_int_matrix(last_moving, height);
  free_c_int_matrix(coloured_cluster, height);
  free_c_int_matrix(moving_cluster, height);
  free_c_int_matrix(back_x, height);
  free_c_int_matrix(back_y, height);
  return frame;}

/* Interruptible usleep */

int wall_clock(void)
{ struct tms buf;
  return times(&buf);}

/* Meteor */

void meteor_tracker(unsigned char *buffer,
		    int frame,
		    int height,
		    int width,
		    int uninterlaced,
		    int capturing,
                    int capturing_code,
		    int tracking_ellipses,
		    int tracking_polygons,
		    int hull,
		    int motion_threshold,
		    int colour_pixel_mass_threshold,
		    int colour_proximity_threshold,
		    int motion_pixel_mass_threshold,
		    int motion_proximity_threshold)
/* Global variables:
   meteor_last_frame
   meteor_starting
   meteor_{r,g,b}{h,s,v}{l,h}
   meteor_last_code
   meteor_conjure_code
   meteor_conjuring
   meteor_code_frames
   gpolygon_x
   gpolygon_y
   gmotion
   gobject_k
   gobject_colour
   gobject_x
   gobject_y
   gobject_theta
   gobject_r
   gobject_s
   */
{ unsigned char *p1, *p2, *p3, *pl;
  int y, x, ymin, ymax, ybest, k, l, r, g, b, rd, gd, bd, h, s, v, p;
  int proximity_threshold;
  int t = motion_threshold*motion_threshold;
  int i, yp, xp, yl, xl, yh, xh, y1, x1, y2, x2, count, state;
  int j[MAX_OBJECTS];
  int blank = 2;
  float mu_x[MAX_OBJECTS], mu_y[MAX_OBJECTS];
  float sigma_xx[MAX_OBJECTS], sigma_xy[MAX_OBJECTS], sigma_yy[MAX_OBJECTS];
  unsigned char colour[MAX_OBJECTS];
  unsigned char code[height/2][width/2];
  unsigned char cluster[height/2][width/2];
  unsigned char map[height/2][width/2];
  short back_y[height/2][width/2];
  short back_x[height/2][width/2];
  short xmin[height], xmax[height];
  if (capturing)
  /* Copy the current frame from buffer (which holds only the current frame)
     to meteor_frames (which holds all of the frames that are captured). */
  { if (uninterlaced)
    /* When uninterlaced, only the even scan lines are stored to save space.
       The meteor buffer has four bytes per pixel and meteor_frames has three
       bytes per pixel (to save space) so packing is done. Two loops are
       needed to allow skipping the odd scan lines. */
    { for (y = 0; y<height/2; y++)
      { for (p1 = &buffer[2*y*width*4],
	     p2 = &meteor_frames[(frame*height/2+y)*width*3],
	     pl = &buffer[(2*y+1)*width*4];
	     p1<pl;)
	{ *p2++ = *p1++;
	  *p2++ = *p1++;
	  *p2++ = *p1++;
	  p1++;}}}
    else
    /* When interlaced, all scan lines are stored. The meteor buffer has four
       bytes per pixel and meteor_frames has three bytes per pixel (to save
       space). In this case, only one loop is needed. */
    { for (p1 = &buffer[0],
	   p2 = &meteor_frames[frame*height*width*3],
	   pl = &buffer[height*width*4];
	   p1<pl;)
      { *p2++ = *p1++;
	*p2++ = *p1++;
	*p2++ = *p1++;
	p1++;}}}
  if (tracking_ellipses||tracking_polygons)
  /* Loop over even numbered scan lines. This is necessary since the odd
     numbered scan lines contain no additional information due to
     deinterlacing. */
  { for (y = 0,
	 p1 = &buffer[0],
	 p2 = &meteor_last_frame[0],
	 p3 = &code[0][0];
	 y<height/2;
	 y++,
	 p1 += 4*width)
    /* Loop over even numbered columns. Since we subsample vertically due to
       deinterlacing, we also subsample horizontally to keep things symmetric.
       */
    { for (x = 0; x<width/2; x++, p3++)
      /* Get the current pixel in r, g, and b. Get the difference between the
         current pixel in the current frame and the current pixel in the
         previous frame in rd, gd, and bd. Copy the current frame into
         meteor_last_frame so that we can do motion detection in the next
         frame. p1 points to the current pixel in the current frame. p2 points
         to the current pixel in the previous frame. */
      { b = *p1++;
	bd = b-*p2;
	*p2++ = b;
	g = *p1++;
	gd = g-*p2;
	*p2++ = g;
	r = *p1++;
	rd = r-*p2;
	*p2++ = r;
	/* Skip the fourth byte and then the next pixel. */
	p1 += 5;
	/* Convert RGB to HSV. */
	if (r>g)
	{ if (g>b)
	  { v = 100*r/255;
	    s = r==0?0:(100*(r-b)/r);
	    h = s==0?0:(60*(g-b)/(r-b));}
	  else if (r>b)
	  { v = 100*r/255;
	    s = r==0?0:(100*(r-g)/r);
	    h = s==0?0:(60*(g-b)/(r-g));}
	  else
	  { v = 100*b/255;
	    s = b==0?0:(100*(b-g)/b);
	    h = s==0?0:(240+60*(r-g)/(b-g));}}
	else if (r>b)
	{ v = 100*g/255;
	  s = g==0?0:(100*(g-b)/g);
	  h = s==0?0:(120+60*(b-r)/(g-b));}
	else if (g>b)
	{ v = 100*g/255;
	  s = g==0?0:(100*(g-r)/g);
	  h = s==0?0:(120+60*(b-r)/(g-r));}
	else
	{ v = 100*b/255;
	  s = b==0?0:(100*(b-r)/b);
	  h = s==0?0:(240+60*(r-g)/(b-r));}
	if (h<0) h += 360;
	/* Classify each pixel. The classification of the pixel y,x is in the
           matrix code[y/2][x/2]. p3 points to the current pixel in this
           matrix. The codes are Moving=0, Red=1, Green=2, Blue=3, Yellow=4,
           other=255. If a pixel can be more than one code, it is assigned a
           unique code in the following priority: Red, Green, Blue, Yellow,
           Moving, other. Note that due to subsampling, only even rows and
           columns are stored in code. */
	if ((h<=meteor_rhl||h>=meteor_rhh)&&
	    s>=meteor_rsl&&s<=meteor_rsh&&
	    v>=meteor_rvl&&v<=meteor_rvh)
	{ *p3 = 1;}
	else if (h>=meteor_ghl&&h<=meteor_ghh&&
		 s>=meteor_gsl&&s<=meteor_gsh&&
		 v>=meteor_gvl&&v<=meteor_gvh)
	{ *p3 = 2;}
	else if (h>=meteor_bhl&&h<=meteor_bhh&&
		 s>=meteor_bsl&&s<=meteor_bsh&&
		 v>=meteor_bvl&&v<=meteor_bvh)
	{ *p3 = 3;}
	else if (h>=meteor_yhl&&h<=meteor_yhh&&
		 s>=meteor_ysl&&s<=meteor_ysh&&
		 v>=meteor_yvl&&v<=meteor_yvh)
	{ *p3 = 4;}
	else if (!meteor_starting&&rd*rd+gd*gd+bd*bd>=t) *p3 = 0;
	else *p3 = 255;}}
    /* Blanking */
    /* Blanking is an attempt to deal with acretion/deletion. */
    /* This blanks moving pixels that are near coloured pixels. */
    for (y = 0; y<height/2; y++)
    { for (x = 0; x<width/2; x++)
      if (code[y][x]>=1&&code[y][x]<=4)
      { for (y1 = MAX(y-blank, 0); y1<=y+blank&&y1<height/2; y1++)
	{ for (x1 = MAX(x-blank, 0); x1<=x+blank&&x1<width/2; x1++)
	  { if (code[y1][x1]==0) code[y1][x1] = 255;}}}}
    /* This blanks moving pixels that were coloured in the previous frame. */
    if (!meteor_starting)
    { for (y = 0; y<height/2; y++)
      { for (x = 0; x<width/2; x++)
	if (meteor_last_code[y][x]>=1&&meteor_last_code[y][x]<=4)
	{ for (y1 = MAX(y-blank, 0); y1<=y+blank&&y1<height/2; y1++)
	  { for (x1 = MAX(x-blank, 0); x1<=x+blank&&x1<width/2; x1++)
	    { if (code[y1][x1]==0) code[y1][x1] = 255;}}}}}
    /* Conjuring */
    for (p = 0, y = 0; y<height/2; y++)
    { for (x = 0; x<width/2; x++) p += (code[y][x]==0);}
    if (p>500&&!meteor_starting)
    { memcpy(&meteor_conjure_code[0][0], &code[0][0], (height/2)*(width/2));
      meteor_conjuring = 5;}
    else if (meteor_conjuring>0)
    { meteor_conjuring--;
      for (y = 0; y<height/2; y++)
      { for (x = 0; x<width/2; x++)
	{ if (meteor_conjure_code[y][x]==0) code[y][x] = 0;}}}
    meteor_starting = FALSE;
    /* This is the proximity clusterer. For each code (Moving, Red, Green,
       Blue, Yellow) form a graph whose vertices are the pixels of that code
       and whose edges connect vertices that are no more than
       proximity_threshold pixels apart both vertically and horizontally.
       There is a separate proximity_threshold for moving and coloured pixels.
       Compute the connected components of each graph. Discard connected
       components with less than pixel_mass_threshold pixels. There is a
       separate pixel_mass_threshold for moving and coloured pixels. Each
       resulting connected component is an object and gets a unique index. The
       index of the object at pixel y,x is in the matrix cluster[y/2][x/2]. It
       is 254 or 255 if the pixel does not belong to any object. Note that due
       to subsampling, only even rows and columns are stored in cluster. The
       vector colour[k] gives the code, i.e. Moving=0, Red=1, Green=2, Blue=3,
       Yellow=4, for object k. */
    colour[0] = 0;
    k = 0;
    for (y = 0; y<height/2; y++)
    { for (x = 0; x<width/2; x++) cluster[y][x] = 255;}
    for (i = 0; i<5; i++)
    { proximity_threshold = (i==0?
			     motion_proximity_threshold:
			     colour_proximity_threshold)/2;
      for (y = 0; y<height/2; y++)
      { for (x = 0; x<width/2; x++)
	{ if (code[y][x]==i&&cluster[y][x]==255)
	  { cluster[y][x] = k;
	    count = 1;
	    back_y[y][x] = -1;
	    back_x[y][x] = -1;
	    yp = y;
	    xp = x;
	    yl = y;
	    xl = x;
	    yh = y;
	    xh = x;
	    while (yp!=-1)
	    { y1 = yp;
	      x1 = xp;
	      yp = back_y[y1][x1];
	      xp = back_x[y1][x1];
	      if (cluster[y1][x1]==k)
	      { for (y2 = MAX(0, y1-proximity_threshold);
		     y2<height/2&&y2<=y1+proximity_threshold;
		     y2++)
		{ for (x2 = MAX(0, x1-proximity_threshold);
		       x2<width/2&&x2<=x1+proximity_threshold;
		       x2++)
		  { if (code[y2][x2]==i&&cluster[y2][x2]==255)
		    { cluster[y2][x2] = k;
		      count++;
		      back_y[y2][x2] = yp;
		      back_x[y2][x2] = xp;
		      yp = y2;
		      xp = x2;
		      yl = MIN(yl, y2);
		      xl = MIN(xl, x2);
		      yh = MAX(yh, y2);
		      xh = MAX(xh, x2);}}}}}
	    if (count>=(i==0?
			motion_pixel_mass_threshold:
			colour_pixel_mass_threshold))
	    { if (k>=MAX_OBJECTS) panic("Too many objects");
	      colour[k] = i;
	      k++;}
	      else
	      { for (y1 = yl; y1<=yh; y1++)
		{ for (x1 = xl; x1<=xh; x1++)
		  if (cluster[y1][x1]==k)
		  { cluster[y1][x1] = 254;
		    code[y1][x1] = 255;}}}}}}}
    memcpy(&meteor_last_code[0][0], &code[0][0], (height/2)*(width/2));
    if (capturing_code)
    { memcpy(&meteor_code_frames[frame*(height/2)*(width/2)],
	     &code[0][0],
	     (height/2)*(width/2));}}
  if (tracking_ellipses)
  /* For each object k, set mu_x[k] and mu_y[k] to the center of mass of the
     object and set sigma_xx[k], sigma_xy[k], and sigma_yy[k] to the variances
     of the pixel coordinates of object k. */
  { i = k;
    gmotion[frame] = FALSE;
    gobject_k[frame] = k;
    for (k = 0; k<i; k++)
    { j[k] = 0;
      mu_x[k] = 0.0;
      mu_y[k] = 0.0;
      sigma_xx[k] = 0.0;
      sigma_xy[k] = 0.0;
      sigma_yy[k] = 0.0;}
    for (y = 0, p3 = &cluster[0][0]; y<height/2; y++)
    { for (x = 0; x<width/2; x++, p3++)
      { k = *p3;
	if (k<254)
	{ j[k]++;
	  mu_x[k] += 2*x;
	  mu_y[k] += 2*y;}}}
    for (k = 0; k<i; k++)
    { mu_x[k] /= j[k];
      mu_y[k] /= j[k];}
    for (y = 0, p3 = &cluster[0][0]; y<height/2; y++)
    { for (x = 0; x<width/2; x++, p3++)
      { k = *p3;
	if (k<254)
	{ sigma_xx[k] += (2*x-mu_x[k])*(2*x-mu_x[k]);
	  sigma_xy[k] += (2*x-mu_x[k])*(2*y-mu_y[k]);
	  sigma_yy[k] += (2*y-mu_y[k])*(2*y-mu_y[k]);}}}
    /* In frame i, for each object k, set gobject_colour[i][k] to the object's
       code, set gellipse_x[i][k] and gellipse_y[i][k] to the object's center
       of mass, set gellipse_theta[i][k] to the orientation of the object's
       major axis, set gellipse_r[i][k] and gellipse_s[i][k] to the lengths
       object's major and minor axes, and set gmotion[i] to TRUE iff there is
       a moving object in that frame. */
    for (k = 0; k<i; k++)
    { sigma_xx[k] /= j[k];
      sigma_xy[k] /= j[k];
      sigma_yy[k] /= j[k];
      gobject_colour[frame][k] = colour[k];
      gellipse_x[frame][k] = mu_x[k];
      gellipse_y[frame][k] = mu_y[k];
      gellipse_theta[frame][k] = eigenvector_angle1(sigma_xx[k],
						    sigma_xy[k],
						    sigma_xy[k],
						    sigma_yy[k]);
      gellipse_r[frame][k] = sqrt(2.0*eigenvalue1(sigma_xx[k],
						  sigma_xy[k],
						  sigma_xy[k],
						  sigma_yy[k]));
      gellipse_s[frame][k] = sqrt(2.0*eigenvalue2(sigma_xx[k],
						  sigma_xy[k],
						  sigma_xy[k],
						  sigma_yy[k]));
      if (colour[k]==0) gmotion[frame] = TRUE;}}
  if (tracking_polygons)
  { i = k;
    gmotion[frame] = FALSE;
    gobject_k[frame] = k;
    for (k = 0; k<i; k++)
    { gobject_colour[frame][k] = colour[k];
      if (hull)
      /* Compute the convex hull of object k. The top of the object is ymin
         and the bottom of the object is ymax. For each even scan line y
         xmin[y/2] is the leftmost pixel in the object and xmax[y/2] is the
         rightmost pixel. Due to subsampling, only even columns are stored in
         xmin and xmax. */
      { ymin = height;
	ymax = -1;
	for (y = 0; y<height/2; y++)
	{ xmin[y] = width;
	  xmax[y] = -1;
	  for (x = 0; x<width/2; x++)
	  { if (cluster[y][x]==k)
	    { ymin = MIN(ymin, 2*y);
	      ymax = MAX(ymax, 2*y);
	      xmin[y] = MIN(xmin[y], 2*x);
	      xmax[y] = MAX(xmax[y], 2*x);}}}
	l = 0;
	/* Go from top to bottom down the right side of the object, splicing
           out concavities. */
	for (ybest = ymin/2; ybest<=ymax/2&&xmax[ybest]==-1; ybest++);
	while (ybest<=ymax/2&&ybest!=-1)
	{ if (l>=MAX_VERTICES) panic("Too many vertices");
	  gpolygon_y[frame][k][l] = 2*ybest;
	  gpolygon_x[frame][k][l] = xmax[ybest];
	  l++;
	  p = FALSE;
	  for (y = gpolygon_y[frame][k][l-1]/2+1; y<=ymax/2; y++)
	  { if (xmax[y]!=-1&&xmax[y]>=gpolygon_x[frame][k][l-1]) p = TRUE;}
	  ybest = -1;
	  for (y = gpolygon_y[frame][k][l-1]/2+1; y<=ymax/2; y++)
	  { if (xmax[y]!=-1&&
		(!p||xmax[y]>=gpolygon_x[frame][k][l-1])&&
		(ybest==-1||
		 (gpolygon_x[frame][k][l-1]-xmax[ybest])*
		 (gpolygon_y[frame][k][l-1]-2*y)<
		 (gpolygon_x[frame][k][l-1]-xmax[y])*
		 (gpolygon_y[frame][k][l-1]-2*ybest)))
	    { ybest = y;}}}
	/* Go from bottom to top up the left side of the object, splicing out
           concavities. */
	for (ybest = ymax/2; ybest>=ymin/2&&xmin[ybest]==-1; ybest--);
	while (ybest>=ymin/2&&ybest!=-1)
	{ if (l>=MAX_VERTICES) panic("Too many vertices");
	  gpolygon_y[frame][k][l] = 2*ybest;
	  gpolygon_x[frame][k][l] = xmin[ybest];
	  l++;
	  p = FALSE;
	  for (y = gpolygon_y[frame][k][l-1]/2-1; y>=ymin/2; y--)
	  { if (xmin[y]!=-1&&xmin[y]<=gpolygon_x[frame][k][l-1]) p = TRUE;}
	  ybest = -1;
	  for (y = gpolygon_y[frame][k][l-1]/2-1; y>=ymin/2; y--)
	  { if (xmin[y]!=-1&&
		(!p||xmin[y]<=gpolygon_x[frame][k][l-1])&&
		(ybest==-1||
		 (xmin[ybest]-gpolygon_x[frame][k][l-1])*
		 (gpolygon_y[frame][k][l-1]-2*y)>
		 (xmin[y]-gpolygon_x[frame][k][l-1])*
		 (gpolygon_y[frame][k][l-1]-2*ybest)))
	    { ybest = y;}}}}
      else
      /* Compute the contour of the object. */
      { proximity_threshold = (colour[k]==0?
			       motion_proximity_threshold:
			       colour_proximity_threshold)/2;
	/* Copy the pixels from cluster that correspond to object k into map.
           */
	for (y = 0; y<height/2; y++)
	{ for (x = 0; x<width/2; x++) map[y][x] = (cluster[y][x]==k);}
	/* Bloat map by the proximity threshold. If there are two pixels in
           map that are separated by less than proximity_threshold both
           vertically and horizontally then fill in the bounding rectangle of
           those two pixels. */
	for (y = 0; y<height/2; y++)
	{ for (x = 0; x<width/2; x++)
	  { if (cluster[y][x]==k)
	    { for (y1 = MAX(0, y-proximity_threshold);
		   y1<height/2&&y1<=y+proximity_threshold;
		   y1++)
	      { for (x1 = MAX(0, x-proximity_threshold);
		     x1<width/2&&x1<=x+proximity_threshold;
		     x1++)
		{ if (cluster[y1][x1]==k)
		  { for (y2 = MIN(y, y1); y2<=MAX(y, y1); y2++)
		    { for (x2 = MIN(x, x1); x2<=MAX(x, x1); x2++)
		      { map[y2][x2] = TRUE;}}}}}}}}
	/* This is a contour follower. */
	for (y = 0; y<height/2; y++)
	{ for (x = 0; x<width/2; x++)
	  { if (map[y][x])
	    { gpolygon_y[frame][k][0] = 2*y;
	      gpolygon_x[frame][k][0] = 2*x;
	      goto l1;}}}
	panic("This shouldn't happen");
	l1:
	if (gpolygon_x[frame][k][0]<width-2&&
	    map[gpolygon_y[frame][k][0]/2]
	    [gpolygon_x[frame][k][0]/2+1])
	{ gpolygon_y[frame][k][1] = gpolygon_y[frame][k][0];
	  gpolygon_x[frame][k][1] = gpolygon_x[frame][k][0]+2;
	  state = 0;}
	else if (gpolygon_y[frame][k][0]<height-2&&
		 map[gpolygon_y[frame][k][0]/2+1]
		 [gpolygon_x[frame][k][0]/2])
	{ gpolygon_y[frame][k][1] = gpolygon_y[frame][k][0]+2;
	  gpolygon_x[frame][k][1] = gpolygon_x[frame][k][0];
	  state = 3;}
	else panic("This shouldn't happen");
	l = 2;
	l2:
	if (gpolygon_y[frame][k][l-1]!=
	    gpolygon_y[frame][k][0]||
	    gpolygon_x[frame][k][l-1]!=
	    gpolygon_x[frame][k][0])
	{ if (l>=MAX_VERTICES) panic("Too many vertices");
	  switch (state)
	  { case 0:
	    if (gpolygon_y[frame][k][l-1]>0&&
		map[gpolygon_y[frame][k][l-1]/2-1]
		[gpolygon_x[frame][k][l-1]/2])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1]-2;
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1];
	      state = 1;
	      l++;
	      goto l2;}
	    if (gpolygon_x[frame][k][l-1]<width-2&&
		map[gpolygon_y[frame][k][l-1]/2]
		[gpolygon_x[frame][k][l-1]/2+1])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1];
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1]+2;
	      state = 0;
	      l++;
	      goto l2;}
	    if (gpolygon_y[frame][k][l-1]<height-2&&
		map[gpolygon_y[frame][k][l-1]/2+1]
		[gpolygon_x[frame][k][l-1]/2])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1]+2;
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1];
	      state = 3;
	      l++;
	      goto l2;}
	    if (gpolygon_x[frame][k][l-1]>0&&
		map[gpolygon_y[frame][k][l-1]/2]
		[gpolygon_x[frame][k][l-1]/2-1])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1];
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1]-2;
	      state = 2;
	      l++;
	      goto l2;}
	    panic("This shouldn't happen");
	    case 1:
	    if (gpolygon_x[frame][k][l-1]>0&&
		map[gpolygon_y[frame][k][l-1]/2]
		[gpolygon_x[frame][k][l-1]/2-1])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1];
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1]-2;
	      state = 2;
	      l++;
	      goto l2;}
	    if (gpolygon_y[frame][k][l-1]>0&&
		map[gpolygon_y[frame][k][l-1]/2-1]
		[gpolygon_x[frame][k][l-1]/2])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1]-2;
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1];
	      state = 1;
	      l++;
	      goto l2;}
	    if (gpolygon_x[frame][k][l-1]<width-2&&
		map[gpolygon_y[frame][k][l-1]/2]
		[gpolygon_x[frame][k][l-1]/2+1])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1];
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1]+2;
	      state = 0;
	      l++;
	      goto l2;}
	    if (gpolygon_y[frame][k][l-1]<height-2&&
		map[gpolygon_y[frame][k][l-1]/2+1]
		[gpolygon_x[frame][k][l-1]/2])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1]+2;
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1];
	      state = 3;
	      l++;
	      goto l2;}
	    panic("This shouldn't happen");
	    case 2:
	    if (gpolygon_y[frame][k][l-1]<height-2&&
		map[gpolygon_y[frame][k][l-1]/2+1]
		[gpolygon_x[frame][k][l-1]/2])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1]+2;
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1];
	      state = 3;
	      l++;
	      goto l2;}
	    if (gpolygon_x[frame][k][l-1]>0&&
		map[gpolygon_y[frame][k][l-1]/2]
		[gpolygon_x[frame][k][l-1]/2-1])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1];
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1]-2;
	      state = 2;
	      l++;
	      goto l2;}
	    if (gpolygon_y[frame][k][l-1]>0&&
		map[gpolygon_y[frame][k][l-1]/2-1]
		[gpolygon_x[frame][k][l-1]/2])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1]-2;
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1];
	      state = 1;
	      l++;
	      goto l2;}
	    if (gpolygon_x[frame][k][l-1]<width-2&&
		map[gpolygon_y[frame][k][l-1]/2]
		[gpolygon_x[frame][k][l-1]/2+1])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1];
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1]+2;
	      state = 0;
	      l++;
	      goto l2;}
	    panic("This shouldn't happen");
	    case 3:
	    if (gpolygon_x[frame][k][l-1]<width-2&&
		map[gpolygon_y[frame][k][l-1]/2]
		[gpolygon_x[frame][k][l-1]/2+1])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1];
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1]+2;
	      state = 0;
	      l++;
	      goto l2;}
	    if (gpolygon_y[frame][k][l-1]<height-2&&
		map[gpolygon_y[frame][k][l-1]/2+1]
		[gpolygon_x[frame][k][l-1]/2])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1]+2;
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1];
	      state = 3;
	      l++;
	      goto l2;}
	    if (gpolygon_x[frame][k][l-1]>0&&
		map[gpolygon_y[frame][k][l-1]/2]
		[gpolygon_x[frame][k][l-1]/2-1])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1];
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1]-2;
	      state = 2;
	      l++;
	      goto l2;}
	    if (gpolygon_y[frame][k][l-1]>0&&
		map[gpolygon_y[frame][k][l-1]/2-1]
		[gpolygon_x[frame][k][l-1]/2])
	    { gpolygon_y[frame][k][l] =
	      gpolygon_y[frame][k][l-1]-2;
	      gpolygon_x[frame][k][l] =
	      gpolygon_x[frame][k][l-1];
	      state = 1;
	      l++;
	      goto l2;}
	    panic("This shouldn't happen");}}}
      gpolygon_l[frame][k] = l;
      if (colour[k]==0) gmotion[frame] = TRUE;}}}

void meteor_handler(int signum)
{ unsigned char *p1, *p2, *pl;
  int y, k, l;
  unsigned char buffer[meteor_height*meteor_width*4];
  /* This is an extra copy. This copy is used three times. First to copy to
     meteor_frames. Second to run the tracker. Third to display. This extra
     copy is made since the actual meteor buffer can change behind your back
     between different accesses. */
  memcpy(&buffer[0],
	 meteor_buffer+meteor_offset.frame_offset[0],
	 meteor_height*meteor_width*4);
  /* meteor_l is the capture downsample factor. If meteor_l==0 then capturing
     is turned off. If meteor_l!=0 then the capture rate is 30Hz/meteor_l. */
  if (meteor_l>0)
  /* meteor_k does the downsampling for capturing. It counts from
     0,...,meteor_l-1 and cycles. Capturing is done when meteor_k==0. */
  { meteor_k = (meteor_k+1)%meteor_l;
    if (meteor_k==0)
    /* meteor_i is the index of the current frame. It is used to index into
       the saved frames as well as the saved tracking data. */
    { if (meteor_i<meteor_n)
      { meteor_tracker(&buffer[0],
		       meteor_i,
		       meteor_height,
		       meteor_width,
		       meteor_uninterlaced,
		       meteor_capturing,
		       meteor_capturing,
		       meteor_tracking_ellipses,
		       meteor_tracking_polygons,
		       meteor_hull,
		       meteor_motion_threshold,
		       meteor_colour_pixel_mass_threshold,
		       meteor_colour_proximity_threshold,
		       meteor_motion_pixel_mass_threshold,
		       meteor_motion_proximity_threshold);
	/* When viewing just the ellipses or polygons, and not the video
	   (since meteor_m==0), draw the ellipses or polygons in black. */
	if ((meteor_tracking_ellipses||meteor_tracking_polygons)&&meteor_m==0)
	{ XClearArea(meteor_display, meteor_window, 0, 0,
		     meteor_width, meteor_height, False);
	  XSetForeground(meteor_display, meteor_gc,
			 XBlackPixel(meteor_display, meteor_screen));
	  /* Draw the ellipses. */
	  if (meteor_tracking_ellipses)
	  { for (k = 0; k<gobject_k[meteor_i]; k++)
	    { meteor_draw_ellipse(gellipse_x[meteor_i][k],
				  gellipse_y[meteor_i][k],
				  gellipse_theta[meteor_i][k],
				  gellipse_r[meteor_i][k],
				  gellipse_s[meteor_i][k]);}}
	  /* Draw the polygons. */
	  if (meteor_tracking_polygons)
	  { for (k = 0; k<gobject_k[meteor_i]; k++)
	    { for (l = 1; l<gpolygon_l[meteor_i][k]; l++)
	      { XDrawLine(meteor_display,
			  meteor_window,
			  meteor_gc,
			  gpolygon_x[meteor_i][k][l-1],
			  gpolygon_y[meteor_i][k][l-1],
			  gpolygon_x[meteor_i][k][l],
			  gpolygon_y[meteor_i][k][l]);}
	      XDrawLine(meteor_display,
			meteor_window,
			meteor_gc,
			gpolygon_x[meteor_i][k][gpolygon_l[meteor_i][k]-1],
			gpolygon_y[meteor_i][k][gpolygon_l[meteor_i][k]-1],
			gpolygon_x[meteor_i][k][0],
			gpolygon_y[meteor_i][k][0]);}}
	  XFlush(meteor_display);}
	meteor_i++;}
      /* Stop capturing when meteor_i>=meteor_n. */
      else meteor_l = 0;}}
  /* meteor_m is the viewfinder refresh downsample factor. If meteor_m==0 then
     the viewfinder is turned off. If meteor_m!=0 then the viewfinder refresh
     rate is 30Hz/meteor_m. */
  if (meteor_m>0)
  /* meteor_j does the downsampling for the viewfinder refresh. It counts from
     0,...,meteor_m-1 and cycles. Viewfinder refresh is done when
     meteor_j==0. */
  { meteor_j = (meteor_j+1)%meteor_m;
    if (meteor_j==0)
    /* Display the current video frame. The meteor buffer has four bytes per
       pixel. Depending on the X server, shmimage might have either three or
       four bytes per pixel. If it has four bytes per pixel then use memcpy
       since it is faster. If it has three bytes per pixel then we need to
       roll our own to do packing. */
    { if (meteor_shmimage->bits_per_pixel==24)
      { for (p1 = &buffer[0],
	     p2 = &meteor_shmimage->data[0],
	     pl = &buffer[meteor_height*meteor_width*4];
	     p1<pl;)
	{ *p2++ = *p1++;
	  *p2++ = *p1++;
	  *p2++ = *p1++;
	  p1++;}}
      else memcpy(&meteor_shmimage->data[0],
		  &buffer[0],
		  meteor_height*meteor_width*4);
      XShmPutImage(meteor_display, meteor_window, meteor_gc, meteor_shmimage,
		   0, 0, 0, 0,
		   MIN(meteor_width, meteor_window_width),
		   MIN(meteor_height, meteor_window_height),
		   False);
      /* When viewing the ellipses or polygons overlayed on the video
	 (since meteor_m>0), draw the ellipses or polygons in white. */
      if (meteor_tracking_ellipses||meteor_tracking_polygons)
      { XSetForeground(meteor_display, meteor_gc,
		       XWhitePixel(meteor_display, meteor_screen));
	/* Draw the ellipses. */
	if (meteor_tracking_ellipses)
	{ for (k = 0; k<gobject_k[meteor_i-1]; k++)
	  { meteor_draw_ellipse(gellipse_x[meteor_i-1][k],
				gellipse_y[meteor_i-1][k],
				gellipse_theta[meteor_i-1][k],
				gellipse_r[meteor_i-1][k],
				gellipse_s[meteor_i-1][k]);}}
	/* Draw the polygons. */
	if (meteor_tracking_polygons)
	{ for (k = 0; k<gobject_k[meteor_i-1]; k++)
	  { for (l = 1; l<gpolygon_l[meteor_i-1][k]; l++)
	    { XDrawLine(meteor_display,
			meteor_window,
			meteor_gc,
			gpolygon_x[meteor_i-1][k][l-1],
			gpolygon_y[meteor_i-1][k][l-1],
			gpolygon_x[meteor_i-1][k][l],
			gpolygon_y[meteor_i-1][k][l]);}
	    XDrawLine(meteor_display,
		      meteor_window,
		      meteor_gc,
		      gpolygon_x[meteor_i-1][k][gpolygon_l[meteor_i-1][k]-1],
		      gpolygon_y[meteor_i-1][k][gpolygon_l[meteor_i-1][k]-1],
		      gpolygon_x[meteor_i-1][k][0],
		      gpolygon_y[meteor_i-1][k][0]);}}}
      XFlush(meteor_display);}}}

void meteor_offline_tracker(int frame,
			    int height,
			    int width,
			    int uninterlaced,
			    int capturing_code,
			    int tracking_ellipses,
			    int tracking_polygons,
			    int hull,
			    int motion_threshold,
			    int colour_pixel_mass_threshold,
			    int colour_proximity_threshold,
			    int motion_pixel_mass_threshold,
			    int motion_proximity_threshold)
{ unsigned char *p1, *p2, *pl;
  int y;
  unsigned char buffer[height*width*4];
  if (uninterlaced)
  /* When uninterlaced, only the even scan lines are stored to save space.
     The meteor buffer has four bytes per pixel and meteor_frames has three
     bytes per pixel (to save space) so packing is done. Two loops are
     needed to allow skipping the odd scan lines. */
  { for (y = 0; y<height/2; y++)
    { for (p1 = &buffer[2*y*width*4],
	   p2 = &meteor_frames[(frame*height/2+y)*width*3],
	   pl = &buffer[(2*y+1)*width*4];
	   p1<pl;)
      { *p1++ = *p2++;
	*p1++ = *p2++;
	*p1++ = *p2++;
	p1++;}}}
    else
    /* When interlaced, all scan lines are stored. The meteor buffer has four
       bytes per pixel and meteor_frames has three bytes per pixel (to save
       space). In this case, only one loop is needed. */
    { for (p1 = &buffer[0],
	   p2 = &meteor_frames[frame*height*width*3],
	   pl = &buffer[height*width*4];
	   p1<pl;)
      { *p1++ = *p2++;
	*p1++ = *p2++;
	*p1++ = *p2++;
	p1++;}}
  meteor_tracker(&buffer[0],
		 frame,
		 height,
		 width,
		 uninterlaced,
		 FALSE,
		 capturing_code,
		 tracking_ellipses,
		 tracking_polygons,
		 hull,
		 motion_threshold,
		 colour_pixel_mass_threshold,
		 colour_proximity_threshold,
		 motion_pixel_mass_threshold,
		 motion_proximity_threshold);}

void meteor_draw_ellipse(float x0, float y0, float t0, float a, float b)
{ float rxx = cos(t0);
  float rxy = -sin(t0);
  float ryx = -rxy;
  float ryy = rxx;
  float ex, ey, tx, ty, px, py;
  int i;
  for (i = 0; i<37; i++)
  { ex = a*sin(DEG_TO_RAD(10*i));
    ey = b*cos(DEG_TO_RAD(10*i));
    tx = rxx*ex+rxy*ey+x0;
    ty = ryx*ex+ryy*ey+y0;
    if (i!=0)
    { XDrawLine(meteor_display, meteor_window, meteor_gc, tx, ty, px, py);}
    px = tx;
    py = ty;}}

void meteor_set_real_time(void)
{ struct sched_param sched_param;
  return;			/* debugging */
  if (meteor_real_time==FALSE)
  { if (meteor_uid=-1) meteor_uid = getuid();
    if (setuid(0)==-1) perror("meteor setuid failed");
    sched_param.sched_priority = 1;
    if (sched_setscheduler(0, SCHED_FIFO, &sched_param)==-1)
    { perror("meteor sched_setscheduler SCHED_FIFO failed");}
    if (mlockall(MCL_FUTURE)==-1) perror("meteor mlockall failed");
    if (setreuid(0, meteor_uid)==-1) perror("meteor setreuid failed");
    meteor_real_time = TRUE;}}

void meteor_set_nonreal_time(void)
{ struct sched_param sched_param;
  if (meteor_real_time==TRUE)
  { if (meteor_uid=-1) meteor_uid = getuid();
    if (setuid(0)==-1) perror("meteor setuid failed");
    sched_param.sched_priority = 0;
    if (sched_setscheduler(0, SCHED_OTHER, &sched_param)==-1)
    { perror("meteor sched_setscheduler SCHED_OTHER failed");}
    if (munlockall()==-1) perror("meteor munlockall failed");
    if (setreuid(0, meteor_uid)==-1) perror("meteor setreuid failed");
    meteor_real_time = FALSE;}}

void meteor_open(void)
{ if ((meteor_port = open("/dev/mmetfgrab0", O_RDONLY))==-1)
  { perror("meteor open failed");
    exit(1);}}

void meteor_close(void)
{ if (close(meteor_port)==-1)
  { perror("meteor close failed");
    exit(1);}}

void meteor_sigaction(void)
{ struct sigaction sigact;
  sigact.sa_handler = meteor_handler;
  sigemptyset(&sigact.sa_mask);
  sigact.sa_flags = SA_RESTART;
  if (sigaction(SIGUSR2, &sigact, NULL))
  { perror("meteor sigaction failed");
    exit(1);}}

void meteor_set_geometry(int height, int width, int frames, int oformat,
			 int uninterlaced)
{ struct meteor_geomet geomet;
  meteor_height = height;
  meteor_width = width;
  meteor_uninterlaced = uninterlaced;
  meteor_deinterlaced = 0;
  geomet.rows = height;
  geomet.columns = width;
  geomet.frames = frames;
  geomet.oformat = oformat;
  if (ioctl(meteor_port, METEORSETGEO, &geomet)==-1)
  { perror("METEORSETGEO ioctl failed");
    exit(1);}}

void meteor_set_video_format(int format)
{ if (ioctl(meteor_port, METEORSFMT, &format)==-1)
  { perror("METEORSFMT ioctl failed");
    exit(1);}}

void meteor_set_input_device(int device)
{ if (ioctl(meteor_port, METEORSINPUT, &device)==-1)
  { perror("METEORSINPUT ioctl failed");
    exit(1);}}

void meteor_get_frame_offset(struct meteor_frame_offset *offset)
{ if (ioctl(meteor_port, METEORGFROFF, offset)==-1)
  { perror("METEORGFROFF ioctl failed");
    exit(1);}}

void meteor_set_signal(int signal)
{ if (ioctl(meteor_port, METEORSSIGNAL, &signal)==-1)
  { perror("METEORSSIGNAL ioctl failed");
    exit(1);}}

void meteor_block(void)
{ sigset_t set;
  sigemptyset(&set);
  sigaddset(&set, SIGUSR2);
  sigprocmask(SIG_BLOCK, &set, NULL);}

void meteor_unblock(void)
{ sigset_t set;
  sigemptyset(&set);
  sigaddset(&set, SIGUSR2);
  sigprocmask(SIG_UNBLOCK, &set, NULL);}

void meteor_capture(int cap)
{ if (ioctl(meteor_port, METEORCAPTUR, &cap)==-1)
  { perror("METEORCAPTUR ioctl failed");
    exit(1);}}

void meteor_capframe(struct meteor_capframe *capframe)
{ if (ioctl(meteor_port, METEORCAPFRM, capframe)==-1)
  { perror("METEORCAPFRM ioctl failed");
    exit(1);}}

void meteor_mmap(void)
{ meteor_buffer = (unsigned char *)mmap((unsigned char *)0,
					meteor_offset.fb_size,
					PROT_READ, MAP_FILE|MAP_SHARED,
					meteor_port, (off_t)0);
  if (meteor_buffer==(unsigned char *)-1)
  { perror("meteor mmap failed");
    exit(1);}}

void meteor_munmap(void)
{ if (munmap(meteor_buffer, meteor_height*meteor_width*4)==-1)
  { perror("meteor munmap failed");
    exit(1);}}

void meteor_malloc(int nframes)
{ int y;
  meteor_frames =
  (unsigned char *)
  malloc(nframes*(meteor_uninterlaced?(meteor_height/2):meteor_height)*
	 meteor_width*3*sizeof(unsigned char));
  if (meteor_frames==NULL)
  { fprintf(stderr, "meteor malloc failed\n");
    exit(1);}
  meteor_code_frames =
  (unsigned char *)
  malloc(nframes*(meteor_height/2)*(meteor_width/2)*sizeof(unsigned char));
  if (meteor_code_frames==NULL)
  { fprintf(stderr, "meteor malloc failed\n");
    exit(1);}
  meteor_last_frame =
  (unsigned char *)
  malloc((meteor_height/2)*meteor_width*3*sizeof(unsigned char));
  if (meteor_last_frame==NULL)
  { fprintf(stderr, "meteor malloc failed\n");
    exit(1);}
  /* This is non-ANSI */
  meteor_last_code =
  (unsigned char **)
  malloc((meteor_height/2)*sizeof(unsigned char *)+
	 ((meteor_height/2)*(meteor_width/2)*sizeof(unsigned char)));
  if (meteor_last_code==NULL)
  { fprintf(stderr, "meteor malloc failed\n");
    exit(1);}
  for (y = 0; y<meteor_height/2; y++)
  { meteor_last_code[y] =
    &((unsigned char *)
      &meteor_last_code[meteor_height/2])[y*meteor_width/2];}
  /* This is non-ANSI */
  meteor_conjure_code =
  (unsigned char **)
  malloc((meteor_height/2)*sizeof(unsigned char *)+
	 ((meteor_height/2)*(meteor_width/2)*sizeof(unsigned char)));
  if (meteor_conjure_code==NULL)
  { fprintf(stderr, "meteor malloc failed\n");
    exit(1);}
  for (y = 0; y<meteor_height/2; y++)
  { meteor_conjure_code[y] =
    &((unsigned char *)
      &meteor_conjure_code[meteor_height/2])[y*meteor_width/2];}}

void meteor_page_in_frames(int iframes, int nframes)
{ int i = iframes*(meteor_uninterlaced?(meteor_height/2):meteor_height)*
          meteor_width*3*sizeof(unsigned char);
  int n = nframes*(meteor_uninterlaced?(meteor_height/2):meteor_height)*
          meteor_width*3*sizeof(unsigned char);
  for (; i<n; i++) meteor_frames[i]++;}

void meteor_set_deinterlaced(int deinterlaced)
{ meteor_deinterlaced = deinterlaced;}

int meteor_get_red(int i, int y, int x)
{ int p;
  if (meteor_uninterlaced) p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
  else p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
  return meteor_frames[p+2];}

int meteor_get_green(int i, int y, int x)
{ int p;
  if (meteor_uninterlaced) p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
  else p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
  return meteor_frames[p+1];}

int meteor_get_blue(int i, int y, int x)
{ int p;
  if (meteor_uninterlaced) p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
  else p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
  return meteor_frames[p];}

int meteor_get_pixel(int i, int y, int x)
{ int p;
  if (meteor_uninterlaced) p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
  else p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
  return 65536*meteor_frames[p+2]+256*meteor_frames[p+1]+meteor_frames[p];}

void meteor_set_red(int i, int y, int x, int v)
{ int p;
  if (meteor_uninterlaced) p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
  else p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
  meteor_frames[p+2] = v;}

void meteor_set_green(int i, int y, int x, int v)
{ int p;
  if (meteor_uninterlaced) p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
  else p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
  meteor_frames[p+1] = v;}

void meteor_set_blue(int i, int y, int x, int v)
{ int p;
  if (meteor_uninterlaced) p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
  else p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
  meteor_frames[p] = v;}

void meteor_set_pixel(int i, int y, int x, int v)
{ int p;
  if (meteor_uninterlaced) p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
  else p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
  meteor_frames[p+2] = v>>16;
  meteor_frames[p+1] = (v>>8)&255;
  meteor_frames[p] = v&255;}

int meteor_get_code_frame_red(int i, int y, int x)
{ int c =  meteor_code_frames[(i*(meteor_height/2)+y/2)*(meteor_width/2)+x/2];
  return (c==0?0:c==1?255:c==2?0:c==3?0:c==4?255:255);}

int meteor_get_code_frame_green(int i, int y, int x)
{ int c =  meteor_code_frames[(i*(meteor_height/2)+y/2)*(meteor_width/2)+x/2];
  return (c==0?0:c==1?0:c==2?255:c==3?0:c==4?255:255);}

int meteor_get_code_frame_blue(int i, int y, int x)
{ int c =  meteor_code_frames[(i*(meteor_height/2)+y/2)*(meteor_width/2)+x/2];
  return (c==0?0:c==1?0:c==2?0:c==3?255:c==4?0:255);}

void meteor_copy_frame_to_ximage(int i, XImage *ximage)
{ int y, x, p;
  for (y = 0; y<meteor_height; y++)
  { for (x = 0; x<meteor_width; x++)
    { if (meteor_uninterlaced)
      p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
      else p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
      XPutPixel(ximage, x, y,
		65536*meteor_frames[p+2]+
		256*meteor_frames[p+1]+
		meteor_frames[p]);}}}

void meteor_motion_detector(int i, XImage *ximage)
{ int y, x, p, q, r, g, b, t = 48*48;
  for (y = 0; y<meteor_height; y++)
  { for (x = 0; x<meteor_width; x++)
    { if (meteor_uninterlaced)
      { p = 3*(((i*meteor_height+y)/2)*meteor_width+x);
	q = 3*((((i+1)*meteor_height+y)/2)*meteor_width+x);}
      else
      { p = 3*((i*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);
	q = 3*(((i+1)*meteor_height+y&~meteor_deinterlaced)*meteor_width+x);}
      r = meteor_frames[q+2]-meteor_frames[p+2];
      g = meteor_frames[q+1]-meteor_frames[p+1];
      b = meteor_frames[q]-meteor_frames[p];
      XPutPixel(ximage, x, y, r*r+g*g+b*b<t?65536*255+256*255+255:0);}}}

void meteor_set_hack_track(int motion_threshold,
			   int colour_pixel_mass_threshold,
			   int colour_proximity_threshold,
			   int motion_pixel_mass_threshold,
			   int motion_proximity_threshold)
{ if (colour_proximity_threshold%2!=0)
  { panic ("colour_proximity_threshold must be a multiple of two");}
  if (motion_proximity_threshold%2!=0)
  { panic ("motion_proximity_threshold must be a multiple of two");}
  meteor_motion_threshold = motion_threshold;
  meteor_colour_pixel_mass_threshold = colour_pixel_mass_threshold;
  meteor_colour_proximity_threshold = colour_proximity_threshold;
  meteor_motion_pixel_mass_threshold = motion_pixel_mass_threshold;
  meteor_motion_proximity_threshold = motion_proximity_threshold;}

void meteor_set_red_hack_track(int hl, int hh,
			       int sl, int sh,
			       int vl, int vh)
{ meteor_rhl = hl;
  meteor_rhh = hh;
  meteor_rsl = sl;
  meteor_rsh = sh;
  meteor_rvl = vl;
  meteor_rvh = vh;}

void meteor_set_green_hack_track(int hl, int hh,
				 int sl, int sh,
				 int vl, int vh)
{ meteor_ghl = hl;
  meteor_ghh = hh;
  meteor_gsl = sl;
  meteor_gsh = sh;
  meteor_gvl = vl;
  meteor_gvh = vh;}

void meteor_set_blue_hack_track(int hl, int hh,
				int sl, int sh,
				int vl, int vh)
{ meteor_bhl = hl;
  meteor_bhh = hh;
  meteor_bsl = sl;
  meteor_bsh = sh;
  meteor_bvl = vl;
  meteor_bvh = vh;}

void meteor_set_yellow_hack_track(int hl, int hh,
				  int sl, int sh,
				  int vl, int vh)
{ meteor_yhl = hl;
  meteor_yhh = hh;
  meteor_ysl = sl;
  meteor_ysh = sh;
  meteor_yvl = vl;
  meteor_yvh = vh;}

void meteor_start_hack_track(int frame)
{ gfirst_frame = frame;
  meteor_first_frame = frame;
  meteor_last_moving_valid = FALSE;
  meteor_last_moving = make_c_int_matrix(meteor_height, meteor_width);}

void meteor_hack_track_step(int frame,
			    int object_frame,
			    int colour_pixel_mass_threshold,
			    int colour_proximity_threshold,
			    int motion_threshold,
			    int motion_pixel_mass_threshold,
			    int motion_proximity_threshold)
{ int i, j, k, x, y, x1, y1, x2, y2, xl, yl, xh, yh, xp, yp, height, width;
  int count, n = 4, no_moving_clusters, once;
  int *colour_pythag, *motion_pythag;
  float mu_x, mu_y, sigma_xx, sigma_xy, sigma_yy;
  int **colour, **moving, **coloured_cluster, **moving_cluster;
  int **back_x, **back_y;
  float r, g, b, h, s, v, m, rc, gc, bc, rd, gd, bd;
  float t = motion_threshold*motion_threshold;
  int p, q, pixel;
  height = meteor_height;
  width = meteor_width;
  /* Allocate storage */
  colour_pythag = make_c_int_vector(colour_proximity_threshold+1);
  motion_pythag = make_c_int_vector(motion_proximity_threshold+1);
  colour = make_c_int_matrix(height, width);
  moving = make_c_int_matrix(height, width);
  coloured_cluster = make_c_int_matrix(height, width);
  moving_cluster = make_c_int_matrix(height, width);
  back_x = make_c_int_matrix(height, width);
  back_y = make_c_int_matrix(height, width);
  /* Set up proximity clusterer */
  for (i = 0; i<=colour_proximity_threshold; i++)
  { colour_pythag[i] = rint(sqrt(SQR(colour_proximity_threshold)-i*i));}
  for (i = 0; i<=motion_proximity_threshold; i++)
  { motion_pythag[i] = rint(sqrt(SQR(motion_proximity_threshold)-i*i));}
  /* Process frame */
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { if (meteor_uninterlaced)
      { p = 3*((((frame-1)*height+y)/2)*width+x);
	q = 3*(((frame*height+y)/2)*width+x);}
      else
      { p = 3*(((frame-1)*height+y&~meteor_deinterlaced)*width+x);
	q = 3*((frame*height+y&~meteor_deinterlaced)*width+x);}
      r = meteor_frames[q+2];
      g = meteor_frames[q+1];
      b = meteor_frames[q];
      if (frame!=meteor_first_frame)
      { rd = r-meteor_frames[p+2];
	gd = g-meteor_frames[p+1];
	bd = b-meteor_frames[p];}
      r = r/255.0;
      g = g/255.0;
      b = b/255.0;
      v = MAX(r, MAX(g, b));
      m = MIN(r, MIN(g, b));
      s = v==0.0?0.0:(v-m)/v;
      if (s==0.0) h = 0.0;
      else
      { rc = (v-r)/(v-m);
	gc = (v-g)/(v-m);
	bc = (v-b)/(v-m);
	h = (r==v?bc-gc:g==v?2.0+rc-bc:4.0+gc-rc)/6.0;
	if (h<0.0) h += 1.0;}
      h *= 360.0;
      s *= 100.0;
      v *= 100.0;
      if ((h<=meteor_rhl||h>=meteor_rhh)&&
	  s>=meteor_rsl&&s<=meteor_rsh&&
	  v>=meteor_rvl&&v<=meteor_rvh)
      { colour[y][x] = 0;
        moving[y][x] = FALSE;}
      else if (h>=meteor_ghl&&h<=meteor_ghh&&
	       s>=meteor_gsl&&s<=meteor_gsh&&
	       v>=meteor_gvl&&v<=meteor_gvh)
      { colour[y][x] = 1;
        moving[y][x] = FALSE;}
      else if (h>=meteor_bhl&&h<=meteor_bhh&&
	       s>=meteor_bsl&&s<=meteor_bsh&&
	       v>=meteor_bvl&&v<=meteor_bvh)
      { colour[y][x] = 2;
        moving[y][x] = FALSE;}
      else if (h>=meteor_yhl&&h<=meteor_yhh&&
	       s>=meteor_ysl&&s<=meteor_ysh&&
	       v>=meteor_yvl&&v<=meteor_yvh)
      { colour[y][x] = 3;
        moving[y][x] = FALSE;}
      else if (frame!=meteor_first_frame&&rd*rd+gd*gd+bd*bd>=t)
      { colour[y][x] = -1;
        moving[y][x] = TRUE;}
      else
      { colour[y][x] = -1;
        moving[y][x] = FALSE;}}}
  /* Cluster coloured objects by proximity, ignoring small clusters */
  k = 0;
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++) coloured_cluster[y][x] = -1;}
  for (i = 0; i<n; i++)
  { for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { if (colour[y][x]==i&&coloured_cluster[y][x]==-1)
	{ coloured_cluster[y][x] = k;
	  count = 1;
	  back_y[y][x] = -1;
	  back_x[y][x] = -1;
	  yp = y;
	  xp = x;
	  yl = y;
	  xl = x;
	  yh = y;
	  xh = x;
	  while (yp!=-1)
	  { y1 = yp;
	    x1 = xp;
	    yp = back_y[y1][x1];
	    xp = back_x[y1][x1];
	    if (coloured_cluster[y1][x1]==k)
	    { for (y2 = MAX(0, y1-colour_proximity_threshold);
		   y2<MIN(height, y1+colour_proximity_threshold+1);
		   y2++)
	      { q = colour_pythag[abs(y2-y1)];
		for (x2 = MAX(0, x1-q); x2<MIN(width, x1+q+1); x2++)
		{ if (colour[y2][x2]==i&&coloured_cluster[y2][x2]==-1)
		  { coloured_cluster[y2][x2] = k;
		    count++;
		    back_y[y2][x2] = yp;
		    back_x[y2][x2] = xp;
		    yp = y2;
		    xp = x2;
		    yl = MIN(yl, y2);
		    xl = MIN(xl, x2);
		    yh = MAX(yh, y2);
		    xh = MAX(xh, x2);}}}}}
	  if (count>=colour_pixel_mass_threshold) k++;
	  else
	  { for (y1 = yl; y1<=yh; y1++)
	    { for (x1 = xl; x1<=xh; x1++)
	      if (coloured_cluster[y1][x1]==k)
	      { coloured_cluster[y1][x1] = -2;
		colour[y1][x1] = -1;}}}}}}}
  /* Cluster moving objects by proximity, ignoring small clusters */
  once = FALSE;
  again:
  no_moving_clusters = TRUE;
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++) moving_cluster[y][x] = -1;}
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { if (moving[y][x]&&moving_cluster[y][x]==-1)
      { moving_cluster[y][x] = k;
	back_y[y][x] = -1;
	back_x[y][x] = -1;
	yp = y;
	xp = x;
	yl = y;
	xl = x;
	yh = y;
	xh = x;
	count = 1;
	while (yp!=-1)
	{ y1 = yp;
	  x1 = xp;
	  yp = back_y[y1][x1];
	  xp = back_x[y1][x1];
	  if (moving_cluster[y1][x1]==k)
	  { for (y2 = MAX(0, y1-motion_proximity_threshold);
		 y2<MIN(height, y1+motion_proximity_threshold+1);
		 y2++)
	    { q = motion_pythag[abs(y2-y1)];
	      for (x2 = MAX(0, x1-q); x2<MIN(width, x1+q+1); x2++)
	      { if (moving[y2][x2]&&moving_cluster[y2][x2]==-1)
		{ moving_cluster[y2][x2] = k;
		  count++;
		  back_y[y2][x2] = yp;
		  back_x[y2][x2] = xp;
		  yp = y2;
		  xp = x2;
		  yl = MIN(yl, y2);
		  xl = MIN(xl, x2);
		  yh = MAX(yh, y2);
		  xh = MAX(xh, x2);}}}}}
        if (count>=motion_pixel_mass_threshold)
	{ k++;
	  no_moving_clusters = FALSE;}
	else
	{ for (y1 = yl; y1<=yh; y1++)
	  { for (x1 = xl; x1<=xh; x1++)
	    if (moving_cluster[y1][x1]==k)
	    { moving_cluster[y1][x1] = -2;
	      moving[y1][x1] = FALSE;}}}}}}
  if (no_moving_clusters)
  { if (meteor_last_moving_valid)
    { if (!once)
      { for (y = 0; y<height; y++)
        { for (x = 0; x<width; x++)
	  { moving[y][x] = meteor_last_moving[y][x];}}
	  once = TRUE;
	  goto again;}}}
  else if (!once)
  { for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { meteor_last_moving[y][x] = moving[y][x];}}
    if (!meteor_last_moving_valid)
    { meteor_last_moving_valid = TRUE;
      gfirst_frame = frame;}
    glast_frame = frame;}
  if (k>MAX_OBJECTS) panic("Too many objects");
  /* Fill object_frame */
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { if (meteor_uninterlaced) q = 3*(((object_frame*height+y)/2)*width+x);
      else q = 3*((object_frame*height+y&~meteor_deinterlaced)*width+x);
      if (moving[y][x])
      { r = 0;
        g = 0;
        b = 0;}
      else switch (colour[y][x])
      { case 0: r = 255; g = 0; b = 0; break;
        case 1: r = 0; g = 255; b = 0; break;
	case 2: r = 0; g = 0; b = 255; break;
        case 3: r = 255; g = 255; b = 0; break;
        default: r = 255; g = 255; b = 255;}
      meteor_frames[q+2] = r;
      meteor_frames[q+1] = g;
      meteor_frames[q] = b;}}
  /* Fit ellipses to clusters */
  gobject_k[frame] = k;
  for (i = 0; i<k; i++)
  { mu_x = 0.0;
    mu_y = 0.0;
    sigma_xx = 0.0;
    sigma_xy = 0.0;
    sigma_yy = 0.0;
    j = 0;
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      if (coloured_cluster[y][x]==i||moving_cluster[y][x]==i)
      { j++;
	mu_x += x;
	mu_y += y;}}
    mu_x /= j;
    mu_y /= j;
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      if (coloured_cluster[y][x]==i||moving_cluster[y][x]==i)
      { sigma_xx += (x-mu_x)*(x-mu_x);
	sigma_xy += (x-mu_x)*(y-mu_y);
	sigma_yy += (y-mu_y)*(y-mu_y);}}
    sigma_xx /= j;
    sigma_xy /= j;
    sigma_yy /= j;
    gellipse_x[frame][i] = mu_x;
    gellipse_y[frame][i] = mu_y;
    gellipse_theta[frame][i] = eigenvector_angle1(sigma_xx,
						  sigma_xy,
						  sigma_xy,
						  sigma_yy);
    gellipse_r[frame][i] = sqrt(2.0*eigenvalue1(sigma_xx,
						sigma_xy,
						sigma_xy,
						sigma_yy));
    gellipse_s[frame][i] = sqrt(2.0*eigenvalue2(sigma_xx,
						sigma_xy,
						sigma_xy,
						sigma_yy));}
  /* Free storage */
  free_c_int_vector(colour_pythag);
  free_c_int_vector(motion_pythag);
  free_c_int_matrix(colour, height);
  free_c_int_matrix(moving, height);
  free_c_int_matrix(coloured_cluster, height);
  free_c_int_matrix(moving_cluster, height);
  free_c_int_matrix(back_x, height);
  free_c_int_matrix(back_y, height);}

void meteor_finish_hack_track(void)
{ free_c_int_matrix(meteor_last_moving, meteor_height);}

void meteor_show_frame(int i)
{ int y, x;
  unsigned char *p, *q, *r;
  if (meteor_shmimage->bits_per_pixel==24)
  { if (meteor_uninterlaced)
    { for (y = 0; y<meteor_height; y++)
      { for (p = &meteor_frames[((i*meteor_height+y)/2)*meteor_width*3],
	     q = &meteor_shmimage->data[y*meteor_width*3],
	     r = &meteor_frames[(((i*meteor_height+y)/2)+1)*meteor_width*3];
	     p<r;)
        { *q++ = *p++;
          *q++ = *p++;
          *q++ = *p++;}}}
    else if (meteor_deinterlaced)
    { for (y = 0; y<meteor_height; y++)
      { for (p = &meteor_frames[(i*meteor_height+y&~meteor_deinterlaced)*
				meteor_width*3],
	     q = &meteor_shmimage->data[y*meteor_width*3],
	     r = &meteor_frames[((i*meteor_height+y&~meteor_deinterlaced)+1)*
				meteor_width*3];
	     p<r;)
	{ *q++ = *p++;
	  *q++ = *p++;
	  *q++ = *p++;}}}
    else
    { for (p = &meteor_frames[i*meteor_height*meteor_width*3],
	   q = &meteor_shmimage->data[0],
	   r = &meteor_frames[(i+1)*meteor_height*meteor_width*3];
	   p<r;)
      { *q++ = *p++;
	*q++ = *p++;
	*q++ = *p++;}}}
  else
  { if (meteor_uninterlaced)
    { for (y = 0; y<meteor_height; y++)
      { for (p = &meteor_frames[((i*meteor_height+y)/2)*meteor_width*3],
	     q = &meteor_shmimage->data[y*meteor_width*4],
	     r = &meteor_frames[(((i*meteor_height+y)/2)+1)*meteor_width*3];
	     p<r;)
        { *q++ = *p++;
          *q++ = *p++;
          *q++ = *p++;
          q++;}}}
    else if (meteor_deinterlaced)
    { for (y = 0; y<meteor_height; y++)
      { for (p = &meteor_frames[(i*meteor_height+y&~meteor_deinterlaced)*
				meteor_width*3],
	     q = &meteor_shmimage->data[y*meteor_width*4],
	     r = &meteor_frames[((i*meteor_height+y&~meteor_deinterlaced)+1)*
				meteor_width*3];
	     p<r;)
	{ *q++ = *p++;
	  *q++ = *p++;
	  *q++ = *p++;
	  q++;}}}
    else
    { for (p = &meteor_frames[i*meteor_height*meteor_width*3],
	   q = &meteor_shmimage->data[0],
	   r = &meteor_frames[(i+1)*meteor_height*meteor_width*3];
	   p<r;)
      { *q++ = *p++;
	*q++ = *p++;
	*q++ = *p++;
	q++;}}}
  XShmPutImage(meteor_display, meteor_window, meteor_gc, meteor_shmimage,
	       0, 0, 0, 0,
	       MIN(meteor_width, meteor_window_width),
	       MIN(meteor_height, meteor_window_height),
	       False);
  XFlush(meteor_display);}

void meteor_show_code_frame(int i)
{ int y, x;
  if (meteor_shmimage->bits_per_pixel==24)
  { for (y = 0; y<meteor_height; y++)
    { for (x = 0; x<meteor_width; x++)
      { meteor_shmimage->data[(y*meteor_width+x)*3] =
	meteor_get_code_frame_blue(i, y, x);
	meteor_shmimage->data[(y*meteor_width+x)*3+1] =
	meteor_get_code_frame_green(i, y, x);
	meteor_shmimage->data[(y*meteor_width+x)*3+2] =
	meteor_get_code_frame_red(i, y, x);}}}
  else
  { for (y = 0; y<meteor_height; y++)
    { for (x = 0; x<meteor_width; x++)
      { meteor_shmimage->data[(y*meteor_width+x)*4] =
	meteor_get_code_frame_blue(i, y, x);
	meteor_shmimage->data[(y*meteor_width+x)*4+1] =
	meteor_get_code_frame_green(i, y, x);
	meteor_shmimage->data[(y*meteor_width+x)*4+2] =
	meteor_get_code_frame_red(i, y, x);}}}
  XShmPutImage(meteor_display, meteor_window, meteor_gc, meteor_shmimage,
	       0, 0, 0, 0,
	       MIN(meteor_width, meteor_window_width),
	       MIN(meteor_height, meteor_window_height),
	       False);
  XFlush(meteor_display);}

void meteor_free(void)
{ free(meteor_frames);
  free(meteor_code_frames);
  free(meteor_last_frame);
  free(meteor_last_code);
  free(meteor_conjure_code);}

void meteor_attach_viewfinder(Display *display, int screen, Window window,
			      int window_width, int window_height)
{ XWindowAttributes attributes;
  Visual *visual;
  int visualid;
  int depth;
  XGCValues gc_values;
  meteor_display = display;
  meteor_screen = screen;
  meteor_window = window;
  meteor_window_width = window_width;
  meteor_window_height = window_height;
  XGetWindowAttributes(display, window, &attributes);
  visual = attributes.visual;
  depth = attributes.depth;
  if (depth!=24&&depth!=32) panic("unrecognized depth");
  visualid = attributes.visual->visualid;
  if (visualid!=32&&visualid!=33&&visualid!=34)
  { panic("unrecognized visual id");}
  gc_values.graphics_exposures = False;
  /* needs work: To check that succeeds */
  meteor_gc = XCreateGC(display, window, GCGraphicsExposures, &gc_values);
  /* needs work: To check that succeeds */
  meteor_shmimage = XShmCreateImage(display, visual, depth, ZPixmap,
				    NULL, &meteor_shminfo,
				    MIN(meteor_width, meteor_window_width),
				    MIN(meteor_height, meteor_window_height));
  if (meteor_shmimage->bits_per_pixel!=24&&meteor_shmimage->bits_per_pixel!=32)
  { panic("unrecognized bits_per_pixel");}
  if (meteor_shmimage->bytes_per_line>4*640||meteor_shmimage->height>480)
  { fprintf(stderr, "you don't want to know\n");
    exit(1);}
  if ((meteor_shminfo.shmid =
       shmget(ftok("/dev/mmetfgrab0", 'v'), 4*640*480, IPC_CREAT|0777))==-1)
  { perror("meteor shmget failed");
    exit(1);}
  if ((meteor_shminfo.shmaddr = shmat(meteor_shminfo.shmid, (char *)0, 0))==
      (char *)(-1))
  { perror("meteor shmat failed");
    exit(1);}
  meteor_shmimage->data = meteor_shminfo.shmaddr;
  /* needs work: To check that succeeds */
  XShmAttach(display, &meteor_shminfo);}

void meteor_detach_viewfinder(void)
/* needs work: To check that succeeds */
{ XShmDetach(meteor_display, &meteor_shminfo);
  if (shmdt(meteor_shminfo.shmaddr)==-1)
  { perror("meteor shmdt failed");
    exit(1);}
  /* needs work: To undo shmget */
  /* needs work: To undo XShmCreateImage */
  /* needs work: To check that succeeds */
  XFreeGC(meteor_display, meteor_gc);}

void meteor_turn_on_viewfinder(int m)
{ if (m<=0) panic("m must be positive");
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = FALSE;
  meteor_l = 0;
  meteor_m = m;}

void meteor_turn_on_capturing(int i, int l, int n)
{ if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = TRUE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = FALSE;
  meteor_i = i;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;}

void meteor_turn_on_ellipse_tracker(int i, int l, int n)
{ if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  if (!meteor_uninterlaced) panic("Must be in uninterlaced mode");
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = TRUE;
  meteor_tracking_polygons = FALSE;
  meteor_starting = TRUE;
  meteor_conjuring = 0;
  meteor_i = i;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;}

void meteor_turn_on_ellipse_tracker_with_capturing(int i, int l, int n)
{ if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  if (!meteor_uninterlaced) panic("Must be in uninterlaced mode");
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = TRUE;
  meteor_tracking_ellipses = TRUE;
  meteor_tracking_polygons = FALSE;
  meteor_starting = TRUE;
  meteor_conjuring = 0;
  meteor_i = i;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;}

void meteor_turn_on_polygon_tracker(int i, int l, int n, int h)
{ if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  if (!meteor_uninterlaced) panic("Must be in uninterlaced mode");
  meteor_hull = h;
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = TRUE;
  meteor_starting = TRUE;
  meteor_conjuring = 0;
  meteor_i = i;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;}

void meteor_turn_on_polygon_tracker_with_capturing(int i, int l, int n, int h)
{ if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  if (!meteor_uninterlaced) panic("Must be in uninterlaced mode");
  meteor_hull = h;
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = TRUE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = TRUE;
  meteor_starting = TRUE;
  meteor_conjuring = 0;
  meteor_i = 0;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;}

void meteor_turn_off(void)
{ meteor_block();
  meteor_l = 0;
  meteor_m = 0;
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = FALSE;
  meteor_set_nonreal_time();}

void meteor_capture_frames(int i, int l, int n)
{ int real_time = meteor_real_time;
  if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = TRUE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = FALSE;
  meteor_i = i;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;
  while (meteor_i<meteor_n) pause();
  if (real_time==FALSE) meteor_set_nonreal_time();
  meteor_block();
  meteor_l = 0;
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = FALSE;}

void meteor_track_ellipses(int i, int l, int n)
{ int real_time = meteor_real_time;
  if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  if (!meteor_uninterlaced) panic("Must be in uninterlaced mode");
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = TRUE;
  meteor_tracking_polygons = FALSE;
  meteor_starting = TRUE;
  meteor_conjuring = 0;
  meteor_i = i;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;
  while (meteor_i<meteor_n) pause();
  if (real_time==FALSE) meteor_set_nonreal_time();
  meteor_block();
  meteor_l = 0;
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = FALSE;}

void meteor_track_ellipses_and_capture_frames(int i, int l, int n)
{ int real_time = meteor_real_time;
  if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  if (!meteor_uninterlaced) panic("Must be in uninterlaced mode");
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = TRUE;
  meteor_tracking_ellipses = TRUE;
  meteor_tracking_polygons = FALSE;
  meteor_starting = TRUE;
  meteor_conjuring = 0;
  meteor_i = i;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;
  while (meteor_i<meteor_n) pause();
  if (real_time==FALSE) meteor_set_nonreal_time();
  meteor_block();
  meteor_l = 0;
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = FALSE;}

void meteor_track_polygons(int i, int l, int n, int h)
{ int real_time = meteor_real_time;
  if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  if (!meteor_uninterlaced) panic("Must be in uninterlaced mode");
  meteor_hull = h;
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = TRUE;
  meteor_starting = TRUE;
  meteor_conjuring = 0;
  meteor_i = i;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;
  while (meteor_i<meteor_n) pause();
  if (real_time==FALSE) meteor_set_nonreal_time();
  meteor_block();
  meteor_l = 0;
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = FALSE;}

void meteor_track_polygons_and_capture_frames(int i, int l, int n, int h)
{ int real_time = meteor_real_time;
  if (i<0||i>=MAX_FRAMES) panic("i out of bounds");
  if (l<=0) panic("l must be positive");
  if (n<=0||n>MAX_FRAMES) panic("n out of bounds");
  if (!meteor_uninterlaced) panic("Must be in uninterlaced mode");
  meteor_hull = h;
  meteor_set_real_time();
  meteor_unblock();
  meteor_capturing = TRUE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = TRUE;
  meteor_starting = TRUE;
  meteor_conjuring = 0;
  meteor_i = i;
  meteor_j = 0;
  meteor_k = 0;
  meteor_n = n;
  meteor_l = l;
  while (meteor_i<meteor_n) pause();
  if (real_time==FALSE) meteor_set_nonreal_time();
  meteor_block();
  meteor_l = 0;
  meteor_capturing = FALSE;
  meteor_tracking_ellipses = FALSE;
  meteor_tracking_polygons = FALSE;}

void meteor_startup(int height, int width, int nframes, int uninterlaced)
{ meteor_open();
  meteor_sigaction();
  meteor_set_geometry(height, width, 1, METEOR_GEO_RGB24, uninterlaced);
  meteor_get_frame_offset(&meteor_offset);
  meteor_set_video_format(METEOR_FMT_NTSC);
  meteor_set_input_device(METEOR_INPUT_DEV0);
  meteor_malloc(nframes);
  meteor_mmap();
  meteor_display = NULL;
  meteor_set_signal(SIGUSR2|METEOR_SIG_FRAME);
  meteor_capture(METEOR_CAP_CONTINOUS);}

void meteor_startup_offline(int height,
			    int width,
			    int nframes,
			    int uninterlaced)
{ meteor_height = height;
  meteor_width = width;
  meteor_uninterlaced = uninterlaced;
  meteor_deinterlaced = 0;
  meteor_malloc(nframes);
  meteor_display = NULL;
  meteor_starting = TRUE;
  meteor_conjuring = 0;}

void meteor_startup_with_viewfinder(int height, int width, int nframes,
				    int uninterlaced,
				    Display *display, int screen,
				    Window window,
				    int window_width, int window_height)
{ meteor_open();
  meteor_sigaction();
  meteor_set_geometry(height, width, 1, METEOR_GEO_RGB24, uninterlaced);
  meteor_get_frame_offset(&meteor_offset);
  meteor_set_video_format(METEOR_FMT_NTSC);
  meteor_set_input_device(METEOR_INPUT_DEV0);
  meteor_malloc(nframes);
  meteor_mmap();
  meteor_attach_viewfinder(display, screen, window,
			   window_width, window_height);
  meteor_set_signal(SIGUSR2|METEOR_SIG_FRAME);
  meteor_capture(METEOR_CAP_CONTINOUS);}

void meteor_restartup(int height, int width, int nframes, int uninterlaced)
{ struct meteor_capframe capframe;
  meteor_l = 0;
  meteor_m = 0;
  /* This is because meteor_capture(METEOR_CAP_STOP_FRAMES) doesn't work */
  capframe.command = METEOR_CAP_STOP_FRAMES;
  meteor_capframe(&capframe);
  meteor_set_signal(0);
  if (meteor_display!=NULL) meteor_detach_viewfinder();
  meteor_munmap();
  meteor_free();
  meteor_close();
  XClearArea(meteor_display, meteor_window, 0, 0, meteor_width, meteor_height,
	     False);
  meteor_open();
  meteor_set_geometry(height, width, 1, METEOR_GEO_RGB24, uninterlaced);
  meteor_get_frame_offset(&meteor_offset);
  meteor_set_video_format(METEOR_FMT_NTSC);
  meteor_set_input_device(METEOR_INPUT_DEV0);
  meteor_malloc(nframes);
  meteor_mmap();
  if (meteor_display!=NULL) meteor_attach_viewfinder(meteor_display,
						     meteor_screen,
						     meteor_window,
						     meteor_window_width,
						     meteor_window_height);
  meteor_set_signal(SIGUSR2|METEOR_SIG_FRAME);
  meteor_capture(METEOR_CAP_CONTINOUS);}

void meteor_shutdown(void)
{ struct meteor_capframe capframe;
  meteor_l = 0;
  meteor_m = 0;
  /* This is because meteor_capture(METEOR_CAP_STOP_FRAMES) doesn't work */
  capframe.command = METEOR_CAP_STOP_FRAMES;
  meteor_capframe(&capframe);
  meteor_set_signal(0);
  if (meteor_display!=NULL) meteor_detach_viewfinder();
  meteor_munmap();
  meteor_free();
  meteor_close();
  meteor_set_nonreal_time();}

void meteor_shutdown_offline(void)
{ meteor_free();}

/* RPC */

void send_frame(FILE *f, int i)
{ int amount = 3*(meteor_height/2)*meteor_width, p;
  if (meteor_uninterlaced) p = 3*(((i+1)*meteor_height)/2)*meteor_width;
  else p = 3*((i+1)*meteor_height)*meteor_width;
  while (amount>0)
  { amount -= write(fileno(f), &meteor_frames[p-amount], amount);}}

void receive_frame(FILE *f, int i)
{ int amount = 3*(meteor_height/2)*meteor_width;
  char *p = (meteor_uninterlaced?
	     &meteor_frames[3*((i*meteor_height)/2)*meteor_width]:
	     &meteor_frames[3*(i*meteor_height)*meteor_width]);
  for (; amount>0; amount--) *(p++) = fgetc(f);}

/* This is trial code */

void set_n_points(int n1, int n2)
{ points1 = (struct point *)malloc(n1*sizeof(struct point));
  if (points1==NULL) panic("set_n_points: out of memory error");
  points2 = (struct point *)malloc(n2*sizeof(struct point));
  if (points2==NULL) panic("set_n_points: out of memory error");
  n_full_points1 = n1;
  n_full_points2 = n2;
  full_points1 = (struct point *)malloc(n1*sizeof(struct point));
  if (full_points1==NULL) panic("set_n_points: out of memory error");
  full_points2 = (struct point *)malloc(n2*sizeof(struct point));
  if (full_points2==NULL) panic("set_n_points: out of memory error");}

void augmented_set_n_points(int n1, int n2,
			    int max_x_velocity, int max_y_velocity)
{ int i, j;
  points1 = (struct point *)malloc(n1*sizeof(struct point));
  if (points1==NULL) panic("augmented_set_n_points: out of memory error");
  points2 = (struct point *)malloc(n2*sizeof(struct point));
  if (points2==NULL) panic("augmented_set_n_points: out of memory error");
  n_full_points1 = n1;
  n_full_points2 = n2;
  full_points1 = (struct point *)malloc(n1*sizeof(struct point));
  if (full_points1==NULL) panic("augmented_set_n_points: out of memory error");
  full_points2 = (struct point *)malloc(n2*sizeof(struct point));
  if (full_points2==NULL) panic("augmented_set_n_points: out of memory error");
  for (i = 0; i<n_full_points1; i++)
  { full_points1[i].a = (int **)malloc((2*max_y_velocity+1)*sizeof(int *));
    if (full_points1[i].a==NULL)
    { panic("augmented_set_n_points: out of memory error");}
    for (j = 0; j<2*max_y_velocity+1; j++)
    { full_points1[i].a[j] =
      (int *)malloc((2*max_x_velocity+1)*sizeof(int));
      if (full_points1[i].a[j]==NULL)
      { panic("augmented_set_n_points: out of memory error");}}}}

void set_point1(int i, int x, int y)
{ full_points1[i].x = x;
  full_points1[i].y = y;}

void set_point2(int i, int x, int y)
{ full_points2[i].x = x;
  full_points2[i].y = y;}

int xor_points_count(int dx, int dy)
{ int point1 = 0, point2 = 0, count = 0;
  while (1)
  { if (point1==n_points1)
    { if (point2==n_points2) return count;
      point2++;
      count++;}
    else if (point2==n_points2)
    { point1++;
      count++;}
    else if (points1[point1].x+dx==points2[point2].x&&
	     points1[point1].y+dy==points2[point2].y)
    { point1++;
      point2++;}
    else if (points1[point1].y+dy<points2[point2].y||
	     (points1[point1].y+dy==points2[point2].y&&
	      points1[point1].x+dx<points2[point2].x))
    { point1++;
      count++;}
    else
    { point2++;
      count++;}}}

void exhaustive_xor_velocity_points(int max_x_velocity, int max_y_velocity)
{ int minimum = -1, dx, dy, count;
  for (dy = -max_y_velocity; dy<=max_y_velocity; dy++)
  { for (dx = -max_x_velocity; dx<=max_x_velocity; dx++)
    { count = xor_points_count(dx, dy);
      if (minimum<0||count<minimum)
      { x_velocity = dx;
	y_velocity = dy;
	minimum = count;}}}}

void hill_climbing_xor_velocity_points(int max_x_velocity, int max_y_velocity)
{ int minimum, dx, dy, bx, by, count;
  /* needs work: This is gcc specific. */
  /* needs work: need to limit dx and dy to maximum_velocity or else there
                 will be an array subscript out of bounds error */
  int cache[2*max_y_velocity+1][2*max_x_velocity+1];
  for (dy = -max_y_velocity; dy<=max_y_velocity; dy++)
  { for (dx = -max_x_velocity; dx<=max_x_velocity; dx++)
    { cache[dy+max_y_velocity][dx+max_x_velocity] = -1;}}
  dy = 0;
  dx = 0;
  while (1)
  { minimum = -1;
    if (cache[dy-1+max_y_velocity][dx-1+max_x_velocity]==-1)
    { cache[dy-1+max_y_velocity][dx-1+max_x_velocity] =
      xor_points_count(dx-1, dy-1);}
    count = cache[dy-1+max_y_velocity][dx-1+max_x_velocity];
    if (minimum<0||count<minimum)
    { bx = dx-1;
      by = dy-1;
      minimum = count;}
    if (cache[dy-1+max_y_velocity][dx+max_x_velocity]==-1)
    { cache[dy-1+max_y_velocity][dx+max_x_velocity] =
      xor_points_count(dx, dy-1);}
    count = cache[dy-1+max_y_velocity][dx+max_x_velocity];
    if (minimum<0||count<minimum)
    { bx = dx;
      by = dy-1;
      minimum = count;}
    if (cache[dy-1+max_y_velocity][dx+1+max_x_velocity]==-1)
    { cache[dy-1+max_y_velocity][dx+1+max_x_velocity] =
      xor_points_count(dx+1, dy-1);}
    count = cache[dy-1+max_y_velocity][dx+1+max_x_velocity];
    if (minimum<0||count<minimum)
    { bx = dx+1;
      by = dy-1;
      minimum = count;}
    if (cache[dy+max_y_velocity][dx-1+max_x_velocity]==-1)
    { cache[dy+max_y_velocity][dx-1+max_x_velocity] =
      xor_points_count(dx-1, dy);}
    count = cache[dy+max_y_velocity][dx-1+max_x_velocity];
    if (minimum<0||count<minimum)
    { bx = dx-1;
      by = dy;
      minimum = count;}
    if (cache[dy+max_y_velocity][dx+max_x_velocity]==-1)
    { cache[dy+max_y_velocity][dx+max_x_velocity] =
      xor_points_count(dx, dy);}
    count = cache[dy+max_y_velocity][dx+max_x_velocity];
    if (minimum<0||count<minimum)
    { bx = dx;
      by = dy;
      minimum = count;}
    if (cache[dy+max_y_velocity][dx+1+max_x_velocity]==-1)
    { cache[dy+max_y_velocity][dx+1+max_x_velocity] =
      xor_points_count(dx+1, dy);}
    count = cache[dy+max_y_velocity][dx+1+max_x_velocity];
    if (minimum<0||count<minimum)
    { bx = dx+1;
      by = dy;
      minimum = count;}
    if (cache[dy+1+max_y_velocity][dx-1+max_x_velocity]==-1)
    { cache[dy+1+max_y_velocity][dx-1+max_x_velocity] =
      xor_points_count(dx-1, dy+1);}
    count = cache[dy+1+max_y_velocity][dx-1+max_x_velocity];
    if (minimum<0||count<minimum)
    { bx = dx-1;
      by = dy+1;
      minimum = count;}
    if (cache[dy+1+max_y_velocity][dx+max_x_velocity]==-1)
    { cache[dy+1+max_y_velocity][dx+max_x_velocity] =
      xor_points_count(dx, dy+1);}
    count = cache[dy+1+max_y_velocity][dx+max_x_velocity];
    if (minimum<0||count<minimum)
    { bx = dx;
      by = dy+1;
      minimum = count;}
    if (cache[dy+1+max_y_velocity][dx+1+max_x_velocity]==-1)
    { cache[dy+1+max_y_velocity][dx+1+max_x_velocity] =
      xor_points_count(dx+1, dy+1);}
    count = cache[dy+1+max_y_velocity][dx+1+max_x_velocity];
    if (minimum<0||count<minimum)
    { bx = dx+1;
      by = dy+1;
      minimum = count;}
    if (bx==dx&&by==dy)
    { x_velocity = dx;
      y_velocity = dy;
      printf("%d %d %d\n%d %d %d\n%d %d %d\n",
	     cache[dy-1+max_y_velocity][dx-1+max_x_velocity],
	     cache[dy-1+max_y_velocity][dx+max_x_velocity],
	     cache[dy-1+max_y_velocity][dx+1+max_x_velocity],
	     cache[dy+max_y_velocity][dx-1+max_x_velocity],
	     cache[dy+max_y_velocity][dx+max_x_velocity],
	     cache[dy+max_y_velocity][dx+1+max_x_velocity],
	     cache[dy+1+max_y_velocity][dx-1+max_x_velocity],
	     cache[dy+1+max_y_velocity][dx+max_x_velocity],
	     cache[dy+1+max_y_velocity][dx+1+max_x_velocity]);
      return;}
    dx = bx;
    dy = by;}}

void xor_velocity_points(int max_x_velocity, int max_y_velocity)
{ int hill_climbing_x, hill_climbing_y;
  hill_climbing_xor_velocity_points(max_x_velocity, max_y_velocity);
  hill_climbing_x = x_velocity;
  hill_climbing_y = y_velocity;
  exhaustive_xor_velocity_points(max_x_velocity, max_y_velocity);
  if (hill_climbing_x!=x_velocity||hill_climbing_y!=y_velocity)
  { printf("bad (%d %d) vs. (%d %d)\n",
	   hill_climbing_x, hill_climbing_y, x_velocity, y_velocity);}
  else
  { printf("good (%d %d) vs. (%d %d)\n",
	   hill_climbing_x, hill_climbing_y, x_velocity, y_velocity);}
  return;}

void full_xor_velocity_points(int max_x_velocity, int max_y_velocity)
{ int i;
  n_points1 = n_full_points1;
  for (i = 0; i<n_points1; i++) points1[i] = full_points1[i];
  n_points2 = n_full_points2;
  for (i = 0; i<n_points2; i++) points2[i] = full_points2[i];
  exhaustive_xor_velocity_points(max_x_velocity, max_y_velocity);}

void clipped_xor_velocity_points(int max_x_velocity, int max_y_velocity,
				 int xl, int yl, int xh, int yh)
{ int i;
  n_points1 = 0;
  for (i = 0; i<n_full_points1; i++)
  { if (full_points1[i].x>=xl&&
	full_points1[i].x<=xh&&
	full_points1[i].y>=yl&&
	full_points1[i].y<=yh)
    { points1[n_points1] = full_points1[i];
      n_points1++;}}
  n_points2 = 0;
  for (i = 0; i<n_full_points2; i++)
  { if (full_points2[i].x>=xl&&
	full_points2[i].x<=xh&&
	full_points2[i].y>=yl&&
	full_points2[i].y<=yh)
    { points2[n_points2] = full_points2[i];
      n_points2++;}}
  exhaustive_xor_velocity_points(max_x_velocity, max_y_velocity);}

void augmented_model(int max_x_velocity, int max_y_velocity)
{ int point1, point2, dx, dy, dxi, dyi, minimum, this;
  for (point1 = 0; point1<n_full_points1; point1++)
  { for (dy = -max_y_velocity, dyi = 0; dy<=max_y_velocity; dy++, dyi++)
    { for (dx = -max_x_velocity, dxi = 0; dx<=max_x_velocity; dx++, dxi++)
      { minimum = -1;
	for (point2 = 0; point2<n_full_points2; point2++)
	{ this = abs(full_points1[point1].x+dx-full_points2[point2].x)+
	         abs(full_points1[point1].y+dy-full_points2[point2].y);
	  if (minimum<0||this<minimum) minimum = this;}
	full_points1[point1].a[dyi][dxi] = minimum;}}}}

void sum_min_velocity_points(int max_x_velocity, int max_y_velocity)
{ int point1, minimum = -1, dx, dy, sum;
  for (dy = -max_y_velocity; dy<=max_y_velocity; dy++)
  { for (dx = -max_x_velocity; dx<=max_x_velocity; dx++)
    { sum = 0;
      for (point1 = 0; point1<n_points1; point1++)
      /* needs work: To make parameter. */
      { sum = sum+MIN(10, points1[point1].a
		      [dy+max_y_velocity][dx+max_x_velocity]);}
      if (minimum<0||sum<minimum)
      { x_velocity = dx;
	y_velocity = dy;
	minimum = sum;}}}}

void full_sum_min_velocity_points(int max_x_velocity, int max_y_velocity)
{ int i;
  n_points1 = n_full_points1;
  for (i = 0; i<n_points1; i++) points1[i] = full_points1[i];
  sum_min_velocity_points(max_x_velocity, max_y_velocity);}

void clipped_sum_min_velocity_points(int max_x_velocity, int max_y_velocity,
				     int xl, int yl, int xh, int yh)
{ int i;
  n_points1 = 0;
  for (i = 0; i<n_full_points1; i++)
  { if (full_points1[i].x>=xl&&
	full_points1[i].x<=xh&&
	full_points1[i].y>=yl&&
	full_points1[i].y<=yh)
    { points1[n_points1] = full_points1[i];
      n_points1++;}}
  sum_min_velocity_points(max_x_velocity, max_y_velocity);}

int get_x_velocity(void) {return x_velocity;}

int get_y_velocity(void) {return y_velocity;}

void free_points(void)
{ free(points1);
  free(points2);
  free(full_points1);
  free(full_points2);}

void augmented_free_points(int max_x_velocity, int max_y_velocity)
{ int i, j;
  for (i = 0; i<n_full_points1; i++)
  { for (j = 0; j<2*max_y_velocity+1; j++) free(full_points1[i].a[j]);
    free(full_points1[i].a);}
  free(points1);
  free(points2);
  free(full_points1);
  free(full_points2);}

/* Edge Flow */

/* needs work: This is a little wastefull since we store RGB values as ints
               when they could be unsigned chars. It is a time-space tradeoff.
               Perhaps this could use MMX. */

/* needs work: For greyscale images, the grey channel is stored in the red
               band. This is a little wastefull of space. Also, we use
               edge_flow_set_rgb1 and edge_flow_set_rgb2 to download the
               greyscale image which is inefficient. */

void edge_flow_set_n_points(int n1, int n2)
{ edge_flow_n_points1 = n1;
  edge_flow_n_points2 = n2;
  edge_flow_points1 =
  (struct edge_flow_point *)malloc(n1*sizeof(struct edge_flow_point));
  if (edge_flow_points1==NULL)
  { panic("edge_flow_set_n_points: out of memory error");}
  edge_flow_points2 =
  (struct edge_flow_point *)malloc(n2*sizeof(struct edge_flow_point));
  if (edge_flow_points2==NULL)
  { panic("edge_flow_set_n_points: out of memory error");}
  edge_flow_map = (int *)malloc(n1*sizeof(int));
  if (edge_flow_map==NULL)
  { panic("edge_flow_set_n_points: out of memory error");}}

void edge_flow_set_point1(int i, int x, int y)
{ edge_flow_points1[i].x = x;
  edge_flow_points1[i].y = y;}

void edge_flow_set_point2(int i, int x, int y)
{ edge_flow_points2[i].x = x;
  edge_flow_points2[i].y = y;}

void edge_flow_set_height_width(int height, int width)
{ int y;
  edge_flow_height = height;
  edge_flow_width = width;
  edge_flow_image1 =
  (struct edge_flow_pixel **)malloc(height*sizeof(struct edge_flow_pixel *));
  if (edge_flow_image1==NULL)
  { panic("edge_flow_set_height_width: out of memory error");}
  edge_flow_image2 =
  (struct edge_flow_pixel **)malloc(height*sizeof(struct edge_flow_pixel *));
  if (edge_flow_image2==NULL)
  { panic("edge_flow_set_height_width: out of memory error");}
  for (y = 0; y<height; y++)
  { edge_flow_image1[y] =
    (struct edge_flow_pixel *)malloc(width*sizeof(struct edge_flow_pixel));
    if (edge_flow_image1[y]==NULL)
    { panic("edge_flow_set_height_width: out of memory error");}
    edge_flow_image2[y] =
    (struct edge_flow_pixel *)malloc(width*sizeof(struct edge_flow_pixel));
    if (edge_flow_image2[y]==NULL)
    { panic("edge_flow_set_height_width: out of memory error");}}}

void edge_flow_set_rgb1(int x, int y, int red, int green, int blue)
{ edge_flow_image1[y][x].red = red;
  edge_flow_image1[y][x].green = green;
  edge_flow_image1[y][x].blue = blue;}

void edge_flow_set_rgb2(int x, int y, int red, int green, int blue)
{ edge_flow_image2[y][x].red = red;
  edge_flow_image2[y][x].green = green;
  edge_flow_image2[y][x].blue = blue;}

float edge_flow_absolute_difference(int x1, int y1, int x2, int y2,
				    int window_x, int window_y)
{ int sum = 0, dx, dy, red, green, blue;
  struct edge_flow_pixel *row1, *row2, *pixel1, *pixel2;
  for (dy = -window_y; dy<=window_y; dy++)
  { if (y1+dy>=0&&y1+dy<edge_flow_height&&y2+dy>=0&&y2+dy<edge_flow_height)
    { row1 = edge_flow_image1[y1+dy];
      row2 = edge_flow_image2[y2+dy];
      dx = -window_x;
      if (dx<-x1) dx = -x1;
      if (dx<-x2) dx = -x2;
      if (dx>=edge_flow_width-x1) dx = edge_flow_width-x1;
      if (dx>=edge_flow_width-x2) dx = edge_flow_width-x2;
      pixel1 = &row1[x1+dx],
      pixel2 = &row2[x2+dx];
      for (;
	   dx<=window_x;
	   dx++,
	   pixel1++,
	   pixel2++)
      { red = pixel1->red-pixel2->red;
        green = pixel1->green-pixel2->green;
        blue = pixel1->blue-pixel2->blue;
        /* note: We no longer compute the square root since that took a long
                 time. And we do the squaring and sumation in fixed point
                 because the fixed-to-float conversion is slow. sum is
                 converted to float only once upon return. The squaring
                 can't overflow because the RGB values are 8 bits. But
                 the summation can overflow if the window is too large. */
        sum += red*red+green*green+blue*blue;}}}
  return sum;}

void edge_flow(int maximum_flow_magnitude, int window_x, int window_y)
{ int p1, p2;
  float best, this;
  for (p1 = 0; p1<edge_flow_n_points1; p1++)
  { edge_flow_map[p1] = -1;
    for (p2 = 0; p2<edge_flow_n_points2; p2++)
    { if (((edge_flow_points1[p1].x-edge_flow_points2[p2].x)*
           (edge_flow_points1[p1].x-edge_flow_points2[p2].x))+
          ((edge_flow_points1[p1].y-edge_flow_points2[p2].y)*
           (edge_flow_points1[p1].y-edge_flow_points2[p2].y))<=
          maximum_flow_magnitude*maximum_flow_magnitude)
      { if (edge_flow_map[p1]==-1)
        { edge_flow_map[p1] = p2;
          best = edge_flow_absolute_difference(edge_flow_points1[p1].x,
                                               edge_flow_points1[p1].y,
                                               edge_flow_points2[p2].x,
                                               edge_flow_points2[p2].y,
					       window_x,
					       window_y);}
        else
        { this = edge_flow_absolute_difference(edge_flow_points1[p1].x,
                                               edge_flow_points1[p1].y,
                                               edge_flow_points2[p2].x,
                                               edge_flow_points2[p2].y,
					       window_x,
					       window_y);
          if (this<best)
          { edge_flow_map[p1] = p2;
            best = this;}}}}}}

float edge_flow_grey_difference(int x1, int y1, int x2, int y2,
				int window_x, int window_y)
{ int sum = 0, dx, dy, grey;
  struct edge_flow_pixel *row1, *row2, *pixel1, *pixel2;
  for (dy = -window_y; dy<=window_y; dy++)
  { if (y1+dy>=0&&y1+dy<edge_flow_height&&y2+dy>=0&&y2+dy<edge_flow_height)
    { row1 = edge_flow_image1[y1+dy];
      row2 = edge_flow_image2[y2+dy];
      dx = -window_x;
      if (dx<-x1) dx = -x1;
      if (dx<-x2) dx = -x2;
      if (dx>=edge_flow_width-x1) dx = edge_flow_width-x1;
      if (dx>=edge_flow_width-x2) dx = edge_flow_width-x2;
      pixel1 = &row1[x1+dx],
      pixel2 = &row2[x2+dx];
      for (;
	   dx<=window_x;
	   dx++,
	   pixel1++,
	   pixel2++)
      { grey = pixel1->red-pixel2->red;
        sum += grey*grey;}}}
  return sum;}

void edge_flow_grey(int maximum_flow_magnitude, int window_x, int window_y)
{ int p1, p2;
  float best, this;
  for (p1 = 0; p1<edge_flow_n_points1; p1++)
  { edge_flow_map[p1] = -1;
    for (p2 = 0; p2<edge_flow_n_points2; p2++)
    { if (((edge_flow_points1[p1].x-edge_flow_points2[p2].x)*
           (edge_flow_points1[p1].x-edge_flow_points2[p2].x))+
          ((edge_flow_points1[p1].y-edge_flow_points2[p2].y)*
           (edge_flow_points1[p1].y-edge_flow_points2[p2].y))<=
          maximum_flow_magnitude*maximum_flow_magnitude)
      { if (edge_flow_map[p1]==-1)
        { edge_flow_map[p1] = p2;
          best = edge_flow_grey_difference(edge_flow_points1[p1].x,
					   edge_flow_points1[p1].y,
					   edge_flow_points2[p2].x,
					   edge_flow_points2[p2].y,
					   window_x,
					   window_y);}
        else
        { this = edge_flow_grey_difference(edge_flow_points1[p1].x,
					   edge_flow_points1[p1].y,
					   edge_flow_points2[p2].x,
					   edge_flow_points2[p2].y,
					   window_x,
					   window_y);
          if (this<best)
          { edge_flow_map[p1] = p2;
            best = this;}}}}}}

int edge_flow_get(int i) {return edge_flow_map[i];}

void edge_flow_free(void)
{ int y;
  free(edge_flow_points1);
  free(edge_flow_points2);
  free(edge_flow_map);
  for (y = 0; y<edge_flow_height; y++)
  { free(edge_flow_image1[y]);
    free(edge_flow_image2[y]);}
  free(edge_flow_image1);
  free(edge_flow_image2);}

/* From cdslib-c.c and cdslib-c.h */

void set_speed_and_format(FILE *file, int speed)
{ int format = AFMT_S16_LE;
  ioctl(fileno(file), SNDCTL_DSP_RESET, 0);
  ioctl(fileno(file), SNDCTL_DSP_SYNC, 0);
  ioctl(fileno(file), SNDCTL_DSP_SPEED, &speed);
  ioctl(fileno(file), SNDCTL_DSP_SETFMT, &format);}

float *make_float_vector(unsigned int length)
{ float *vector;
  if (length==0) length = 1;
  vector = (float *)malloc(length*sizeof(float));
  if (vector==NULL) panic("Out of memory");
  return vector;}

void free_float_vector(float *vector) {free(vector);}

float float_vector_ref(float *vector, unsigned int i) {return vector[i];}

void float_vector_set(float *vector, unsigned int i, float x) {vector[i] = x;}

unsigned int file_length(FILE *file)
{ long current = ftell(file), length;
  if (current==-1) panic("ftell error");
  if (fseek(file, 0, SEEK_END)!=0) panic("fseek error");
  length = ftell(file);
  if (length==-1) panic("ftell error");
  if (fseek(file, current, SEEK_SET)!=0) panic("fseek error");
  return length;}

unsigned char *make_mapped_file(FILE *file, unsigned int length)
{ unsigned char *result =
  mmap(0, length, PROT_READ, MAP_SHARED, fileno(file), 0);
  if ((int)result==-1) panic("mmap error");
  return result;}

void free_mapped_file(unsigned char *mapped_file, unsigned int length)
{ if (munmap(mapped_file, length)!=0) panic("munmap error");}

unsigned int mapped_file_ref(unsigned char *mapped_file, unsigned int i)
{ return mapped_file[i];}

/* Ratio Regions */

#define CF_OUT(y, x) (v*w[(y)][(x)]-f_out[(y)][(x)])
#define CF_RIGHT(y, x) (c_right[(y)][(x)]-f_right[(y)][(x)])
#define CF_LEFT(y, x) (c_right[(y)][(x)-1]+f_right[(y)][(x)-1])
#define CF_DOWN(y, x) (c_down[(y)][(x)]-f_down[(y)][(x)])
#define CF_UP(y, x) (c_down[(y)-1][(x)]+f_down[(y)-1][(x)])

#define EF_OUT(y, x) (CF_OUT((y), (x))>0)
#define EF_RIGHT(y, x) (CF_RIGHT((y), (x))>0)
#define EF_LEFT(y, x) (CF_LEFT((y), (x))>0)
#define EF_DOWN(y, x) (CF_DOWN((y), (x))>0)
#define EF_UP(y, x) (CF_UP((y), (x))>0)

#define CAN_PUSH_OUT(y, x)\
 (e[(y)][(x)]!=0&&\
  EF_OUT((y), (x))&&\
  h[(y)][(x)]==1)
#define CAN_PUSH_RIGHT(y, x)\
 ((x)<width-1&&\
  e[(y)][(x)]!=0&&\
  mask[y][x+1]&&\
  EF_RIGHT((y), (x))&&\
  h[(y)][(x)]==h[(y)][(x)+1]+1)
#define CAN_PUSH_LEFT(y, x)\
 ((x)>0&&\
  e[(y)][(x)]!=0&&\
  mask[y][x-1]&&\
  EF_LEFT((y), (x))&&\
  h[(y)][(x)]==h[(y)][(x)-1]+1)
#define CAN_PUSH_DOWN(y, x)\
 ((y)<height-1&&\
  e[(y)][(x)]!=0&&\
  mask[y+1][x]&&\
  EF_DOWN((y), (x))&&\
  h[(y)][(x)]==h[(y)+1][(x)]+1)
#define CAN_PUSH_UP(y, x)\
 ((y)>0&&\
  e[(y)][(x)]!=0&&\
  mask[y-1][x]&&\
  EF_UP((y), (x))&&\
  h[(y)][(x)]==h[(y)-1][(x)]+1)
#define CAN_LIFT(y, x)\
 (e[(y)][(x)]!=0&&\
  h[(y)][(x)]<m1&&\
  (!EF_OUT((y), (x))||h[(y)][(x)]==0)&&\
  ((x)==width-1||\
   !mask[y][(x)+1]||\
   !EF_RIGHT((y), (x))||\
   h[(y)][(x)]<=h[(y)][(x)+1])&&\
  ((x)==0||\
   !mask[y][(x)-1]||\
   !EF_LEFT((y), (x))||\
   h[(y)][(x)]<=h[(y)][(x)-1])&&\
  ((y)==height-1||\
   !mask[(y)+1][x]||\
   !EF_DOWN((y), (x))||\
   h[(y)][(x)]<=h[(y)+1][(x)])&&\
  ((y)==0||\
   !mask[(y)-1][x]||\
   !EF_UP((y), (x))||\
   h[(y)][(x)]<=h[(y)-1][(x)]))

int **f_right, **f_down, **f_out, **h, **e, **marked;
int **next_y, **next_x, **next2_y, **next2_x, *qy, *qx;

void enqueue(int y, int x)
{ int t;
  if (!marked[y][x])
  { t = h[y][x];
    next_y[y][x] = qy[t];
    next_x[y][x] = qx[t];
    qy[t] = y;
    qx[t] = x;
    marked[y][x] = TRUE;}}

void relabel(int height,
	     int width,
	     int **mask,
	     int **c_right,
	     int **c_down,
	     int **w,
	     int v)
{ int y, x, t, ty, tx, qy = -1, qx;
  int m1 = height*width+1;
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { h[y][x] = m1;
      marked[y][x] = FALSE;}}
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { if (mask[y][x]&&EF_OUT(y, x)&&h[y][x]>1)
      { h[y][x] = 1;
	marked[y][x] = TRUE;
	if (qy==-1)
	{ qy = y;
	  qx = x;
	  ty = y;
	  tx = x;
	  next_y[y][x] = -1;}
	else
	{ next_y[ty][tx] = y;
	  next_x[ty][tx] = x;
	  next_y[y][x] = -1;
	  ty = y;
	  tx = x;}}}}
  y = qy;
  x = qx;
  while (y!=-1)
  { marked[y][x] = FALSE;
    t = h[y][x]+1;
    if (x>0&&mask[y][x-1]&&EF_RIGHT(y, x-1)&&h[y][x-1]>t)
    { if (t>m1) panic("fuck up");
      h[y][x-1] = t;
      if (!marked[y][x-1])
      { marked[y][x-1] = TRUE;
	next_y[ty][tx] = y;
	next_x[ty][tx] = x-1;
	next_y[y][x-1] = -1;
	ty = y;
	tx = x-1;}}
    if (x<width-1&&mask[y][x+1]&&EF_LEFT(y, x+1)&&h[y][x+1]>t)
    { if (t>m1) panic("fuck up");
      h[y][x+1] = t;
      if (!marked[y][x+1])
      { marked[y][x+1] = TRUE;
	next_y[ty][tx] = y;
	next_x[ty][tx] = x+1;
	next_y[y][x+1] = -1;
	ty = y;
	tx = x+1;}}
    if (y>0&&mask[y-1][x]&&EF_DOWN(y-1, x)&&h[y-1][x]>t)
    { if (t>m1) panic("fuck up");
      h[y-1][x] = t;
      if (!marked[y-1][x])
      { marked[y-1][x] = TRUE;
	next_y[ty][tx] = y-1;
	next_x[ty][tx] = x;
	next_y[y-1][x] = -1;
	ty = y-1;
	tx = x;}}
    if (y<height-1&&mask[y+1][x]&&EF_UP(y+1, x)&&h[y+1][x]>t)
    { if (t>m1) panic("fuck up");
      h[y+1][x] = t;
      if (!marked[y+1][x])
      { marked[y+1][x] = TRUE;
	next_y[ty][tx] = y+1;
	next_x[ty][tx] = x;
	next_y[y+1][x] = -1;
	ty = y+1;
	tx = x;}}
    t = next_y[y][x];
    x = next_x[y][x];
    y = t;}}

void preflow_push(int height,
		  int width,
		  int **mask,
		  int **c_right,
		  int **c_down,
		  int **w,
		  int v)
{ int y, x, t, i = 0, k, flag = FALSE, pushes = 0, lifts = 0, relabels = 0;
  int m0 = height*width;
  int m1 = height*width+1;
  int total_flow = 0;
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { e[y][x] = 0;
      f_out[y][x] = 0;}}
  for (y = 0; y<height; y++) for (x = 0; x<width-1; x++) f_right[y][x] = 0;
  for (y = 0; y<height-1; y++) for (x = 0; x<width; x++) f_down[y][x] = 0;
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    if (x==width-1||x==0||y==height-1||y==0||
	mask[y][x]&&
	(!mask[y][x+1]||!mask[y][x-1]||!mask[y+1][x]||!mask[y-1][x]))
    { e[y][x] = -1;}}
  loop:
  if (i%6==0&&!flag)
  { relabel(height, width, mask, c_right, c_down, w, v);
    relabels++;
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { if (mask[y][x]&&e[y][x]!=0&&h[y][x]!=m1)
	{ for (k = 0; k<m1; k++) qy[k] = -1;
	  for (y = 0; y<height; y++)
	  { for (x = 0; x<width; x++)
	    { marked[y][x] = FALSE;}}
	  for (y = 0; y<height; y++)
	  { for (x = 0; x<width; x++)
	    { if (mask[y][x]&&e[y][x]!=0&&h[y][x]!=m1) enqueue(y, x);}}
	  flag = TRUE;
	  goto loop;}}}
    printf("%d push%s, %d lift%s, %d relabel%s, %d wave%s, flow=%d, e.t.\n",
	   pushes,
	   pushes==1?"":"es",
	   lifts,
	   lifts==1?"":"s",
	   relabels,
	   relabels==1?"":"s",
	   i,
	   i==1?"":"s",
	   total_flow);
    return;}
  for (k = 0; k<m1; k++)
  { y = qy[k];
    x = qx[k];
    while (y!=-1)
    { if (e[y][x]!=0)
      { for (k = m0; k>=0; k--)
	{ y = qy[k];
	  x = qx[k];
	  while (y!=-1)
	  { marked[y][x] = FALSE;
	    next2_y[y][x] = next_y[y][x];
	    next2_x[y][x] = next_x[y][x];
	    t = next_y[y][x];
	    x = next_x[y][x];
	    y = t;}
	  y = qy[k];
	  x = qx[k];
	  qy[k] = -1;
	  while (y!=-1)
	  { if (CAN_PUSH_OUT(y, x))
	    { pushes++;
	      t = CF_OUT(y, x);
	      if (e[y][x]>=0&&e[y][x]<t) t = e[y][x];
	      f_out[y][x] += t;
	      if (e[y][x]>=0) e[y][x] -= t;
	      total_flow += t;}
	    if (CAN_PUSH_RIGHT(y, x))
	    { pushes++;
	      t = CF_RIGHT(y, x);
	      if (e[y][x]>=0&&e[y][x]<t) t = e[y][x];
	      f_right[y][x] += t;
	      if (e[y][x]>=0) e[y][x] -= t;
	      if (e[y][x+1]>=0) {e[y][x+1] += t; enqueue(y, x+1);}}
	    if (CAN_PUSH_LEFT(y, x))
	    { pushes++;
	      t = CF_LEFT(y, x);
	      if (e[y][x]>=0&&e[y][x]<t) t = e[y][x];
	      f_right[y][x-1] -= t;
	      if (e[y][x]>=0) e[y][x] -= t;
	      if (e[y][x-1]>=0) {e[y][x-1] += t; enqueue(y, x-1);}}
	    if (CAN_PUSH_DOWN(y, x))
	    { pushes++;
	      t = CF_DOWN(y, x);
	      if (e[y][x]>=0&&e[y][x]<t) t = e[y][x];
	      f_down[y][x] += t;
	      if (e[y][x]>=0) e[y][x] -= t;
	      if (e[y+1][x]>=0) {e[y+1][x] += t; enqueue(y+1, x);}}
	    if (CAN_PUSH_UP(y, x))
	    { pushes++;
	      t = CF_UP(y, x);
	      if (e[y][x]>=0&&e[y][x]<t) t = e[y][x];
	      f_down[y-1][x] -= t;
	      if (e[y][x]>=0) e[y][x] -= t;
	      if (e[y-1][x]>=0) {e[y-1][x] += t; enqueue(y-1, x);}}
	    if (e[(y)][(x)]!=0&&h[(y)][(x)]<m1)
	    { lifts++;
	      if (EF_OUT(y, x)) h[y][x] = 1;
	      else
	      { t = m0;
		if (x<width-1&&mask[y][x+1]&&EF_RIGHT(y, x)&&t>h[y][x+1])
		{ t = h[y][x+1];}
		if (x>0&&mask[y][x-1]&&EF_LEFT(y, x)&&t>h[y][x-1])
		{ t = h[y][x-1];}
		if (y<height-1&&mask[y+1][x]&&EF_DOWN(y, x)&&t>h[y+1][x])
		{ t = h[y+1][x];}
		if (y>0&&mask[y-1][x]&&EF_UP(y, x)&&t>h[y-1][x])
		{ t = h[y-1][x];}
		if (t>=m1) panic("fuck up");
		h[y][x] = t+1;}}
	    if (e[y][x]!=0&&h[y][x]!=m1) enqueue(y, x);
	    t = next2_y[y][x];
	    x = next2_x[y][x];
	    y = t;}}
	i++;
	flag = FALSE;
	goto loop;}
      t = next_y[y][x];
      x = next_x[y][x];
      y = t;}}
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { if (mask[y][x]&&h[y][x]!=m1&&e[y][x]!=0) panic("fuck up");}}
  relabel(height, width, mask, c_right, c_down, w, v);
  relabels++;
  printf("%d push%s, %d lift%s, %d relabel%s, %d wave%s, flow=%d\n",
	 pushes,
	 pushes==1?"":"es",
	 lifts,
	 lifts==1?"":"s",
	 relabels,
	 relabels==1?"":"s",
	 i,
	 i==1?"":"s",
	 total_flow);
  return;}

int **ratio_region(int height,
		   int width,
		   int **mask,
		   int **c_right,
		   int **c_down,
		   int **w,
		   int lg_max_v)
{ int lg_v = lg_max_v, v_max = 0, v, y, x, i;
  int m1 = height*width+1;
  f_right = make_c_int_matrix(height, width-1);
  f_down = make_c_int_matrix(height-1, width);
  f_out = make_c_int_matrix(height, width);
  h = make_c_int_matrix(height, width);
  e = make_c_int_matrix(height, width);
  marked = make_c_int_matrix(height, width);
  next_y = make_c_int_matrix(height, width);
  next_x = make_c_int_matrix(height, width);
  next2_y = make_c_int_matrix(height, width);
  next2_x = make_c_int_matrix(height, width);
  qy = make_c_int_vector(m1);
  qx = make_c_int_vector(m1);
  while (lg_v>=0)
  { v = 1;
    for (i = 0; i<lg_v; i++) v += v;
    v += v_max;
    printf("LG-V=%d, V-MAX=%d, V=%d\n", lg_v, v_max, v);
    preflow_push(height, width, mask, c_right, c_down, w, v);
    lg_v--;
    for (y = 0; y<height; y++)
    { for (x = 0; x<width; x++)
      { if (mask[y][x]&&h[y][x]!=m1) goto l1;}}
    v_max = v;
    l1:;}
  printf("V-MAX=%d\n", v_max);
  preflow_push(height, width, mask, c_right, c_down, w, v_max+1);
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { if (mask[y][x]&&h[y][x]!=m1) goto l2;}}
  panic("fuck up");
  l2:
  for (y = 0; y<height; y++)
  { for (x = 0; x<width; x++)
    { marked[y][x] = mask[y][x]&&h[y][x]!=m1;}}
  free_c_int_matrix(f_right, height);
  free_c_int_matrix(f_down, height-1);
  free_c_int_matrix(f_out, height);
  free_c_int_matrix(h, height);
  free_c_int_matrix(e, height);
  free_c_int_matrix(next_y, height);
  free_c_int_matrix(next_x, height);
  free_c_int_matrix(next2_y, height);
  free_c_int_matrix(next2_x, height);
  free_c_int_vector(qy);
  free_c_int_vector(qx);
  return marked;}

/* Tam V'Nishlam Shevah L'El Borei Olam */
